

#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "mbuf.h"
#include "rtl_utils.h"
#include "rtl_glue.h"
#if defined(RTL865X_TEST)||defined(RTL865X_MODEL_USER)
#include "ctype.h"
#else
#include "linux/ctype.h"
#endif

#define RTL_STATIC_INLINE   static __inline__

#define MBUF_COMPACT   




#define UNINIT_CLUSTER_SIZE (-1)
int32 m_clusterSize=UNINIT_CLUSTER_SIZE;	


#include "mbuf_local.h"

static int32 _mbInited=0;
struct extCluster { 
	void *buf;
	uint32 id;
};
static struct extCluster *_clusters;
#define DEFAULT_MAX_RXDESC_IDX  (32767)
static int16 _maxRxDescidx= DEFAULT_MAX_RXDESC_IDX;

#define UNCACHE_MASK		0x20000000
#define UNCACHE(x)		((uint32)x|UNCACHE_MASK)
#define CACHED(x)		((uint32)x & ~UNCACHE_MASK)


static int32 _iTryGet(uint32 * piAvailableBuffers, uint32 iRequestedBuffers,  uint32 * piWaitID, int32 iHow, uint32 * piHaveWaited);
static uint32 _iFreeMbufChain(struct rtl_mBuf *pFirstMbuf, int8 flags);
static void _vWakeup(uint32 * piWaitID);
static void _vInitBufferPools(void *pBufPool, int32 iSize, int32 iNum);
RTL_STATIC_INLINE  void _vFreePkthdr(struct rtl_pktHdr *ph);
RTL_STATIC_INLINE uint8 _iGetClusterDataByte(struct rtl_mBuf *m, uint32 offset, uint32 *offsetTbl, uint32 *lenTbl, uint32 mbufs);
RTL_STATIC_INLINE void _vReturnFreeBuffer(void *pvFreeBuffer,  struct poolstat *pBufPoolStat);
RTL_STATIC_INLINE void *_pvGetFreeBuffer(struct poolstat *pBufPoolStat);
static struct rtl_mBuf *_pCopyPacket(struct rtl_mBuf *pMbuf, uint32 iHow);

#if defined(RTL865X_TEST)||defined(RTL865X_MODEL_USER)
#define UNCACHED_MALLOC(x)  (void *) ((uint32) rtlglue_malloc(x))
#else
#define UNCACHED_MALLOC(x)  (void *) (0x20000000 | (uint32) rtlglue_malloc(x))
#endif 

int32 mBuf_init(uint32 nmbufs, uint32 nmbclusters, uint32 nmpkthdrs, uint32 clusterSize, memaddr logId)
{
	uint32 exponent;

	if ( _mbInited!=0)
		return FAILED;


	rtlglue_mbufMutexLock();

	_iTotalMbufs = nmbufs;

	_iTotalPkthdrs = nmpkthdrs;
	_mbufMsgLogId = logId;

	_MbufPool= UNCACHED_MALLOC(_iTotalMbufs *sizeof(struct rtl_mBuf));	

	if(nmbclusters){
		
		if((isPowerOf2(clusterSize, &exponent)==FALSE)||(clusterSize < sizeof(struct mcluster_tag)))
			return FAILED;

		m_clusterSize = clusterSize;
		m_clusterShift = exponent;

		_ClusterDataPool= UNCACHED_MALLOC(nmbclusters * clusterSize); 
		_ClusterTag= (struct mcluster_tag *)rtlglue_malloc(nmbclusters * sizeof(struct mcluster_tag)); 
		_iTotalClusters = nmbclusters;
	}else{
		
		
		
		m_clusterSize = clusterSize;
		m_clusterShift =0;
		 _clusters= (struct extCluster *)rtlglue_malloc(sizeof(struct extCluster)*nmbufs);
		 if(!_clusters)
		 	return FAILED;
		
		memset(_clusters, 0, sizeof(struct extCluster)*nmbufs);
		_iTotalClusters = 0;
		_ClusterDataPool =NULL;
		_ClusterTag=NULL;
	}

	_PkthdrPool= UNCACHED_MALLOC(_iTotalPkthdrs * sizeof(struct rtl_pktHdr)); 
	if(!_MbufPool || (nmbclusters&&(!_ClusterDataPool || !_ClusterDataPool)) || !_PkthdrPool)
		return FAILED;

	if(nmbclusters){
		_vInitBufferPools(_ClusterTag, sizeof(struct mcluster_tag),  _iTotalClusters);
		memset((void *) &_ClStatus, (uint8) 0,   sizeof(struct poolstat));
		
		memset((void *) _ClusterDataPool, (uint8) 0,
			   _iTotalClusters * m_clusterSize);
		_ClStatus.pvAddr = _ClStatus.pvFreelist = _ClusterTag;
		_ClStatus.iTotalSize = _ClStatus.iFreebufs = _iTotalClusters;

	}



	
	_vInitBufferPools(_MbufPool, sizeof(struct rtl_mBuf), _iTotalMbufs);
	_vInitBufferPools(_PkthdrPool, sizeof(struct rtl_pktHdr), _iTotalPkthdrs);

	
	memset((void *) &_MbufStatus, (uint8) 0,   sizeof(struct poolstat));
	memset((void *) &_PHStatus, (uint8) 0,   sizeof(struct poolstat));

	_MbufStatus.pvAddr = _MbufStatus.pvFreelist = _MbufPool;
	_MbufStatus.iTotalSize = _MbufStatus.iFreebufs = _iTotalMbufs;


	_PHStatus.pvAddr = _PHStatus.pvFreelist = _PkthdrPool;
	_PHStatus.iTotalSize = _PHStatus.iFreebufs = _iTotalPkthdrs;


	memset((void *) &_Statistics, (uint8) 0, sizeof(struct rtl_mBufStatus));
	_Statistics.m_msize = sizeof(struct rtl_mBuf);	   
	_Statistics.m_pkthdrsize = sizeof(struct rtl_pktHdr);
	_Statistics.m_mclbytes=m_clusterSize;
	_mbInited = 1;
	rtlglue_mbufMutexUnlock();
	return SUCCESS;
}

#if defined(RTL865X_TEST)||defined(RTL865X_MODEL_USER)

int32 mBuf_Reinit(void)
{

	if ( _mbInited!=1)
		return FAILED;

	rtlglue_mbufMutexLock();

	
	memset(_MbufPool, 0, (_iTotalMbufs*sizeof(struct rtl_mBuf)));

	
	if (_iTotalClusters)
	{	
		memset(_ClusterDataPool, 0, (m_clusterSize*_iTotalClusters));
		memset(_ClusterTag, 0, (_iTotalClusters*sizeof(struct mcluster_tag)));
		memset((void *) &_ClStatus, (uint8) 0,   sizeof(struct poolstat));

		_vInitBufferPools(_ClusterTag, sizeof(struct mcluster_tag), _iTotalClusters);

		_ClStatus.pvAddr = _ClStatus.pvFreelist = _ClusterTag;
		_ClStatus.iTotalSize = _ClStatus.iFreebufs = _iTotalClusters;	
	}
	else
	{	
		if (_clusters)
			memset(_clusters, 0, (_iTotalMbufs*sizeof(struct extCluster)));
	}
	
	memset(_PkthdrPool, 0 , (_iTotalPkthdrs*sizeof(struct rtl_pktHdr)));

	
	_vInitBufferPools(_MbufPool, sizeof(struct rtl_mBuf), _iTotalMbufs);
	_vInitBufferPools(_PkthdrPool, sizeof(struct rtl_pktHdr), _iTotalPkthdrs);

	
	memset((void *) &_MbufStatus, (uint8) 0,   sizeof(struct poolstat));
	_MbufStatus.pvAddr = _MbufStatus.pvFreelist = _MbufPool;
	_MbufStatus.iTotalSize = _MbufStatus.iFreebufs = _iTotalMbufs;

	
	memset((void *) &_PHStatus, (uint8) 0,   sizeof(struct poolstat));
	_PHStatus.pvAddr = _PHStatus.pvFreelist = _PkthdrPool;
	_PHStatus.iTotalSize = _PHStatus.iFreebufs = _iTotalPkthdrs;

	
	memset((void *) &_Statistics, (uint8) 0, sizeof(struct rtl_mBufStatus));
	_Statistics.m_msize = sizeof(struct rtl_mBuf);	   
	_Statistics.m_pkthdrsize = sizeof(struct rtl_pktHdr);
	_Statistics.m_mclbytes=m_clusterSize;
	_mbInited = 1;
	rtlglue_mbufMutexUnlock();
	return SUCCESS;
}
#endif 


int32 mBuf_getBufStat(struct rtl_mBufStatus *mbs){
	if (!mbs || _mbInited!=1)
		return FAILED;
	mbs->m_totalmbufs=_iTotalMbufs;
	mbs->m_totalclusters=_iTotalClusters;
	mbs->m_totalpkthdrs=_iTotalPkthdrs;
	mbs->m_freembufs=  _MbufStatus.iFreebufs;	   
	mbs->m_freeclusters=_ClStatus.iFreebufs;	   
	mbs->m_freepkthdrs=_PHStatus.iFreebufs;	   
	mbs->m_msize = _Statistics.m_msize;		   
	mbs->m_mclbytes=_Statistics.m_mclbytes;	   
	mbs->m_pkthdrsize= _Statistics.m_pkthdrsize;   
	mbs->m_wait= _Statistics.m_wait;	  
	return SUCCESS;
}


int32 mBuf_leadingSpace(struct rtl_mBuf * m)
{
	int32 retval;

	assert(m &&	ISSET(m->m_flags, MBUF_USED) && (m->m_flags & MBUF_EXT));
	retval = (int32)(m->m_data - m->m_extbuf);

	return retval;
}

int32 mBuf_reserve(struct rtl_mBuf * m, uint16 headroom){
	assert(m);
	assert(headroom);
	if(m->m_len){
		return FAILED;
	}
	if(m->m_extsize<headroom){
		return FAILED;
	}
	m->m_data=m->m_extbuf;
	MBUF_RESERVE(m, headroom);
	return SUCCESS;
}



int32 mBuf_trailingSpace(struct rtl_mBuf * m)
{
	int32 retval;

	assert(m && ISSET(m->m_flags, MBUF_USED) &&(m->m_flags & MBUF_EXT));
	retval =  (( (memaddr)m->m_extbuf + m->m_extsize) - ((memaddr)m->m_data + m->m_len));

	return retval;

}


int32 mBuf_clusterIsWritable(struct rtl_mBuf * m)
{
	int32 retval;
	
	if(!_iTotalClusters)
		return TRUE;
	if (CLDataAddr2TagAddr(m->m_extbuf)->cltag_mbuf == m)
		retval = (CLDataAddr2TagAddr(m->m_extbuf)->cltag_refcnt == 1);
	else
		retval = 0;

	return retval;

}


static uint32 _m_getPktlen(struct rtl_mBuf *mp){
	struct rtl_mBuf *m=mp;
	uint32 pktlen=0;
	assert(mp);
	assert(ISSET(m->m_flags, MBUF_USED));

	
	if(m->m_pkthdr)	
		pktlen = m->m_pkthdr->ph_len;
	else{
		pktlen = 0;	
		for (;m;)		
		{
			pktlen += m->m_len;
			if (m->m_next == (struct rtl_mBuf *) 0){
				m=mp;
				break;
			}
			m = m->m_next;
		}
	}
	return pktlen;
}

uint32 mBuf_getPktlen(struct rtl_mBuf *mp){
	uint32 ret;
	rtlglue_mbufMutexLock();
	
	ret = _m_getPktlen(mp);
	rtlglue_mbufMutexUnlock();
	return ret;
}

#ifndef MBUF_COMPACT

struct rtl_mBuf *mBuf_data2Mbuf(int8 * x)
{
	struct rtl_mBuf *retval;
	assert(x);
	assert(_iTotalClusters);
	retval = CLDataAddr2TagAddr(x)->cltag_mbuf;
	return retval;
}
#endif

static struct rtl_mBuf *_m_get(int32 how, uint32 Nbuf)
{
	register int32      iIndex=0, iHow, extClusterAllocated=0;
	register struct rtl_mBuf *pThisMbuf, *pFirstMbuf, *pPrevMbuf;
	register struct mcluster_tag *pThisClusterTag;
	uint32 iMbufWaited;
	uint32 iClusterWaited;
	assert(how);
	assert(Nbuf > 0);

	pThisMbuf = pFirstMbuf = pPrevMbuf = NULL;
	pThisClusterTag = NULL;
	iMbufWaited = iClusterWaited = FALSE;

	iHow = how;

	
	
	if (Nbuf>_MbufStatus.iFreebufs){
		goto out;
	}
	if(Nbuf*m_clusterSize>65536){
		goto out;
	}
	if(_iTryGet(&_MbufStatus.iFreebufs, Nbuf, &_MbufStatus.iWaiting, iHow,&iMbufWaited) == FALSE)
		goto out;

	
	if (ISCLEARED(iHow, MBUF_ONLY))	{
		if(_iTotalClusters==0){
			
			
			if(ISSET(iHow, MBUF_WAITOK))
				goto out;
			for(extClusterAllocated=0; extClusterAllocated < Nbuf; extClusterAllocated++){		
				if(rtlglue_mbufAllocCluster(&_clusters[extClusterAllocated].buf,m_clusterSize,&_clusters[extClusterAllocated].id)==FAILED){
					int32 release;
					
					for(release=0; release < extClusterAllocated; release++)
						rtlglue_mbufFreeCluster(_clusters[release].buf,m_clusterSize,_clusters[release].id);
					goto out;
				}
				assert(_clusters[extClusterAllocated].buf);
			}
		}else{
			
			if ((Nbuf>_iTotalClusters)||
				(_iTryGet(&_ClStatus.iFreebufs, Nbuf, &_ClStatus.iWaiting, iHow,&iClusterWaited) == FALSE))
					goto out;
		}
	}
	CLEARBITS(iHow, MBUF_WAITOK);	   

	

	if (ISSET(iHow, MBUF_ONLY)){	
		for (iIndex = 0; iIndex < Nbuf; iIndex++){		   
			assert(_MbufStatus.pvFreelist);
			pThisMbuf = (struct rtl_mBuf *) _pvGetFreeBuffer(&_MbufStatus);
			assert(pThisMbuf);
			memset((void *) pThisMbuf, (uint8) 0, sizeof(struct rtl_mBuf));	
			CLEARBITS(pThisMbuf->m_flags, MBUF_EXT);
			SETBITS(pThisMbuf->m_flags, MBUF_USED);
			pThisMbuf->m_next = NULL;
			_MbufStatus.iUsedbufs++;

			if (iIndex == 0)
				pFirstMbuf = pThisMbuf;	
			else
				pPrevMbuf->m_next = pThisMbuf;	
			pPrevMbuf = pThisMbuf;
		}
	}else{
		
		for (iIndex = 0; iIndex < Nbuf; iIndex++)
		{							   
			assert(_MbufStatus.pvFreelist);
			pThisMbuf = (struct rtl_mBuf *) _pvGetFreeBuffer(&_MbufStatus);

			assert(pThisMbuf);

			
			memset((void *) pThisMbuf, (uint8) 0, sizeof(struct rtl_mBuf));
			SETBITS(pThisMbuf->m_flags, MBUF_EXT | MBUF_USED);
			pThisMbuf->m_next = NULL;
			_MbufStatus.iUsedbufs++;

			if(_iTotalClusters){
				assert(_ClStatus.pvFreelist);
				pThisClusterTag =	(struct mcluster_tag *) _pvGetFreeBuffer(&_ClStatus);
				assert(pThisClusterTag);
				pThisMbuf->m_extbuf = pThisMbuf->m_data =
					(uint8 *) CLTagAddr2DataAddr(pThisClusterTag);
				
				pThisClusterTag->cltag_refcnt = 1;
				SETBITS(pThisClusterTag->cltag_flags, CLUSTER_USED);
				pThisClusterTag->cltag_mbuf = pThisMbuf;	
			}else{
				pThisMbuf->m_extbuf = pThisMbuf->m_data =  (uint8 *)_clusters[iIndex].buf;
				pThisMbuf->m_extClusterId = _clusters[iIndex].id;
			}				
			pThisMbuf->m_len = 0;
			pThisMbuf->m_extsize = m_clusterSize;

			if (iIndex == 0)
				pFirstMbuf = pThisMbuf;	
			else
				pPrevMbuf->m_next = pThisMbuf;	
			pPrevMbuf = pThisMbuf;

		}
	}

	if(_iTotalClusters){
	
		if (iClusterWaited == TRUE && _ClStatus.pvFreelist)	{
			_Statistics.m_wait++;
			_vWakeup(&_ClStatus.iWaiting); 
		}
		else if (iMbufWaited == TRUE && _MbufStatus.pvFreelist){
			_Statistics.m_wait++;
			_vWakeup(&_MbufStatus.iWaiting);	
		}
	}
out:

	if (iIndex == Nbuf)
		return (pFirstMbuf);		   

	return NULL;					   
}


 struct rtl_mBuf *mBuf_get(int32 how, int32 type, uint32 Nbuf){
	
	struct rtl_mBuf *m;
	rtlglue_mbufMutexLock();
	m=_m_get( how, Nbuf);
	rtlglue_mbufMutexUnlock();
	return m;
}



struct rtl_mBuf *mBuf_attachHeader(void *buffer, uint32 id, uint32 bufsize,uint32 datalen, uint16 align){
	struct rtl_mBuf *m=NULL;
	struct rtl_pktHdr *ph=NULL;	
	rtlglue_mbufMutexLock();

	
	if (!_MbufStatus.iFreebufs||!_PHStatus.iFreebufs||(align &&align>=bufsize)) 
		goto out;
	m = (struct rtl_mBuf *) _pvGetFreeBuffer(&_MbufStatus);
	_MbufStatus.iUsedbufs++;
	
	ph = (struct rtl_pktHdr *) _pvGetFreeBuffer(&_PHStatus);
	_PHStatus.iUsedbufs++;
	
	ph->ph_mbuf = m;
	ph->ph_len = datalen;
	ph->ph_flags= PKTHDR_USED;
	*((uint16 *)((uint8 *)ph)+8)=0;

	ph->ph_iphdrOffset=0;
	ph->ph_pppeTagged = 0;
	ph->ph_LLCTagged = 0;
	ph->ph_vlanTagged = 0;

	
	ph->ph_rxdesc=_maxRxDescidx;
	m->m_next = NULL;
	m->m_pkthdr = ph;
	m->m_len=datalen;
	m->m_flags=MBUF_USED|MBUF_EXT|MBUF_PKTHDR;
	m->m_data = ((uint8 *)buffer)+align;
	m->m_extbuf=buffer;
	m->m_extClusterId = id;
	m->m_extsize= bufsize;
out:
	rtlglue_mbufMutexUnlock();
	return m;
 }


int mBuf_attachHeaderLSO(uint32 skb_len,uint32 skb_data_len,uint8 *skb_data,uint32 skb,uint8 protocol, struct rtl_mBuf *m[32]){
	
	struct rtl_pktHdr *ph[32],*phi;
	struct rtl_mBuf *mi;
	int loop=0,size=skb_data_len;
	uint32 header_len=(skb_len-skb_data_len);
	uint32 mss=1514-header_len;
	rtlglue_mbufMutexLock();


	if(protocol==0x06) 
	{
		while(1)
		{
			size-=mss;		
			loop++;
			if(size<=0) break;
		}	
		
		if((loop>1) && ((loop&1)==1))
		{
			mss=(skb_data_len/(loop+1))+1;		
			
			if(mss&1) mss++;
		}	
		loop=0;
		size=skb_data_len;
	}

	while(1)
	{
		if (!_MbufStatus.iFreebufs||!_PHStatus.iFreebufs) 
			goto out;
		
		m[loop] = (struct rtl_mBuf *) _pvGetFreeBuffer(&_MbufStatus);
		_MbufStatus.iUsedbufs++;

		ph[loop] = (struct rtl_pktHdr *) _pvGetFreeBuffer(&_PHStatus);
		_PHStatus.iUsedbufs++;

		mi=m[loop];
		phi=ph[loop];

		phi->ph_mbuf = mi;
		phi->ph_flags= PKTHDR_USED;
		phi->ph_iphdrOffset=0;
		phi->ph_pppeTagged = 0;
		phi->ph_LLCTagged = 0;
		phi->ph_vlanTagged = 0;

		
		phi->ph_pkt_property = 0;	
		phi->ph_pkt_property |= PH_PKT_PROPER_IPIP_L3CSM;
		phi->ph_pkt_property |= PH_PKT_PROPER_IPIP_L4CSM;

		*((uint16 *)((uint8 *)ph)+8)=0;
		phi->ph_rxdesc=_maxRxDescidx;
		mi->m_next = NULL;
		mi->m_pkthdr = phi;		
		mi->m_flags=MBUF_USED|MBUF_EXT|MBUF_PKTHDR;		

		mi->m_len = phi->ph_len = header_len+((size>mss)?mss:size);		
		size-=mss;
		mi->m_extbuf = mi->m_data = (uint8*)UNCACHE(skb_data)+mss*loop;
		
		if(size<=0) 
		{
			mi->m_extClusterId = (uint32)skb;
			break;
		}
		mi->m_extClusterId = 0;
		mi->m_extsize= mi->m_len;
		loop++;		
	}
	
	rtlglue_mbufMutexUnlock();
	return SUCCESS;
	
out:	
	rtlglue_mbufMutexUnlock();
	return FAILED;

 }

#if 0
int mBuf_attachHeaderLSO_2mbuf(uint32 skb_len,uint32 skb_data_len,uint8 *skb_data,uint32 skb,uint8 protocol, struct rtl_mBuf *m[64]){
	
	struct rtl_pktHdr *ph[32],*phi;
	struct rtl_mBuf *mi=NULL,*mih;
	int loop=0,loop2=0,size=skb_len;
	uint32 header_len=(skb_len-skb_data_len);
	uint32 mss=1514-header_len;
	rtlglue_mbufMutexLock();


	if(protocol==0x06) 
	{
		while(1)
		{
			size-=mss;		
			loop++;
			if(size<=0) break;
		}	
		
		if((loop>1) && ((loop&1)==1))
		{
			mss=(skb_data_len/(loop+1))+1;		
			
			if(mss&1) mss++;
		}	
		loop=0;
		size=skb_len;
	}

	while(1)
	{
		if (!_MbufStatus.iFreebufs)
			goto out;		
		m[loop2] = (struct rtl_mBuf *) _pvGetFreeBuffer(&_MbufStatus);
		_MbufStatus.iUsedbufs++;
		mih=m[loop2];	

		if(!_PHStatus.iFreebufs) 
			goto out;
		ph[loop] = (struct rtl_pktHdr *) _pvGetFreeBuffer(&_PHStatus);
		_PHStatus.iUsedbufs++;
		phi=ph[loop];
	
		
		if(loop!=0)
		{
			if (!_MbufStatus.iFreebufs)
				goto out;
			m[loop2+1] = (struct rtl_mBuf *) _pvGetFreeBuffer(&_MbufStatus);
			_MbufStatus.iUsedbufs++;	
			mi=m[loop2+1];
			mih->m_next=mi;
			mi->m_next = NULL;			
			mih->m_len=header_len;
			mi->m_len = ((size>mss)?mss:size);	
			phi->ph_len = mi->m_len+mih->m_len;
			mi->m_pkthdr = phi;
			mi->m_flags=MBUF_USED|MBUF_EXT;
			mi->m_extbuf = mi->m_data = (uint8*)UNCACHE(skb_data)+mss*loop;
			mi->m_extClusterId = 0;
			mi->m_extsize= mi->m_len;
		}
		else
		{
			mih->m_next=NULL;			
			mih->m_len=phi->ph_len =((skb_len>mss)?mss:skb_len);				

		}

		
		mih->m_pkthdr = phi;
		mih->m_flags=MBUF_USED|MBUF_EXT|MBUF_PKTHDR;			
		mih->m_extbuf =( mih->m_data = (uint8*)UNCACHE(skb_data));			


		phi->ph_mbuf = mih;
		phi->ph_flags= PKTHDR_USED;
		phi->ph_iphdrOffset=0;
		phi->ph_pppeTagged = 0;
		phi->ph_LLCTagged = 0;
		phi->ph_vlanTagged = 0;

		*((uint16 *)((uint8 *)ph)+8)=0;
		phi->ph_rxdesc=_maxRxDescidx;		
		
		size-=mss;		
		
		if(size<=0) 
		{
			if(loop==0)
				{
				mih->m_extClusterId = (uint32)skb;
				}
			else
				mi->m_extClusterId = (uint32)skb;
			break;
		}
		
		mih->m_extClusterId = 0;										
		mih->m_extsize=mih->m_len;
	
		loop++;		
		loop2+=2;
	}

	
	rtlglue_mbufMutexUnlock();
	return SUCCESS;
	
out:		
	rtlglue_mbufMutexUnlock();
	return FAILED;

 }
#endif

#if 0
struct rtl_mBuf *mBuf_attachHeaderSG(struct sk_buff * skb,uint32 bufsize, uint16 align){
	
	struct rtl_mBuf *m=NULL;
	struct rtl_pktHdr *ph=NULL;
	rtlglue_mbufMutexLock();

	
	if (!_MbufStatus.iFreebufs||!_PHStatus.iFreebufs||(align &&align>=bufsize)) 
		goto out;

	ph = (struct rtl_pktHdr *) _pvGetFreeBuffer(&_PHStatus);
	_PHStatus.iUsedbufs++;
	
	
	m = (struct rtl_mBuf *) _pvGetFreeBuffer(&_MbufStatus);
	_MbufStatus.iUsedbufs++;
	
	
	ph->ph_mbuf = m;
	ph->ph_len = skb->len;
	ph->ph_flags= PKTHDR_USED;
	*((uint16 *)((uint8 *)ph)+8)=0;
	
	
	ph->ph_rxdesc=_maxRxDescidx;
	
	
	m->m_pkthdr = ph;
	m->m_len=skb->tail-skb->data;
	m->m_flags=MBUF_USED|MBUF_EXT|MBUF_PKTHDR;
	m->m_data = ((uint8 *)UNCACHE(skb->data))+align;
	m->m_extbuf=(uint8 *)UNCACHE(skb->data);
	m->m_extClusterId = (uint32)skb;
	m->m_extsize= bufsize;

	if(m->m_len<ph->ph_len)
	{
		int i;
		struct skb_shared_info *shinfo;	
		struct rtl_mBuf *old_m=m;
		struct rtl_mBuf *new_m=NULL;
		uint8 *v;
#ifdef SG_DEBUG			
		printk("send it.........................\n");
#endif
		
		shinfo=(struct skb_shared_info *)skb->end;
		for(i=0;i<shinfo->nr_frags;i++)
		{				

			if (!_MbufStatus.iFreebufs)
			{
				old_m->m_next = NULL;
				goto out; 
			}

			new_m = (struct rtl_mBuf *) _pvGetFreeBuffer(&_MbufStatus);
			_MbufStatus.iUsedbufs++;

			
			v=page_address(shinfo->frags[i].page);
			v=(uint8*)UNCACHE(v)+shinfo->frags[i].page_offset;
			
			old_m->m_next=new_m;
		
			new_m->m_pkthdr = ph;
			new_m->m_len=shinfo->frags[i].size;			
			new_m->m_flags=MBUF_USED|MBUF_EXT;
			new_m->m_data = v+align;
			new_m->m_extbuf=v;
			new_m->m_extClusterId = (uint32)(skb);
			new_m->m_extsize=bufsize;
			old_m=new_m;
#ifdef SG_DEBUG			
			printk("nr_frags=%d frags[i].size=%d .page_offset=%d .size=%d \n",i,shinfo->frags[i].size,shinfo->frags[i].page_offset,shinfo->frags[i].size);
			memDump(v,32,"test");
#endif			
		}		
		old_m->m_next=NULL;
		old_m->m_flags|=MBUF_EOR;


#ifdef SG_DEBUG			

		
		{
			int must_pad=128-(int)m->m_len;
			if(must_pad>0)
			{
				new_m=m->m_next;
				if(new_m->m_len>=must_pad)
				{					
					memcpy(m->m_data+m->m_len,new_m->m_data,must_pad);
					m->m_len+=must_pad;
					new_m->m_data+=must_pad;
					new_m->m_len-=must_pad;
				}
			}
		}
		

		
		printk("--------------------------\n");
		new_m=ph->ph_mbuf;
		while(new_m!=NULL)
		{
			memDump(new_m->m_data,30,"test");
		
			new_m=new_m->m_next;
		}
#endif		
		
		

	}
	else	
		m->m_next = NULL;
out:
	rtlglue_mbufMutexUnlock();
	return m;
 }
#endif


#ifndef MBUF_COMPACT

struct rtl_mBuf *mBuf_getCleared(int32 how, int32 type, uint32 Nbuf)
{
	register struct rtl_mBuf *m, *n;
	int32     i = 0;

	assert(Nbuf);
	rtlglue_mbufMutexLock();
	
	m = _m_get(how, Nbuf);
	if (m){
		assert(ISSET(m->m_flags, MBUF_USED));

		if(ISSET(how, MBUF_ONLY))
			goto out;

		
		assert(ISCLEARED(how, MBUF_ONLY));
		for (n = m, i = 0; i < Nbuf; i++, n = n->m_next)
			memset(MBUF2DATAPTR(n, int8 *), 0, m->m_extsize);
	}
out:
	rtlglue_mbufMutexUnlock();
	return (m);
}
#endif


static struct rtl_mBuf *_m_gethdr(struct rtl_mBuf *pMbuf, int32 how){
	int32 iLen = 0;
	struct rtl_mBuf *pThisMbuf = NULL;
	struct rtl_pktHdr *pPkthdr = NULL;
	uint32 PkthdrWaited;

	assert(how &&pMbuf);
	assert(ISCLEARED(pMbuf->m_flags, MBUF_PKTHDR));

	iLen = pMbuf->m_len;
	
	for (pThisMbuf = pMbuf->m_next; pThisMbuf; pThisMbuf = pThisMbuf->m_next)
	{
		
		if (ISSET(pThisMbuf->m_flags, MBUF_PKTHDR)){
			assert(ISSET(pPkthdr->ph_flags, PKTHDR_USED));
			assert(pPkthdr->ph_mbuf == pThisMbuf);
			pPkthdr = pThisMbuf->m_pkthdr;
			pMbuf->m_pkthdr = pPkthdr;
			pPkthdr->ph_mbuf = pMbuf;
			pPkthdr->ph_len += iLen;   
			SETBITS(pMbuf->m_flags, MBUF_PKTHDR);	
			for (; pThisMbuf && (pThisMbuf->m_next != NULL); pThisMbuf = pThisMbuf->m_next)
				CLEARBITS(pThisMbuf->m_flags, MBUF_PKTHDR);	
			return pMbuf;
		}
		iLen += pThisMbuf->m_len;
	}


	

	
	if ((!_PHStatus.pvFreelist) &&
		(_iTryGet(&_PHStatus.iFreebufs, 1, &_PHStatus.iWaiting, how, &PkthdrWaited) == FALSE))
			goto out;

	
	assert(_PHStatus.pvFreelist);
	pPkthdr = (struct rtl_pktHdr *) _pvGetFreeBuffer(&_PHStatus);
	assert(pPkthdr);
	memset((void *) pPkthdr, (uint8) 0, sizeof(struct rtl_pktHdr));	
	SETBITS(pPkthdr->ph_flags, PKTHDR_USED);
	pPkthdr->ph_mbuf = pMbuf;
	pPkthdr->ph_len += iLen;
	pMbuf->m_pkthdr = pPkthdr;
	SETBITS(pMbuf->m_flags, MBUF_PKTHDR);	

	
	for (pThisMbuf = pMbuf->m_next; pThisMbuf;pThisMbuf = pThisMbuf->m_next)
		pThisMbuf->m_pkthdr = pPkthdr;

	if (PkthdrWaited == TRUE && _PHStatus.pvFreelist){
		_Statistics.m_wait++;
		_vWakeup(&_PHStatus.iWaiting);	
	}

out:
	if (pPkthdr)
		return pMbuf;				   
	else
		return NULL;
}


 struct rtl_mBuf *mBuf_getPkthdr(struct rtl_mBuf *pMbuf, int32 how){
 	struct rtl_mBuf *m;
	rtlglue_mbufMutexLock();
	m = _m_gethdr( pMbuf,  how);
	
	m->m_pkthdr->ph_rxdesc=_maxRxDescidx;
	m->m_pkthdr->ph_priority = 0;	
	rtlglue_mbufMutexUnlock();
	return m;
	
 }



void mBuf_freePkthdr(struct rtl_pktHdr *ph){
	rtlglue_mbufMutexLock();
	_vFreePkthdr(ph);
	rtlglue_mbufMutexUnlock();
}





struct rtl_mBuf *mBuf_getm(struct rtl_mBuf *m, uint32 len, int32 how, int32 type){
	register struct rtl_mBuf *pMbuf = m, *mtail = NULL, *p=NULL;
	register int32     trailing_len = 0, Nbuf = 0;

	assert(how);
	assert(len);

	
	if(len> 65535)
		return pMbuf;

	rtlglue_mbufMutexLock();

	if (m != NULL){	   
		assert(ISSET(m->m_flags, MBUF_USED));
		for (mtail = m; mtail->m_next != NULL; mtail = mtail->m_next) ;
		trailing_len = mBuf_trailingSpace(mtail);
		
		if (trailing_len >= len)	   
			goto out;
		len -= trailing_len;
	}

	
	Nbuf = ((len - 1) / m_clusterSize) + 1;
	assert(Nbuf > 0);

	CLEARBITS(how, MBUF_ONLY);

	
	pMbuf = _m_get(how, Nbuf);

	if (!pMbuf)
		goto out;

	if (m != NULL)	{
		if(m->m_pkthdr){
			for (p = pMbuf; p; p = p->m_next)
				p->m_pkthdr = m->m_pkthdr;
		}
		mtail->m_next = pMbuf;
		pMbuf = m;
	}else{
		
		
		if (ISSET(how, MBUF_ALLOCPKTHDR) && (pMbuf != _m_gethdr(pMbuf, how))){
			_iFreeMbufChain(m, MBUF_FREEALL);
			pMbuf = NULL;
			goto out;
		}
		pMbuf->m_pkthdr->ph_rxdesc=_maxRxDescidx;		
		
		for (p = pMbuf; p; p = p->m_next)
			p->m_pkthdr = pMbuf->m_pkthdr;
	}
out:
	rtlglue_mbufMutexUnlock();
	return pMbuf;

}


struct rtl_mBuf *mBuf_prepend(struct rtl_mBuf *m, uint32 plen, int32 how){
	struct rtl_mBuf *pMbuf=m, *mm;
	int32     leading_len, Nbuf, start;
	int32	count=0;


	assert(m);
	assert(ISSET(m->m_flags, MBUF_USED));

	rtlglue_mbufMutexLock();
	
	
	if (ISSET(how, MBUF_GETNEWMBUF))
		leading_len = 0;
	else
		leading_len = mBuf_leadingSpace(m);

	
	if (leading_len >= plen){
		m->m_data -= plen;
		m->m_len += plen;
		if(m->m_pkthdr)
			m->m_pkthdr->ph_len +=plen;
		assert(pMbuf==m);
		goto out;
	}
	if(leading_len){
				
		
		m->m_data -= leading_len;	
		m->m_len += leading_len;	
		if(m->m_pkthdr)
			m->m_pkthdr->ph_len += leading_len;
		plen -= leading_len;
	}
	count = leading_len;


	
	Nbuf = ((plen - 1) / m_clusterSize) + 1;
	start = (Nbuf*m_clusterSize)-plen;

	CLEARBITS(how, MBUF_ONLY);

	
	pMbuf = _m_get(how, Nbuf);
	if (!pMbuf)
		goto out;

	assert(pMbuf->m_data == pMbuf->m_extbuf);

	
	for (mm = pMbuf; mm; mm = mm->m_next){
		
		assert(mm->m_len==0);
		if(mm==pMbuf){
			pMbuf->m_data += start;
			pMbuf->m_len = m->m_extsize - start; 	
			count+= pMbuf->m_len;
		}else{
			mm->m_len = mm->m_extsize;
			count+=mm->m_extsize;
		}
		if(m->m_pkthdr){	
			mm->m_pkthdr = m->m_pkthdr;
			m->m_pkthdr->ph_len+=mm->m_len;
		}
		if(mm->m_next==NULL){
			mm->m_next = m;			
			if(m->m_pkthdr){
				SETBITS(pMbuf->m_flags, MBUF_PKTHDR);
				m->m_pkthdr->ph_mbuf = pMbuf;		
				CLEARBITS(m->m_flags, MBUF_PKTHDR);
				assert(count==plen+leading_len);
				
			}
			break;
		}
	}
out:
	rtlglue_mbufMutexUnlock();
	return pMbuf;
}



struct rtl_mBuf *mBuf_padding(struct rtl_mBuf *m, uint32 plen, int32 how){
	struct rtl_mBuf *pLastMbuf=NULL, *pLastMbufNext, *pNewMbuf, *mm;
	int32     trailing_len, Nbuf, end;
	int32	count=0;


	assert(m);
	assert(ISSET(m->m_flags, MBUF_USED));

	rtlglue_mbufMutexLock();
	
	for(pLastMbuf=m, pLastMbufNext=m->m_next; pLastMbufNext; pLastMbuf=pLastMbuf->m_next, pLastMbufNext=pLastMbufNext->m_next); 

	trailing_len = mBuf_trailingSpace(pLastMbuf);

	
	if (trailing_len >= plen){
		
		pLastMbuf->m_len += plen;
		if(m->m_pkthdr)
			m->m_pkthdr->ph_len +=plen;
		goto out;
	}

	
	
	pLastMbuf->m_len += trailing_len;	
	if(m->m_pkthdr)
		m->m_pkthdr->ph_len += plen;
	count = trailing_len;

	plen -= trailing_len;

	
	Nbuf = ((plen - 1) / m_clusterSize) + 1;
	end = (Nbuf*m_clusterSize)-plen;

	CLEARBITS(how, MBUF_ONLY);
	pNewMbuf = _m_get(how, Nbuf);
	if (!pNewMbuf){
		rtlglue_mbufMutexUnlock();
		return NULL;
	}

	assert(pNewMbuf->m_data == pNewMbuf->m_extbuf);
	
	pLastMbuf->m_next = pNewMbuf;

	
	for (mm = pNewMbuf; mm; mm = mm->m_next)
	{
		
		assert(mm->m_len==0);

		if(m->m_pkthdr){	
			mm->m_pkthdr = m->m_pkthdr;
		}
		if(mm->m_next != NULL) {
			mm->m_len = mm->m_extsize;
			memset((void *)mm->m_data, 0, mm->m_len); 
			count+=mm->m_len;
		} else {
			mm->m_len = mm->m_extsize-end;
			memset((void *)mm->m_data, 0, mm->m_len); 
			count+=mm->m_len;
			assert(count==plen);
			pLastMbuf = mm;
			break;
		}
	}
out:
	rtlglue_mbufMutexUnlock();
	return pLastMbuf;
}


int32 mBuf_freeOneMbufPkthdr(struct rtl_mBuf *m, void **buffer, uint32 *id, uint16 *size){
	uint32 ret;
	assert(m);
	rtlglue_mbufMutexLock();
	if(buffer)
		*buffer =(void *)m->m_extbuf;
	if(id)
		*id =m->m_extClusterId;
	if(size)		
		*size = m->m_extsize;
	ret=_iFreeMbufChain(m, MBUF_KEEPCLUSTER | MBUF_FREEONE);
	rtlglue_mbufMutexUnlock();
	return ret;
}


int32 mBuf_attachCluster(struct rtl_mBuf *m, void *buffer, uint32 id, uint32 size, uint16 datalen, uint16 align){
	int32 retval = FAILED;
	struct rtl_pktHdr *ph;
	ph=m->m_pkthdr;
	rtlglue_mbufMutexLock();

	if(ISCLEARED(m->m_flags, MBUF_USED|MBUF_PKTHDR))
		goto out;
	SETBITS(m->m_flags, MBUF_EXT);
	
	m->m_extbuf=buffer;
	m->m_extClusterId = id;
	m->m_extsize= size;
	
	if(datalen){
		if(datalen+align>size)
			goto out;
		m->m_len=datalen;
		m->m_data = ((uint8 *)buffer)+align;
	}else{
		m->m_len=0;
		m->m_data = ((uint8 *)buffer);
	}
	
	retval = SUCCESS;
out:
	rtlglue_mbufMutexUnlock();
	return retval;
}




struct rtl_mBuf *mBuf_freeOne(struct rtl_mBuf *m)
{

	struct rtl_mBuf *n = NULL;
	rtlglue_mbufMutexLock();
	assert(m);
	assert(ISSET(m->m_flags, MBUF_USED));
	if (m->m_next)
		n = m->m_next;
	 _iFreeMbufChain(m, MBUF_FREEONE);
	rtlglue_mbufMutexUnlock();
	return n;
}



uint32 mBuf_freeMbufChain(register struct rtl_mBuf * m){
	int32     i = 0;

	assert(m);
	assert(ISSET(m->m_flags, MBUF_USED));

	rtlglue_mbufMutexLock();
	i = _iFreeMbufChain(m, MBUF_FREEALL);
	rtlglue_mbufMutexUnlock();
	if (i > 0)
		return i;
	else
		return 0;

}







int32 mBuf_copyToMbuf(struct rtl_mBuf * m0,  uint32 offset, register uint32 len,  int8 * cp)
{
	 int32 mlen;
	 struct rtl_mBuf *m = m0, *n;
	 uint32 off = offset;
	int32     totlen = 0;


	assert(m);
	rtlglue_mbufMutexLock();
	if(mBuf_clusterIsWritable(m)==0){	
		rtlglue_mbufMutexUnlock();
		return -1;
	}

	assert(ISSET(m->m_flags, MBUF_USED));
	mlen = m->m_len;
	while (off > mlen){
		off -= mlen;
		totlen += mlen;
		if (m->m_next == 0){
			if(totlen < offset) 
			{
				rtlglue_mbufMutexUnlock();
				return -1;
			}
		}
		m = m->m_next;
	}
 	while (len > 0){
 		int32 maxLen;
		assert(ISSET(m->m_flags, MBUF_EXT));
		if (m->m_len)
			mlen= min(mBuf_trailingSpace(m)+m->m_len - off, len);
		else  
			mlen = min((uint32)m->m_extsize,(uint32) len);

		
		maxLen = totlen + m->m_len + mlen;
		if(maxLen > 65535)
			mlen= 65535-totlen-m->m_len;

		rtl8651_memcpy( (void *) (off + MBUF2DATAPTR(m, int8 *)),(void *) cp,(uint32) mlen);
		cp += mlen;
		len -= mlen;
		mlen += off;
		off = 0;
		totlen += mlen;
		if(m->m_len < mlen+off)
			m->m_len = mlen+off;

		if (len == 0 || maxLen>65535)
			break;
		if (m->m_next == 0){
			n = _m_get(MBUF_DONTWAIT,  1);
			if (n == 0)
				break;
			n->m_len = min(m_clusterSize, (int32)len);
			m->m_next = n;
		}
		m = m->m_next;
	}

  	if (((m = m0)->m_flags & MBUF_PKTHDR)&& (m->m_pkthdr->ph_len < totlen))
		m->m_pkthdr->ph_len = totlen;
	rtlglue_mbufMutexUnlock();
	return totlen;
}



void mBuf_getBMjmpTable(uint8 *pat,  uint16 *jump_tbl,uint16 patLen, uint8 caseSensitive) {
	uint32 i;
	rtlglue_mbufMutexLock();
	for(i=0; i<256; i++)
		jump_tbl[i]= patLen;

	if (caseSensitive == FALSE)
	{
		for(i = 0; i < patLen - 1; i++)
		{
			jump_tbl[(uint16) tolower(pat[i])] = patLen -i -1;
			jump_tbl[(uint16) toupper(pat[i])] = patLen -i -1;
		}
	}
	else
	{
		for(i=0; i< patLen-1; i++)
			jump_tbl[(uint16) pat[i]]= patLen -i -1;
	}
	rtlglue_mbufMutexUnlock();
}










int32 mBuf_BMpatternMatch(struct rtl_mBuf *m, uint32 len, uint8 *delimiter, uint32 delimitLen, uint16 *jmp_tbl, uint8 caseSensitive) {
	int32 retval= -1;
	

	assert(m && len && delimiter && delimitLen && jmp_tbl);
	rtlglue_mbufMutexLock();
	if(m->m_data + len <= m->m_extbuf + m->m_extsize){
		
		int32 i,j,k;
		uint8 *mem= (uint8 *)m->m_data;

		if (caseSensitive == FALSE)
		{
			for(i = delimitLen -1; i < len; ){  
				for(j = delimitLen - 1, k = i; j >= 0 && (tolower(mem[k]) == delimiter[j]); k--, j--); 
				if (j < 0) {	
					retval = k+1;	
					goto out;
				}else
					i+=jmp_tbl[(uint16)mem[i]];	
			}
		}
		else 
		{	
			for(i=delimitLen -1; i < len; ){  
				for(j=delimitLen-1, k=i; j>=0 && (mem[k]==delimiter[j]); k--, j--); 
				if (j<0){	
					retval = k+1;	
					goto out;
				}else
					i+=jmp_tbl[(uint16)mem[i]];	
			}
		}
	}else{

		
		struct rtl_mBuf *p;
		int32 i,j,k;
		uint32 offsetTbl[16], lenTbl[16],mbufs, num=0;

		
		for(mbufs=0, p=m; p; p=p->m_next, mbufs++){
			offsetTbl[mbufs]=num+p->m_len;
			lenTbl[mbufs]=p->m_len;
			num +=p->m_len;
		}

		if (caseSensitive == FALSE)
		{
			for(i=delimitLen -1; i < len; ){  
				for(j=delimitLen-1, k=i; j>=0 && tolower(_iGetClusterDataByte(m, k,offsetTbl,lenTbl,mbufs)) ==delimiter[j]; k--, j--); 
				if (j<0){	
					retval = k+1;	
					goto out;
				}else
					i+=jmp_tbl[(uint16)_iGetClusterDataByte(m, i, offsetTbl,lenTbl,mbufs)];	
			}
		}
		else
		{	
			for(i=delimitLen -1; i < len; ){  
				for(j=delimitLen-1, k=i; j>=0 && _iGetClusterDataByte(m, k,offsetTbl,lenTbl,mbufs) ==delimiter[j]; k--, j--); 
				if (j<0){	
					retval = k+1;	
					goto out;
				}else
					i+=jmp_tbl[(uint16)_iGetClusterDataByte(m, i, offsetTbl,lenTbl,mbufs)];	
			}
		}
	}
out:
	rtlglue_mbufMutexUnlock();
	return retval;
}



int32 mBuf_copyToUserBuffer(struct rtl_mBuf * m, uint32 off, uint32 len, int8 * cp)
{
	register uint32 count;
	register uint32 copied = 0;


	rtlglue_mbufMutexLock();
	while (off > 0)
	{
		assert(m != NULL);
		assert(ISSET(m->m_flags, MBUF_USED));
		if (off < m->m_len)
			break;
		off -= m->m_len;
		m = m->m_next;
	}
	while (len > 0)
	{
		assert(m != NULL);
		assert(ISSET(m->m_flags, MBUF_USED));
		count = min(m->m_len - off, len);
		rtl8651_memcpy( cp,MBUF2DATAPTR(m, int8 *) + off, count);
		len -= count;
		copied += count;
		cp += count;
		off = 0;
		m = m->m_next;
	}
	rtlglue_mbufMutexUnlock();
	return copied;
}

#ifndef MBUF_COMPACT

struct rtl_mBuf *mBuf_cloneMbufChain(struct rtl_mBuf *pMbufChain, int32 iOffset, int32 iLength,int32 flag){
	struct rtl_mBuf *m;
	rtlglue_mbufMutexLock();
	m =  _pCopyMbufChain(pMbufChain,  iOffset,  iLength, flag | MBUF_ONLY);
	rtlglue_mbufMutexUnlock();
	return m;
}



struct rtl_mBuf *mBuf_dupMbufChain(struct rtl_mBuf *pMbufChain, int32 iOffset, int32 iLength,  int32 flag){
	struct rtl_mBuf *m;
	rtlglue_mbufMutexLock();
	m =  _pCopyMbufChain( pMbufChain,  iOffset,  iLength, flag);
	rtlglue_mbufMutexUnlock();
	return m;
}


struct rtl_mBuf *mBuf_clonePacket(struct rtl_mBuf *pMbuf, int32 iHow){
	struct rtl_mBuf *m;
	rtlglue_mbufMutexLock();
	m =  _pCopyPacket(pMbuf, iHow | MBUF_ONLY);
	rtlglue_mbufMutexUnlock();
	return m;
}
#endif


struct rtl_mBuf *mBuf_dupPacket(struct rtl_mBuf *pMbuf, int32 iHow){
	struct rtl_mBuf *m;
	rtlglue_mbufMutexLock();
	m =  _pCopyPacket(pMbuf, iHow);
	rtlglue_mbufMutexUnlock();
	return m;
}

#ifndef MBUF_COMPACT


struct rtl_mBuf *mBuf_cat(struct rtl_mBuf *pHeadMbuf, struct rtl_mBuf *pTailMbuf)
{
	struct rtl_mBuf *pHeadLastMbuf = pHeadMbuf;
	struct rtl_mBuf *pMbuf;
	uint32    iTailLength = 0, pktlen=0;

	assert(pTailMbuf && pHeadMbuf);
	assert(ISSET(pHeadMbuf->m_flags, MBUF_USED));
	assert(ISSET(pTailMbuf->m_flags, MBUF_USED));

	rtlglue_mbufMutexLock();

	while (pHeadLastMbuf->m_next){
		pktlen+= pHeadLastMbuf->m_len;
		pHeadLastMbuf = pHeadLastMbuf->m_next;
	}
	pktlen+=pHeadLastMbuf->m_len;
	
	
	if((_m_getPktlen(pTailMbuf)+pktlen)>=65536){
		pHeadMbuf=NULL;
		goto out;
	}

	pHeadLastMbuf->m_next = pTailMbuf;

	
	if ISSET(pTailMbuf->m_flags, MBUF_PKTHDR){
		assert(pTailMbuf->m_pkthdr && ISSET(pTailMbuf->m_pkthdr->ph_flags, PKTHDR_USED));
		iTailLength = pTailMbuf->m_pkthdr->ph_len;
		_vFreePkthdr(pTailMbuf->m_pkthdr);
		pTailMbuf->m_pkthdr=0;
		CLEARBITS(pTailMbuf->m_flags, MBUF_PKTHDR);
	}

	if ISSET(pHeadMbuf->m_flags, MBUF_PKTHDR){
		assert(pHeadMbuf->m_pkthdr
			   && ISSET(pHeadMbuf->m_pkthdr->ph_flags, PKTHDR_USED));
		if (iTailLength > 0)
			pHeadMbuf->m_pkthdr->ph_len += iTailLength;
		pMbuf = pTailMbuf;
		while (pMbuf)	{
			CLEARBITS(pMbuf->m_flags, MBUF_PKTHDR);
			pMbuf->m_pkthdr = pHeadMbuf->m_pkthdr;	
			pMbuf = pMbuf->m_next;
		}
	}

out:
	rtlglue_mbufMutexUnlock();
	return pHeadMbuf;
}
#endif
struct rtl_mBuf *mBuf_adjHead(struct rtl_mBuf *mp, uint32 req_len){
	 uint32 len = req_len;
	 struct rtl_mBuf *m=mp;
	 uint32 adjusted=0, toomuch=0;


	assert(mp);
	assert(ISSET(m->m_flags, MBUF_USED));

	if(ISCLEARED(m->m_flags, MBUF_EXT))
		return NULL;

	rtlglue_mbufMutexLock();

	while (m != NULL && len > 0)	{
		assert(m && (m->m_flags & MBUF_EXT));
		if (m->m_len <= len){		
			len -= m->m_len;
			adjusted+= m->m_len;
			m->m_len = 0;		   
			m->m_data = m->m_extbuf; 
			m = m->m_next;
		}else{
			m->m_len -= len;
			adjusted+= len;

			if(m->m_len==0)
			{
				m->m_data = m->m_extbuf;
			}
			else
				m->m_data += len;
			len = 0;
			assert(m->m_len>0 );
			break;
		}
	}
	if(len>0)
		toomuch=1;
	if (mp->m_pkthdr)
		mp->m_pkthdr->ph_len -= adjusted;
	rtlglue_mbufMutexUnlock();
	if(toomuch)
		return mp;
	else
		return m;
}

struct rtl_mBuf *mBuf_adjTail(struct rtl_mBuf *mp, uint32 req_len){
	 uint32 len = req_len;
	 struct rtl_mBuf *m=mp, *n=NULL;
	 uint32 pktlen, drop;
	 uint32 adjusted=0;


	assert(mp);
	assert(req_len);
	assert(ISSET(m->m_flags, MBUF_USED));

	if(ISCLEARED(m->m_flags, MBUF_EXT)){
		return NULL;
	}

	rtlglue_mbufMutexLock();
	
	pktlen = _m_getPktlen(mp);

	if(len>pktlen)
		len = pktlen;
	drop = pktlen - len;	


	
	
	
	

	

	m = mp;		
	n = NULL;
	adjusted=0;
	for (; m; n=m, m = m->m_next)
	{
		assert(m && (m->m_flags & MBUF_EXT));
		if (m->m_len >= drop)			
		{
			adjusted+= (m->m_len - drop);
			if (drop==0){
				assert(m->m_extbuf);
				m->m_data = m->m_extbuf;
			}
			m->m_len = drop;		
			break;
		}
		drop -= m->m_len;			
	}

	
	

	n = m->m_next;
	if(!n && (m->m_len==0)){	
		m = mp;
		goto done;
	}

	while (n){
		adjusted+= n->m_len;
		n->m_len = 0;
		n->m_data = n->m_extbuf;
		n = n->m_next;
	}

done:
	if (mp->m_pkthdr)
		mp->m_pkthdr->ph_len -= adjusted;

	rtlglue_mbufMutexUnlock();
	return m;
}



#ifndef MBUF_COMPACT


static struct rtl_mBuf *_m_split(register struct rtl_mBuf *m0, uint32 len0, int32 wait){

	register struct rtl_mBuf *m = NULL, *n = NULL;
	uint32  len = len0, remain = 0;


	assert(len);
	assert(m0);
	assert(ISSET(m0->m_flags, MBUF_USED));

	rtlglue_mbufMutexLock();
	for (m = m0; m && len > m->m_len; m = m->m_next){  
		assert(ISSET(m->m_flags, MBUF_EXT));
		len -= m->m_len;
	}
	if (m == 0)	   
		goto out;

	assert(m->m_len >= len);
	remain = m->m_len - len;		   


	if (ISSET(m0->m_flags, MBUF_PKTHDR))	{
		if (remain == 0)	{							   
			if (!m->m_next)
				goto out;
			n = m->m_next;
			n->m_pkthdr = NULL;
		}else if ((n = _m_get(wait | MBUF_ONLY,  1)) == NULL)	
			goto out;

		if (!_m_gethdr(n, wait)){
			_iFreeMbufChain(n, MBUF_FREEALL);
			n=NULL;
			goto out;
		}


		
		rtl8651_memcpy((void *) (((int32 *) n->m_pkthdr) + 1), (void *) (((int32 *) m0->m_pkthdr) + 1),sizeof(struct rtl_pktHdr) - 4);

		n->m_pkthdr->ph_len = m0->m_pkthdr->ph_len - len0;	
		n->m_pkthdr->ph_rxdesc=_maxRxDescidx;
		
		m0->m_pkthdr->ph_len = len0;
		assert(ISSET(m->m_flags, MBUF_EXT));
		goto extpacket;

	}else if (remain == 0){	 
		n = m->m_next;
		m->m_next = 0;				   
		goto out;
	}else{	   
		m->m_len -= remain;
		if ((n = _m_get(wait | MBUF_ONLY,  1)) == NULL)
			goto out;
	}

  extpacket:
	SETBITS(n->m_flags, MBUF_EXT);
	if (remain != 0)	{
		n->m_extbuf = m->m_extbuf;
		n->m_extsize = m->m_extsize;
		if (_iTotalClusters)
			CLDataAddr2TagAddr(m->m_extbuf)->cltag_refcnt++;
		else{
			n->m_extClusterId= m->m_extClusterId;
			rtlglue_mbufClusterRefcnt(m->m_extbuf, m->m_extClusterId, NULL, 1);
		}
		n->m_data = m->m_data + len;
		n->m_len = remain;
		n->m_next = m->m_next;
		m->m_len = len;
	}
	m->m_next = 0;
out:
	rtlglue_mbufMutexUnlock();
	return n;
}


struct rtl_mBuf *mBuf_split(register struct rtl_mBuf *m0, uint32 len0, int32 wait){
	struct rtl_mBuf *m;
	rtlglue_mbufMutexLock();
	m =  _m_split( m0, len0, wait);
	rtlglue_mbufMutexUnlock();
	return m;
}



extern struct rtl_mBuf *mBuf_trimHead(struct rtl_mBuf *mp, uint32 req_len){
	struct rtl_mBuf *n=NULL;

	assert(mp);
	assert(req_len);
	assert(ISSET(mp->m_flags, MBUF_USED));
	rtlglue_mbufMutexLock();
	if(_m_getPktlen(mp)<= req_len){
		_iFreeMbufChain(mp,MBUF_FREEALL);
	}else{
		n = _m_split(mp, req_len, MBUF_DONTWAIT);
		if(_iFreeMbufChain(mp,MBUF_FREEALL)<1)
			;
	}
	rtlglue_mbufMutexUnlock();
	return n;

}



struct rtl_mBuf *mBuf_trimTail(struct rtl_mBuf *mp, uint32 req_len){
	struct rtl_mBuf *n;
	int32 len;

	assert(mp);
	assert(req_len);
	assert(ISSET(mp->m_flags, MBUF_USED));

	rtlglue_mbufMutexLock();
	len = _m_getPktlen(mp);

	if(len==0){
		mp=NULL;
		goto out;
	}else
		len -= req_len;

	if(len <= 0){ 
		_iFreeMbufChain(mp,MBUF_FREEALL);
		mp=NULL;
		goto out;

	}

	n = _m_split(mp, len, MBUF_DONTWAIT);
	assert(n);
	if(_iFreeMbufChain(n, MBUF_FREEALL)<1){
		;
		mp=NULL;
	}
	
out:
	rtlglue_mbufMutexUnlock();
	return mp;
}




int32 mBuf_pullup(register struct rtl_mBuf * pMbuf, int32 iLen)
{
	int32     iBytesLeft = 0, iRestSpace = 0, iCount = 0;
	uint8     *pChar = NULL;

	struct rtl_mBuf *pThisMbuf = NULL;

	assert(pMbuf && (iLen > 0));
	assert(ISSET(pMbuf->m_flags, MBUF_USED));

	rtlglue_mbufMutexLock();

	if (ISCLEARED(pMbuf->m_flags, MBUF_EXT) || (!pMbuf->m_data))	
		goto quit;

	if ((iBytesLeft = iLen) <= pMbuf->m_len)	
		goto quit;

	if ((pMbuf->m_pkthdr) && (pMbuf->m_pkthdr->ph_len < iBytesLeft))	
		goto quit;

	

	iRestSpace = mBuf_trailingSpace(pMbuf);

	iBytesLeft -= pMbuf->m_len;

	if (iRestSpace >= iBytesLeft)
	{								   
		pThisMbuf = pMbuf->m_next;	   
		assert(pThisMbuf != NULL);
		pChar = pMbuf->m_data + pMbuf->m_len;	
		while (iBytesLeft > 0)
		{
			while (pThisMbuf->m_len == 0)
				pThisMbuf = pThisMbuf->m_next;
			assert(pThisMbuf != NULL);
			iCount = min(pThisMbuf->m_len, iBytesLeft);
			if (!rtl8651_memcpy( pChar, pThisMbuf->m_data,iCount))
				goto quit;

			pThisMbuf->m_len -= iCount;	
			pThisMbuf->m_data += iCount;	

			pMbuf->m_len += iCount;
			pChar += iCount;

			iBytesLeft -= iCount;
		}
	}
	else {
	
	
		uint8 *cp1, *cp2;
		int32 ii = 0;
		iRestSpace = mBuf_leadingSpace(pMbuf);
		if (iRestSpace >= iBytesLeft){ 
		pThisMbuf = pMbuf->m_next;	   
		assert(pThisMbuf != NULL);

		cp1 = pMbuf->m_data;
		pMbuf->m_data -= iBytesLeft;
		cp2 = pMbuf->m_data;
		for(ii = 0 ; ii < pMbuf->m_len ; ii++){
			*cp2 = *cp1;
			 cp2++;
			 cp1++;
		}

		pChar = pMbuf->m_data + pMbuf->m_len;	


		while (iBytesLeft > 0)
		{
			while (pThisMbuf->m_len == 0)
				pThisMbuf = pThisMbuf->m_next;
			assert(pThisMbuf != NULL);
			iCount = min(pThisMbuf->m_len, iBytesLeft);
			if (!rtl8651_memcpy( pChar, pThisMbuf->m_data,iCount))
				goto quit;

			pThisMbuf->m_len -= iCount;	
			pThisMbuf->m_data += iCount;	

			pMbuf->m_len += iCount;
			pChar += iCount;

			iBytesLeft -= iCount;
		}
		}

	}

  quit:
	rtlglue_mbufMutexUnlock();
	return pMbuf->m_len;
}
#endif









uint32 mBuf_driverGetPkthdr(uint32 Npkthdr, struct rtl_pktHdr **ppFirstPkthdr, struct rtl_pktHdr **ppTailPkthdr){
	
	 int32 iIndex;
	 struct rtl_pktHdr *pThisPkthdr=NULL,*pPrevPkthdr=NULL;
	rtlglue_mbufMutexLock();

	Npkthdr = min(Npkthdr, _PHStatus.iFreebufs);
	if (Npkthdr==0){
		rtlglue_mbufMutexUnlock();
		return 0;
	}

	
	for (iIndex = 0; iIndex < Npkthdr; iIndex++){	
		assert(_PHStatus.pvFreelist);
		pThisPkthdr = _PHStatus.pvFreelist;
		_PHStatus.iFreebufs--;

		
		
		_PHStatus.pvFreelist = (uint32 *) (*((uint32 *) pThisPkthdr));

		
		
		pThisPkthdr->ph_flags= PKTHDR_USED|PKTHDR_DRIVERHOLD;
		pThisPkthdr->ph_nextHdr=NULL;
		pThisPkthdr->ph_len=0;
		pThisPkthdr->ph_rxdesc=_maxRxDescidx;
		
		if (iIndex != 0)
			pPrevPkthdr->ph_nextHdr = pThisPkthdr;	
		else
			*ppFirstPkthdr = pThisPkthdr;	
		pPrevPkthdr = pThisPkthdr;
	}
	assert(iIndex==Npkthdr);
	rtlglue_mbufMutexUnlock();
	*ppTailPkthdr = pThisPkthdr;
	return (iIndex);		   
}




uint32 mBuf_driverFreePkthdr(struct rtl_pktHdr *ph, uint32 Npkthdr, struct rtl_pktHdr **ppHeadPkthdr)
{
	register struct rtl_pktHdr *pThisPkthdr = ph, *pNextPkthdr = NULL;
	register uint32     iFreed = 0;
	rtlglue_mbufMutexLock();
	while (iFreed<Npkthdr && pThisPkthdr && ISSET(pThisPkthdr->ph_flags, PKTHDR_USED))
	{
		
		CLEARBITS(ph->ph_flags, PKTHDR_USED);
		
		*((uint32 *) pThisPkthdr) = (uint32) _PHStatus.pvFreelist;	
		_PHStatus.pvFreelist = pThisPkthdr;	
		_PHStatus.iFreebufs++;
		pNextPkthdr = pThisPkthdr->ph_nextHdr;
		iFreed++;
		pThisPkthdr = pNextPkthdr;
	}
	if(ppHeadPkthdr)
		*ppHeadPkthdr = pNextPkthdr;
	rtlglue_mbufMutexUnlock();
	return iFreed;
}





uint32 mBuf_driverGet(uint32 Nmbuf,struct rtl_mBuf **ppFirstMbuf, struct rtl_mBuf **ppTailMbuf)
{
	
	register uint32 iIndex, extClusterAllocated=0;
	register struct rtl_mBuf *pThisMbuf=NULL,*pPrevMbuf=NULL;
	register struct mcluster_tag *pThisClusterTag=NULL;
	rtlglue_mbufMutexLock();

	Nmbuf = min(Nmbuf, _MbufStatus.iFreebufs);
	if(_iTotalClusters){
		Nmbuf = min(Nmbuf, _ClStatus.iFreebufs);
		if (Nmbuf==0){
			rtlglue_mbufMutexUnlock();
			return 0;
		}
	}else{
		for(extClusterAllocated=0; extClusterAllocated < Nmbuf; extClusterAllocated++){		
			if(rtlglue_mbufAllocCluster(&_clusters[extClusterAllocated].buf,m_clusterSize,&_clusters[extClusterAllocated].id)==FAILED){
				int32 release;
				
				for(release=0; release < extClusterAllocated; release++)
					rtlglue_mbufFreeCluster(_clusters[extClusterAllocated].buf,m_clusterSize,_clusters[release].id);
				rtlglue_mbufMutexUnlock();
				return 0;
			}
		}
	}

	
	for (iIndex = 0; iIndex < Nmbuf; iIndex++){	
		

		
		assert(_MbufStatus.pvFreelist);
		pThisMbuf = _MbufStatus.pvFreelist;
		_MbufStatus.iFreebufs--;
		_MbufStatus.pvFreelist = (uint32 *) (*((uint32 *) pThisMbuf));

		if(_iTotalClusters){
			
			assert(_ClStatus.pvFreelist);
			pThisClusterTag = _ClStatus.pvFreelist;
			_ClStatus.iFreebufs--;
			_ClStatus.pvFreelist = (uint32 *) (*((uint32 *) pThisClusterTag));
		}
		pThisMbuf->m_next = NULL;
		pThisMbuf->m_pkthdr= 0;		
		pThisMbuf->m_len = 0;
		pThisMbuf->m_flags =  MBUF_EXT | MBUF_USED;
		_MbufStatus.iUsedbufs++;

		if(_iTotalClusters){
			pThisMbuf->m_extbuf = pThisMbuf->m_data =
				(uint8 *) CLTagAddr2DataAddr(pThisClusterTag);
			pThisClusterTag->cltag_mbuf = pThisMbuf;	
			pThisClusterTag->cltag_flags =  CLUSTER_USED;
			pThisClusterTag->cltag_refcnt = 1;
		}else{
			pThisMbuf->m_extbuf = pThisMbuf->m_data =  (uint8 *)_clusters[iIndex].buf;	
			pThisMbuf->m_extClusterId = _clusters[iIndex].id;
		}
		pThisMbuf->m_extsize = _Statistics.m_mclbytes;

		if (iIndex != 0)
			pPrevMbuf->m_next = pThisMbuf;	
		else
			*ppFirstMbuf = pThisMbuf;	
		pPrevMbuf = pThisMbuf;
	}
	assert(iIndex==Nmbuf);
	rtlglue_mbufMutexUnlock();
	*ppTailMbuf = pThisMbuf;
	return (iIndex);		   
}






uint32 mBuf_driverFreeMbufChain(struct rtl_mBuf *pFirstMbuf)
{
	register struct rtl_mBuf *pThisMbuf = pFirstMbuf, *pNextMbuf = NULL;
	register struct mcluster_tag *pThisClusterTag = NULL;
	struct rtl_pktHdr * pThisPkthdr=NULL;
	register uint32     iFreed = 0;

	rtlglue_mbufMutexLock();
	if(pThisMbuf->m_flags & MBUF_PKTHDR){
		int32 rxDescIdx;
		pThisPkthdr=pThisMbuf->m_pkthdr;
		rxDescIdx=pThisPkthdr->ph_rxdesc;
		if(rxDescIdx>=0&&rxDescIdx<_maxRxDescidx){
			if(rtlglue_reclaimRxBD(rxDescIdx, pThisPkthdr,  pThisMbuf))
				goto free;
			iFreed=1;
			goto out;
		}
	}

free:	
	
	while (pThisMbuf && ISSET(pThisMbuf->m_flags, MBUF_USED))
	{
		
		if (pThisMbuf->m_flags & MBUF_EXT)
		{
			if(_iTotalClusters){
				pThisClusterTag = CLDataAddr2TagAddr(pThisMbuf->m_data);

				pThisClusterTag->cltag_refcnt--;
				if (pThisClusterTag->cltag_refcnt == 0)
				{					   
					CLEARBITS(pThisClusterTag->cltag_flags, CLUSTER_USED);

					
					*((uint32 *) pThisClusterTag) = (uint32) _ClStatus.pvFreelist;	
					_ClStatus.pvFreelist = pThisClusterTag;	
					_ClStatus.iFreebufs++;
				}
				else if (pThisClusterTag->cltag_mbuf == pThisMbuf)
				{
					
					
					
					pThisClusterTag->cltag_mbuf = NULL;
				}
				
			}else{
				uint32 count;
				if(pThisMbuf->m_extClusterId){
					rtlglue_mbufClusterRefcnt(pThisMbuf->m_extbuf, pThisMbuf->m_extClusterId, &count, 0);
					if(count==1){
						
						rtlglue_mbufFreeCluster(pThisMbuf->m_extbuf, m_clusterSize, pThisMbuf->m_extClusterId);
					}else 
						rtlglue_mbufClusterRefcnt(pThisMbuf->m_extbuf, pThisMbuf->m_extClusterId, &count, 2);
				}

			}
		}

		if (pThisMbuf->m_pkthdr)
		{
			if (pThisMbuf->m_flags & MBUF_PKTHDR)
			{
				CLEARBITS(pThisMbuf->m_pkthdr->ph_flags, PKTHDR_USED);
				*((uint32 *) pThisMbuf->m_pkthdr) = (uint32) _PHStatus.pvFreelist;	
				_PHStatus.pvFreelist = pThisMbuf->m_pkthdr;	
				_PHStatus.iFreebufs++;

				pThisMbuf->m_pkthdr = NULL;
				CLEARBITS(pThisMbuf->m_flags, MBUF_PKTHDR);
			}
			else					   
				pThisMbuf->m_pkthdr = NULL;
		}

		
		_MbufStatus.iUsedbufs--;
		pThisMbuf->m_flags = MBUF_FREE;
		pNextMbuf = pThisMbuf->m_next;
		*((uint32 *) pThisMbuf) = (uint32) _MbufStatus.pvFreelist;	
		_MbufStatus.pvFreelist = pThisMbuf;	
		_MbufStatus.iFreebufs++;
		iFreed++;
		pThisMbuf = pNextMbuf;
	}
out:
	rtlglue_mbufMutexUnlock();
	return iFreed;
}

int32 mBuf_setNICRxRingSize(uint32 size){
	if ( _mbInited==0)
		return FAILED; 
	if (size>_iTotalPkthdrs)
		return FAILED;
	_maxRxDescidx=size;
	return SUCCESS;
}






static void _vWakeup(uint32 * piWaitID)
{
	
	assert(piWaitID);

	if (piWaitID){
		(*piWaitID)--;
		rtlglue_mbufTwakeup((uint32)piWaitID);
	}
	return;
}

RTL_STATIC_INLINE  void _vFreePkthdr(struct rtl_pktHdr *ph){
		assert(ph->ph_flags & PKTHDR_USED);
		CLEARBITS(ph->ph_flags, PKTHDR_USED);
		_vReturnFreeBuffer(ph, &_PHStatus);	
}





RTL_STATIC_INLINE uint8 _iGetClusterDataByte(struct rtl_mBuf *m, uint32 offset, uint32 *offsetTbl, uint32 *lenTbl, uint32 mbufs){
	uint32 i, len=offset;
	struct rtl_mBuf *p;
	for(i=0, p=m; i<mbufs && offset>offsetTbl[i]; len-=lenTbl[i], p=p->m_next, i++);  
	return (uint8)p->m_data[len];
}



static int32 _iTryGet(register uint32 * piAvailableBuffers, register uint32 iRequestedBuffers,
					register uint32 * piWaitID, register int32 iHow, register uint32 * piHaveWaited)
{
	
	int32     iWaited = FALSE;

	assert(piAvailableBuffers && (iRequestedBuffers > 0) && (piWaitID)
		   && (iHow) && (piHaveWaited));

	while (*piAvailableBuffers < iRequestedBuffers)	{

		
		if (ISSET(iHow , MBUF_DONTWAIT)){
			*piHaveWaited = iWaited;
			return FALSE;   
		}

		
		(*piWaitID)++;
		rtlglue_mbufMutexLock();
		
		rtlglue_mbufTsleep((uint32)piWaitID);
		(*piWaitID)--;
		iWaited = TRUE;
		rtlglue_mbufMutexUnlock();
	}

	*piHaveWaited = iWaited;
	return TRUE;
}





static void _vInitBufferPools(void *pBufPool, int32 iSize, int32 iNum)
{
	
	int32     i;
	int8     *pBuf;

	assert(pBufPool);
	assert(iSize && iNum);
	pBuf = (int8 *) pBufPool;

	memset((void *) pBuf, (uint8) 0, iSize * iNum);	

	
	for (i = 0; i < (iNum - 1); i++, pBuf += iSize)
		*((uint32 *) pBuf) = (uint32) (pBuf + iSize);	
}




RTL_STATIC_INLINE void *_pvGetFreeBuffer(struct poolstat *pBufPoolStat)
{
	
	void     *pvFreeBuffer;
	

	pvFreeBuffer = pBufPoolStat->pvFreelist;
	pBufPoolStat->iFreebufs--;

	
	pBufPoolStat->pvFreelist = (uint32 *) (*((uint32 *) pvFreeBuffer));
	return pvFreeBuffer;
};



RTL_STATIC_INLINE void _vReturnFreeBuffer(register void *pvFreeBuffer,
										  register struct poolstat *pBufPoolStat)
{
	

	
	*((uint32 *) pvFreeBuffer) = (uint32) pBufPoolStat->pvFreelist;	
	pBufPoolStat->pvFreelist = pvFreeBuffer;	
	pBufPoolStat->iFreebufs++;
}



uint32 _iFreeMbufChain(struct rtl_mBuf *pFirstMbuf, int8 flags)
{
	register struct rtl_mBuf *pThisMbuf = pFirstMbuf, *pNextMbuf = NULL;
	register struct mcluster_tag *pThisClusterTag = NULL;
	register uint32     iFreed = 0;
	register uint32     iFreedPkthdr = 0, iFreedCLusters = 0;

	assert(pFirstMbuf && ISCLEARED(pFirstMbuf->m_flags, MBUF_FREE));
	if(pThisMbuf->m_flags & MBUF_PKTHDR){		
		struct rtl_pktHdr *pThisPkthdr;
		int32 rxDescIdx;
		pThisPkthdr=pThisMbuf->m_pkthdr;
		rxDescIdx=pThisPkthdr->ph_rxdesc;
		if(rxDescIdx>=0&&rxDescIdx<_maxRxDescidx){
			if(rtlglue_reclaimRxBD(rxDescIdx, pThisPkthdr,  pThisMbuf))
				goto free;
			iFreed=1;
			goto out;
		}
	}
free:

	while (pThisMbuf && ISSET(pThisMbuf->m_flags, MBUF_USED))
	{
		
		if (pThisMbuf->m_flags & MBUF_EXT&& ISCLEARED(flags,MBUF_KEEPCLUSTER)){
			if(_iTotalClusters){
				
				pThisClusterTag = CLDataAddr2TagAddr(pThisMbuf->m_data);
				assert(pThisClusterTag);
				assert(pThisClusterTag->cltag_flags & CLUSTER_USED);
				pThisClusterTag->cltag_refcnt--;
				if (pThisClusterTag->cltag_refcnt == 0){   
					CLEARBITS(pThisClusterTag->cltag_flags, CLUSTER_USED);
					_vReturnFreeBuffer(pThisClusterTag, &_ClStatus);
					iFreedCLusters++;
				}else if (pThisClusterTag->cltag_mbuf == pThisMbuf){
					
					
					
					pThisClusterTag->cltag_mbuf = NULL;
				}
				
			}else{
				uint32 count;
				if(pThisMbuf->m_extClusterId){
					rtlglue_mbufClusterRefcnt(pThisMbuf->m_extbuf, pThisMbuf->m_extClusterId, &count, 0);
					if(count==1){
						
						rtlglue_mbufFreeCluster(pThisMbuf->m_extbuf, m_clusterSize, pThisMbuf->m_extClusterId);
					}else 
						rtlglue_mbufClusterRefcnt(pThisMbuf->m_extbuf, pThisMbuf->m_extClusterId, &count, 2);
				}
			}	
		}
		
		if (pThisMbuf->m_pkthdr){
			if (pThisMbuf->m_flags & MBUF_PKTHDR){
				assert(pThisMbuf->m_pkthdr->ph_mbuf == pThisMbuf);
				

				if ((pThisMbuf->m_next) && (flags&MBUF_FREEONE))	{
					pThisMbuf->m_pkthdr->ph_mbuf = pThisMbuf->m_next;
					assert(pThisMbuf->m_len);
					pThisMbuf->m_pkthdr->ph_len -= pThisMbuf->m_len;	
					SETBITS(pThisMbuf->m_next->m_flags, MBUF_PKTHDR);	
				}else{
					
					CLEARBITS(pThisMbuf->m_pkthdr->ph_flags, PKTHDR_USED);
					_vReturnFreeBuffer(pThisMbuf->m_pkthdr, &_PHStatus);	
					iFreedPkthdr++;
				}
				pThisMbuf->m_pkthdr = NULL;
				CLEARBITS(pThisMbuf->m_flags, MBUF_PKTHDR);
			}else if (ISSET(flags, MBUF_FREEONE))		   
				pThisMbuf->m_pkthdr->ph_len -= pThisMbuf->m_len;	
			else					   
				pThisMbuf->m_pkthdr = NULL;
		}

		
 		_MbufStatus.iUsedbufs--;
		pThisMbuf->m_flags = MBUF_FREE;
		pNextMbuf = pThisMbuf->m_next;
		CLEARBITS(pThisMbuf->m_flags, MBUF_USED);
		_vReturnFreeBuffer(pThisMbuf, &_MbufStatus);	

		iFreed++;

		if (ISSET(flags, MBUF_FREEALL))
			pThisMbuf = pNextMbuf;
		else
			break;
	}

	if (_iTotalClusters&&iFreedCLusters&&_ClStatus.iWaiting)
		_vWakeup(&_ClStatus.iWaiting);
	if (iFreedPkthdr&&_PHStatus.iWaiting)
		_vWakeup(&_PHStatus.iWaiting);
	if(_MbufStatus.iWaiting)
		_vWakeup(&_MbufStatus.iWaiting);

	pFirstMbuf = pNextMbuf;
out:
	return iFreed;
}



static struct rtl_mBuf *_pCopyPacket(struct rtl_mBuf *pMbuf, uint32 iHow)
{
	struct rtl_mBuf *pFirstMbuf = NULL, *pThisMbuf = NULL, *pNewMbuf =
		NULL, *pPrevNewMbuf = NULL;
	struct rtl_pktHdr *pNewPkthdr = NULL;
	uint32    iCopied = 0;

	assert(pMbuf);
	assert(iHow);
	pThisMbuf = pMbuf;

	while (pThisMbuf)
	{
		assert(ISSET(pThisMbuf->m_flags, MBUF_USED));
		pNewMbuf = _m_get(iHow,  1);	
		

		if (!pNewMbuf)
			goto nospace;

		if (pPrevNewMbuf)
			pPrevNewMbuf->m_next = pNewMbuf;	

		pPrevNewMbuf = pNewMbuf;	   

		if (!pFirstMbuf && ISSET(pThisMbuf->m_flags, MBUF_PKTHDR))
		{							   
			pFirstMbuf = pNewMbuf;
			if (!_m_gethdr(pFirstMbuf, iHow))	
				goto nospace;
			assert(pFirstMbuf->m_pkthdr);
			pNewPkthdr = pFirstMbuf->m_pkthdr;
			assert(pNewPkthdr->ph_mbuf == pFirstMbuf);

			
			rtl8651_memcpy
				((void *) (((int32 *) pNewPkthdr) + 1),
				  (void *) (((int32 *) pThisMbuf->m_pkthdr) + 1),
				 sizeof(struct rtl_pktHdr) - 4);
			
			pNewPkthdr->ph_rxdesc=_maxRxDescidx;
			assert(pFirstMbuf->m_pkthdr);
			pNewPkthdr = pFirstMbuf->m_pkthdr;
			assert(pNewPkthdr->ph_mbuf == pFirstMbuf);
			assert(pThisMbuf->m_pkthdr->ph_mbuf == pMbuf);
		}

		
		if ISSET
			(iHow, MBUF_ONLY)
		{							   
			rtl8651_memcpy
				((void *) (((int32 *) pNewMbuf) + 1),
				  (void *) (((int32 *) pThisMbuf) + 1),
				  sizeof(struct rtl_mBuf) - 4);
		}
		else
		{							   
			
			pNewMbuf->m_len = pThisMbuf->m_len;
			pNewMbuf->m_flags = pThisMbuf->m_flags;
		}

		
		if (pNewPkthdr)
			pNewMbuf->m_pkthdr = pNewPkthdr;

		assert(ISSET(pThisMbuf->m_flags, MBUF_EXT));
		assert(ISSET(pNewMbuf->m_flags, MBUF_EXT));

		if ISSET
			(iHow, MBUF_ONLY)
		{							   
			
			if (_iTotalClusters)
			{
				
				CLDataAddr2TagAddr(pThisMbuf->m_extbuf)->cltag_refcnt++;
			}else{
				pNewMbuf->m_extClusterId= pThisMbuf->m_extClusterId;				
				rtlglue_mbufClusterRefcnt(pThisMbuf->m_extbuf, pThisMbuf->m_extClusterId, NULL, 1);
			}
	
		}
		else
		{							   
			int32 align=UNCACHE(pThisMbuf->m_data)-UNCACHE(pThisMbuf->m_extbuf);
			if(align){				
				assert(align>=0&&align<pNewMbuf->m_extsize);
				pNewMbuf->m_data+=align;
			}
			rtl8651_memcpy
				( (void *) CACHED(pNewMbuf->m_data),(void *) pThisMbuf->m_data,
				 pThisMbuf->m_len);
			pNewMbuf->m_data=(uint8 *)CACHED(pNewMbuf->m_data);
			pNewMbuf->m_extbuf=(uint8 *)CACHED(pNewMbuf->m_extbuf);
			assert(CACHED(pNewMbuf->m_data)==(uint32)pNewMbuf->m_data);
		}

		iCopied++;
		pThisMbuf = pThisMbuf->m_next;
	}
	assert(pFirstMbuf);

	return pFirstMbuf;

  nospace:
  	if(pFirstMbuf)
		_iFreeMbufChain(pFirstMbuf, MBUF_FREEALL);

	return NULL;
}

#ifndef MBUF_COMPACT

static struct rtl_mBuf *_pCopyMbufChain(struct rtl_mBuf *pMbufChain, int32 iOffset, int32 iLength,	 int32 flag)
{
	register struct rtl_mBuf *pNewMbuf, *pThisMbuf, **ppNewMbuf;
	int32     off = iOffset;
	struct rtl_mBuf *pTopMbuf;
	struct rtl_pktHdr *copiedNewHdr=NULL;
	int32     copyhdr = 0;

	pThisMbuf = pMbufChain;

	assert(pThisMbuf && ISSET(pThisMbuf->m_flags, MBUF_USED));
	assert(iLength > 0 && off >= 0);

	if (ISSET(flag, MBUF_ALLOCPKTHDR))
		copyhdr = 1;				   

	if (pThisMbuf->m_len == 0){		   
		return NULL;
	}

	while (off > 0)
	{								   
		assert(pThisMbuf && pThisMbuf->m_len);
		if (off < pThisMbuf->m_len)
			break;
		off -= pThisMbuf->m_len;
		pThisMbuf = pThisMbuf->m_next;
	}

	
	if (pThisMbuf->m_len < off){
		return NULL;
	}

	ppNewMbuf = &pTopMbuf;
	pTopMbuf = 0;

	while ((iLength > 0) && (pThisMbuf))
	{
		if(ISSET(flag, MBUF_ONLY))  
			pNewMbuf = _m_get(MBUF_ONLY | flag,  1);	
		else  
			pNewMbuf = _m_get(flag,  1);	

		if (pNewMbuf == 0)
			goto nospace;

		*ppNewMbuf = pNewMbuf;

		if (copyhdr){	 
			if (!_m_gethdr(pNewMbuf, flag))	
				goto nospace;
			
			rtl8651_memcpy
				((void *) (((int32 *) pNewMbuf->m_pkthdr) + 1),
				  (void *) (((int32 *) pThisMbuf->m_pkthdr) + 1),
				 sizeof(struct rtl_pktHdr) - 4);
			pNewMbuf->m_pkthdr->ph_rxdesc=_maxRxDescidx;	 
			
			pNewMbuf->m_flags = pThisMbuf->m_flags;	
			SETBITS(pNewMbuf->m_flags, MBUF_PKTHDR);
			if (iLength == MBUF_COPYALL)
				pNewMbuf->m_pkthdr->ph_len -= iOffset;
			else
				pNewMbuf->m_pkthdr->ph_len = iLength;
			copiedNewHdr = pNewMbuf->m_pkthdr;
			copyhdr = 0;
		}

		pNewMbuf->m_len = min(iLength, pThisMbuf->m_len - off);

		if(ISSET(flag, MBUF_ONLY)){
			
			assert(ISSET(pThisMbuf->m_flags, MBUF_EXT));
			pNewMbuf->m_data = pThisMbuf->m_data + off;

			if (_iTotalClusters)
				CLDataAddr2TagAddr(pThisMbuf->m_extbuf)->cltag_refcnt++;
			else{
				pNewMbuf->m_extClusterId= pThisMbuf->m_extClusterId;
				rtlglue_mbufClusterRefcnt(pThisMbuf->m_extbuf, pThisMbuf->m_extClusterId, NULL, 1);
			}
			pNewMbuf->m_extbuf = pThisMbuf->m_extbuf;
			pNewMbuf->m_extsize = pThisMbuf->m_extsize;
			SETBITS(pNewMbuf->m_flags, MBUF_EXT);
		}else{
			int32 align;
			
			assert(ISSET(pNewMbuf->m_flags, MBUF_EXT));

			if(_iTotalClusters){
#ifdef RTL865X_DEBUG
				assert(CLDataAddr2TagAddr(pThisMbuf->m_extbuf)->cltag_refcnt==1);
#endif
			}else{
				int32 count;
				rtlglue_mbufClusterRefcnt(pThisMbuf->m_extbuf, pThisMbuf->m_extClusterId, &count, 0);
#ifdef RTL865X_DEBUG
				assert(count==1);
#endif
			}

			align=UNCACHE(pThisMbuf->m_data)-UNCACHE(pThisMbuf->m_extbuf);
			if(align){				
				assert(align>=0&&align<pNewMbuf->m_extsize);
				pNewMbuf->m_data+=align;
			}
			rtl8651_memcpy
				( (void *) pNewMbuf->m_data,(void *) pThisMbuf->m_data+off,
				 pThisMbuf->m_len);
			
			
			
		}
		if(copiedNewHdr)
			pNewMbuf->m_pkthdr = copiedNewHdr;

		iLength -= pNewMbuf->m_len;	   
		off = 0;
		pThisMbuf = pThisMbuf->m_next;
		ppNewMbuf = &pNewMbuf->m_next;
	}

	return (pTopMbuf);
  nospace:
	if (pTopMbuf)
		_iFreeMbufChain(pTopMbuf, MBUF_FREEALL);
	return (0);
}
#endif



