

#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_h323.h"

#define H225_TUNNEL_SUPPORT 1

#define _RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE 576

#undef RTL_H323_RASDEBUG

static uint8 *_rtl8651_l4_h323_alias_buf[2]; 

static int32 init_h323 = FALSE;
static uint16 tempPort = 0;

static uint16 lastPort[4]={0};
static uint32 RsvpLocalIp = 0;
static uint32 RsvpGlobalIp = 0;


static int32 find_h245_id(uint8 *data, int32 datlen);
static int32 find_port (int32, uint8 *, int32 , int32 *, uint16 *);


#if 0

#define H323_TPKT_VERSION		0x03
struct h323_tpktHdr {
	uint8	tpkt_ver;
	uint8	tpkt_reserv;
	uint16	tpkt_len;
};

#define H323_DISCRIMINATOR_Q931	0x08

#define H225_MSGTYPE_ALERTING					0x01
#define H225_MSGTYPE_CALLPROCEEDING			0x02
#define H225_MSGTYPE_CONNECT					0x07
#define H225_MSGTYPE_CONNECTACK				0x0f
#define H225_MSGTYPE_PROGRESS					0x03
#define H225_MSGTYPE_SETUP					0x05
#define H225_MSGTYPE_SETUPACK					0x0d

#define H225_IEI_UUIE							0x7e


#define PTR_SHIFT_EXIT(rem, ptr, shift, exit) \
	do { \
		if ((rem) <= (shift)) \
		{ \
			goto exit; \
		} \
		(rem) -= (shift); \
		(uint32)(ptr) += (uint32)(shift); \
	} while (0);





static void * _rtl8651_H225_getConnectPeerAddr(uint8 *data)
{
	void *ret = NULL;
	uint32 totalLen;
	uint8 *ptr = data;
	struct h323_tpktHdr *tpkt_p;

	
	tpkt_p = (struct h323_tpktHdr*)ptr;

	if (	(tpkt_p->tpkt_ver != H323_TPKT_VERSION) ||
		(tpkt_p->tpkt_reserv != 0) ||
		(htons(tpkt_p->tpkt_len) <= sizeof(struct h323_tpktHdr)))
	{
		goto out;
	}
	totalLen = htons(tpkt_p->tpkt_len);

	PTR_SHIFT_EXIT(totalLen, ptr, sizeof(struct h323_tpktHdr), out);

	
	{
		uint8 protoDiscriminator = *((uint8*)ptr);

		switch (protoDiscriminator)
		{
			case H323_DISCRIMINATOR_Q931:	
				PTR_SHIFT_EXIT(totalLen, ptr, sizeof(uint8), out);	
				break;
			default:
				goto out;
		}
	}
	
	{
		uint8 ref_len = (*((uint8 *)ptr)) & 0x0f;	

		PTR_SHIFT_EXIT(totalLen, ptr, sizeof(uint8), out);	
		PTR_SHIFT_EXIT(totalLen, ptr, (sizeof(uint8)*ref_len), out);	
	}

	
	{
		uint8 callMsgType = (*((uint8*)ptr)) & 0x7f;	

		switch (callMsgType)
		{
			case H225_MSGTYPE_CONNECT:
				PTR_SHIFT_EXIT(totalLen, ptr, sizeof(uint8), out);
				break;
			default:
				goto out;
		}
	}

	
	{
		uint8 IEI = *((uint8*)ptr);

		switch (IEI)
		{
			case H225_IEI_UUIE:
				PTR_SHIFT_EXIT(totalLen, ptr, sizeof(uint8), out);
				
				{
					uint16 ie_len = ntohs(*((uint16*)ptr));

					if (ie_len > totalLen)
					{
						goto out;
					}
				}
				
			default:
				goto out;
		}
	}

	
	{
		uint8 h225Magic[] = {0x00, 0x08, 0x91, 0x4a};
		uint32 idx;

		for (idx = 3 ; idx < (totalLen - sizeof(h225Magic)) ; idx ++)
		{
			if (memcmp((&(ptr[idx])), h225Magic, sizeof(h225Magic)) == 0)
			{
				break;
			}
		}
		if (idx >= (totalLen - sizeof(h225Magic)))
			goto out;

		if (*(uint8*)(&ptr[idx - 3]) == 0x02)
		{
			PTR_SHIFT_EXIT(totalLen, ptr, (idx + sizeof(h225Magic)), out);
			if (*ptr == 0)
			{
				PTR_SHIFT_EXIT(totalLen, ptr, 1, out);
				if (totalLen >= 6)
				{
					ret = ptr;
					goto out;
				}
			}
		}
	}

out:
	return ret;
}
#endif

static int32 find_port(ipaddr, data, datlen, off, port)
int32 ipaddr;
uint8 *data;
int32 datlen, *off;
uint16 *port;
{
	uint32 addr;
	uint8 *dp;
	int32 offset;

	if (datlen < 6 || data == NULL)
		return -1;
	
	*port = 0;
	offset = *off;
	dp = (uint8 *)data;	

	for (offset = 0; offset <= datlen - 6; offset++, dp++) {
		addr = (dp[0] << 24) | (dp[1] << 16) | (dp[2] << 8) | dp[3];
		if (ipaddr == addr)
		{			
			*port = (uint16)((*(dp + 4) << 8) | *(dp + 5));
			
			break;
		}
	}
	*off = offset;
  	return (offset > datlen - 6) ? -1 : 0;
}


#ifdef H225_TUNNEL_SUPPORT

static int32 find_h245_id(uint8 *data, int32 datlen)
{
	uint8 h245_id[] = { 0x08, 0x81, 0x75, 0x00  };

	uint32 id1, id2;
	uint8 *dp;
	int32 offset;

	if (datlen < 5 || data == NULL)
		return 0;
	
	offset = 0;
	dp = (uint8 *)data;
	id1 = (h245_id[0] << 24) | (h245_id[1] << 16) | (h245_id[2] << 8) | h245_id[3];

	for (offset = 0; offset <= datlen - 5; offset++, dp++) {
		id2 = (dp[0] << 24) | (dp[1] << 16) | (dp[2] << 8) | dp[3];
		if (id1 == id2 )
		{
			return 1;
		}
	}
	
  	return 0;
  }
#endif 



#define RTL865X_ALG_H323_H225

#define RTL865X_ALG_H323_H225_RAS


int32 rtl8651_l4AliasHandleClientH225RASIn (
					struct rtl_pktHdr* pkthdrPtr,
					struct ip *iphdr,
					rtl8651_tblDrv_naptTcpUdpFlowEntry_t *link)
{
	struct rtl_mBuf *mbuf = pkthdrPtr->ph_mbuf;
	struct udphdr *ud = NULL;
	struct ip *pip = NULL;
	uint8 *cp = NULL, *cp_end = NULL, *data = NULL;
	uint32 mins = mbuf->m_len, copied = 0, datalen = 0, hlen = 0, chksum_flag = FALSE;
	int32 offset = 0;
	uint16 port = 0;
	

	if ( (alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING) == 0 )
	{
		goto out;
	}

	if (IS_FRAGMENT(iphdr->ip_off))
	{
		goto out;
	}

	if (iphdr->ip_p != IPPROTO_UDP)
	{
		assert(0);
		goto out;
	}

	ud = (struct udphdr *)L4HDR_ADDR(iphdr);

	

	
	if (mins > _RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE)
	{
		mins = _RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE;
	}

	cp = _rtl8651_l4_h323_alias_buf[0];

	copied = (int32)mBuf_copyToUserBuffer(	mbuf,
											0,
											mins,
											(int8 *)(_rtl8651_l4_h323_alias_buf[0]));
	cp_end = cp + copied;

	pip = (struct ip *)(&_rtl8651_l4_h323_alias_buf[0][pkthdrPtr->ph_iphdrOffset]);

	switch (pip->ip_p)
	{
		case IPPROTO_UDP:
			ud = (struct udphdr *)L4HDR_ADDR(pip);
			hlen = ((pip->ip_vhl & 0xf) + sizeof(struct udphdr));				
			break;
		default:
			assert(0);
			return SUCCESS;
	}

	

	data = &_rtl8651_l4_h323_alias_buf[0][pkthdrPtr->ph_iphdrOffset + hlen];
	datalen = copied - hlen;

	while (find_port(link->insideGlobalIpAddr, data, datalen, &offset, &port) == 0 )
	{
		
		chksum_flag = TRUE;
#ifdef RTL_H323_RASDEBUG
		rtlglue_printf("[DATA IN] %u.%u.%u.%u(%d) -- connect back to LAN-endpoint.\n\t=>We modify IP to %u.%u.%u.%u.\n",
						NIPQUAD(link->insideGlobalIpAddr), port,
						NIPQUAD(link->insideLocalIpAddr));
#endif
		
		PKTSET_UINT32_UNALIGNED(link->insideLocalIpAddr, (uint8*)&(data[offset]));

		
		if (datalen < (offset + 6))
		{
			break;
		}
		data += (6+offset);
		datalen -= (6+offset);
		offset = 0;
	}

	

	if (chksum_flag)
	{
		
		if (!IPHDRLEN_IS_TWENTY(pip))
		{
			pip->ip_sum = 0;
			if (ud->uh_sum)
			{
				ud->uh_sum = 0;
				ud->uh_sum = _rtl8651_tcpChecksum(pip);
			}
			pip->ip_sum = _rtl8651_ipChecksum(pip);
		}
	}

	mBuf_copyToMbuf(mbuf, 0,copied,(int8 *)_rtl8651_l4_h323_alias_buf[0]);

out:
	return SUCCESS;
}

int32 rtl8651_l4AliasHandleClientH225RASOut (
					struct rtl_pktHdr* pkthdrPtr,
					struct ip *iphdr,
					rtl8651_tblDrv_naptTcpUdpFlowEntry_t *link)
{
	struct rtl_mBuf *mbuf = pkthdrPtr->ph_mbuf;
	struct udphdr *ud = NULL;
	struct ip *pip = NULL;
	uint8 *cp = NULL, *cp_end = NULL, *data = NULL;
	uint32 mins = mbuf->m_len, copied = 0, datalen = 0, hlen = 0, chksum_flag = FALSE;
	int32 offset = 0;
	uint16 port = 0;

	
	if ( (alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING) == 0 )
	{
		goto out;
	}
	if (IS_FRAGMENT(iphdr->ip_off))
	{
		goto out;
	}
	if (iphdr->ip_p != IPPROTO_UDP)
	{
		assert(0);
		goto out;
	}

	

	
	if (mins > _RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE)
	{
		mins = _RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE;
	}

	cp = _rtl8651_l4_h323_alias_buf[0];

	copied = (int32)mBuf_copyToUserBuffer(	mbuf,
											0,
											mins,
											(int8 *)(_rtl8651_l4_h323_alias_buf[0]));
	cp_end = cp + copied;

	pip = (struct ip *)(&_rtl8651_l4_h323_alias_buf[0][pkthdrPtr->ph_iphdrOffset]);

	switch (pip->ip_p)
	{
		case IPPROTO_UDP:
			ud = (struct udphdr *)L4HDR_ADDR(pip);
			hlen = ((pip->ip_vhl & 0xf) + sizeof(struct udphdr));				
			break;
		default:
			assert(0);
			return SUCCESS;
	}

	

	
	data = &_rtl8651_l4_h323_alias_buf[0][pkthdrPtr->ph_iphdrOffset + hlen];
	datalen = copied - hlen;

	while (find_port (link->insideLocalIpAddr, data, datalen, &offset, &port) == 0)
	{
		chksum_flag = TRUE;

		if (port == link->insideGlobalPort)
		{
			
#ifdef RTL_H323_RASDEBUG
			rtlglue_printf("[DATA OUT] %u.%u.%u.%u(%d)\n\t=>We modify IP to %u.%u.%u.%u.\n",
							NIPQUAD(link->insideLocalIpAddr), port,
							NIPQUAD(link->insideGlobalIpAddr));
#endif
		} else if (port == link->insideLocalPort)
		{
			
#ifdef RTL_H323_RASDEBUG
			rtlglue_printf("[DATA OUT] %u.%u.%u.%u(%d)\n\t=>We modify IP/Port to %u.%u.%u.%u(%d).\n",
							NIPQUAD(link->insideLocalIpAddr), port,
							NIPQUAD(link->insideGlobalIpAddr), link->insideGlobalPort);
#endif
			data[offset + sizeof(ipaddr_t)] = (htons(link->insideGlobalPort)) >> 8;
			data[offset + sizeof(ipaddr_t) + 1] = (htons(link->insideGlobalPort)) & 0xff;
		} else
		{
			

			uint32 retval;
#ifdef RTL_H323_RASDEBUG
			rtlglue_printf("[ALLOW IN] %u.%u.%u.%u(%d) -- allow other H.323 endpoints to connect back.\n\t=>We modify IP to %u.%u.%u.%u and add NAPT for new calls\n",
							NIPQUAD(link->insideLocalIpAddr), port,
							NIPQUAD(link->insideGlobalIpAddr));
#endif
#if 0
			rtl8651_addUpnpMap(	UPNP_TCP,
									0,
									0,
									link->insideGlobalIpAddr, port,
									link->insideLocalIpAddr, port);
#else
			if ((retval = _rtl8651_addNaptConnection(	TRUE, (TRUE|PURE_SW),
													TCP_LISTEN_FLOW,
													link->insideLocalIpAddr, port,
													&link->insideGlobalIpAddr,&port,
													0, 0)) != SUCCESS)
			{
				if (retval != TBLDRV_EDRVNAPTEXIST)
				{
					goto out;
				}
#ifdef RTL_H323_RASDEBUG
				rtlglue_printf("[INFO] Napt connection INT %u.%u.%u.%u(%d) : GLB %u.%u.%u.%u(%d) exists\n",
							NIPQUAD(link->insideLocalIpAddr), port,
							NIPQUAD(link->insideGlobalIpAddr), port);
#endif
			}
#endif
		}

		
		PKTSET_UINT32_UNALIGNED(link->insideGlobalIpAddr, ((uint8*)&(data[offset])));

		
		if (datalen < (offset + 6))
		{
			break;
		}
		data += (6+offset);
		datalen -= (6+offset);
		offset = 0;
	}

	

	if (chksum_flag)
	{
		
		if (!IPHDRLEN_IS_TWENTY(pip))
		{
			pip->ip_sum = 0;
			if (ud->uh_sum)
			{
				ud->uh_sum = 0;
				ud->uh_sum = _rtl8651_tcpChecksum(pip);
			}
			pip->ip_sum = _rtl8651_ipChecksum(pip);
		}
	}

	mBuf_copyToMbuf(mbuf, 0,copied,(int8 *)_rtl8651_l4_h323_alias_buf[0]);

out:
	return SUCCESS;
}

int32 rtl8651_l4AliasHandleClientH225CSIn(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
	struct in_addr ipaddr;
	int32 off;
	uint16 port;
	uint8 *data;
	struct tcphdr *tcp;
	int32     tlen;
	uint8 *cp;
	uint8 *cp_end;
	int32 copied = 0;
	struct rtl_mBuf *m=NULL;
	uint8 chksum_flag=0;
	uint32 mins;
	int32     hlen,datalen;

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING)==0) return SUCCESS;	

	m = (struct rtl_mBuf *)pkthdrPtr->ph_mbuf;
	mins=m->m_len;

	

	ipaddr.s_addr = PKTGET_UINT32_UNALIGNED(&pip->ip_src);

	
	tlen = ntohs(pip->ip_len);
	
	

	if(mins>_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE) mins=_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE;
	cp = _rtl8651_l4_h323_alias_buf[0];
	copied = (int32)mBuf_copyToUserBuffer((struct rtl_mBuf *)m, (uint32)0, mins, (int8 *)(_rtl8651_l4_h323_alias_buf[0]));
	cp_end = cp + copied;		

	
	pip = (struct ip *)(&_rtl8651_l4_h323_alias_buf[0][pkthdrPtr->ph_iphdrOffset]);
	tcp = (struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));

	hlen = ((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2 ;	

	
	data = ((uint8 *)pip)+hlen;
	
	datalen = copied - hlen;

	

#ifdef H225_TUNNEL_SUPPORT	





		
	if( link->h245tunnel || 		
		(find_h245_id(data,datalen)) ){
		
		link->h245tunnel = 1;
		return SUCCESS;
	}
#endif 
		
		
	off = cp - data;

	

	if (find_port(ipaddr.s_addr, data, datalen, &off, &port) == 0) {

	{
		uint32 addr_1;
		uint16 port_1;
		struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb = &entry;
		int32 r;
		
		addr_1 = link->insideGlobalIpAddr;
		port_1 = link->insideLocalPort+1;
		
		
		
			
			
				

			r = _rtl8651_addNaptConnection(TRUE,(TRUE|PURE_SW), TCP_LISTEN_FLOW,
				link->insideLocalIpAddr,link->insideLocalPort+1,
				&addr_1,&port_1,
				link->dstIpAddr,port);	
				
			
			if(r){
				goto end_proc;
			}

		if(_rtl8651_getOutboundNaptFlow(TRUE,link->insideLocalIpAddr,link->insideLocalPort+1, 
				link->dstIpAddr,port,&tb)!=SUCCESS) 
			{			
				
				goto end_proc;
			}					
			
			tb->alg_client_out = rtl8651_l4AliasHandleClientH245Out;
			tb->alg_client_in = rtl8651_l4AliasHandleClientH245In;

	}

}
end_proc:


	while( find_port(link->insideGlobalIpAddr, data, datalen, &off, &port) ==0 ) {
		

		
		
		chksum_flag = 1;
		data[off]=(link->insideLocalIpAddr>>24)&0xff;
		data[off+1]=(link->insideLocalIpAddr>>16)&0xff;
		data[off+2]=(link->insideLocalIpAddr>>8)&0xff;
		data[off+3]=(link->insideLocalIpAddr)&0xff;		

		
		
		data+=(off+6);
		datalen-=(off+6);		
		off = 0;
	}	

	if(chksum_flag){
		
		
		if((pip->ip_vhl & 0xf)!=5)
		{
			tcp->th_sum = 0;
			pip->ip_sum = 0;				
			tcp->th_sum = _rtl8651_tcpChecksum(pip);
			pip->ip_sum = _rtl8651_ipChecksum(pip);		
		}
	}

	mBuf_copyToMbuf(m, 0,copied,(int8 *)_rtl8651_l4_h323_alias_buf[0]);


	return SUCCESS;
}


int32 rtl8651_l4AliasHandleClientH225CSOut(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{

	uint8 *data;
	struct tcphdr *tcp;
	int32  tlen;
	uint8 *cp;
	uint8 *cp_end;
	struct rtl_mBuf *m = pkthdrPtr->ph_mbuf;
	int32     hlen,datalen;
	int32 copied = 0;
	int32 off=0;
	uint16 result_port=0;
	uint8 chksum_flag=0;
	uint32 mins=m->m_len;

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING)==0) return SUCCESS;	

	
	tcp = (struct tcphdr *)(struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));

	data = (uint8 *)tcp + ((tcp->th_off_x & 0xf0) >> 2);
	tlen = ntohs(pip->ip_len);
	
	



	if(mins>_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE) mins=_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE;
	
	cp = _rtl8651_l4_h323_alias_buf[0];
	copied = (int32)mBuf_copyToUserBuffer((struct rtl_mBuf *)m, (uint32)0, mins, (int8 *)(_rtl8651_l4_h323_alias_buf[0]));
	cp_end = cp + copied;		


	pip = (struct ip *)(&_rtl8651_l4_h323_alias_buf[0][pkthdrPtr->ph_iphdrOffset]);
	tcp = (struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));

	hlen = ((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2 ;	
	
	data = &_rtl8651_l4_h323_alias_buf[0][pkthdrPtr->ph_iphdrOffset+hlen];
	datalen = copied - hlen;

	
	while( find_port(link->insideLocalIpAddr, data, datalen, &off, &result_port) ==0 ) {
	
		
		
		chksum_flag = 1;
		data[off]=(link->insideGlobalIpAddr>>24)&0xff;
		data[off+1]=(link->insideGlobalIpAddr>>16)&0xff;
		data[off+2]=(link->insideGlobalIpAddr>>8)&0xff;
		data[off+3]=(link->insideGlobalIpAddr)&0xff;		
		
		data+=(6+off);
		datalen-=(6+off);		
		off=0;
	}	

	data = &_rtl8651_l4_h323_alias_buf[0][pkthdrPtr->ph_iphdrOffset+hlen];
	datalen = copied - hlen;
	while( find_port(link->insideGlobalIpAddr, data, datalen, &off, &result_port) ==0 ) {
	

		_rtl8651_addAlgQosNaptConnection(TRUE, TRUE|ALG_QOS|DO_REDUNDANCY_CHECK,UDP_FLOW, link->insideLocalIpAddr, result_port, &link->insideGlobalIpAddr, &result_port,0,0,RTL8651_ALG_NETMEETING_IDX);
		
		data+=(6+off);
		datalen-=(6+off);		
		off=0;
	}


#ifdef H225_TUNNEL_SUPPORT
	
	
	if( link->h245tunnel || 
		find_h245_id(cp,cp_end-cp) ){
		
		
		link->h245tunnel = 1;
		
		rtl8651_l4AliasHandleClientH245Out(pkthdrPtr,pip,link);
	}
#endif	


	if(chksum_flag){
		
		
		if((pip->ip_vhl & 0xf)!=5)
		{
			tcp->th_sum = 0;
			pip->ip_sum = 0;				
			tcp->th_sum = _rtl8651_tcpChecksum(pip);
			pip->ip_sum = _rtl8651_ipChecksum(pip);		
		}	
	}

	mBuf_copyToMbuf(m, 0,copied,(int8 *)_rtl8651_l4_h323_alias_buf[0]);

	
	RsvpLocalIp = link->insideLocalIpAddr;
	RsvpGlobalIp = link->insideGlobalIpAddr;

	return SUCCESS;
}


int32 rtl8651_l4AliasHandleClientH245In(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
				struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
#if 0
	int32 off=0, datlen;
	uint16 result_port;
	uint8 *data = NULL;
	struct tcphdr *tcp = NULL;
	int32     hlen;
	int portIdx=0;
	

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING)==0) return SUCCESS;	

	tcp = (struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));
	hlen = ((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2 ;		
	data = (uint8*)pip+hlen;
	datlen =ntohs(pip->ip_len)  - hlen;

	memset(lastPort,0,sizeof(uint16)*4);
	
	while( find_port(link->dstIpAddr, data, datlen, &off, &result_port) ==0 ) {

		if(ntohs(result_port)==1503)
		goto ClientH245In_break;
					

		lastPort[portIdx]=result_port;
		portIdx++;
		
		


ClientH245In_break:
		data+=(off+6);
		datlen-=(off+6);
		result_port = 0;
		off = 0;
	}
	
#endif
	return SUCCESS;

}


int32 rtl8651_l4AliasHandleClientH245Out(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
				struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
	struct in_addr ipaddr_org;
	int32 off=0, datlen;
	uint16 result_port;
	uint8 *data = NULL;
	struct tcphdr *tcp = NULL;
	int32     hlen;
	int32 chksum_flag = 0;
	int32 copied = 0;
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb = &entry;

	uint32 mins;
	int portIdx=0;
	struct rtl_mBuf *m;

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING)==0) return SUCCESS;	


	m  = pkthdrPtr->ph_mbuf;
	mins=m->m_len;
	if(mins>_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE) mins=_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE;
	copied = (int32)mBuf_copyToUserBuffer((struct rtl_mBuf *)m, (uint32)0, mins, (int8 *)(_rtl8651_l4_h323_alias_buf[0]));
	pip = (struct ip *)(&_rtl8651_l4_h323_alias_buf[0][pkthdrPtr->ph_iphdrOffset]);
	tcp = (struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));
	
	hlen = ((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2 ;	
	
	data = &_rtl8651_l4_h323_alias_buf[0][pkthdrPtr->ph_iphdrOffset+hlen];
	datlen = copied - hlen;
	
	ipaddr_org.s_addr  = link->insideLocalIpAddr;
			
	while( find_port(ipaddr_org.s_addr, data, datlen, &off, &result_port) ==0 ) {

		
		chksum_flag = 1;
		data[off]=(link->insideGlobalIpAddr>>24)&0xff;
		data[off+1]=(link->insideGlobalIpAddr>>16)&0xff;
		data[off+2]=(link->insideGlobalIpAddr>>8)&0xff;
		data[off+3]=(link->insideGlobalIpAddr)&0xff;	

		if ( result_port == 1503 )  goto end_aliash245Out;
	
		if(link->natlink == 0){

			if(_rtl8651_getOutboundNaptFlow(FALSE,link->insideLocalIpAddr,result_port,link->dstIpAddr,lastPort[portIdx],  &tb)==SUCCESS){
				_rtl8651_delNaptConnection(tb);							
			}

#if 0
			if( lastPort[portIdx])
				_rtl8651_addNaptConnection(TRUE, TRUE,UDP_FLOW, link->insideLocalIpAddr,result_port,  &link->insideGlobalIpAddr, &result_port,link->dstIpAddr, lastPort[portIdx]);
			else
#else
				
				_rtl8651_removeNaptConnectionByIntIPandPort(FALSE,link->insideLocalIpAddr,result_port);
#endif
				_rtl8651_addAlgQosNaptConnection(TRUE, TRUE|ALG_QOS,UDP_FLOW, link->insideLocalIpAddr, result_port, &link->insideGlobalIpAddr, &result_port,0,0,RTL8651_ALG_NETMEETING_IDX);
				

			
			
			portIdx++;
		}			
				
		
end_aliash245Out:		
		data+=(off+6);
		datlen-=(off+6);
		result_port = 0;
		off = 0;
	}


	if(chksum_flag){
		
		
		tcp->th_sum = 0;
		pip->ip_sum = 0;				
		tcp->th_sum = _rtl8651_tcpChecksum(pip);
		pip->ip_sum = _rtl8651_ipChecksum(pip);		
	}

	mBuf_copyToMbuf(m, 0,copied,(int8 *)_rtl8651_l4_h323_alias_buf[0]);
	
	return SUCCESS;

}

static int32 _rtl8651_l4RsvpNaptAliasIn(struct rtl_pktHdr *pkthdrPtr,struct ip * pip)
{
	int32 retval = FWDENG_RSVP_IGNORE;
	uint8 *data=NULL,*dp;
	uint32 dip,addr,offset;
	uint16 datlen,port;

	
	if ((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING) == 0 )
	{
		goto out;
	}

	if (RsvpLocalIp == 0)
	{
		goto out;
	}

	dip = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

	data = (uint8*)((uint8 *) pip + ((pip->ip_vhl &0xf) << 2));
	datlen = (uint32)data-(uint32)pip;
	dp = data;

	for (offset = 0; offset < datlen - 8; offset++, dp++)
	{
		addr = (dp[0] << 24) | (dp[1] << 16) | (dp[2] << 8) | dp[3];
		if (dip == addr)
		{
			port = (uint16)((*(dp + 6) << 8) | *(dp + 7));
			if (	_rtl8651_addAlgQosNaptConnection(	TRUE, 
													TRUE|ALG_QOS|DO_REDUNDANCY_CHECK,
													UDP_FLOW,
													RsvpLocalIp, port,
													&addr, &port,
													0,0,
													RTL8651_ALG_NETMEETING_IDX) == SUCCESS)
			{
				PKTSET_UINT32_UNALIGNED(RsvpLocalIp, &(pip->ip_dst));
				retval = SUCCESS;
				
			}
			goto out;
		}
	}	

out:
	return retval;
}


static int32 _rtl8651_l4RsvpNaptAliasOut(struct rtl_pktHdr *phdr,struct ip * pip)
{

	if ((alg_supports[phdr->ph_dsid]&RTL8651_ALG_NETMEETING) == 0 )
	{
		return FWDENG_RSVP_IGNORE;
	}

	if (RsvpGlobalIp != 0)
	{
		PKTSET_UINT32_UNALIGNED(RsvpGlobalIp, &pip->ip_src);
		return SUCCESS;
	}

	return FWDENG_RSVP_IGNORE;	
}

void rtl8651_l4AliasH323ReInit(void)
{
	
	tempPort = 0;
	RsvpLocalIp = 0;
	RsvpGlobalIp = 0;

	rtl8651_l4RsvpNaptAliasIn = _rtl8651_l4RsvpNaptAliasIn;
	rtl8651_l4RsvpNaptAliasOut = _rtl8651_l4RsvpNaptAliasOut;

}

int32 rtl8651_l4AliasH323Init(void *arg)
{
	
	
	
	if (init_h323 == TRUE)
	{
		goto init_para;
		return SUCCESS;
	}
		
	init_h323 = TRUE;
	_rtl8651_l4_h323_alias_buf[0] = (uint8 *)rtlglue_malloc(_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE+64);
	_rtl8651_l4_h323_alias_buf[1] = (uint8 *)rtlglue_malloc(_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE+64);
	assert(_rtl8651_l4_h323_alias_buf[0]);
	assert(_rtl8651_l4_h323_alias_buf[1]);

	memset(_rtl8651_l4_h323_alias_buf[0], 0, sizeof(_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE+64));
	memset(_rtl8651_l4_h323_alias_buf[1], 0, sizeof(_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE+64));


init_para:

	rtl8651_l4AliasH323ReInit();	

	return SUCCESS;
}

int32 rtl8651_l4AliasHandleServerH225RASIn (
					struct rtl_pktHdr* pkthdrPtr,
					struct ip *pip,
					rtl8651_tblDrv_naptTcpUdpFlowEntry_t *link)
{
	if ( (alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING) == 0 )
	{
		goto out;
	}

	

out:
	return SUCCESS;
}

int32 rtl8651_l4AliasHandleServerH225RASOut (
					struct rtl_pktHdr* pkthdrPtr,
					struct ip *pip,
					rtl8651_tblDrv_naptTcpUdpFlowEntry_t *link)
{
	if ( (alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING) == 0 )
	{
		goto out;
	}

	

out:
	return SUCCESS;
}



int32 rtl8651_l4AliasHandleServerH225CSOut(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
	
	struct tcphdr *tcp;
	int32  tlen,hlen,dlen;
	struct rtl_mBuf *m = pkthdrPtr->ph_mbuf;
	int32 copied = 0;
	int  off;
	uint16 result_port;
	uint32 mins;
	uint8 *data=NULL;
	uint32 datalen=0;
	uint8 chksum_flag=0;

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING)==0) return SUCCESS;	
	
	tcp = (struct tcphdr *)(struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));
	hlen =((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2;
	tlen = ntohs(pip->ip_len);
	dlen = tlen - hlen;		


	mins=m->m_len;
	if(mins>_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE) mins=_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE;
	
	copied = mBuf_copyToUserBuffer(m, 0,
		mins, 
		(int8 *)_rtl8651_l4_h323_alias_buf[1]);


	pip = (struct ip *)(&_rtl8651_l4_h323_alias_buf[1][pkthdrPtr->ph_iphdrOffset]);
	tcp = (struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));
	hlen = ((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2 ;	
		
	data = &_rtl8651_l4_h323_alias_buf[1][pkthdrPtr->ph_iphdrOffset+hlen];
	datalen = copied - hlen;

	while(find_port(link->insideLocalIpAddr, data, datalen, &off, &result_port) ==0)
	{
		data[off]=(link->insideGlobalIpAddr>>24)&0xff;
		data[off+1]=(link->insideGlobalIpAddr>>16)&0xff;
		data[off+2]=(link->insideGlobalIpAddr>>8)&0xff;
		data[off+3]=(link->insideGlobalIpAddr)&0xff;				

		tempPort=result_port;				

#ifdef _RTL_UPNP
		rtl8651_addUpnpMap(UPNP_ONESHOT|UPNP_TCP , link->dstIpAddr, 0, 
			link->insideGlobalIpAddr, 17200,
			link->insideLocalIpAddr, result_port);
#endif 

		
		result_port=17200;
		data[off+4]=(result_port>>8)&0xff;
		data[off+5]=(result_port)&0xff;
		chksum_flag=1;		
		data+=(off+6);
		datalen-=(off+6);
		off = 0;
	}		


	data = &_rtl8651_l4_h323_alias_buf[1][pkthdrPtr->ph_iphdrOffset+hlen];
	datalen = copied - hlen;

	while(find_port(link->insideGlobalIpAddr, data, datalen, &off, &result_port) ==0)
	{

		_rtl8651_addAlgQosNaptConnection(TRUE, TRUE|ALG_QOS|DO_REDUNDANCY_CHECK,UDP_FLOW, link->insideLocalIpAddr, result_port, &link->insideGlobalIpAddr, &result_port,0,0,RTL8651_ALG_NETMEETING_IDX);
		data+=(off+6);
		datalen-=(off+6);
		off = 0;
	}


	if(chksum_flag){
		
		
		if((pip->ip_vhl & 0xf)!=5)
		{
			tcp->th_sum = 0;
			pip->ip_sum = 0;				
			tcp->th_sum = _rtl8651_tcpChecksum(pip);
			pip->ip_sum = _rtl8651_ipChecksum(pip);		
		}		
	}

	mBuf_copyToMbuf(m, 0,copied,(int8 *)_rtl8651_l4_h323_alias_buf[1]);
	
	return SUCCESS;

}

int32 rtl8651_l4AliasHandleServerH225CSIn(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{

	
	struct tcphdr *tcp;
	int32  tlen,hlen,dlen;
	struct rtl_mBuf *m = pkthdrPtr->ph_mbuf;
	int32 copied = 0;
	int  off;
	uint16 result_port;
	uint32 mins;
	uint8 chksum_flag=0;
	uint8 *data;
	uint32 datalen=0;


	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING)==0) return SUCCESS;	
	
	tcp = (struct tcphdr *)(struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));
	hlen =((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2;
	tlen = ntohs(pip->ip_len);
	dlen = tlen - hlen;		
	

	mins=m->m_len;
	if(mins>_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE) mins=_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE;

	copied = mBuf_copyToUserBuffer(m, 0,
		mins, 
		(int8 *)_rtl8651_l4_h323_alias_buf[1]);

	pip = (struct ip *)(&_rtl8651_l4_h323_alias_buf[1][pkthdrPtr->ph_iphdrOffset]);
	tcp = (struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));
	hlen = ((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2 ;	
	
	data = &_rtl8651_l4_h323_alias_buf[1][pkthdrPtr->ph_iphdrOffset+hlen];
	datalen = copied - hlen;
	

	while(find_port(link->insideGlobalIpAddr, data, datalen, &off, &result_port) ==0)
	{
		data[off]=(link->insideLocalIpAddr>>24)&0xff;
		data[off+1]=(link->insideLocalIpAddr>>16)&0xff;
		data[off+2]=(link->insideLocalIpAddr>>8)&0xff;
		data[off+3]=(link->insideLocalIpAddr)&0xff;	


		
		chksum_flag=1;
		data+=(6+off);
		datalen-=(6+off);			
		off=0;
	}


	if(chksum_flag){
		
		
		if((pip->ip_vhl & 0xf)!=5)
		{
			tcp->th_sum = 0;
			pip->ip_sum = 0;				
			tcp->th_sum = _rtl8651_tcpChecksum(pip);
			pip->ip_sum = _rtl8651_ipChecksum(pip);		
		}
	}

	mBuf_copyToMbuf(m, 0,copied,(int8 *)_rtl8651_l4_h323_alias_buf[1]);

	
	return SUCCESS;
}


int32 rtl8651_l4AliasHandleServerH245Out(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
				struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{

	
	struct tcphdr *tcp;
	int32  tlen,hlen,dlen;
	struct rtl_mBuf *m = pkthdrPtr->ph_mbuf;
	int32 copied = 0;
	int  off;
	uint16 result_port;
	int offset=0;
	int mins;
	uint8 *data;
	int portIdx=0;

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING)==0) return SUCCESS;	

	tcp = (struct tcphdr *)(struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));
	hlen =((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2;
	tlen = ntohs(pip->ip_len);
	dlen = tlen - hlen;		


	mins=m->m_len;
	if(mins>_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE) mins=_RTL8651_L4_MAX_H323_ALIAS_BUF_SIZE;
	
	copied = mBuf_copyToUserBuffer(m, 0,
		mins, 
		(int8 *)_rtl8651_l4_h323_alias_buf[1]);

	data = &_rtl8651_l4_h323_alias_buf[1][pkthdrPtr->ph_iphdrOffset+hlen];

	while(find_port(link->insideLocalIpAddr, &data[offset], mins-pkthdrPtr->ph_iphdrOffset-hlen-offset, &off, &result_port) ==0)
	{
		struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb = &entry;
		data[off+offset]=(link->insideGlobalIpAddr>>24)&0xff;
		data[1+off+offset]=(link->insideGlobalIpAddr>>16)&0xff;
		data[2+off+offset]=(link->insideGlobalIpAddr>>8)&0xff;
		data[3+off+offset]=(link->insideGlobalIpAddr)&0xff;



		if(result_port==1503) goto H245Out_break;

		if(_rtl8651_getOutboundNaptFlow(FALSE,link->insideLocalIpAddr,result_port,link->dstIpAddr,lastPort[portIdx],  &tb)==SUCCESS){
			_rtl8651_delNaptConnection(tb);							
		}

#if 0
		if(lastPort[portIdx])
			_rtl8651_addNaptConnection(TRUE, TRUE,UDP_FLOW, link->insideLocalIpAddr, result_port, &link->insideGlobalIpAddr, &result_port,link->dstIpAddr, lastPort[portIdx]);
		else
#else
			
			_rtl8651_removeNaptConnectionByIntIPandPort(FALSE,link->insideLocalIpAddr,result_port);			
			
			
#endif
			_rtl8651_addAlgQosNaptConnection(TRUE, TRUE|ALG_QOS|DO_REDUNDANCY_CHECK,UDP_FLOW, link->insideLocalIpAddr, result_port, &link->insideGlobalIpAddr, &result_port,0,0,RTL8651_ALG_NETMEETING_IDX);

				

		
		portIdx++;

H245Out_break:

		mBuf_copyToMbuf(m, 0,copied,(int8 *)_rtl8651_l4_h323_alias_buf[1]);
		offset+=(6+off);
	}			
	return SUCCESS;


}


int32 rtl8651_l4AliasHandleServerH245In(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
				struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
#if 0
	int32 off=0, datlen;
	uint16 result_port;
	uint8 *data = NULL;
	struct tcphdr *tcp = NULL;
	int32     hlen;
	int portIdx=0;

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_NETMEETING)==0) return SUCCESS;	

	tcp = (struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));
	hlen = ((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2 ;		
	data = (uint8*)pip+hlen;
	datlen =ntohs(pip->ip_len)  - hlen;

	memset(lastPort,0,sizeof(uint16)*4);
			
	while( find_port(link->dstIpAddr, data, datlen, &off, &result_port) ==0 ) {

		if(ntohs(result_port)==1503)
		goto ServerH245In_break;
					
#ifdef _RTL_UPNP

		
		if(link->natlink == 0){
			
			lastPort[portIdx]=result_port;
			portIdx++;
			
			
		}
#endif 

ServerH245In_break:
		data+=(off+6);
		datlen-=(off+6);
		result_port = 0;
		off = 0;
	}

#endif	
	return SUCCESS;

}



