

#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_ipsec.h"


#define IPSEC_IDLE_TIME 600 

#define  IPSEC_FREE 0
#define  IPSEC_USED 1

struct isakmp_data_s {
	uint64	icookie;	
	uint64	rcookie;	
	
	uint32 peer_ip;
	uint32 local_ip;
	uint32 alias_ip;

	uint32	ospi;
	uint32	ispi;

	uint16 state;
	uint16 idle_timer;

	uint32 dsid;
};

static struct isakmp_data_s *_rtl8651_isakmpDb=NULL;



static struct isakmp_data_s* 
_rtl8651_findIsakmpIn(uint32 peer_ip,uint32 alias_ip,uint64 icookie,uint64 rcookie){ 
	int i;
	struct isakmp_data_s *tp;
	tp = _rtl8651_isakmpDb;
	
	for(i = 0 ; i < algMaxSession.ipsecMaxSession ; i++){
		if(tp->peer_ip == peer_ip &&
		   tp->alias_ip == alias_ip &&
		   ((tp->icookie == icookie )||(tp->icookie==0)||(icookie==0)) &&
		   ((tp->rcookie == rcookie)||(tp->rcookie==0)||(rcookie==0)) && 
		   tp->state == IPSEC_USED){	
		   	tp->idle_timer = 0;	 
		   	tp->rcookie = rcookie;  	
		   	tp->icookie = icookie;  
		   	return tp;
		}
		tp++;
	}
	
	return NULL;
}



static struct isakmp_data_s* 
_rtl8651_findIsakmpOut(uint32 local_ip,uint32 peer_ip,uint64 icookie,uint64 rcookie){ 
	int i;
	struct isakmp_data_s *tp;
	tp = _rtl8651_isakmpDb;
	
	for(i = 0 ; i < algMaxSession.ipsecMaxSession ; i++){
		if(tp->peer_ip == peer_ip &&
		   tp->local_ip == local_ip &&
		   ((tp->icookie== icookie)||(icookie==0) ||(tp->icookie==0)) &&
		   ((tp->rcookie== rcookie)||(rcookie==0) ||(tp->rcookie==0)) &&

		   tp->state == IPSEC_USED){	
		   	tp->idle_timer = 0;	   
		   	tp->rcookie = rcookie;  	
		   	tp->icookie = icookie;  		   		
		   	return tp;
		}
		tp++;
	}
	
	return NULL;
}




static struct isakmp_data_s* 
_rtl8651_addIsakmp(uint32 local_ip,uint32 peer_ip,uint32 alias_ip, 
		uint64 icookie,uint64 rcookie,uint32 dsid){ 
	int i;
	struct isakmp_data_s *tp;
	tp = _rtl8651_isakmpDb;
	
	for(i = 0 ; i < algMaxSession.ipsecMaxSession ; i++){
		if(tp->state == IPSEC_FREE){
		    memset(tp,0,sizeof(struct isakmp_data_s));
		    tp->idle_timer = 0;
		    tp->peer_ip = peer_ip;
		    tp->local_ip = local_ip;
		    tp->alias_ip = alias_ip;
		    tp->icookie = icookie;
		    tp->rcookie = rcookie;
		    tp->state = IPSEC_USED;
		    tp->dsid = dsid;
		   	return tp;
		}
		tp++;
	}
	return NULL;
}


static struct isakmp_data_s* 
_rtl8651_findEspIn(uint32 peer_ip,uint32 alias_ip,uint32 ispi){ 
	int i;
	struct isakmp_data_s *tp;
	tp = _rtl8651_isakmpDb;
	
	for(i = 0 ; i < algMaxSession.ipsecMaxSession ; i++){
		if(tp->peer_ip == peer_ip &&
		   tp->alias_ip == alias_ip &&
		   tp->state == IPSEC_USED &&
		   ((tp->ispi == ispi) || (tp->ispi == 0 ) || (ispi==0)) ){	
		   	tp->idle_timer = 0;	   	
		   	tp->ispi = ispi;  
		   	return tp;
		}
		tp++;
	}
	
	return NULL;
}



static struct isakmp_data_s* 
_rtl8651_findEspOut(uint32 local_ip,uint32 peer_ip, uint32 ospi){ 
	int i;
	struct isakmp_data_s *tp;
	tp = _rtl8651_isakmpDb;
	
	for(i = 0 ; i < algMaxSession.ipsecMaxSession ; i++){
		if(tp->peer_ip == peer_ip &&
		   tp->local_ip == local_ip &&
		   ((tp->ospi ==  ospi )||(ospi==0)||(tp->ospi==0)) &&
		   tp->state == IPSEC_USED ){	
		   	tp->idle_timer = 0;	  
		   	tp->ospi = ospi;   	
		   	return tp;
		}
		tp++;
	}

	return NULL;
}




static struct isakmp_data_s* 
_rtl8651_addEsp(uint32 local_ip,uint32 peer_ip,uint32 ospi,uint32 ispi){ 

	struct isakmp_data_s *tp;	
	int i;


	tp = _rtl8651_isakmpDb;

	for(i = 0 ; i < algMaxSession.ipsecMaxSession ; i++){
		if(tp->state == IPSEC_USED && tp->peer_ip == peer_ip
		   && tp->local_ip == local_ip){
			tp->idle_timer = 0;
		    tp->ospi = ospi;
		    tp->ispi = ispi;
		   	return tp;
		}
		tp++;
	}

	
	{
		ipaddr_t extip;
		extip=_rtl8651_getExtIpAddrByDsid(0);
		if(extip)
		{
			tp = _rtl8651_addIsakmp(local_ip,peer_ip,extip,0,0,0);
			return tp;
		}
	}	
	return NULL;
}



static int32 _rtl8651_l4EspNaptAliasIn(struct rtl_pktHdr *phdr,struct ip * pip)
{
	uint32	i_spi;
	struct isakmp_data_s *link;
	if((alg_supports[phdr->ph_dsid]&(RTL8651_ALG_IPSEC|RTL8651_ALG_IPSEC_SERVER))==0) return SUCCESS;

	i_spi = PKTGET_UINT32_UNALIGNED((uint32 *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2)));

	link = _rtl8651_findEspIn(PKTGET_UINT32_UNALIGNED(&pip->ip_src), PKTGET_UINT32_UNALIGNED(&pip->ip_dst), i_spi);

	if(link == NULL){
		if (alg_serverips[0][RTL8651_ALG_IPSEC_IDX] == 0)	
		{
			return FWDENG_ESP_IGNORE;
		}
		
		link = _rtl8651_addEsp(alg_serverips[0][RTL8651_ALG_IPSEC_IDX], PKTGET_UINT32_UNALIGNED(&pip->ip_src),0, i_spi);
	}
	
	if (link != NULL)
	{			
		uint8 *dst;
		dst=(uint8 *)&pip->ip_dst;
		dst[0]=(link->local_ip>>24)&0xff;
		dst[1]=(link->local_ip>>16)&0xff;
		dst[2]=(link->local_ip>>8)&0xff;
		dst[3]=link->local_ip&0xff;		
		phdr->ph_dip = link->local_ip;

		


		return SUCCESS;		
	}
	
	return FWDENG_ESP_IGNORE;
}



static int32 _rtl8651_l4EspNaptAliasOut(struct rtl_pktHdr *phdr,struct ip *pip)
{
	
	uint32	o_spi;
	struct isakmp_data_s *link;
	if((alg_supports[phdr->ph_dsid]&(RTL8651_ALG_IPSEC|RTL8651_ALG_IPSEC_SERVER))==0) return SUCCESS;
	o_spi = PKTGET_UINT32_UNALIGNED((uint32 *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2)));

	link = _rtl8651_findEspOut(PKTGET_UINT32_UNALIGNED(&pip->ip_src), PKTGET_UINT32_UNALIGNED(&pip->ip_dst), o_spi);

	if(link == NULL){
		link = _rtl8651_addEsp(PKTGET_UINT32_UNALIGNED(&pip->ip_src), PKTGET_UINT32_UNALIGNED(&pip->ip_dst), o_spi,0);
	}
	
	if (link != NULL)
	{
		uint8 *ptr;
				 
	
		ptr=(uint8 *)&pip->ip_src;
		ptr[0]=(link->alias_ip>>24)&0xff;
		ptr[1]=(link->alias_ip>>16)&0xff;
		ptr[2]=(link->alias_ip>>8)&0xff;
		ptr[3]=(link->alias_ip)&0xff;
		phdr->ph_sip = link->alias_ip;
	

		


		return SUCCESS;	
	}	
	return FWDENG_ESP_IGNORE;
}


int32 rtl8651_l4AliasHandleIpsecClientInbound(struct rtl_pktHdr *pkt, struct ip *pip, 
					   struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s * link)	
{
	struct isakmp_data_s *tb;	
	struct 	udphdr *ud;
	uint64	*dptr, icookie, rcookie;	
	char compstr[12]={0};

	if((alg_supports[pkt->ph_dsid]&(RTL8651_ALG_IPSEC|RTL8651_ALG_IPSEC_SERVER))==0) return SUCCESS;
		
	ud = (struct udphdr *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2));

	
	if (ud->uh_ulen < 16) {
		return SUCCESS;
	}
	dptr = (uint64 *) (ud+1);
	rtl8651_memcpy(&icookie,dptr,8);
	dptr++;
	rtl8651_memcpy(&rcookie,dptr,8);

	if(memcmp((uint8*)dptr,compstr,12)!=0)
	{
	
		tb = _rtl8651_findIsakmpIn(link->dstIpAddr,link->insideGlobalIpAddr,icookie,rcookie);
		
		if(tb == NULL){

			if((alg_supports[pkt->ph_dsid]&RTL8651_ALG_IPSEC_SERVER)==0) return FWDENG_USER_DROP;
			
#ifdef _RTL_UPNP
			
			rtl8651_addUpnpMap(UPNP_PERSIST, 0, 0, 
					link->insideGlobalIpAddr, 4500,
					link->insideLocalIpAddr, 4500);	
#endif 

			tb = _rtl8651_addIsakmp(alg_serverips[pkt->ph_dsid][RTL8651_ALG_IPSEC_IDX],link->dstIpAddr,
									link->insideGlobalIpAddr,icookie,rcookie,pkt->ph_dsid);
			if(tb == NULL){			
				return SUCCESS;
			}		
	       }
		
		link->insideLocalIpAddr = tb->local_ip;	
	}

	return SUCCESS;
}




int32 rtl8651_l4AliasHandleIpsecClientOutbound(
		struct rtl_pktHdr *pkt,
		struct ip *pip,
		struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
	struct isakmp_data_s *tb;	
	struct 	udphdr *ud;
	uint64	*dptr, icookie, rcookie;	
	char compstr[12]={0};

	if((alg_supports[pkt->ph_dsid]&(RTL8651_ALG_IPSEC|RTL8651_ALG_IPSEC_SERVER))==0) return SUCCESS;
	
	ud = (struct udphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));

		
	if ( ud->uh_ulen < 16) {	
		return SUCCESS;
	}

	dptr = (uint64 *) (ud+1);
	rtl8651_memcpy(&icookie,dptr,8);
	dptr++;
	rtl8651_memcpy(&rcookie,dptr,8);

	if(memcmp((uint8*)dptr,compstr,12)!=0)
	{
	
		tb = _rtl8651_findIsakmpOut(link->insideLocalIpAddr,link->dstIpAddr,icookie,rcookie);
		if(tb == NULL){
			
			if((alg_supports[pkt->ph_dsid]&RTL8651_ALG_IPSEC)==0) return FWDENG_USER_DROP;
			
			tb = _rtl8651_addIsakmp(link->insideLocalIpAddr,link->dstIpAddr,
									link->insideGlobalIpAddr,icookie,rcookie,pkt->ph_dsid);
			if(tb == NULL){	
				return SUCCESS;
			}		
	       }

		if(link->insideLocalPort==500)
		{
			
			if(link->insideGlobalPort!=500)
			{
#ifdef _RTL_UPNP			
		    	rtl8651_addUpnpMap(UPNP_ONESHOT, link->dstIpAddr, 500, 
					link->insideGlobalIpAddr, 500,link->insideLocalIpAddr, 500);		
#endif 	
				link->insideGlobalPort = 500;
			}
		}		
	}
	return SUCCESS;
}





int32 rtl8651_l4AliasIpsecInit(void *arg){
	
	
	
	if(_rtl8651_isakmpDb) rtlglue_free(_rtl8651_isakmpDb);
	_rtl8651_isakmpDb = (struct isakmp_data_s *)rtlglue_malloc(algMaxSession.ipsecMaxSession*(sizeof(struct isakmp_data_s)));
	memset(_rtl8651_isakmpDb,0,algMaxSession.ipsecMaxSession*(sizeof(struct isakmp_data_s)));	
	rtl8651_l4EspNaptAliasIn = _rtl8651_l4EspNaptAliasIn;
	rtl8651_l4EspNaptAliasOut = _rtl8651_l4EspNaptAliasOut;
	return SUCCESS;
}


void rtl8651_l4AliasIpsecReInit(void){
	
	if(_rtl8651_isakmpDb) memset(_rtl8651_isakmpDb,0,algMaxSession.ipsecMaxSession*(sizeof(struct isakmp_data_s)));	
	rtl8651_l4EspNaptAliasIn = _rtl8651_l4EspNaptAliasIn;
	rtl8651_l4EspNaptAliasOut = _rtl8651_l4EspNaptAliasOut;	
}



void 
rtl8651_l4AliasIpsecHouseKeeping(void){ 
	int i,j;
	struct isakmp_data_s *tp;
	tp = _rtl8651_isakmpDb;
	
	
	for(i = 0 ; i < algMaxSession.ipsecMaxSession ; i++){
		
		if(tp == NULL || _rtl8651_isakmpDb == NULL)
				return;
		if(tp->state == IPSEC_FREE)
			continue;

		
		for(j=0;j<RTL8651_MAX_DIALSESSION;j++)
			if((alg_supports[j]&RTL8651_ALG_IPSEC)==0)
			{
				if(tp->dsid==j) tp->state = IPSEC_FREE;
			}
			
		tp->idle_timer++;
		if(tp->idle_timer > IPSEC_IDLE_TIME ){	
			tp->state = IPSEC_FREE;
		}
		tp++;
	}	
}



