


#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_mms.h"

#define _RTL8651_L4_MAX_MMS_ALIAS_BUF_SIZE 576

static uint8 *_rtl8651_l4_mms_alias_buf[2]; 

#define IS_DIGIT(c)    (c >= '0' && c <= '9')
#define IS_DIGITDOT(c) ((c >= '0' && c <= '9') || c == '.')
#define IS_ALPHA(c)    ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z'))
#define IS_WSPACE(c)   (c == ' ' || c == '\t' || c == '\r' || c == '\n')
#define SKIP_WSPACE(p,n) while( n > 0 && IS_WSPACE(*p) ) { n--; p++; }




static int32
str_to_u16( int8* pbuf, uint16* pval )
{
    int32 n = 0;

    *pval = 0;
    while( pbuf[n] >= '0' && pbuf[n] <= '9' )
    {
        *pval = (*pval * 10) + (pbuf[n] - '0');
        n++;
    }

    return n;
}

static int32
str_to_u32( int8* pbuf, uint32* pval )
{
    int32 n = 0;

    *pval = 0;
    while( pbuf[n] >= '0' && pbuf[n] <= '9' )
    {
        *pval = (*pval * 10) + (pbuf[n] - '0');
        n++;
    }

    return n;
}

static int32
next_line( int8** ppbuf, uint32* pbuflen, int8** ppline, uint32* plinelen )
{
    int8* pbuf = *ppbuf;
    uint32 buflen = *pbuflen;
    uint32 linelen = 0;

    do
    {
        while( *pbuf != '\r' && *pbuf != '\n' )
        {
            if( buflen <= 1 )
            {
                return 0;
            }

            pbuf++;
            linelen++;
            buflen--;
        }

        if( buflen > 1 && *pbuf == '\r' && *(pbuf+1) == '\n' )
        {
            pbuf++;
            buflen--;
        }

        pbuf++;
        buflen--;
    }
    while( buflen > 0 && (*pbuf == ' ' || *pbuf == '\t') );

    *ppline = *ppbuf;
    *plinelen = linelen;
    *ppbuf = pbuf;
    *pbuflen = buflen;

    return linelen;
}


static  struct tcphdr*
get_mbuf_data( struct ip *iph, struct rtl_mBuf** ppm, int8** ppdata, uint32* plen )
{
    struct tcphdr*  th;
    struct rtl_mBuf *pm;
    int8* pbuf;
    int8* pend;
    
    pm = *ppm;
    mBuf_copyToUserBuffer(pm,0,mBuf_getPktlen(pm),(int8 *)_rtl8651_l4_mms_alias_buf[0]);
    
    th = (struct tcphdr*)&(_rtl8651_l4_mms_alias_buf[0][(iph->ip_vhl & 0x0f)<<2]);

    pbuf = (int8*)th + (((th->th_off_x>>4) & 0xf) << 2);
    pend = (int8 *)(pm->m_data) + pm->m_len;    
    *ppdata = pbuf;
    *plen = pend - pbuf;
    return th;
}



int32 rtl8651_l4AliasHandleMmsCleintOutbound(
	struct rtl_pktHdr *indat, 
	struct ip *pip,
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t *tb	
)
{


    int32   diff     = 0;
    int8* mbuf_data_p  = NULL;
    uint32 mbuf_len   = 0;

    int8* pdata    = NULL;
    uint32   datalen  = 0;

    int8* pline;
    uint32 linelen;
    int8  buf[32];
    int32   buflen;

    uint32 entitylen = 0;
    int32 isinvite = 0;
    int32 accepted = 0;
    int32 addrpos = 0;
    int32 addrlen = 0;
    int32 portpos = 0;
    int32 portlen = 0;
    uint16 s_port;
 
    struct in_addr t_in_addr ;
    struct tcphdr *tc;
    struct rtl_mBuf *m;
    
    ipaddr_t ip_s,ip_d = 0;	
		
    ip_s = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
    ip_d = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);    
        
    m = indat->ph_mbuf;
    
  


   
    tc = get_mbuf_data(pip, &m, &mbuf_data_p, &mbuf_len );
    pip = (struct ip*)_rtl8651_l4_mms_alias_buf[0];
    pdata = mbuf_data_p;
    datalen = mbuf_len;

    if( !next_line( &pdata, &datalen, &pline, &linelen ) )
    {
        return FAILED;
    }

    
    if( strncmp( pline, "MSG ", 4 ) != 0 ) 
    	return FAILED;
    
    pline += 4;
    
    if( ! IS_DIGIT(*pline) ) 
    	return FAILED;
    
    while( IS_DIGIT(*pline) ) 
    	pline++;
    
    if( *pline != ' ' ) 
    	return FAILED;
    
    pline++;
    
    if( ! IS_ALPHA(*pline) ) 
    	return FAILED;
    
    pline++;
    
    if( *pline != ' ' ) 
    	return FAILED;
    
    pline++;
    
    if( ! IS_DIGIT(*pline) ) 
    	return FAILED;
    
    pline += str_to_u32( pline, &entitylen );

    
    while( next_line( &pdata, &datalen, &pline, &linelen ) )
    {
        if( linelen == 0 )
        {
            break;
        }
        if( memcmp( pline, "Content-Type:", 13 ) == 0 )
        {
            pline += 13;
            while( *pline == ' ' ) 
            	pline++;
            
            if( memcmp( pline, "text/x-msmsgsinvite;", 20 ) == 0 )
                isinvite = 1;
        }
    }
    
    if( !isinvite ) 
    	return FAILED;

    
    while( next_line( &pdata, &datalen, &pline, &linelen ) )
    {
        if( memcmp( pline, "Invitation-Command:", 19 ) == 0 )
        {
            pline += 19;
            while( *pline == ' ' ) pline++;
            if( memcmp( pline, "ACCEPT", 6 ) == 0 )
                accepted = 1;
        }
        if( memcmp( pline, "IP-Address:", 11 ) == 0 )
        {
            pline += 11;
            while( *pline == ' ' ) pline++;
            addrpos = pline - mbuf_data_p;
            addrlen = 0;
            while( IS_DIGITDOT(pline[addrlen]) ) addrlen++;
        }
        if( memcmp( pline, "Port:", 5 ) == 0 )
        {
            pline += 5;
            while( *pline == ' ' ) pline++;
            portpos = pline - mbuf_data_p;
            portlen = 0;
            while( IS_DIGIT(pline[portlen]) ) portlen++;
        }
    }

    if( !accepted || !addrpos || !portpos ) 
    	return FAILED;

    str_to_u16( mbuf_data_p+portpos, &s_port );


#ifdef _RTL_UPNP
		
    if(tb->natlink == 0){
    	rtl8651_addUpnpMap(UPNP_TCP|UPNP_ONESHOT, ip_d, 0, 
			tb->insideGlobalIpAddr, s_port,
			ip_s, s_port);		
    }
#endif     
    
    
    
    
    {    	
    	uint8 *cp ;
    	t_in_addr.s_addr = tb->insideGlobalIpAddr ;
    	NTOHL(t_in_addr.s_addr);
   		cp = (uint8 *)(&t_in_addr);
    	
	    buflen = sprintf( buf, "%d.%d.%d.%d", cp[0],cp[1],cp[2],cp[3]);
    }
    
    if( buflen == addrlen )
    {
        rtl8651_memcpy( mbuf_data_p+addrpos, buf, buflen );
    }
    else
    {
        get_mbuf_data( pip, &m, &mbuf_data_p, &mbuf_len );
        diff += buflen - addrlen;
        if( portpos > addrpos )
        {
            portpos += buflen - addrlen;
        }
    }

    
    
    buflen = sprintf( buf, "%hu", ntohs(s_port) );
    
    if( buflen == portlen )
    {
        rtl8651_memcpy( mbuf_data_p+portpos, buf, buflen );
    }
    else
    {
		
		
			int32     delta;

            
            
			tb->tcpAckModified = 1;
			delta = _rtl8651_l4GetDeltaSeqOut(pip, tc, tb);
			diff += buflen - portlen;
			_rtl8651_l4AddSeq(pip, tb, delta + diff);
                        
    }

    	
	{
		uint16    new_len;
		
		new_len = ntohs(pip->ip_len)+ diff;
		HTONS(new_len);
		pip->ip_len = new_len;
		
		
		tc->th_sum = 0;
		pip->ip_sum = 0;				
		tc->th_sum = _rtl8651_tcpChecksum(pip);
		pip->ip_sum = _rtl8651_ipChecksum(pip);			

		mBuf_copyToMbuf(m, 0, new_len, (int8 *)_rtl8651_l4_mms_alias_buf[0]);
	}

    return SUCCESS;
}


static int32 init_mms = 0;
int32 rtl8651_l4AliasMmsInit(void *arg){
	
	if(init_mms)
		return SUCCESS;
		
	init_mms = 1;	
	
	_rtl8651_l4_mms_alias_buf[0] = (uint8 *)rtlglue_malloc(_RTL8651_L4_MAX_MMS_ALIAS_BUF_SIZE+64);
	_rtl8651_l4_mms_alias_buf[1] = (uint8 *)rtlglue_malloc(_RTL8651_L4_MAX_MMS_ALIAS_BUF_SIZE+64);
	assert(_rtl8651_l4_mms_alias_buf[0]);
	assert(_rtl8651_l4_mms_alias_buf[1]);

	return SUCCESS;
}

