


#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_nbt.h"


typedef struct
{
	struct in_addr oldaddr;
	uint16    oldport;
	struct in_addr newaddr;
	uint16    newport;
	uint16   *uh_sum;
	struct ip *pip;
} NBTArguments;

typedef struct
{
	uint8     type;
	uint8     flags;
	uint16    id;
	struct in_addr source_ip;
	uint16    source_port;
	uint16    len;
	uint16    offset;
} NbtDataHeader;

#define OpQuery		0
#define OpUnknown	4
#define OpRegist	5
#define OpRelease	6
#define OpWACK		7
#define OpRefresh	8
typedef struct
{
	uint16    nametrid;
	uint16    dir:1, opcode:4, nmflags:7, rcode:4;
	uint16    qdcount;
	uint16    ancount;
	uint16    nscount;
	uint16    arcount;
}
NbtNSHeader;

#define FMT_ERR		0x1
#define SRV_ERR		0x2
#define IMP_ERR		0x4
#define RFS_ERR		0x5
#define ACT_ERR		0x6
#define CFT_ERR		0x7


#define DGM_DIRECT_UNIQ		0x10
#define DGM_DIRECT_GROUP	0x11
#define DGM_BROADCAST		0x12
#define DGM_ERROR			0x13
#define DGM_QUERY			0x14
#define DGM_POSITIVE_RES	0x15
#define DGM_NEGATIVE_RES	0x16



#define QS_TYPE_NB		0x0020
#define QS_TYPE_NBSTAT	0x0021
#define QS_CLAS_IN		0x0001
typedef struct
{
	uint16    type;					   
	uint16    class;				   
}
NBTNsQuestion;



#define RR_TYPE_A		0x0001
#define RR_TYPE_NS		0x0002
#define RR_TYPE_NULL	0x000a
#define RR_TYPE_NB		0x0020
#define RR_TYPE_NBSTAT	0x0021
#define RR_CLAS_IN		0x0001
#define SizeOfNsResource	8
typedef struct
{
	uint16    type;
	uint16    class;
	uint32    ttl;
	uint16    rdlen;
}
NBTNsResource;

#define SizeOfNsRNB			6
typedef struct
{
	uint16    g:1, ont:2, resv:13;
	struct in_addr addr;
}
NBTNsRNB;

#define SizeOfResourceA		6
typedef struct
{
	struct in_addr addr;
}
NBTNsResourceA;

typedef struct
{
	uint16    opcode:4, flags:8, resv:4;
}
NBTNsResourceNULL;

typedef struct
{
	uint16    numnames;
}
NBTNsResourceNBSTAT;


#define MAX_NBT_NAME_LEN 16 

static uint8 *_rtl8651_l4AliasHandleName(uint8 * p, int8 * pmax, int8 *nbtname, int32 nbtlength);
static uint8 *_rtl8651_l4AliasHandleQuestion(uint16 , NBTNsQuestion *, int8 * , NBTArguments * );
static uint8 *_rtl8651_l4AliasHandleResource(uint16 , NBTNsResource * , int8 * , NBTArguments * );
static int32 _rtl8651_l4AliasHandleUdpNbt(struct ip *,struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *,struct in_addr *, uint16 );
static int32 _rtl8651_l4AliasHandleUdpNbtNS(struct ip * ,struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s * , struct in_addr * , uint16 * , struct in_addr * , uint16 * );



int32 rtl8651_l4AliasHandleClientUdpNbtOut(struct rtl_pktHdr *pkthdrPtr, struct ip *pip, struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{

	uint16    alias_port;
	struct in_addr alias_address;

	alias_address.s_addr = link->insideGlobalIpAddr;
	alias_port = link->insideGlobalPort;

	_rtl8651_l4AliasHandleUdpNbt(pip, link, &alias_address, alias_port);
	return SUCCESS;
}



int32 rtl8651_l4AliasHandleClientUdpNbtNSOut(struct rtl_pktHdr *pkthdrPtr, struct ip *pip, struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
	struct udphdr *ud;
	uint16    alias_port;
	struct in_addr alias_address;

	ud = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl & 0xf) << 2));

	alias_address.s_addr = link->insideGlobalIpAddr;
	alias_port = link->insideGlobalPort;

	{
	struct in_addr addr_x;
	 addr_x.s_addr= PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	_rtl8651_l4AliasHandleUdpNbtNS(pip, link, &addr_x, &ud->uh_sport,
								&alias_address, &alias_port);
	}
	return SUCCESS;
}



int32 rtl8651_l4AliasHandleClientUdpNbtIn(struct rtl_pktHdr *pkthdrPtr,
						struct ip *pip, 
					   struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)	
{

	struct udphdr *ud;
	struct in_addr original_address;

	ud = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl & 0xf) << 2));

	original_address.s_addr = link->insideLocalIpAddr;
	ud->uh_dport = link->insideLocalPort;

	return _rtl8651_l4AliasHandleUdpNbt(pip, link, &original_address, ud->uh_dport);
}



int32 rtl8651_l4AliasHandleClientUdpNbtNSIn(struct rtl_pktHdr *pkthdrPtr,
						struct ip *pip, 
					   struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)	
{
	struct udphdr *ud;
	uint16    alias_port;
	struct in_addr alias_address;
	struct in_addr original_address;


	ud = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl & 0xf) << 2));

	alias_address.s_addr = link->insideGlobalIpAddr;
	original_address.s_addr = link->insideLocalIpAddr;
	alias_port = ud->uh_dport;
	ud->uh_dport = link->insideLocalPort;
	
	return _rtl8651_l4AliasHandleUdpNbtNS(pip, link, &alias_address, &alias_port,
									&original_address, &ud->uh_dport);
}




static int32 _rtl8651_l4AliasHandleUdpNbt(struct ip *pip,	
						struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link,
						struct in_addr *alias_address, uint16 alias_port)
{
	struct udphdr *uh;
	NbtDataHeader *ndh;
	uint8    *p = NULL;
	int8     *pmax;
	int8	     nbtname[MAX_NBT_NAME_LEN]; 

	
	uh = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl & 0xf) << 2));
	pmax = (int8 *) uh + ntohs(uh->uh_ulen);

	ndh = (NbtDataHeader *) ((int8 *) uh + (sizeof(struct udphdr)));
	if ((int8 *) (ndh + 1) > pmax)
		return (-1);

	switch (ndh->type)
	{
		case DGM_DIRECT_UNIQ:		
		case DGM_DIRECT_GROUP:
		case DGM_BROADCAST:
			p = (uint8 *) ndh + 14;			
			p = _rtl8651_l4AliasHandleName(p, pmax,nbtname,MAX_NBT_NAME_LEN);	
			p = _rtl8651_l4AliasHandleName(p, pmax,nbtname,MAX_NBT_NAME_LEN );	
			break;
		case DGM_ERROR:
			p = (uint8 *) ndh + 11;
			break;
		case DGM_QUERY:
		case DGM_POSITIVE_RES:
		case DGM_NEGATIVE_RES:
			p = (uint8 *) ndh + 10;
			p = _rtl8651_l4AliasHandleName(p, pmax,nbtname,MAX_NBT_NAME_LEN);	
			break;
	}
	if (p == NULL || (int8 *) p > pmax)
		p = NULL;
	
	
	if (uh->uh_sum != 0)
	{
		int32     acc,tmp_val;
		uint16   *sptr;

		
		sptr = (uint16 *) & (ndh->source_ip);
		acc = *sptr++;
		acc += *sptr;
		sptr = (uint16 *) alias_address;
		acc -= *sptr++;
		acc -= *sptr;
		tmp_val = acc;
		_RTL8651_ADJUST_CHECKSUM(acc, pip->ip_sum);

		
		tmp_val += ndh->source_port;
		tmp_val -= alias_port;
		_RTL8651_ADJUST_CHECKSUM(tmp_val, uh->uh_sum);
	}
	
	rtl8651_memcpy(&ndh->source_ip , alias_address,4);	
	ndh->source_port = alias_port;

	return ((p == NULL) ? -1 : 0);
}



static int32 _rtl8651_l4AliasHandleUdpNbtNS(struct ip * pip,	
						  struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s * link,
						  struct in_addr * alias_address,
						  uint16 * alias_port,
						  struct in_addr * original_address,
						  uint16 * original_port)
{
	struct udphdr *uh;
	NbtNSHeader *nsh;
	uint8    *p;
	int8     *pmax;
	NBTArguments nbtarg;

	
	nbtarg.oldaddr = *alias_address;
	nbtarg.oldport = *alias_port;
	nbtarg.newaddr = *original_address;
	nbtarg.newport = *original_port;
	nbtarg.pip = pip;

	
	uh = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl & 0xf) << 2));
	nbtarg.uh_sum = &(uh->uh_sum);
	nsh = (NbtNSHeader *) ((int8 *) uh + (sizeof(struct udphdr)));
	p = (uint8 *) (nsh + 1);
	pmax = (int8 *) uh + ntohs(uh->uh_ulen);

	if ((int8 *) (nsh + 1) > pmax)
		return (-1);

	
	if (ntohs(nsh->qdcount) != 0)
	{
		p = _rtl8651_l4AliasHandleQuestion(ntohs(nsh->qdcount),
								(NBTNsQuestion *) p, pmax, &nbtarg);
	}

	
	if (ntohs(nsh->ancount) != 0)
	{
		p = _rtl8651_l4AliasHandleResource(ntohs(nsh->ancount),
								(NBTNsResource *) p, pmax, &nbtarg);
	}

	
	if (ntohs(nsh->nscount) != 0)
	{
		p = _rtl8651_l4AliasHandleResource(ntohs(nsh->nscount),
								(NBTNsResource *) p, pmax, &nbtarg);
	}

	
	if (ntohs(nsh->arcount) != 0)
	{
		p = _rtl8651_l4AliasHandleResource(ntohs(nsh->arcount),
								(NBTNsResource *) p, pmax, &nbtarg);
	}

	return ((p == NULL) ? -1 : 0);
}



static uint8 *_rtl8651_l4AliasHandleName(uint8 * p, int8 * pmax, int8 *nbtname, int32 nbtnamelen)
{

	uint8    *s;
	uint8     c;
	int32     compress;
	int32     ret_len = 0;
	int32     i;

	

	if (p == NULL || (int8 *) p >= pmax)
		return (NULL);

	if (*p & 0xc0)
	{
		p = p + 2;
		if ((int8 *) p > pmax)
			return (NULL);
		return ((uint8 *) p);
	}
	while ((*p & 0x3f) != 0x00)
	{
		s = p + 1;
		if (*p == 0x20)
			compress = 1;
		else
			compress = 0;

		
		p = (uint8 *) (p + (*p & 0x3f) + 1);
		if ((int8 *) p > pmax)
		{
			p = NULL;
			break;
		}
		while (s < p)
		{
			if (compress == 1)
			{
				
				

				c = ( (( (int)(*s) - 'A' ) << 4) +  ( (int)(*(s+1)) - 'A' ) );
								
				if(ret_len < (nbtnamelen-1) )
					nbtname[ret_len] = c;
				else
					nbtname[nbtnamelen-1] = '\0';
				
				ret_len++;
				s += 2;

				
				nbtname[15] = '\0';
								
				
			}
			else
			{
				if(ret_len < (nbtnamelen-1) )
					nbtname[ret_len] = (int8)(*s);
				else
					nbtname[nbtnamelen-1] = '\0';		
				
				ret_len++;				
				s++;

				
				nbtname[15] = '\0';
				
			}
		}
	}

	
	if (p == NULL || (int8 *) p >= pmax)
		p = NULL;
	else
		p++;

	
	if(ret_len < (nbtnamelen-1) )
		nbtname[ret_len] = '\0';
	else
		nbtname[nbtnamelen-1] = '\0';		

	
	  for( i = (nbtnamelen-1); (i >= 0) && (' ' == nbtname[i]); i-- )
	    nbtname[i] = '\0';
	
	
	return ((uint8 *) p);
}



static uint8 *_rtl8651_l4AliasHandleQuestion(uint16 count,
								  NBTNsQuestion * q, int8 * pmax,
								  NBTArguments * nbtarg)
{

	int8 nbtname[MAX_NBT_NAME_LEN];
	
	while (count != 0)
	{
		
		q = (NBTNsQuestion *) _rtl8651_l4AliasHandleName((uint8 *) q, pmax,nbtname,MAX_NBT_NAME_LEN);

		if (q == NULL || (int8 *) (q + 1) > pmax)
		{
			q = NULL;
			break;
		}

		
		switch (ntohs(q->type))
		{
			case QS_TYPE_NB:
			case QS_TYPE_NBSTAT:
				q = q + 1;
				break;
			default:
				break;
		}
		count--;
	}

	
	return ((uint8 *) q);
}



static uint8 *AliasHandleResourceNB(NBTNsResource * q, int8 * pmax,
									NBTArguments * nbtarg)
{
	NBTNsRNB *nb;
	uint16    bcount;

	if (q == NULL || (int8 *) (q + 1) > pmax)
		return (NULL);
	
	bcount = ntohs(q->rdlen);

	
	nb = (NBTNsRNB *) ((uint8 *) q + SizeOfNsResource);

	
	while (nb != NULL && bcount != 0)
	{
		if ((int8 *) (nb + 1) > pmax)
		{
			nb = NULL;
			break;
		}
		if (!memcmp(&nbtarg->oldaddr, &nb->addr, sizeof(struct in_addr)))
		{
			if (*nbtarg->uh_sum != 0)
			{
				int32     acc,tmp_val;
				uint16   *sptr;

				
				sptr = (uint16 *) & (nb->addr);
				acc = *sptr++;
				acc += *sptr;
				sptr = (uint16 *) & (nbtarg->newaddr);
				acc -= *sptr++;
				acc -= *sptr;
				tmp_val = acc;
				_RTL8651_ADJUST_CHECKSUM(acc, nbtarg->uh_sum);			

				
				_RTL8651_ADJUST_CHECKSUM(tmp_val, nbtarg->pip->ip_sum);	
			}
			
			rtl8651_memcpy(&nb->addr ,&nbtarg->newaddr,4);
		}
		nb = (NBTNsRNB *) ((uint8 *) nb + SizeOfNsRNB);
		bcount -= SizeOfNsRNB;
	}
	if (nb == NULL || (int8 *) (nb + 1) > pmax)
	{
		nb = NULL;
	}

	return ((uint8 *) nb);
}



static uint8 *AliasHandleResourceA(NBTNsResource * q, int8 * pmax,
								   NBTArguments * nbtarg)
{
	NBTNsResourceA *a;
	uint16    bcount;

	if (q == NULL || (int8 *) (q + 1) > pmax)
		return (NULL);

	
	a = (NBTNsResourceA *) ((uint8 *) q + sizeof(NBTNsResource));

	
	bcount = ntohs(q->rdlen);

	
	while (bcount != 0)
	{
		if (a == NULL || (int8 *) (a + 1) > pmax)
			return (NULL);

		if (!memcmp(&nbtarg->oldaddr, &a->addr, sizeof(struct in_addr)))
		{
			if (*nbtarg->uh_sum != 0)
			{
				int32     acc,tmp_val;
				uint16   *sptr;

				
				sptr = (uint16 *) & (a->addr);	
				acc = *sptr++;
				acc += *sptr;
				sptr = (uint16 *) & nbtarg->newaddr;	
				acc -= *sptr++;
				acc -= *sptr;
				tmp_val = acc;
				_RTL8651_ADJUST_CHECKSUM(acc, nbtarg->uh_sum);	
				
				
				_RTL8651_ADJUST_CHECKSUM(tmp_val, nbtarg->pip->ip_sum);					
				
			}
			rtl8651_memcpy(&a->addr, &nbtarg->newaddr,4);
		}
		a++;
		  bcount -= SizeOfResourceA;
	}
	if (a == NULL || (int8 *) (a + 1) > pmax)
		a = NULL;
	return ((uint8 *) a);
}



static uint8 *AliasHandleResourceNULL(NBTNsResource * q,
									  int8 * pmax, NBTArguments * nbtarg)
{
	NBTNsResourceNULL *n;
	uint16    bcount;

	if (q == NULL || (int8 *) (q + 1) > pmax)
		return (NULL);

	
	n = (NBTNsResourceNULL *) ((uint8 *) q + sizeof(NBTNsResource));

	
	bcount = ntohs(q->rdlen);

	
	while (bcount != 0)
	{
		if ((int8 *) (n + 1) > pmax)
		{
			n = NULL;
			break;
		}
		n++;
		bcount -= sizeof(NBTNsResourceNULL);
	}
	if ((int8 *) (n + 1) > pmax)
		n = NULL;

	return ((uint8 *) n);
}



static uint8 *AliasHandleResourceNS(NBTNsResource * q, int8 * pmax,
									NBTArguments * nbtarg)
{
	NBTNsResourceNULL *n;
	uint16    bcount;
	int8 nbtname[MAX_NBT_NAME_LEN];

	if (q == NULL || (int8 *) (q + 1) > pmax)
		return (NULL);

	
	n = (NBTNsResourceNULL *) ((uint8 *) q + sizeof(NBTNsResource));

	
	bcount = ntohs(q->rdlen);

	
	q = (NBTNsResource *) _rtl8651_l4AliasHandleName((uint8 *) n, pmax,nbtname,MAX_NBT_NAME_LEN);	

	if (q == NULL || (int8 *) ((uint8 *) n + bcount) > pmax)
		return (NULL);
	else
		return ((uint8 *) n + bcount);
}



static uint8 *AliasHandleResourceNBSTAT(NBTNsResource * q,
										int8 * pmax, NBTArguments * nbtarg)
{
	NBTNsResourceNBSTAT *n;
	uint16    bcount;

	if (q == NULL || (int8 *) (q + 1) > pmax)
		return (NULL);

	
	n = (NBTNsResourceNBSTAT *) ((uint8 *) q + sizeof(NBTNsResource));

	
	bcount = ntohs(q->rdlen);

	if (q == NULL || (int8 *) ((uint8 *) n + bcount) > pmax)
		return (NULL);
	else
		return ((uint8 *) n + bcount);
}



static uint8 *_rtl8651_l4AliasHandleResource(uint16 count,
								  NBTNsResource * q, int8 * pmax,
								  NBTArguments * nbtarg)
{
	int8 nbtname[MAX_NBT_NAME_LEN];
	while (count != 0)
	{
		
		q = (NBTNsResource *) _rtl8651_l4AliasHandleName((uint8 *) q, pmax,nbtname,MAX_NBT_NAME_LEN);

		if (q == NULL || (int8 *) (q + 1) > pmax)
			break;

		
		switch (ntohs(q->type))
		{
			case RR_TYPE_NB:
				q = (NBTNsResource *) AliasHandleResourceNB(q, pmax, nbtarg);
				break;
			case RR_TYPE_A:
				q = (NBTNsResource *) AliasHandleResourceA(q, pmax, nbtarg);
				break;
			case RR_TYPE_NS:
				q = (NBTNsResource *) AliasHandleResourceNS(q, pmax, nbtarg);
				break;
			case RR_TYPE_NULL:
				q = (NBTNsResource *) AliasHandleResourceNULL(q, pmax,
															  nbtarg);
				break;
			case RR_TYPE_NBSTAT:
				q = (NBTNsResource *) AliasHandleResourceNBSTAT(q, pmax,
																nbtarg);
				break;
			default:
				break;
		}
		count--;
	}

	return ((uint8 *) q);
}


