




#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_pptp.h"
#include "rtl8651_tblAsicDrv.h"


#define RTL8650B_PPTP_HASH



struct grehdr						   
{
	uint16    gh_flags;				   
	uint16    gh_protocol;			   
	uint16    gh_length;			   
	uint16    gh_call_id;			   
	uint32    gh_seq_no;			   
	uint32    gh_ack_no;			   
};
typedef struct grehdr GreHdr;


#define PPTP_GRE_PROTO          0x880b


#define PPTP_INIT_VALUE		((0x2001 << 16) | PPTP_GRE_PROTO)
#define PPTP_INIT_MASK		0xef7fffff

#define PPTP_MAGIC		0x1a2b3c4d
#define PPTP_CTRL_MSG_TYPE	1

enum
{
	PPTP_StartCtrlConnRequest = 1,
	PPTP_StartCtrlConnReply = 2,
	PPTP_StopCtrlConnRequest = 3,
	PPTP_StopCtrlConnReply = 4,
	PPTP_EchoRequest = 5,
	PPTP_EchoReply = 6,
	PPTP_OutCallRequest = 7,
	PPTP_OutCallReply = 8,
	PPTP_InCallRequest = 9,
	PPTP_InCallReply = 10,
	PPTP_InCallConn = 11,
	PPTP_CallClearRequest = 12,
	PPTP_CallDiscNotify = 13,
	PPTP_WanErrorNotify = 14,
	PPTP_SetLinkInfo = 15
};

int reInit=0;

  
struct pptpMsgHead
{
	uint16    length;				   
	uint16    msgType;				   
	uint32    magic;				   
	uint16    type;					   
	uint16    resv0;				   
};
typedef struct pptpMsgHead *PptpMsgHead;

struct pptpCodes
{
	uint8     resCode;				   
	uint8     errCode;				   
};
typedef struct pptpCodes *PptpCode;

struct pptpCallIds
{
	uint16    cid1;					   
	uint16    cid2;					   
};
typedef struct pptpCallIds *PptpCallId;

typedef struct _rtl8651_pptp_passthrough_para_s {
	uint16 idelTimeout;
	uint16 heavyLoadSwappingThreshold;
} _rtl8651_pptp_passthrough_para_t;

_rtl8651_pptp_passthrough_para_t pptpPassthroughSysPara;

typedef struct _rtl8651_pptpGreEntry_s {
	uint32 outsideGlobalIpAddr;
	uint32 insideGlobalIpAddr;
	uint32 insideLocalIpAddr;

	uint16 outboundGlobalId;
	uint16 outboundLocalId;
	uint16 inboundGlobalId;
	uint16 inboundLocalId;

	uint16 inboundIdleTimer;	
	uint16 inboundPktCnt;	
	uint16 outboundIdleTimer;	
	uint16 outboundPktCnt;	

	uint32 dsid;
	
	uint16 inboundIdInitiated:1;
	uint16 inboundIdClosed:1;
	uint16 inboundToAsic:1;
	uint16 outboundIdInitiated:1;
	uint16 outboundIdClosed:1;
	uint16 outboundToAsic:1;
	uint16 controlIdleTimer;
	struct _rtl8651_pptpGreEntry_s * inboundNext, * outboundNext;
}_rtl8651_pptpGreEntry_t;

_rtl8651_pptpGreEntry_t * pptpGreInboundHeadPtr[RTL8651_ICMPTBL_SIZE], * pptpGreOutboundHeadPtr[RTL8651_ICMPTBL_SIZE], 
	* freePptpGreEntryHeadPtr=NULL ,*softwarePptpGreEntryHeadPtr=NULL;
uint16 pptpPassthroughControlShortTimeout = 10, pptpPassthroughControlMiddleTimeout = 60;
uint8 pptpPassthroughClientInboundMsg = 0, pptpPassthroughClientOutboundMsg = 0,
	 pptpPassthroughServerInboundMsg = 0, pptpPassthroughServerOutboundMsg = 0,
	 pptpPassthroughGreInboundMsg = 0, pptpPassthroughGreOutboundMsg = 0;


int32 _rtl8651_addAsicPptpPassthroughInboundEntry(_rtl8651_pptpGreEntry_t * entryPtr) {
	_rtl8651_pptpGreEntry_t *trackPtr;
	rtl865x_tblAsicDrv_naptIcmpParam_t pptpAsicEntry;
	uint32 tblIdx = entryPtr->inboundGlobalId&((1<<RTL8651_ICMPTBL_BITS)-1);
	
	if(entryPtr->inboundToAsic)
		return SUCCESS;

	trackPtr = pptpGreInboundHeadPtr[tblIdx];
	while(trackPtr) {
		if(trackPtr->inboundIdInitiated == 1 && trackPtr->inboundToAsic==1)
			break;
		trackPtr = trackPtr->inboundNext;
	}
	if(trackPtr == NULL) {
		trackPtr = pptpGreOutboundHeadPtr[tblIdx];
		while(trackPtr) {
			if(trackPtr->outboundIdInitiated == 1 && trackPtr->outboundToAsic==1)
				break;
			trackPtr = trackPtr->outboundNext;
		}
	}
	if(trackPtr==NULL) {
		memset(&pptpAsicEntry, 0, sizeof(rtl865x_tblAsicDrv_naptIcmpParam_t));
		pptpAsicEntry.insideLocalId = entryPtr->inboundLocalId;
		pptpAsicEntry.insideLocalIpAddr = entryPtr->insideLocalIpAddr;
		pptpAsicEntry.offset = (entryPtr->inboundGlobalId>>10) & 0x3F;
		pptpAsicEntry.isCollision = 1;
		pptpAsicEntry.isStatic = 1;
		pptpAsicEntry.isValid = 1;
		pptpAsicEntry.isPptp = 1;
		pptpAsicEntry.isSpi = 0;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx , &pptpAsicEntry);
		entryPtr->inboundToAsic = 1;
		return SUCCESS;
	}
	return FAILED;
}


int32 _rtl8651_delAsicPptpPassthroughInboundEntry(_rtl8651_pptpGreEntry_t * entryPtr) {
	_rtl8651_pptpGreEntry_t *trackPtr;
	rtl865x_tblAsicDrv_naptIcmpParam_t pptpAsicEntry;
	uint32 tblIdx = entryPtr->inboundGlobalId&((1<<RTL8651_ICMPTBL_BITS)-1);
	memset(&pptpAsicEntry, 0, sizeof(rtl865x_tblAsicDrv_naptIcmpParam_t));

	if(entryPtr->inboundToAsic == 0)
		return SUCCESS;

	entryPtr->inboundToAsic = 0;
	trackPtr = pptpGreInboundHeadPtr[tblIdx];
	while(trackPtr) {
		if(trackPtr->inboundIdInitiated == 1 && trackPtr->inboundToAsic==0) {
			pptpAsicEntry.insideLocalId = trackPtr->inboundLocalId;
			pptpAsicEntry.offset = (trackPtr->inboundGlobalId>>10) & 0x3F;
			entryPtr->inboundToAsic = 1;
			break;
		}
		trackPtr = trackPtr->inboundNext;
	}
	if(trackPtr == NULL) {
		trackPtr = pptpGreOutboundHeadPtr[tblIdx];
		while(trackPtr) {
			if(trackPtr->outboundIdInitiated == 1 && trackPtr->outboundToAsic==0) {
				pptpAsicEntry.insideLocalId = trackPtr->outboundLocalId;
				pptpAsicEntry.offset = (trackPtr->outboundGlobalId>>10) & 0x3F;
				entryPtr->outboundToAsic = 1;
				break;
			}
			trackPtr = trackPtr->outboundNext;
		}
	}
	if(trackPtr) {
		pptpAsicEntry.insideLocalIpAddr = trackPtr->insideLocalIpAddr;
		pptpAsicEntry.isCollision = 1;
		pptpAsicEntry.isStatic = 1;
		pptpAsicEntry.isValid = 1;
		pptpAsicEntry.isPptp = 1;
		pptpAsicEntry.isSpi = 0;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx , &pptpAsicEntry);
	} else {
		memset(&pptpAsicEntry, 0, sizeof(rtl865x_tblAsicDrv_naptIcmpParam_t));
		pptpAsicEntry.isCollision = 1;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx, &pptpAsicEntry);
	}
	return SUCCESS;
}

int32 _rtl8651_addAsicPptpPassthroughOutboundEntry(_rtl8651_pptpGreEntry_t * entryPtr) {
	_rtl8651_pptpGreEntry_t *trackPtr;
	rtl865x_tblAsicDrv_naptIcmpParam_t pptpAsicEntry;
	uint32 tblIdx = entryPtr->outboundGlobalId&((1<<RTL8651_ICMPTBL_BITS)-1);

	if(entryPtr->outboundToAsic)
		return SUCCESS;
	
	trackPtr = pptpGreInboundHeadPtr[tblIdx];
	while(trackPtr) {
		if(trackPtr->inboundIdInitiated == 1 && trackPtr->inboundToAsic==1)
			break;
		trackPtr = trackPtr->inboundNext;
	}
	if(trackPtr == NULL) {
		trackPtr = pptpGreOutboundHeadPtr[tblIdx];
		while(trackPtr) {
			if(trackPtr->outboundIdInitiated == 1 && trackPtr->outboundToAsic==1)
				break;
			trackPtr = trackPtr->outboundNext;
		}
	}
	if(trackPtr==NULL) {
		memset(&pptpAsicEntry, 0, sizeof(rtl865x_tblAsicDrv_naptIcmpParam_t));
		pptpAsicEntry.insideLocalId = entryPtr->outboundLocalId;
		pptpAsicEntry.insideLocalIpAddr = entryPtr->insideLocalIpAddr;
		pptpAsicEntry.offset = (entryPtr->outboundGlobalId>>10) & 0x3F;
		pptpAsicEntry.isCollision = 1;
		pptpAsicEntry.isStatic = 1;
		pptpAsicEntry.isValid = 1;
		pptpAsicEntry.isPptp = 1;
		pptpAsicEntry.isSpi = 0;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx, &pptpAsicEntry);
		entryPtr->outboundToAsic = 1;
		
		return SUCCESS;
	}
	
	return FAILED;
}


int32 _rtl8651_delAsicPptpPassthroughOutboundEntry(_rtl8651_pptpGreEntry_t * entryPtr) {
	_rtl8651_pptpGreEntry_t *trackPtr;
	rtl865x_tblAsicDrv_naptIcmpParam_t pptpAsicEntry;
	uint32 tblIdx = entryPtr->outboundGlobalId&((1<<RTL8651_ICMPTBL_BITS)-1);

	if(entryPtr->outboundToAsic == 0)
		return SUCCESS;

	entryPtr->outboundToAsic = 0;
	trackPtr = pptpGreOutboundHeadPtr[tblIdx];
	while(trackPtr) {
		if(trackPtr->outboundIdInitiated == 1 && trackPtr->outboundToAsic==0) {
			pptpAsicEntry.insideLocalId = trackPtr->outboundLocalId;
			pptpAsicEntry.offset = (trackPtr->outboundGlobalId>>10) & 0x3F;
			trackPtr->outboundToAsic = 1;
			break;
		}
		trackPtr = trackPtr->outboundNext;
	}
	if(trackPtr == NULL) {
		trackPtr = pptpGreInboundHeadPtr[tblIdx];
		while(trackPtr) {
			if(trackPtr->inboundIdInitiated == 1 && trackPtr->inboundToAsic==0) {
				pptpAsicEntry.insideLocalId = trackPtr->inboundLocalId;
				pptpAsicEntry.offset = (trackPtr->inboundGlobalId>>10) & 0x3F;
				trackPtr->inboundToAsic = 1;
				break;
			}
			trackPtr = trackPtr->inboundNext;
		}
	}
	if(trackPtr) {
		pptpAsicEntry.insideLocalIpAddr = trackPtr->insideLocalIpAddr;
		pptpAsicEntry.isCollision = 1;
		pptpAsicEntry.isStatic = 1;
		pptpAsicEntry.isValid = 1;
		pptpAsicEntry.isPptp = 1;
		pptpAsicEntry.isSpi = 0;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx , &pptpAsicEntry);
	} else {
		memset(&pptpAsicEntry, 0, sizeof(rtl865x_tblAsicDrv_naptIcmpParam_t));
		pptpAsicEntry.isCollision = 1;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx, &pptpAsicEntry);
	}
	return SUCCESS;
}


static int32 _rtl8651_l4GreNaptAliasIn(struct rtl_pktHdr *phdr,struct ip * pip) {
	_rtl8651_pptpGreEntry_t * tempPtr, *checkPtr, * asicPtr;
	GreHdr   *gr;
	uint32 hdr_bit;
	uint16 gh_call_id;
	uint32 srcIp, dstIp;
	uint8 software=0;
	int i;

	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;


	
	gr = (GreHdr *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
	hdr_bit = PKTGET_UINT32_UNALIGNED(gr);
	gh_call_id = ntohs(gr->gh_call_id);
	srcIp = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	dstIp = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

	if(pptpPassthroughGreInboundMsg)
		printk("PPTP-PTH: Data GRE In ID=%d DIP %d.%d.%d.%d SIP %d.%d.%d.%d\n",
			gh_call_id, (srcIp>>24)&0xff, (srcIp>>16)&0xff, (srcIp>>8)&0xff, srcIp&0xff,
			(dstIp>>24)&0xff, (dstIp>>16)&0xff, (dstIp>>8)&0xff, dstIp&0xff);	

	
	if ((hdr_bit & PPTP_INIT_MASK) != PPTP_INIT_VALUE)
		return FWDENG_GRE_IGNORE;


	if(software)
	{

		tempPtr=softwarePptpGreEntryHeadPtr;
		for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
		{
			if(tempPtr->inboundIdInitiated &&  tempPtr->inboundGlobalId == gh_call_id && tempPtr->insideGlobalIpAddr == dstIp && tempPtr->outsideGlobalIpAddr == srcIp)
				break;
		}
		if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;

		if(tempPtr)
		{
			uint8 *dip;
			tempPtr->inboundIdleTimer = 0;
								 
			dip=(uint8 *)&pip->ip_dst;
			dip[0]=(tempPtr->insideLocalIpAddr>>24)&0xff;
			dip[1]=(tempPtr->insideLocalIpAddr>>16)&0xff;
			dip[2]=(tempPtr->insideLocalIpAddr>>8)&0xff;
			dip[3]=tempPtr->insideLocalIpAddr&0xff;
			return SUCCESS;
		}
		
	}

	else 
	{
		tempPtr = pptpGreInboundHeadPtr[gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1)];


		while(tempPtr) {

			if(tempPtr->inboundIdInitiated &&  tempPtr->inboundGlobalId == gh_call_id && tempPtr->insideGlobalIpAddr == dstIp && tempPtr->outsideGlobalIpAddr == srcIp)
				break;
			tempPtr = tempPtr->inboundNext;
		}
		if (tempPtr)	{
			uint8 *dip;
			
			gr->gh_call_id = htons(tempPtr->inboundLocalId);

								 
			dip=(uint8 *)&pip->ip_dst;
			dip[0]=(tempPtr->insideLocalIpAddr>>24)&0xff;
			dip[1]=(tempPtr->insideLocalIpAddr>>16)&0xff;
			dip[2]=(tempPtr->insideLocalIpAddr>>8)&0xff;
			dip[3]=tempPtr->insideLocalIpAddr&0xff;
			phdr->ph_dip = tempPtr->insideLocalIpAddr;
			

			tempPtr->inboundIdleTimer = 0;
			tempPtr->inboundPktCnt++;
			if(tempPtr->inboundPktCnt == pptpPassthroughSysPara.heavyLoadSwappingThreshold) {
				asicPtr = NULL;
				checkPtr = pptpGreInboundHeadPtr[gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1)];
				while(checkPtr) {
					if(checkPtr->inboundToAsic == 1) {
						assert(asicPtr == NULL);
						asicPtr = checkPtr;
					}
					if(checkPtr != tempPtr && 
						checkPtr->inboundPktCnt >= pptpPassthroughSysPara.heavyLoadSwappingThreshold)
						break;
					checkPtr = checkPtr->inboundNext;
				}
				if(checkPtr == NULL) {
					checkPtr = pptpGreOutboundHeadPtr[gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1)];
					while(checkPtr) {
						if(checkPtr->outboundToAsic == 1) {
							assert(asicPtr == NULL);
							asicPtr = checkPtr;
						}
						if(checkPtr->outboundPktCnt >= pptpPassthroughSysPara.heavyLoadSwappingThreshold)
							break;
						checkPtr = checkPtr->outboundNext;
					}
				}
				if(checkPtr == NULL) {
					assert(asicPtr);
					
					if(asicPtr->inboundToAsic==1 && 
						(asicPtr->inboundGlobalId & ((1<<RTL8651_ICMPTBL_BITS)-1)) == (gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1))) {
						asicPtr->inboundToAsic=0;
						
						
						
						
					}
					else {
						assert(asicPtr->outboundToAsic==1);
						assert((asicPtr->outboundGlobalId& ((1<<RTL8651_ICMPTBL_BITS)-1)) == (gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1)));
						asicPtr->outboundToAsic=0;
						
						
						
						
					}
					
					
					
					
					_rtl8651_addAsicPptpPassthroughInboundEntry(tempPtr);
				}
			}
			
			
			
			
			
			
			
			return SUCCESS;
		}
		else {
			if(pptpPassthroughGreInboundMsg) printk("PPTP-PTH: GRE IN no existed entry \n");
		}
	}
	return FWDENG_GRE_IGNORE;
}



static int32 _rtl8651_l4GreNaptAliasOut(struct rtl_pktHdr *phdr,struct ip *pip)
{
	_rtl8651_pptpGreEntry_t * tempPtr, *checkPtr, * asicPtr;
	GreHdr   *gr;
	uint32 hdr_bit, tblIdx;
	uint16 gh_call_id;
	uint32 srcIp, dstIp;
	uint8 software=0;
	int i;

	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;
	gr = (GreHdr *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
	hdr_bit = PKTGET_UINT32_UNALIGNED(gr);
	gh_call_id =  ntohs(gr->gh_call_id) ;
	srcIp = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	dstIp = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

	if(pptpPassthroughGreOutboundMsg)
		printk("PPTP-PTH: Data GRE Out ID=%d DIP %d.%d.%d.%d SIP %d.%d.%d.%d\n",
			gh_call_id, (dstIp>>24)&0xff, (dstIp>>16)&0xff, (dstIp>>8)&0xff, dstIp&0xff,
			(srcIp>>24)&0xff, (srcIp>>16)&0xff, (srcIp>>8)&0xff, srcIp&0xff );	

		
	if ((hdr_bit & PPTP_INIT_MASK) != PPTP_INIT_VALUE)
		return FWDENG_GRE_IGNORE;

	if(software)
	{
		tempPtr=softwarePptpGreEntryHeadPtr;
		for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
		{
			if(tempPtr->outboundIdInitiated && tempPtr->insideLocalIpAddr == srcIp && tempPtr->outboundLocalId == gh_call_id && tempPtr->outsideGlobalIpAddr == dstIp)
				break;
		}
		if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;

		if(tempPtr)
		{
			uint8 *sip;
			tempPtr->outboundIdleTimer = 0;
			sip=(uint8 *)&pip->ip_src;		
			sip[0]=(tempPtr->insideGlobalIpAddr>>24)&0xff;
			sip[1]=(tempPtr->insideGlobalIpAddr>>16)&0xff;
			sip[2]=(tempPtr->insideGlobalIpAddr>>8)&0xff;
			sip[3]=tempPtr->insideGlobalIpAddr&0xff;
			
			return SUCCESS;
		}
		
	}
	else
	{

		rtl8651_naptIcmpTableIndex(srcIp, gh_call_id, dstIp, &tblIdx);	
		
		tempPtr = pptpGreOutboundHeadPtr[tblIdx];


		while(tempPtr) {
			
			
			
			if(tempPtr->outboundIdInitiated && tempPtr->insideLocalIpAddr == srcIp && tempPtr->outboundLocalId == gh_call_id && tempPtr->outsideGlobalIpAddr == dstIp)
				break;
			tempPtr = tempPtr->outboundNext;
		}
		if (tempPtr) {
			uint8 *sip;

			gr->gh_call_id = htons(tempPtr->outboundGlobalId);

			sip=(uint8 *)&pip->ip_src;		
			sip[0]=(tempPtr->insideGlobalIpAddr>>24)&0xff;
			sip[1]=(tempPtr->insideGlobalIpAddr>>16)&0xff;
			sip[2]=(tempPtr->insideGlobalIpAddr>>8)&0xff;
			sip[3]=tempPtr->insideGlobalIpAddr&0xff;
			phdr->ph_sip = tempPtr->insideGlobalIpAddr;
			

			tempPtr->outboundIdleTimer = 0;
			tempPtr->outboundPktCnt++;
			if(tempPtr->outboundPktCnt == pptpPassthroughSysPara.heavyLoadSwappingThreshold) {			
				asicPtr = NULL;
				checkPtr = pptpGreOutboundHeadPtr[tblIdx];
				while(checkPtr) {
					if(checkPtr->outboundToAsic == 1) {
						assert(asicPtr == NULL);
						asicPtr = checkPtr;
					}
					if(checkPtr != tempPtr && 
						checkPtr->outboundPktCnt >= pptpPassthroughSysPara.heavyLoadSwappingThreshold)
						break;
					checkPtr = checkPtr->outboundNext;
				}
				if(checkPtr == NULL) {
					checkPtr = pptpGreInboundHeadPtr[tblIdx];
					while(checkPtr) {
						if(checkPtr->inboundToAsic == 1) {
							assert(asicPtr == NULL);
							asicPtr = checkPtr;
						}
						if(checkPtr->inboundPktCnt >= pptpPassthroughSysPara.heavyLoadSwappingThreshold)
							break;
						checkPtr = checkPtr->inboundNext;
					}
				}
				if(checkPtr == NULL) {
					assert(asicPtr);
					
					if(asicPtr->inboundToAsic==1 && (asicPtr->inboundGlobalId & ((1<<RTL8651_ICMPTBL_BITS)-1)) == tblIdx) {
						asicPtr->inboundToAsic=0;
						
						
						
						
					}
					else {
						assert(asicPtr->outboundToAsic==1);
						assert((asicPtr->outboundGlobalId & ((1<<RTL8651_ICMPTBL_BITS)-1)) == tblIdx);
						asicPtr->outboundToAsic=0;
						
						
						
						
					}
					
					
					
					
					_rtl8651_addAsicPptpPassthroughOutboundEntry(tempPtr);
				}
			}
			
			
			
			
			
			
			
			return SUCCESS;	
		}
	}

	return FWDENG_GRE_IGNORE;
}




static PptpCallId _rtl8651_aliasVerifyPptp(struct ip *pip, uint16 * ptype)	
{
	int32     hlen, tlen, dlen;
	PptpMsgHead hptr;
	struct tcphdr *tc;

	uint32 magic;

	
	tc =  (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
	hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >>4)&0xf)) << 2;
	tlen = ntohs(pip->ip_len);
	dlen = tlen - hlen;

	
	if (dlen < (sizeof(struct pptpMsgHead) + sizeof(struct pptpCallIds)))
		return (NULL);

	
	hptr = (PptpMsgHead) (((int8 *) pip) + hlen);

	
	*ptype = ntohs(hptr->type);
	
	

	magic = PKTGET_UINT32_UNALIGNED(&hptr->magic);
	
	if ((ntohs(hptr->msgType) != PPTP_CTRL_MSG_TYPE) ||
		(magic != PPTP_MAGIC))
		return (NULL);

	
	if ((*ptype == PPTP_OutCallReply || *ptype == PPTP_InCallReply) &&
		(dlen < sizeof(struct pptpMsgHead) + sizeof(struct pptpCallIds) +
		 sizeof(struct pptpCodes)))
		return (NULL);
	else
		return (PptpCallId) (hptr + 1);

}




int32 rtl8651_l4AliasHandlePptpClientOutbound(
		struct rtl_pktHdr *pkt,
		struct ip *pip,
		struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
	_rtl8651_pptpGreEntry_t * tempPtr=NULL;
	PptpCallId cptr;
	uint16    ctl_type;				   
	uint32 i, entryIdx, tblIdx;
	uint8 software=0;
	static uint32 searchIdx = 0;
	
	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;
	if((alg_supports[pkt->ph_dsid]&RTL8651_ALG_PPTP)==0) return FWDENG_USER_DROP;

	
	if ((cptr = _rtl8651_aliasVerifyPptp(pip, &ctl_type)) == NULL)
		return SUCCESS;


	
	switch (ctl_type)
	{
		case PPTP_OutCallRequest:
			if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH PKT: Client Outbound OutCallRequest\n");
			if(!software)
			{
				for(i=0, tempPtr = NULL; tempPtr==NULL && i<RTL8651_ICMPTBL_SIZE; i++) {
					tempPtr = pptpGreInboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideLocalIpAddr == link->insideLocalIpAddr &&
							tempPtr->inboundLocalId == ntohs(cptr->cid1) &&	tempPtr->inboundIdInitiated == 1)
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
			}				
			else 
			{

				tempPtr=softwarePptpGreEntryHeadPtr;

				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{	
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideLocalIpAddr == link->insideLocalIpAddr &&
					   tempPtr->inboundLocalId == ntohs(cptr->cid1) &&	tempPtr->inboundIdInitiated == 1)
							break;
				}
				
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				else
				{

					memset(tempPtr,0,sizeof(tempPtr));
					tempPtr=NULL;
				}
				
			
			}
	
			
			if(tempPtr == NULL) {	
				if(!software)
				{
					if(freePptpGreEntryHeadPtr == NULL)
						return FWDENG_USER_DROP; 
					tempPtr = freePptpGreEntryHeadPtr;
					freePptpGreEntryHeadPtr = freePptpGreEntryHeadPtr->inboundNext;
				}
				else 
				{
					_rtl8651_pptpGreEntry_t empty;
					memset(&empty,0,sizeof(empty));
	
					tempPtr=softwarePptpGreEntryHeadPtr;
					for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
					{
						if(memcmp(tempPtr,&empty,sizeof(empty))==0)
								break;
					}
					if(i==algMaxSession.pptpMaxSession) return SUCCESS; 
					
				}
				
				tempPtr->outsideGlobalIpAddr = link->dstIpAddr;
				tempPtr->insideGlobalIpAddr = link->insideGlobalIpAddr;
				tempPtr->insideLocalIpAddr = link->insideLocalIpAddr;
				tempPtr->inboundLocalId = ntohs(cptr->cid1);
				tempPtr->inboundIdInitiated = 1;
				tempPtr->outboundIdInitiated = 0;
				tempPtr->inboundToAsic = 0;
				tempPtr->outboundToAsic = 0;
				tempPtr->inboundIdClosed = 0;
				tempPtr->outboundIdClosed = 0;
				tempPtr->controlIdleTimer = 0;
				tempPtr->inboundIdleTimer = 0;
				tempPtr->inboundPktCnt = 0;
				tempPtr->outboundIdleTimer = 0;
				tempPtr->outboundPktCnt = 0;
				tempPtr->dsid = pkt->ph_dsid;

				if(software==1)
				{
					tempPtr->inboundGlobalId = tempPtr->inboundLocalId;
				}
				else
				{
					_rtl8651_pptpGreEntry_t * trackPtr;

					tempPtr->inboundGlobalId = rtl8651_naptIcmpTableIndex(tempPtr->insideLocalIpAddr, tempPtr->inboundLocalId, tempPtr->outsideGlobalIpAddr, &tblIdx);
					if(pptpPassthroughClientOutboundMsg)
						printk("PPTP-PH: Client from LAN OutCallRequest SIP: %d.%d.%d.%d DIP:%d.%d.%d.%d ID %d\n",
							(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff,
							(tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff,
							(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff,
							(tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
							tempPtr->inboundGlobalId);
					if(pptpGreInboundHeadPtr[tblIdx]) {
						uint32 offsetMask[2], i;
					
						trackPtr = pptpGreInboundHeadPtr[tblIdx];
						offsetMask[0] = offsetMask[1] = 0;
						while(trackPtr) {
							assert(trackPtr->inboundIdInitiated);
							offsetMask[(trackPtr->inboundGlobalId>>10)&0x1] |= (1<<((trackPtr->inboundGlobalId>>11)&0x1F));
							trackPtr = trackPtr->inboundNext;
						}
						for(i=searchIdx; i!=((searchIdx+63)&0x3F); i=(i+1)&0x3F) {
							if((offsetMask[i&0x1] & (1<<((i>>1)&0x1F))) == 0)
								break;
						}
						searchIdx = (i+1)&0x3F;
						assert(i<64);
						tempPtr->inboundGlobalId |= (i<<10);
					}
					if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH: Add inbound GID:%d LID:%d\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					_rtl8651_addAsicPptpPassthroughInboundEntry(tempPtr);
					tempPtr->inboundNext = pptpGreInboundHeadPtr[tempPtr->inboundGlobalId &((1<<RTL8651_ICMPTBL_BITS)-1)];
					pptpGreInboundHeadPtr[tempPtr->inboundGlobalId &((1<<RTL8651_ICMPTBL_BITS)-1)] = tempPtr;

				}
				
			}
			else {
				if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH: Established inbound GID:%d LID:%d\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
				tempPtr->inboundIdClosed = 0;
				tempPtr->controlIdleTimer = 0;
			}
			cptr->cid1 = htons(tempPtr->inboundGlobalId);
			
			
			break;
		case PPTP_CallClearRequest:
			if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH PKT: Client Outbound CallClearRequest\n");
			if(!software)
			{
			
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreInboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->inboundLocalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
				if(tempPtr) {
#if 0					
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreInboundHeadPtr[entryIdx] == tempPtr) {
						pptpGreInboundHeadPtr[entryIdx] = tempPtr->inboundNext;
					} else {
						trackPtr = pptpGreInboundHeadPtr[entryIdx];
						while(trackPtr && trackPtr->inboundNext != tempPtr)
							trackPtr = trackPtr->inboundNext;
						assert(trackPtr);
						trackPtr->inboundNext = tempPtr->inboundNext;
					}
					tempPtr->inboundIdInitiated = 0;
					
					
					
					
					_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
					if(tempPtr->outboundIdInitiated)
						tempPtr->inboundNext = NULL;
					else {
						tempPtr->inboundNext = freePptpGreEntryHeadPtr;
						freePptpGreEntryHeadPtr = tempPtr;
					}
#endif
					if(tempPtr->inboundIdClosed == 1 && pptpPassthroughClientOutboundMsg)
						printk("PPTP-PTH:(Call Clear Request) Inbound Closed ID GID:%d LID: %d\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					else if (pptpPassthroughClientOutboundMsg)
						printk("PPTP-PTH:(Call Clear Request) Inbound ID GID: %d LID: %d Closed\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					tempPtr->inboundIdClosed = 1;
					tempPtr->controlIdleTimer = 0;
					cptr->cid1 = htons(tempPtr->inboundGlobalId);
				}
				
			}
			else
			{

				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->inboundLocalId == ntohs(cptr->cid1))
					break;
				}
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				
				if(tempPtr)
				{
					tempPtr->inboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
					tempPtr->outboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
				}
			}
			break;
		case PPTP_WanErrorNotify:
		case PPTP_SetLinkInfo:
			if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH PKT: Client Outbound SetLinkInfo\n");
			if(!software)
			{
				
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->outboundLocalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}

				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				if(tempPtr)
				{
					cptr->cid1=htons(tempPtr->outboundGlobalId);
					tempPtr->controlIdleTimer = 0;
				}					
			}
			break;
		case PPTP_StopCtrlConnRequest:
		case PPTP_StopCtrlConnReply:
			if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH PKT: Client Outbound StopCtrlConn\n");
			if(!software)	{
				do {
					for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
						entryIdx = i;
						tempPtr = pptpGreInboundHeadPtr[i];
						while(tempPtr) {
							if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && tempPtr->inboundIdClosed == 0)
								break;
							tempPtr = tempPtr->inboundNext;
						}
					}
					if(tempPtr) {
						if (pptpPassthroughClientOutboundMsg)
							printk("PPTP-PTH:(Stop Control Connection) Inbound ID GID: %d LID: %d Closed\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
						tempPtr->inboundIdClosed = 1;
						tempPtr->controlIdleTimer = 0;
						cptr->cid1 = htons(tempPtr->inboundGlobalId);
					}
				}while(tempPtr);
			}				
			break;
		case PPTP_InCallRequest:
		case PPTP_InCallReply:
		case PPTP_InCallConn: 
			
		default:
			return SUCCESS;
	}

	return SUCCESS;
}


int32 rtl8651_l4AliasHandlePptpServerInbound(
	struct rtl_pktHdr *pkt, 
	struct ip *pip,
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s * link)	
{

	_rtl8651_pptpGreEntry_t * tempPtr=NULL;
	PptpCallId cptr;
	uint16    ctl_type;				   
	uint32 i, entryIdx, tblIdx;
	uint8 software=0;
	static uint32 searchIdx = 0, lastId=0;
	
	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;

	if((alg_supports[pkt->ph_dsid]&RTL8651_ALG_PPTP_SERVER)==0) return FWDENG_USER_DROP;

	
	if ((cptr = _rtl8651_aliasVerifyPptp(pip, &ctl_type)) == NULL)
		return SUCCESS;

	
	switch (ctl_type)
	{
		case PPTP_OutCallRequest:
			if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH PKT: Server Inbound OutCallRequest\n");
			if(!software)
			{
				for(i=0, tempPtr = NULL; tempPtr==NULL && i<RTL8651_ICMPTBL_SIZE; i++) {
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideLocalIpAddr == link->insideLocalIpAddr &&
							tempPtr->outboundGlobalId == ntohs(cptr->cid1) &&	tempPtr->outboundIdInitiated == 1)
							break;
						tempPtr = tempPtr->outboundNext;
					}
				}
			}
			else
			{				
				tempPtr=softwarePptpGreEntryHeadPtr;

				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{	
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideLocalIpAddr == link->insideLocalIpAddr &&
						tempPtr->outboundGlobalId == ntohs(cptr->cid1) &&	tempPtr->outboundIdInitiated == 1)
						break;
				}
				
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				
			}
			
			if(tempPtr == NULL) {	
				
				
				if(!software)
				{						
					if(freePptpGreEntryHeadPtr == NULL)
						return FWDENG_USER_DROP;
					tempPtr = freePptpGreEntryHeadPtr;
					freePptpGreEntryHeadPtr = freePptpGreEntryHeadPtr->inboundNext;
				}
				else
				{
					_rtl8651_pptpGreEntry_t empty;
					memset(&empty,0,sizeof(empty));
	
					tempPtr=softwarePptpGreEntryHeadPtr;
					for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
					{
						if(memcmp(tempPtr,&empty,sizeof(empty))==0)
							break;
					}
					if(i==algMaxSession.pptpMaxSession) return SUCCESS; 
				}
				
				tempPtr->outsideGlobalIpAddr = link->dstIpAddr;
				tempPtr->insideGlobalIpAddr = link->insideGlobalIpAddr;
				tempPtr->insideLocalIpAddr = link->insideLocalIpAddr;
				tempPtr->outboundGlobalId = ntohs(cptr->cid1);
				tempPtr->outboundIdInitiated = 1;
				
				tempPtr->inboundIdInitiated = 0;
				tempPtr->inboundToAsic = 0;
				tempPtr->outboundToAsic = 0;
				tempPtr->inboundIdClosed = 0;
				tempPtr->outboundIdClosed = 0;
				tempPtr->inboundIdleTimer = 0;
				tempPtr->inboundPktCnt = 0;
				tempPtr->outboundIdleTimer = 0;
				tempPtr->outboundPktCnt = 0;
				tempPtr->controlIdleTimer = 0; 
				
				tempPtr->dsid = pkt->ph_dsid;

				if(software)
				{
					tempPtr->outboundLocalId = tempPtr->outboundGlobalId;
					
				}
				else
				{
					{
						_rtl8651_pptpGreEntry_t * trackPtr;
						uint16 localid;
						uint16 targetHashIdx=tempPtr->outboundGlobalId & 0x3ff;
						lastId = ((lastId +1)*(tempPtr->outboundGlobalId+1))&0xFFFF;
						if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH: Server Inbound try ID start at %d\n", lastId);
						for(localid = lastId;localid != ((lastId+65535)&0xFFFF);localid = (localid +1)&0xFFFF) 	{
							uint16 hashIdx;
							
							hashIdx=rtl8651_naptIcmpTableIndex(tempPtr->insideLocalIpAddr, localid, tempPtr->outsideGlobalIpAddr, &tblIdx);
							if(hashIdx==targetHashIdx)
								{
									tempPtr->outboundLocalId=localid;
									break;
								}
						}
						lastId = localid;
						if(pptpGreOutboundHeadPtr[tblIdx]) {
							uint32 offsetMask[2], i;
							
							trackPtr = pptpGreOutboundHeadPtr[tblIdx];
							offsetMask[0] = offsetMask[1] = 0;
							while(trackPtr) {
								assert(trackPtr->outboundIdInitiated);
								offsetMask[(trackPtr->outboundLocalId>>10)&0x1] |= (1<<((trackPtr->outboundLocalId>>11)&0x1F));
								trackPtr = trackPtr->outboundNext;
							}
							if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH: Server Inbound try ID offset start at %d\n", searchIdx);
							for(i=searchIdx; i!=((searchIdx+63)&0x3F); i=(i+1)&0x3f) {
								if((offsetMask[i&0x1] & (1<<((i>>1)&0x1F))) == 0)
									break;
							}
							searchIdx = (i+1)&0x3F;
							assert(i<64);
							tempPtr->outboundLocalId |= (i<<10);
							lastId = tempPtr->outboundLocalId;
						}
						
						if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH: Add outbound GID:%d LID:%d to table index %d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId, tblIdx);
						_rtl8651_addAsicPptpPassthroughOutboundEntry(tempPtr);
					}
					
					
					
					rtl8651_naptIcmpTableIndex(tempPtr->insideLocalIpAddr, tempPtr->outboundLocalId, tempPtr->outsideGlobalIpAddr, &tblIdx);
					tempPtr->outboundNext = pptpGreOutboundHeadPtr[tblIdx];
					pptpGreOutboundHeadPtr[tblIdx] = tempPtr;
				}
			}
			else {
				if(tempPtr->outboundIdClosed == 1 && pptpPassthroughServerInboundMsg)
					printk("PPTP-PTH: Closed Outbound ID GID:%d LID:%d refreshed to normal state\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
				tempPtr->outboundIdClosed = 0;
				tempPtr->controlIdleTimer = 0; 
			}
			cptr->cid1 = htons(tempPtr->outboundLocalId);
			if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH: Modify Server Inbound  Outcall Request from GID:%d to LID:%d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
			break;
		case PPTP_CallClearRequest:
			if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH PKT: Server Inbound CallClearRequest\n");
			if(!software)
			{
				
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->outboundGlobalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->outboundNext;
					}
				}
				if(tempPtr) {
#if 0					
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreOutboundHeadPtr[entryIdx] == tempPtr) {
						pptpGreOutboundHeadPtr[entryIdx] = tempPtr->outboundNext;
					} else {
						trackPtr = pptpGreOutboundHeadPtr[entryIdx];
						while(trackPtr && trackPtr->outboundNext != tempPtr)
							trackPtr = trackPtr->outboundNext;
						assert(trackPtr);
						trackPtr->outboundNext = tempPtr->outboundNext;
					}
					tempPtr->outboundIdInitiated = 0;
					printk("PPTP-PTH: Del outbound GIP:%d.%d.%d.%d LIP:%d. %d.%d.%d GID:%d LID:%d\n", 
						(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
						(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
						tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
					_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
					if(tempPtr->inboundIdInitiated)
						tempPtr->outboundNext = NULL;
					else {
						tempPtr->inboundNext = freePptpGreEntryHeadPtr;
						freePptpGreEntryHeadPtr = tempPtr;
					}
#endif
					if(tempPtr->outboundIdClosed == 1 && pptpPassthroughServerInboundMsg)
						printk("PPTP-PTH:(Call Clear Request) Outbound Closed ID GID:%d LID: %d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
					else if (pptpPassthroughServerInboundMsg)
						printk("PPTP-PTH:(Call Clear Request) Outbound ID GID: %d LID: %d Closed\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
					tempPtr->outboundIdClosed = 1;
					tempPtr->controlIdleTimer = 0; 
					cptr->cid1 = htons(tempPtr->outboundLocalId);
				}
			}			
			else
			{

				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->outboundGlobalId == ntohs(cptr->cid1))
					break;
				}
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
			
				if(tempPtr)
				{
					tempPtr->inboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
					tempPtr->outboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
				}
			}			
			break;
		case PPTP_StopCtrlConnRequest:
		case PPTP_StopCtrlConnReply:
			if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH PKT: Server Inbound StopCtrlConnRequest and StopCtrlConnReply\n");
			if(!software)
			{
				do {
					
					for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
						entryIdx = i;
						tempPtr = pptpGreOutboundHeadPtr[i];
						while(tempPtr) {
							if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && tempPtr->outboundIdClosed == 0)
								break;
							tempPtr = tempPtr->outboundNext;
						}
					}
					if(tempPtr) {
#if 0
						_rtl8651_pptpGreEntry_t * trackPtr;
						if(pptpGreOutboundHeadPtr[entryIdx] == tempPtr) {
							pptpGreOutboundHeadPtr[entryIdx] = tempPtr->outboundNext;
						} else {
							trackPtr = pptpGreOutboundHeadPtr[entryIdx];
							while(trackPtr && trackPtr->outboundNext != tempPtr)
								trackPtr = trackPtr->outboundNext;
							assert(trackPtr);
							trackPtr->outboundNext = tempPtr->outboundNext;
						}
						tempPtr->outboundIdInitiated = 0;
						printk("PPTP-PTH: Del outbound GIP:%d.%d.%d.%d LIP:%d. %d.%d.%d GID:%d LID:%d\n", 
							(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
							(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
							tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
						_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
						if(tempPtr->inboundIdInitiated)
							tempPtr->outboundNext = NULL;
						else {
							tempPtr->inboundNext = freePptpGreEntryHeadPtr;
							freePptpGreEntryHeadPtr = tempPtr;
						}
#endif
						if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH:(Stop Control Connection) Outbound ID GID: %d LID: %d Closed\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
						tempPtr->outboundIdClosed = 1;
						tempPtr->controlIdleTimer = 0; 
					}
				} while(tempPtr);
			}			
			break;
		case PPTP_WanErrorNotify:
		case PPTP_SetLinkInfo:

			if(!software)
			{
				
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {

						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->inboundGlobalId == ntohs(cptr->cid1))						
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
				
				if(tempPtr)
				{
					if(tempPtr->outboundIdClosed == 1 && pptpPassthroughServerInboundMsg)
						printk("PPTP-PTH: Closed Inbound ID GID:%d LID:%d packet keep send Call Clear Request\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					cptr->cid1=htons(tempPtr->inboundLocalId);
					tempPtr->controlIdleTimer = 0; 
				}					
			}


			break;
		case PPTP_InCallRequest:
		case PPTP_InCallReply:
		case PPTP_InCallConn: 
			
		default:
			return SUCCESS;
	}

	return SUCCESS;
}


int32 rtl8651_l4AliasHandlePptpClientInbound(struct rtl_pktHdr *pkt, struct ip *pip, 
					   struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s * link)	
{
	_rtl8651_pptpGreEntry_t * tempPtr=NULL, *trackPtr;
	PptpCallId cptr;
	uint16    ctl_type;	
	uint32 i, entryIdx;
	PptpCode  codes;
	uint8 software=0;
	static uint32 lastId=0;
	
	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;
	if((alg_supports[pkt->ph_dsid]&RTL8651_ALG_PPTP)==0) return FWDENG_USER_DROP;

	
	if ((cptr = _rtl8651_aliasVerifyPptp(pip, &ctl_type)) == NULL)
		return SUCCESS;
	
	switch (ctl_type)
	{
		case PPTP_OutCallReply:
			if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH PKT: Client Inbound OutCallReply\n");
			
			if(!software)
			{
				tempPtr = pptpGreInboundHeadPtr[ntohs(cptr->cid2) &((1<<RTL8651_ICMPTBL_BITS)-1)];
				while(tempPtr) {
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->inboundIdInitiated == 1 && tempPtr->inboundGlobalId == ntohs(cptr->cid2))
						break;
					tempPtr = tempPtr->inboundNext;
				}
			}
			else
			{
				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->inboundIdInitiated == 1 && tempPtr->inboundGlobalId == ntohs(cptr->cid2))
						break;
				}
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
			}

			
			if(tempPtr) {
				uint32 tblIdx=0;

				codes = (PptpCode) (cptr + 1);
				if(codes->resCode == 1) {
				
					cptr->cid2 = htons(tempPtr->inboundLocalId);
					if(tempPtr->outboundIdInitiated == 0) 
					{					
						if(pptpPassthroughClientInboundMsg)
							printk("PPTP-PH: Server from WAN OutCallReply SIP: %d.%d.%d.%d DIP: %d.%d.%d.%d ID %d\n",
								(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff,
								(tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
								(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff,
								(tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,
								tempPtr->inboundGlobalId);
						tempPtr->outboundGlobalId = ntohs(cptr->cid1);
						lastId = ((lastId +1)*(tempPtr->outboundGlobalId+1))&0xFFFF;
						if(!software)
						{
							uint32 localId, targetHashIdx, hashIdx, tempTblIdx;
							tblIdx = tempPtr->outboundGlobalId & ((1<<RTL8651_ICMPTBL_BITS)-1);
							targetHashIdx = tempPtr->outboundGlobalId & 0x3FF;
							for(localId = lastId; localId!=((lastId+65535)&0xFFFF); localId=(localId+1)&0xFFFF) {
								hashIdx = rtl8651_naptIcmpTableIndex(tempPtr->insideLocalIpAddr, localId, tempPtr->outsideGlobalIpAddr, &tempTblIdx);
								if(hashIdx == targetHashIdx && tempTblIdx == tblIdx) {
									trackPtr =  pptpGreOutboundHeadPtr[tblIdx];
									while(trackPtr) {
										if(trackPtr->outboundIdInitiated == 1 && trackPtr->outboundLocalId == localId)
											break;
										trackPtr = trackPtr->outboundNext;
									}
									if(trackPtr == NULL)
										break;
								}
							}
							lastId = localId;
							tempPtr->outboundLocalId =  localId;
							cptr->cid1 = htons(tempPtr->outboundLocalId);
							tempPtr->outboundIdInitiated = 1;
							tempPtr->inboundIdClosed = 0;
							tempPtr->controlIdleTimer = 0;
							if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH: Add outbound from GID:%d to LID:%d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
							_rtl8651_addAsicPptpPassthroughOutboundEntry(tempPtr);
						}
						else 
						{

							tempPtr->outboundLocalId =  tempPtr->outboundGlobalId;							
							tempPtr->outboundIdInitiated = 1;
						}




						if(!software)
						{
							tempPtr->outboundNext = pptpGreOutboundHeadPtr[tblIdx];
							pptpGreOutboundHeadPtr[tblIdx] = tempPtr;
						}
					}
					else {
						if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH: Established outbound from GID:%d to LID:%d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
						cptr->cid1 = htons(tempPtr->outboundLocalId);
						tempPtr->inboundIdClosed = 0;
						tempPtr->controlIdleTimer = 0;
					}
				}
				else 
				{

					if(software)
					{
						memset(tempPtr,0,sizeof(_rtl8651_pptpGreEntry_t));						
					}
					else 
					{
						
						for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
							entryIdx = i;
							tempPtr = pptpGreInboundHeadPtr[i];
							while(tempPtr) {
								if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
									tempPtr->inboundGlobalId == ntohs(cptr->cid2))
									break;
								tempPtr = tempPtr->inboundNext;
							}
						}
						if(tempPtr) 
						{
#if 0						
							_rtl8651_pptpGreEntry_t * trackPtr;
							if(pptpGreInboundHeadPtr[entryIdx] == tempPtr) {
								pptpGreInboundHeadPtr[entryIdx] = tempPtr->inboundNext;
							} else {
								trackPtr = pptpGreInboundHeadPtr[entryIdx];
								while(trackPtr && trackPtr->inboundNext != tempPtr)
									trackPtr = trackPtr->inboundNext;
								assert(trackPtr);
								trackPtr->inboundNext = tempPtr->inboundNext;
							}
							tempPtr->inboundIdInitiated = 0;
							
							_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
							if(tempPtr->outboundIdInitiated)
								tempPtr->inboundNext = NULL;
							else {
								tempPtr->inboundNext = freePptpGreEntryHeadPtr;
								freePptpGreEntryHeadPtr = tempPtr;
							}
#endif
							tempPtr->inboundIdClosed = 1;
							tempPtr->controlIdleTimer = 0;
						}
					}
				}
			}
			break;
		case PPTP_CallDiscNotify:	   
			if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH PKT: Client Inbound Call Disconnect Notify\n");
			if(!software)
			{
				
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->outboundGlobalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
				if(tempPtr) {
#if 0
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreOutboundHeadPtr[entryIdx] == tempPtr) {
						pptpGreOutboundHeadPtr[entryIdx] = tempPtr->outboundNext;
					} else {
						trackPtr = pptpGreOutboundHeadPtr[entryIdx];
						while(trackPtr && trackPtr->outboundNext != tempPtr)
							trackPtr = trackPtr->outboundNext;
						assert(trackPtr);
						trackPtr->outboundNext = tempPtr->outboundNext;
					}
					tempPtr->outboundIdInitiated = 0;
					
					
					
					
					_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
					if(tempPtr->inboundIdInitiated)
						tempPtr->outboundNext = NULL;
					else {
						tempPtr->inboundNext = freePptpGreEntryHeadPtr;
						freePptpGreEntryHeadPtr = tempPtr;
					}
#endif
					tempPtr->outboundIdClosed = 1;
					tempPtr->controlIdleTimer = 0;
					cptr->cid1 = htons(tempPtr->outboundLocalId);
				}
			}
			else
			{

			printk("PPTP_CallDiscNotify.....client in\n");
				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->outboundLocalId == ntohs(cptr->cid1))
						break;
				}

				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				if(tempPtr)
				{
				printk("kill this entry..........link->dstIpAddr=%x link->insideGlobalIpAddr=%x localid=%x\n", link->dstIpAddr, link->insideGlobalIpAddr ,ntohs(cptr->cid1));
					tempPtr->inboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
					tempPtr->outboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
				}				
			}
			break;
		case PPTP_StopCtrlConnRequest:
		case PPTP_StopCtrlConnReply:
			if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH PKT: Client Inbound Stop Control Connection\n");
			if(!software)
			{
				do {
					for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
						entryIdx = i;
						tempPtr = pptpGreOutboundHeadPtr[i];
						while(tempPtr) {
							if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
								tempPtr->outboundIdClosed == 0)
								break;
							tempPtr = tempPtr->inboundNext;
						}
					}
					if(tempPtr) {
						tempPtr->outboundIdClosed = 1;
						tempPtr->controlIdleTimer = 0;
						cptr->cid1 = htons(tempPtr->outboundLocalId);
					}
				} while (tempPtr);
			}
			break;
		case PPTP_WanErrorNotify:
		case PPTP_SetLinkInfo:
			if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH PKT: Client Inbound SetLinkInfo\n");
			
			if(!software)
			{
				
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->inboundGlobalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
				tempPtr->controlIdleTimer = 0;
				if(tempPtr) cptr->cid1=htons(tempPtr->inboundLocalId);
			}
			break;			

		case PPTP_InCallRequest:
		case PPTP_InCallReply:
		case PPTP_InCallConn: 
			
		default:
			return SUCCESS;
	}

	return SUCCESS;
}



int32 rtl8651_l4AliasHandlePptpServerOutbound(
		struct rtl_pktHdr *pkt,
		struct ip *pip,
		struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link){
		
	_rtl8651_pptpGreEntry_t * tempPtr=NULL, *trackPtr;
	PptpCallId cptr;
	uint16    ctl_type;	
	uint32 i, entryIdx , tblIdx;
	PptpCode  codes;
	uint8 software=0;
	static uint32 searchIdx = 0;

	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;

	if((alg_supports[pkt->ph_dsid]&RTL8651_ALG_PPTP_SERVER)==0) return FWDENG_USER_DROP;
		

	
	if ((cptr = _rtl8651_aliasVerifyPptp(pip, &ctl_type)) == NULL)
		return SUCCESS;
	
	switch (ctl_type)
	{
		case PPTP_OutCallReply:
			if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH PKT: Server Outbound OutCallReply\n");
			if(!software)
			{
				
				
				rtl8651_naptIcmpTableIndex(link->insideLocalIpAddr, ntohs(cptr->cid2), link->dstIpAddr, &tblIdx);
				tempPtr = pptpGreOutboundHeadPtr[tblIdx];			
						
				

				while(tempPtr) {

					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->outboundIdInitiated == 1 && tempPtr->outboundLocalId == ntohs(cptr->cid2))
						break;
					tempPtr = tempPtr->outboundNext;
				}
			}
			else
			{
				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->outboundIdInitiated == 1 && tempPtr->outboundLocalId == ntohs(cptr->cid2))
						break;
				}
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				
			}
			
			if(tempPtr) {
				uint32 tblIdx;

				if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH: Exist outbound GID:%d LID:%d inboundInitiated %d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId, tempPtr->inboundIdInitiated);
				codes = (PptpCode) (cptr + 1);
				if(codes->resCode == 1) {
			
					cptr->cid2 = htons(tempPtr->outboundGlobalId);


					if(tempPtr->inboundIdInitiated == 0) 
					{
						tempPtr->inboundLocalId = ntohs(cptr->cid1);
						if(pptpPassthroughServerOutboundMsg)
							printk("PPTP-PH: Server from LAN reply DIP:%d.%d.%d.%d SIP:%d.%d.%d.%d ID:%d\n",
								(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff,
								(tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
								(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff,
								(tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff,
								tempPtr->inboundLocalId);
						if(!software)
						{


							tempPtr->inboundGlobalId = rtl8651_naptIcmpTableIndex(tempPtr->insideLocalIpAddr, tempPtr->inboundLocalId, tempPtr->outsideGlobalIpAddr, &tblIdx);
							
							if(pptpGreInboundHeadPtr[tblIdx]) {
								uint32 offsetMask[2], i;
								
								trackPtr = pptpGreInboundHeadPtr[tblIdx];
								offsetMask[0] = offsetMask[1] = 0;
								while(trackPtr) {
									assert(trackPtr->inboundIdInitiated);
									offsetMask[(trackPtr->inboundGlobalId>>10)&0x1] |= (1<<((trackPtr->inboundGlobalId>>11)&0x1F));
									trackPtr = trackPtr->inboundNext;
								}
								if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH: Server Inbound try ID offset start at %d\n", searchIdx);
								for(i=searchIdx; i!=((searchIdx+63)&0x3F); i=(i+1)&0x3f) {
									if((offsetMask[i&0x1] & (1<<((i>>1)&0x1F))) == 0)
										break;
								}
								searchIdx = (i+1)&0x3F;
								assert(i<64);
								tempPtr->inboundGlobalId |= (i<<10);
							}

							cptr->cid1 = htons(tempPtr->inboundGlobalId);
							if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH: Modify inbound from LID:%d to GID:%d inseart to table idx: %d\n", tempPtr->inboundLocalId, tempPtr->inboundGlobalId, tblIdx); 
							_rtl8651_addAsicPptpPassthroughInboundEntry(tempPtr);
							tempPtr->inboundNext = pptpGreInboundHeadPtr[tblIdx];
							pptpGreInboundHeadPtr[tblIdx] = tempPtr;
							tempPtr->inboundIdInitiated = 1;
							tempPtr->inboundIdClosed = 0;
							tempPtr->controlIdleTimer = 0;
						}
						else {
							tempPtr->inboundGlobalId =  tempPtr->inboundLocalId;
							tempPtr->inboundIdInitiated = 1;
						}
					}
					else {
						cptr->cid1 = htons(tempPtr->inboundGlobalId);
						tempPtr->inboundIdClosed = 0;
						tempPtr->controlIdleTimer = 0;
					}
				}
				else {

					if(software)
					{
						memset(tempPtr,0,sizeof(_rtl8651_pptpGreEntry_t));						
					}
					else 
					{
						
						for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
							entryIdx = i;
							tempPtr = pptpGreOutboundHeadPtr[i];
							while(tempPtr) {
								if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
									tempPtr->outboundGlobalId == ntohs(cptr->cid2))
									break;
								tempPtr = tempPtr->outboundNext;
							}
						}
					
					
						if(tempPtr) {
#if 0
							_rtl8651_pptpGreEntry_t * trackPtr;
							if(pptpGreOutboundHeadPtr[entryIdx] == tempPtr) {
								pptpGreOutboundHeadPtr[entryIdx] = tempPtr->outboundNext;
							} else {
								trackPtr = pptpGreOutboundHeadPtr[entryIdx];
								while(trackPtr && trackPtr->outboundNext != tempPtr)
									trackPtr = trackPtr->outboundNext;
								assert(trackPtr);
								trackPtr->outboundNext = tempPtr->outboundNext;
							}
							tempPtr->outboundIdInitiated = 0;

							_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
							if(tempPtr->inboundIdInitiated)
								tempPtr->outboundNext = NULL;
							else {
								tempPtr->inboundNext = freePptpGreEntryHeadPtr;
								freePptpGreEntryHeadPtr = tempPtr;
							}
#endif							
							tempPtr->outboundIdClosed = 1;
							tempPtr->controlIdleTimer = 0;
						}
					}
				}
			}
			break;
		case PPTP_CallDiscNotify:	   
			if(pptpPassthroughServerOutboundMsg)  printk("PPTP-PTH PKT: Server Outbound CallDiscNotify\n");
			if(!software)
			{
				
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreInboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->inboundLocalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->outboundNext;
					}
				}
				if(tempPtr) {
#if 0					
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreInboundHeadPtr[entryIdx] == tempPtr) {
						pptpGreInboundHeadPtr[entryIdx] = tempPtr->inboundNext;
					} else {
						trackPtr = pptpGreInboundHeadPtr[entryIdx];
						while(trackPtr && trackPtr->inboundNext != tempPtr)
							trackPtr = trackPtr->inboundNext;
						assert(trackPtr);
						trackPtr->inboundNext = tempPtr->inboundNext;
					}
					tempPtr->inboundIdInitiated = 0;
					printk("PPTP-PTH: Del inbound GIP:%d.%d.%d.%d LIP:%d.%d.%d.%d GID:%d LID:%d\n", 
						(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
						(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
						tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
					if(tempPtr->outboundIdInitiated)
						tempPtr->inboundNext = NULL;
					else {
						tempPtr->inboundNext = freePptpGreEntryHeadPtr;
						freePptpGreEntryHeadPtr = tempPtr;
					}
#endif
					if(tempPtr->inboundIdClosed == 1 && pptpPassthroughServerOutboundMsg)
						printk("PPTP-PTH:(Call Disconnect Notify) Inbound closed session GID: %d LID: %d\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					else if (pptpPassthroughServerOutboundMsg)
						printk("PPTP-PTH:(Call Disconnect Notify) Inbound session GID: %d LID: %d closed\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					tempPtr->inboundIdClosed = 1;
					tempPtr->controlIdleTimer = 0;
					
					cptr->cid1 = htons(tempPtr->inboundGlobalId);
				}
			}
			else
			{
				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->inboundLocalId == ntohs(cptr->cid1))
					break;
				}
				
				if(tempPtr)
				{
					tempPtr->inboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
					tempPtr->outboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
				}
			}
			break;
		case PPTP_StopCtrlConnRequest:
		case PPTP_StopCtrlConnReply:
			if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH PKT: Server Outbound StopCtrlConnRequest and StopCtrlConnReply\n");
			if(!software)
			{
				do {
					for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
						entryIdx = i;
						tempPtr = pptpGreInboundHeadPtr[i];
						while(tempPtr) {
							if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && tempPtr->inboundIdClosed == 0)
								break;
							tempPtr = tempPtr->outboundNext;
						}
					}
					if(tempPtr) {
#if 0						
						_rtl8651_pptpGreEntry_t * trackPtr;
						if(pptpGreInboundHeadPtr[entryIdx] == tempPtr) {
							pptpGreInboundHeadPtr[entryIdx] = tempPtr->inboundNext;
						} else {
							trackPtr = pptpGreInboundHeadPtr[entryIdx];
							while(trackPtr && trackPtr->inboundNext != tempPtr)
								trackPtr = trackPtr->inboundNext;
							assert(trackPtr);
							trackPtr->inboundNext = tempPtr->inboundNext;
						}
						tempPtr->inboundIdInitiated = 0;
						printk("PPTP-PTH: Del inbound GIP:%d.%d.%d.%d LIP:%d.%d.%d.%d GID:%d LID:%d\n", 
							(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
							(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
							tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
						_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
						if(tempPtr->outboundIdInitiated)
							tempPtr->inboundNext = NULL;
						else {
							tempPtr->inboundNext = freePptpGreEntryHeadPtr;
							freePptpGreEntryHeadPtr = tempPtr;
						}
#endif
						if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH:(Stop Control Connection) Inbound session GID: %d LID: %d closed\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
						tempPtr->inboundIdClosed = 1;
						tempPtr->controlIdleTimer = 0;
					}
				} while(tempPtr);
			}
			break;
		case PPTP_WanErrorNotify:
		case PPTP_SetLinkInfo:
			if(!software)
			{
				
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {

						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->outboundLocalId == ntohs(cptr->cid1))						
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
				
				if(tempPtr)
				{
					cptr->cid1=htons(tempPtr->outboundGlobalId);
				}					
			}
			break;
			
		case PPTP_InCallRequest:
		case PPTP_InCallReply:
		case PPTP_InCallConn: 
			
		default:
			return SUCCESS;
	}

	return SUCCESS;
}

void rtl8651_l4AliasPptpReInit(void)
{
	reInit=1;
	pptpPassthroughSysPara.idelTimeout=0;
	rtl8651_l4AliasPptpHouseKeeping();
	pptpPassthroughSysPara.idelTimeout = 1800; 
}




static int algPptpInited=0;
int32 rtl8651_l4AliasPptpInit(void *arg){
	_rtl8651_pptpGreEntry_t * pptpPassThroughEntryPtr;
	uint32 i;

	if(algPptpInited) return SUCCESS;

	algPptpInited=1;
	for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
		pptpGreInboundHeadPtr[i] = NULL;
		pptpGreOutboundHeadPtr[i] = NULL;
	}

	freePptpGreEntryHeadPtr = NULL;

	for(i=0; i<algMaxSession.pptpMaxSession; i++) {
		pptpPassThroughEntryPtr = (_rtl8651_pptpGreEntry_t *) rtlglue_malloc(sizeof(_rtl8651_pptpGreEntry_t));
		
		if(pptpPassThroughEntryPtr) {
			pptpPassThroughEntryPtr->inboundNext = freePptpGreEntryHeadPtr;
			freePptpGreEntryHeadPtr = pptpPassThroughEntryPtr;
		}
	}

	if(softwarePptpGreEntryHeadPtr!=NULL) rtlglue_free(softwarePptpGreEntryHeadPtr);
	softwarePptpGreEntryHeadPtr = (_rtl8651_pptpGreEntry_t *) rtlglue_malloc(sizeof(_rtl8651_pptpGreEntry_t)*algMaxSession.pptpMaxSession);
	memset(softwarePptpGreEntryHeadPtr,0,sizeof(_rtl8651_pptpGreEntry_t)*algMaxSession.pptpMaxSession);
	
	pptpPassthroughSysPara.heavyLoadSwappingThreshold = 100;
	pptpPassthroughSysPara.idelTimeout = 1800; 
	rtl8651_l4GreNaptAliasIn = _rtl8651_l4GreNaptAliasIn;
	rtl8651_l4GreNaptAliasOut = _rtl8651_l4GreNaptAliasOut;
	return SUCCESS;
}



void  rtl8651_l4AliasPptpHouseKeeping(){ 
	_rtl8651_pptpGreEntry_t * trackEntryPtr, * tempPtr, *nextPtr;
	uint32 i,j;
	uint8 software=0;

	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;

	
	
	if(software)
	{
		tempPtr=softwarePptpGreEntryHeadPtr;
		for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
		{
			if((tempPtr->inboundIdInitiated==1)&&(tempPtr->outboundIdInitiated==1))
			{				
				tempPtr->inboundIdleTimer++;
				tempPtr->outboundIdleTimer++;
				if((tempPtr->inboundIdleTimer >= pptpPassthroughSysPara.idelTimeout) &&
					(tempPtr->outboundIdleTimer >= pptpPassthroughSysPara.idelTimeout))
				{
					memset(tempPtr,0,sizeof(_rtl8651_pptpGreEntry_t));					
				}

				for(j=0;j<RTL8651_MAX_DIALSESSION;j++)
				{
					if((alg_supports[j]&RTL8651_ALG_PPTP)==0)
					{
						if(tempPtr->dsid==j)
						{
							memset(tempPtr,0,sizeof(_rtl8651_pptpGreEntry_t));							
						}
					}
				}
				
			}
		}
	}
	else
	{
		
		for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
			
			tempPtr = pptpGreInboundHeadPtr[i];
			while(tempPtr) {

				
				uint8 disabled=FALSE;
				assert(tempPtr->inboundIdInitiated);
				tempPtr->inboundPktCnt = 0;

				for(j=0;j<RTL8651_MAX_DIALSESSION;j++)
				{
					if((alg_supports[j]&RTL8651_ALG_PPTP)==0)
					{
						if(tempPtr->dsid==j)
						{
							disabled=TRUE;
							break;
						}
					}
				}

				if((tempPtr->inboundToAsic==1)&&(!disabled)&&(reInit==0))
                                       tempPtr = tempPtr->inboundNext;
                               else				
				{
					reInit=0;
					if(tempPtr->inboundToAsic==1)
					{
						
						tempPtr->inboundIdInitiated = 0;
						_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
					}
					
					tempPtr->inboundIdleTimer++;	
					if((tempPtr->inboundIdleTimer >= pptpPassthroughSysPara.idelTimeout)||(disabled)) {
						
						nextPtr = tempPtr->inboundNext;
						if(pptpGreInboundHeadPtr[i] == tempPtr) {
							pptpGreInboundHeadPtr[i] = tempPtr->inboundNext;
						} else {
							trackEntryPtr = pptpGreInboundHeadPtr[i];
							while(trackEntryPtr) {
								if(trackEntryPtr->inboundNext == tempPtr) {
									trackEntryPtr->inboundNext = tempPtr->inboundNext;
									break;
								}
								trackEntryPtr = trackEntryPtr->inboundNext;
							}
						}
						tempPtr->inboundIdInitiated = 0;
						if(tempPtr->outboundIdInitiated == 0) {
							tempPtr->inboundNext = freePptpGreEntryHeadPtr;
							freePptpGreEntryHeadPtr = tempPtr;
						}
						tempPtr = nextPtr;
					} 
					else 
						tempPtr = tempPtr->inboundNext;
				}
			}
			
			tempPtr = pptpGreOutboundHeadPtr[i];
			while(tempPtr) {
				uint8 disabled=FALSE;
				assert(tempPtr->outboundIdInitiated);
				tempPtr->outboundPktCnt = 0;

				for(j=0;j<RTL8651_MAX_DIALSESSION;j++)
				{
					if((alg_supports[j]&RTL8651_ALG_PPTP)==0)
					{
						if(tempPtr->dsid==j)
						{
							disabled=TRUE;
							break;
						}
					}
				}				
				
			      if((tempPtr->outboundToAsic == 1) &&(!disabled)&&(reInit==0))
                                     tempPtr = tempPtr->outboundNext;
                              else 
				{
					reInit=0;
					if(tempPtr->outboundToAsic == 1)
					{
						
						tempPtr->outboundIdInitiated = 0;						
						_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
					}
					tempPtr->outboundIdleTimer++;					
					
					if((tempPtr->outboundIdleTimer >= pptpPassthroughSysPara.idelTimeout)||(disabled)) {
						
						nextPtr = tempPtr->outboundNext;
						if(pptpGreOutboundHeadPtr[i] == tempPtr) {
							pptpGreOutboundHeadPtr[i] = tempPtr->outboundNext;
						} else {
							trackEntryPtr = pptpGreOutboundHeadPtr[i];
							while(trackEntryPtr) {
								if(trackEntryPtr->outboundNext == tempPtr) {
									trackEntryPtr->outboundNext = tempPtr->outboundNext;
									break;
								}
								trackEntryPtr = trackEntryPtr->outboundNext;
							}
						}
						tempPtr->outboundIdInitiated = 0;
						if(tempPtr->inboundIdInitiated == 0) {
												
							tempPtr->inboundNext = freePptpGreEntryHeadPtr;
							freePptpGreEntryHeadPtr = tempPtr;
						}
						tempPtr = nextPtr;
					} 
					else 
						tempPtr = tempPtr->outboundNext;
				}
			}
			
			tempPtr = pptpGreInboundHeadPtr[i];
			while(tempPtr) {
				uint8 inboundNotOpen, outboundNotOpen;
				inboundNotOpen = (tempPtr->inboundIdClosed || !tempPtr->inboundIdInitiated )? 1: 0;
				outboundNotOpen = (tempPtr->outboundIdClosed  || !tempPtr->outboundIdInitiated )? 1: 0;
				if( inboundNotOpen || outboundNotOpen)
					tempPtr->controlIdleTimer++;
				if(	((inboundNotOpen && outboundNotOpen) && tempPtr->controlIdleTimer >= pptpPassthroughControlShortTimeout) ||
					((inboundNotOpen || outboundNotOpen) &&tempPtr->controlIdleTimer >= pptpPassthroughControlShortTimeout)) {
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreInboundHeadPtr[i] == tempPtr) {
						pptpGreInboundHeadPtr[i] = tempPtr->inboundNext;
					} else {
						trackPtr = pptpGreInboundHeadPtr[i];
						while(trackPtr && trackPtr->inboundNext != tempPtr)
							trackPtr = trackPtr->inboundNext;
						assert(trackPtr);
						trackPtr->inboundNext = tempPtr->inboundNext;
					}
					tempPtr->inboundIdInitiated = 0;
					if(pptpPassthroughClientInboundMsg ||  pptpPassthroughServerOutboundMsg)
						printk("PPTP-PTH: Del inbound GIP:%d.%d.%d.%d LIP:%d.%d.%d.%d GID:%d LID:%d\n", 
							(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
							(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
							tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
					if(tempPtr->outboundIdInitiated)
						tempPtr->inboundNext = NULL;
					else {
						tempPtr->inboundNext = freePptpGreEntryHeadPtr;
						freePptpGreEntryHeadPtr = tempPtr;
					}
					tempPtr = pptpGreInboundHeadPtr[i];
				}
				else 
					tempPtr = tempPtr->inboundNext;
			}
			
			if(pptpGreInboundHeadPtr[i])
				_rtl8651_addAsicPptpPassthroughInboundEntry(pptpGreInboundHeadPtr[i]);
			
			tempPtr = pptpGreOutboundHeadPtr[i];
			while(tempPtr) {
				uint8 inboundNotOpen, outboundNotOpen;
				inboundNotOpen = (tempPtr->inboundIdClosed == 1 || tempPtr->inboundIdInitiated == 0)? 1: 0;
				outboundNotOpen = (tempPtr->outboundIdClosed == 1 || tempPtr->outboundIdInitiated == 0)? 1: 0;
				if( !tempPtr->inboundIdInitiated)
					tempPtr->controlIdleTimer++;
				
				if(	((inboundNotOpen && outboundNotOpen) && tempPtr->controlIdleTimer >= pptpPassthroughControlShortTimeout) ||
					((inboundNotOpen || outboundNotOpen) &&tempPtr->controlIdleTimer >= pptpPassthroughControlShortTimeout)) {
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreOutboundHeadPtr[i] == tempPtr) {
						pptpGreOutboundHeadPtr[i] = tempPtr->outboundNext;
					} else {
						trackPtr = pptpGreOutboundHeadPtr[i];
						while(trackPtr && trackPtr->outboundNext != tempPtr)
							trackPtr = trackPtr->outboundNext;
						assert(trackPtr);
						trackPtr->outboundNext = tempPtr->outboundNext;
					}
					tempPtr->outboundIdInitiated = 0;
					if(pptpPassthroughClientOutboundMsg || pptpPassthroughServerInboundMsg)
						printk("PPTP-PTH: Del outbound GIP:%d.%d.%d.%d LIP:%d. %d.%d.%d GID:%d LID:%d\n", 
							(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
							(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
							tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
					_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
					assert(tempPtr->inboundIdInitiated==0);
					tempPtr->inboundNext = freePptpGreEntryHeadPtr;
					freePptpGreEntryHeadPtr = tempPtr;
					tempPtr = pptpGreOutboundHeadPtr[i];
				}
				else
					tempPtr = tempPtr->outboundNext;
			}
			
			if(pptpGreOutboundHeadPtr[i])
				_rtl8651_addAsicPptpPassthroughInboundEntry(pptpGreOutboundHeadPtr[i]);

		}
	}

}


void rtl8651_l4AliasPptpMsg( int8 clientInbound, int8 clientOutbound, int8 serverInbound, int8 serverOutbound, int8 greInbound, int8 greOutbound) {
	if(clientInbound>=0)
		pptpPassthroughClientInboundMsg = clientInbound==0?0:1;
	if(clientOutbound>=0)
		pptpPassthroughClientOutboundMsg = clientOutbound==0?0:1;
	if(serverInbound>=0)
		pptpPassthroughServerInboundMsg = serverInbound==0?0:1;
	if(serverOutbound>=0)
		pptpPassthroughServerOutboundMsg = serverOutbound==0?0:1;
	if(greInbound>=0)
		pptpPassthroughGreInboundMsg = greInbound==0?0:1;
	if(greOutbound>=0)
		pptpPassthroughGreOutboundMsg = greOutbound==0?0:1;	
}

void rtl8651_l4AliasPptpDump(void) {
	_rtl8651_pptpGreEntry_t * tempPtr;
	uint32 i;
	uint8 software=0;

	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;

	if(software) {
		if(softwarePptpGreEntryHeadPtr!=NULL)
		{
			printk("PPTP pure software current session information\n");

			tempPtr=softwarePptpGreEntryHeadPtr;
			for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
			{
				if(tempPtr->outboundIdInitiated&&tempPtr->inboundIdInitiated)
				{
					printk("outsideGlobal:%d.%d.%d.%d:%d insideGlobal:%d.%d.%d.%d:%d  outsideLocal:%d insideLocal %d.%d.%d.%d:%d Age:%u inPkt:%u outPkt:%u\n",
						(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,tempPtr->outboundGlobalId
						,(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,tempPtr->inboundGlobalId,tempPtr->outboundLocalId
						,(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff,tempPtr->inboundLocalId
						, tempPtr->inboundIdleTimer, tempPtr->inboundPktCnt, tempPtr->outboundPktCnt);
				}
			}	
		}
	}
	else {
		printk("PPTP Passthrough Software session information\n");
		printk("Inbound session\n");
		for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
			tempPtr = pptpGreInboundHeadPtr[i];
			if(tempPtr)
				printk("[%02d]:\n", i);
			while(tempPtr) {
				if(!tempPtr->inboundToAsic)
					printk("Src %d.%d.%d.%d Dst %d.%d.%d.%d GID:%d LID:%d %s\n\tAge:%u Pkt:%u Inbound:%c%c Outbound:%c%c CAge: %u\n",
						(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
						(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,
						tempPtr->inboundGlobalId, tempPtr->inboundLocalId, 
						tempPtr->inboundToAsic?"ASIC":"", tempPtr->inboundIdleTimer, tempPtr->inboundPktCnt,
						tempPtr->inboundIdInitiated?'I':' ', tempPtr->inboundIdClosed?'C':' ', 
						tempPtr->outboundIdInitiated?'I':' ', tempPtr->outboundIdClosed?'C':' ', 
						tempPtr->controlIdleTimer);
				tempPtr = tempPtr->inboundNext;
			}
		}
		printk("Outbound session\n");
		for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
			tempPtr = pptpGreOutboundHeadPtr[i];
			if(tempPtr)
				printk("[%02d]:\n", i);
			while(tempPtr) {
				if(!tempPtr->inboundToAsic)
					printk("Src %d.%d.%d.%d Dst %d.%d.%d.%d LID %d GID %d %s\n\tAge:%u Pkt:%u Inbound:%c%c Outbound:%c%cCAge: %u\n",
						(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff,
						(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
						tempPtr->outboundLocalId, tempPtr->outboundGlobalId, 
						tempPtr->outboundToAsic?"ASIC":"", tempPtr->outboundIdleTimer, tempPtr->outboundPktCnt,
						tempPtr->inboundIdInitiated?'I':' ', tempPtr->inboundIdClosed?'C':' ', 
						tempPtr->outboundIdInitiated?'I':' ', tempPtr->outboundIdClosed?'C':' ', 
						tempPtr->controlIdleTimer);
				tempPtr = tempPtr->outboundNext;
			}
		}
		printk("PPTP Pass-through current session information\n");
		printk("Inbound session\n");
		for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
			tempPtr = pptpGreInboundHeadPtr[i];
			if(tempPtr)
				printk("[%02d]:\n", i);
			while(tempPtr) {
				if(tempPtr->inboundToAsic)
					printk("Src %d.%d.%d.%d Dst %d.%d.%d.%d GID:%d LID:%d %s\n\tAge:%u Pkt:%u Inbound:%c%c Outbound:%c%c CAge: %u\n",
						(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
						(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,
						tempPtr->inboundGlobalId, tempPtr->inboundLocalId, 
						tempPtr->inboundToAsic?"ASIC":"", tempPtr->inboundIdleTimer, tempPtr->inboundPktCnt,
						tempPtr->inboundIdInitiated?'I':' ', tempPtr->inboundIdClosed?'C':' ', 
						tempPtr->outboundIdInitiated?'I':' ', tempPtr->outboundIdClosed?'C':' ', 
						tempPtr->controlIdleTimer);

				tempPtr = tempPtr->inboundNext;
			}
		}
		printk("Outbound session\n");
		for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
			tempPtr = pptpGreOutboundHeadPtr[i];
			if(tempPtr)
				printk("[%02d]:\n", i);
			while(tempPtr) {
				if(tempPtr->inboundToAsic)
					printk("Src %d.%d.%d.%d Dst %d.%d.%d.%d LID %d GID %d %s\n\tAge:%u Pkt:%u Inbound:%c%c Outbound:%c%cCAge: %u\n",
						(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff,
						(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
						tempPtr->outboundLocalId, tempPtr->outboundGlobalId, 
						tempPtr->outboundToAsic?"ASIC":"", tempPtr->outboundIdleTimer, tempPtr->outboundPktCnt,
						tempPtr->inboundIdInitiated?'I':' ', tempPtr->inboundIdClosed?'C':' ', 
						tempPtr->outboundIdInitiated?'I':' ', tempPtr->outboundIdClosed?'C':' ', 
						tempPtr->controlIdleTimer);
				tempPtr = tempPtr->outboundNext;
			}
		}
		tempPtr = freePptpGreEntryHeadPtr;
		i = 0;
		while(tempPtr) {
			i++;
			tempPtr = tempPtr->inboundNext;
		}
		printk("Total %u free entry in the free list\n\n", i);
	}
}


