
#include "rtl_types.h"
#include "rtl_queue.h"
#include "rtl8651_alg_qos.h"

static uint32 _algQosTblSize = RTL8651_ALG_QOS_DEFAULT_TBL_SIZE;


typedef struct rtl8651_alg_qos_entry_s {
	rtl8651_alg_qos_t algQos;		
	int8 inUse;					
	CLIST_ENTRY( rtl8651_alg_qos_entry_s ) link;
} rtl8651_alg_qos_entry_t;


uint8 rtl8651_algQos_isInit = FALSE;


static rtl8651_alg_qos_entry_t *_algQosBuf = NULL;


static CLIST_HEAD( , rtl8651_alg_qos_entry_s ) _algQosTbl = {NULL, 0};


static CLIST_HEAD( , rtl8651_alg_qos_entry_s ) _algQosFree = {NULL, 0};





static int32 _rtl8651_algQos_compareEntry( uint8 isTCP, uint32 srcIP, uint16 srcPort, uint32 extIP, uint16 extPort, 
								 uint32 dstIP, uint16 dstPort, rtl8651_alg_qos_t *entry );



int32 _rtl8651_algQos_init( uint32 tblSize )
{
	rtl8651_alg_qos_entry_t *entry;
	int n;

	
	if (tblSize == 0)
		_algQosTblSize = RTL8651_ALG_QOS_DEFAULT_TBL_SIZE;
	else
		_algQosTblSize = tblSize;

	
	CLIST_INIT( &_algQosTbl );

	
	if ( (_algQosBuf = (rtl8651_alg_qos_entry_t *)rtlglue_malloc( _algQosTblSize * sizeof(rtl8651_alg_qos_entry_t) )) == NULL )
	{
		printk( "Failed to allocate %d bytes memory in file %s, line %d\n", 
			_algQosTblSize * sizeof(rtl8651_alg_qos_entry_t), __FILE__, __LINE__ );
		return FAILED;
	}

	
	memset( _algQosBuf, 0, sizeof(rtl8651_alg_qos_entry_t) * _algQosTblSize );

	
	CLIST_INIT( &_algQosFree );
	entry = (rtl8651_alg_qos_entry_t *)_algQosBuf;
	for (n = 0; n < _algQosTblSize; n++)
	{
		
		CLIST_INSERT_HEAD( &_algQosFree, entry, link );
		entry++;
	}

	rtl8651_algQos_isInit = TRUE;
	
	return SUCCESS;
}



int32 _rtl8651_algQos_destroy( void )
{
	if (rtl8651_algQos_isInit != TRUE)
		return FAILED;

	
	CLIST_INIT( &_algQosTbl );
	CLIST_INIT( &_algQosFree );
		
	
	rtlglue_free( _algQosBuf );
	_algQosBuf = NULL;

	rtl8651_algQos_isInit = FALSE;
	
	return SUCCESS;
}





int32 _rtl8651_algQos_addEntry( rtl8651_alg_qos_t *qosEntry )
{
	rtl8651_alg_qos_entry_t *entry;
	
	if (rtl8651_algQos_isInit != TRUE)
		return FAILED;	

	
	if (qosEntry == NULL)
		return FAILED;
	
	
	CLIST_FOREACH( entry, &_algQosTbl, link )
	{
		
		if ((qosEntry->algType == entry->algQos.algType) && _rtl8651_algQos_compareEntry( qosEntry->isTCP, qosEntry->srcIP, qosEntry->srcPort, qosEntry->extIP, qosEntry->extPort, qosEntry->dstIP, qosEntry->dstPort, &entry->algQos ))
		{
			
			memcpy( &entry->algQos, qosEntry, sizeof(rtl8651_alg_qos_t) );
			return SUCCESS;
		}
	}

	

	
	if ((entry = CLIST_FIRST( &_algQosFree )) == NULL)
		return FAILED;

	
	CLIST_REMOVE( &_algQosFree, entry, link );

	
	memcpy( &entry->algQos, qosEntry, sizeof(rtl8651_alg_qos_t) );
	entry->inUse = 1;

	
	CLIST_INSERT_HEAD( &_algQosTbl, entry, link );

	return SUCCESS;
}



int32 _rtl8651_algQos_match( uint8 isTCP, uint32 srcIP, uint16 srcPort, uint32 extIP, uint16 extPort, 
								 uint32 dstIP, uint16 dstPort, rtl8651_alg_qos_t *algQos )
{
	rtl8651_alg_qos_entry_t *entry;

	if (rtl8651_algQos_isInit != TRUE)
		return FAILED;	

	
	if (algQos == NULL) 
		return FAILED;

	
	if ((entry = CLIST_FIRST( &_algQosTbl )) == NULL)
		return FAILED;

	
	CLIST_FOREACH( entry, &_algQosTbl, link )
	{
		if (_rtl8651_algQos_compareEntry( isTCP, srcIP, srcPort, extIP, extPort, dstIP, dstPort, &entry->algQos ) == TRUE)
		{
			
			memcpy( algQos, &entry->algQos, sizeof(rtl8651_alg_qos_t) );
			return SUCCESS;
		}
	}

	return FAILED;
}



int32 _rtl8651_algQos_matchAndDelete( uint8 isTCP, uint32 srcIP, uint16 srcPort, uint32 extIP, uint16 extPort, 
								 uint32 dstIP, uint16 dstPort, rtl8651_alg_qos_t *algQos )
{
	rtl8651_alg_qos_entry_t *entry;

	if (rtl8651_algQos_isInit != TRUE)
		return FAILED;	

	
	if (algQos == NULL) 
		return FAILED;

	
	if ((entry = CLIST_FIRST( &_algQosTbl )) == NULL)
		return FAILED;

	
	CLIST_FOREACH( entry, &_algQosTbl, link )
	{
		if (_rtl8651_algQos_compareEntry( isTCP, srcIP, srcPort, extIP, extPort, dstIP, dstPort, &entry->algQos ) == TRUE)
		{
			
			memcpy( algQos, &entry->algQos, sizeof(rtl8651_alg_qos_t) );

			
			if (entry->algQos.isOneShot == 1)
			{
				CLIST_REMOVE( &_algQosTbl, entry, link );
				memset( entry, 0, sizeof(rtl8651_alg_qos_entry_t) );
				CLIST_INSERT_HEAD( &_algQosFree, entry, link );
			}

			return SUCCESS;
		}
	}

	return FAILED;
}



int32 _rtl8651_algQos_timeUpdate( uint32 secPassed )
{
	rtl8651_alg_qos_entry_t *entry = _algQosBuf;
	int n;

	if (rtl8651_algQos_isInit == FALSE)
		return FAILED;

	
	if ( secPassed <= 0 )
		return FAILED;

	
	for (n = 0; n < _algQosTblSize; n++)
	{
		
		if (entry->inUse == 1) 
		{
			
			if (entry->algQos.liveTime <= secPassed)
			{
				CLIST_REMOVE( &_algQosTbl, entry, link );
				memset( entry, 0, sizeof(rtl8651_alg_qos_entry_t) );
				CLIST_INSERT_HEAD( &_algQosFree, entry, link );
			}
			else
			{
				
				entry->algQos.liveTime -= secPassed;
			}
		}

		entry++;
	}

	return SUCCESS;
}



int32 _rtl8651_algQos_flush( void )
{
	rtl8651_alg_qos_entry_t *entry;
	int n;

	
	if (rtl8651_algQos_isInit == FALSE)
		return FAILED;

	
	CLIST_INIT( &_algQosTbl );

	
	CLIST_INIT( &_algQosFree );
	memset( _algQosBuf, 0, sizeof(rtl8651_alg_qos_entry_t) * _algQosTblSize );
	entry = (rtl8651_alg_qos_entry_t *)_algQosBuf;
	for (n = 0; n < _algQosTblSize; n++)
	{
		
		CLIST_INSERT_HEAD( &_algQosFree, entry, link );
		entry++;
	}

	return SUCCESS;
}



static int32 _rtl8651_algQos_compareEntry( uint8 isTCP, uint32 srcIP, uint16 srcPort, uint32 extIP, uint16 extPort, 
								 uint32 dstIP, uint16 dstPort, rtl8651_alg_qos_t *entry )
{
	uint8 algType = entry->algType;
	uint8 typeMask;

	
	if (entry == NULL)
		return FALSE;

	

	
	typeMask = 1;
	if ((algType & typeMask) > 0)
	{
		if (entry->isTCP != (isTCP > 0))
		{
			return FALSE;
		}
	};

	
	typeMask <<= 1;
	if ((algType & typeMask) > 0)
	{
		if (entry->srcIP != srcIP)
		{
			return FALSE;
		}
	};

	
	typeMask <<= 1;
	if ((algType & typeMask) > 0)
	{
		if (entry->srcPort != srcPort)
		{
			return FALSE;
		}
	};

	
	typeMask <<= 1;
	if ((algType & typeMask) > 0)
	{
		if (entry->extIP != extIP)
		{
			return FALSE;
		}
	};

	
	typeMask <<= 1;
	if ((algType & typeMask) > 0)
	{
		if (entry->extPort != extPort)
		{
			return FALSE;
		}
	};

	
	typeMask <<= 1;
	if ((algType & typeMask) > 0)
	{
		if (entry->dstIP != dstIP)
		{
			return FALSE;
		}
	};

	
	typeMask <<= 1;
	if ((algType & typeMask) > 0)
	{
		if (entry->dstPort != dstPort)
		{
			return FALSE;
		}
	};

	return TRUE;
}
