


#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_errno.h"
#include "rtl_glue.h"
#include "mbuf.h"

#include "mbuf.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_dns_local.h"
#include "rtl8651_dns_domainBlock_local.h"

static _rtl8651_domainBlockingBlockList_t *_rtl8651_domainBlocking_findBlockList(_rtl8651_domainBlockingInfo_t *info, ipaddr_t mapped_ip);
static int32 _rtl8651_domainBlocking_delBlockInfo(_rtl8651_domainBlockingInfo_t *info, uint8 fromDbModule);
static int32 _rtl8651_domainBlocking_delBlockList(_rtl8651_domainBlockingBlockList_t *ptr, uint8 refreshACL);




static uint8							_rtl8651_domainBlocking_procSequence		= 0;
static uint32							 _rtl8651_domainBlocking_blockListTimeout	= 0;
static _rtl8651_domainBlockingInfo_t		*_rtl8651_domainBlocking_info_inused		= NULL;
static _rtl8651_domainBlockingInfo_t		*_rtl8651_domainBlocking_info_freeList		= NULL;
static _rtl8651_domainBlockingBlockList_t	*_rtl8651_domainBlocking_blockList_h		= NULL;
static _rtl8651_domainBlockingBlockList_t	*_rtl8651_domainBlocking_blockList_t		= NULL;
static _rtl8651_domainBlockingBlockList_t	*_rtl8651_domainBlocking_blockList_freeList	= NULL;


int32 _rtl8651_registerDomainBlock(void)
{
	return _rtl8651_dns_registerProcess(	&_rtl8651_domainBlocking_procSequence,
										RTL8651_DNSDB_PROPERTY_KEYWORDMATCH,
										_rtl8651_domainBlocking_init,
										_rtl8651_domainBlocking_flushBlockList,
										_rtl8651_domainBlocking_procMapping,
										_rtl8651_domainBlocking_timeUpdate,
										_rtl8651_domainBlocking_blockInfoDestructor);
}



#define RTL8651_DOMAINBLOCK_CALLBACK


int32 _rtl8651_domainBlocking_init(void *_para)
{
	rtl8651_fwdEngineInitPara_t *para = (rtl8651_fwdEngineInitPara_t *)_para;

	_rtl8651_domainBlocking_info_inused = NULL;
	_rtl8651_domainBlocking_info_freeList = NULL;
	_rtl8651_domainBlocking_blockList_h = NULL;
	_rtl8651_domainBlocking_blockList_t = NULL;
	_rtl8651_domainBlocking_blockList_freeList = NULL;
	_rtl8651_domainBlocking_blockListTimeout = 0;

	if (	(para->domainBlock_tblSize == 0) ||
		(para->domainBlock_maxBlockEntryCnt == 0) ||
		(para->domainBlock_blockEntryTimeout == 0))
	{
		RTL_DNS_INFO("Disable [domain blocking]:\n\tdomainBlock_tblSize (%d) domainBlock_maxBlockEntryCnt (%d) domainBlock_blockEntryTimeout (%d)\n",
			para->domainBlock_tblSize,
			para->domainBlock_maxBlockEntryCnt,
			para->domainBlock_blockEntryTimeout);
		return SUCCESS;
	}

	MEM_ALLOC(_rtl8651_domainBlocking_info_freeList, sizeof(_rtl8651_domainBlockingInfo_t) * para->domainBlock_tblSize, FAILED);
	DNS_LIST_INIT(_rtl8651_domainBlocking_info_freeList, _rtl8651_domainBlockingInfo_t, para->domainBlock_tblSize, next);

	MEM_ALLOC(_rtl8651_domainBlocking_blockList_freeList, sizeof(_rtl8651_domainBlockingBlockList_t) * para->domainBlock_maxBlockEntryCnt, FAILED);
	DNS_LIST_INIT(_rtl8651_domainBlocking_blockList_freeList, _rtl8651_domainBlockingBlockList_t, para->domainBlock_maxBlockEntryCnt, next);

	_rtl8651_domainBlocking_blockListTimeout = para->domainBlock_blockEntryTimeout;

	return SUCCESS;	

}


void _rtl8651_domainBlocking_flushBlockList(void)
{
	uint8 refreshAcl = FALSE;

	while (_rtl8651_domainBlocking_blockList_h)
	{
		refreshAcl = TRUE;
		_rtl8651_domainBlocking_delBlockList(_rtl8651_domainBlocking_blockList_h, FALSE);
	}

	if (refreshAcl == TRUE)
	{
		_rtl8651_arrangeAllACL();
	}
}


int32 _rtl8651_domainBlocking_procMapping(
	char *orig_domain,
	char *mapped_domain,
	ipaddr_t mapped_ip,
	void *_info)
{
	_rtl8651_domainBlockingInfo_t *info = (_rtl8651_domainBlockingInfo_t *)_info;
	_rtl8651_domainBlockingBlockList_t *blkList;

	
	if ((blkList = _rtl8651_domainBlocking_findBlockList(info, mapped_ip)) != NULL)
	{
		
		blkList->updateTime = _sysUpSeconds;
		HTLIST_MOVE2TAIL(blkList, next, prev, _rtl8651_domainBlocking_blockList_h, _rtl8651_domainBlocking_blockList_t);
	} else
	{
		
		if (_rtl8651_domainBlocking_blockList_freeList == NULL)
		{
			if (_rtl8651_domainBlocking_blockList_h)
			{
				_rtl8651_domainBlocking_delBlockList(_rtl8651_domainBlocking_blockList_h, FALSE);
			}
		}
		
		if ((blkList = _rtl8651_domainBlocking_blockList_freeList) != NULL)
		{
			_rtl8651_domainBlocking_blockList_freeList = blkList->next;

			
			blkList->next = NULL;
			blkList->prev = NULL;
			blkList->info = info;
			blkList->updateTime = _sysUpSeconds;
			blkList->dip = mapped_ip;

			HTLIST_ADD2TAIL(blkList, next, prev, _rtl8651_domainBlocking_blockList_h, _rtl8651_domainBlocking_blockList_t);
		}
	}

	_rtl8651_arrangeAllACL();
	
	return SUCCESS;
}


void _rtl8651_domainBlocking_timeUpdate(uint32 secPassed)
{
	uint8 refreshAcl = FALSE;

	while (	_rtl8651_domainBlocking_blockList_h &&
			((_rtl8651_domainBlocking_blockList_h->updateTime + _rtl8651_domainBlocking_blockListTimeout) < _sysUpSeconds))
	{
		refreshAcl = TRUE;
		_rtl8651_domainBlocking_delBlockList(_rtl8651_domainBlocking_blockList_h, FALSE);
	}

	if (refreshAcl == TRUE)
	{
		_rtl8651_arrangeAllACL();
	}
}



int32 _rtl8651_domainBlocking_blockInfoDestructor(void *info)
{
	return _rtl8651_domainBlocking_delBlockInfo((_rtl8651_domainBlockingInfo_t *)info, FALSE);
}



#define RTL8651_DOMAINBLOCK_EXTERNAL


int32 rtl8651_domainBlocking_addDomainBlockingEntry(char *domain_name, rtl8651_domainBlockingInfo_t *usrInfo)
{
	int32 retval;

	if (!domain_name)
		return FAILED;

	rtlglue_drvMutexLock();
	retval = _rtl8651_domainBlocking_addDomainBlockingEntry(domain_name, usrInfo);
	rtlglue_drvMutexUnlock();

	return retval;
}

int32 _rtl8651_domainBlocking_addDomainBlockingEntry(char *domain_name, rtl8651_domainBlockingInfo_t *usrInfo)
{
	void *tmp = NULL;
	char *domain;
	rtl8651_domainBlockingInfo_t dummy, *usedUsrInfo = usrInfo;
	_rtl8651_domainBlockingInfo_t *info;

	if (domain_name == NULL)
	{
		return FAILED;
	}

	if (usrInfo == NULL)
	{
		memset(&dummy, 0, sizeof(dummy));
		usedUsrInfo = &dummy;
	}
	
	while (_rtl8651_dns_findProcessEntry(	_rtl8651_domainBlocking_procSequence,
											domain_name,
											NULL,
											(char**)&domain,
											(void*)&info,
											(void*)&tmp) == TRUE)
	{
		assert(info);
		if (	(strcmp(domain, domain_name) == 0) &&
			(memcmp(&(info->info), usedUsrInfo, sizeof(rtl8651_domainBlockingInfo_t)) == 0))
		{	
			return SUCCESS;
		}
	}

	
	if (_rtl8651_domainBlocking_info_freeList)
	{
		info = _rtl8651_domainBlocking_info_freeList;
		if (_rtl8651_dns_addProcessEntry(	_rtl8651_domainBlocking_procSequence,
											domain_name,
											(void*)info,
											&(info->ctrlEntry)) == SUCCESS)
		{
			HLIST_REMOVE(info, next, prev, _rtl8651_domainBlocking_info_freeList);
			HLIST_ADD2HDR(info, next, prev, _rtl8651_domainBlocking_info_inused);
			rtl8651_memcpy(&(info->info), usedUsrInfo, sizeof(rtl8651_domainBlockingInfo_t));
			return SUCCESS;
		}
	}

	return FAILED;
}



int32 rtl8651_domainBlocking_delDomainBlockingEntry(char *domain_name, rtl8651_domainBlockingInfo_t *usrInfo)
{
	int32 retval;

	if (!domain_name)
		return FAILED;

	rtlglue_drvMutexLock();
	retval = _rtl8651_domainBlocking_delDomainBlockingEntry(domain_name, usrInfo);
	rtlglue_drvMutexUnlock();

	return retval;
}


int32 _rtl8651_domainBlocking_delDomainBlockingEntry(char *domain_name, rtl8651_domainBlockingInfo_t *usrInfo)
{
	void *tmp = NULL;
	char *domain;
	rtl8651_domainBlockingInfo_t dummy, *usedUsrInfo = usrInfo;
	_rtl8651_domainBlockingInfo_t *info;

	if (domain_name == NULL)
	{
		return FAILED;
	}

	if (usrInfo == NULL)
	{
		memset(&dummy, 0, sizeof(dummy));
		usedUsrInfo = &dummy;
	}

	
	while (_rtl8651_dns_findProcessEntry(	_rtl8651_domainBlocking_procSequence,
											domain_name,
											NULL,
											(char**)&domain,
											(void*)&info,
											(void*)&tmp) == TRUE)
	{
		assert(info);
		if (	(strcmp(domain, domain_name) == 0) &&
			(memcmp(&(info->info), usedUsrInfo, sizeof(rtl8651_domainBlockingInfo_t)) == 0))
		{	
			if (_rtl8651_domainBlocking_delBlockInfo(info, TRUE) != SUCCESS)
			{
				return FAILED;
			}
			
			tmp = NULL;
		}
	}
	return SUCCESS;
}


int32 rtl8651_domainBlocking_flushDomainBlockingEntry(void)
{
	int32 ret;

	rtlglue_drvMutexLock();
	ret = _rtl8651_domainBlocking_flushDomainBlockingEntry();
	rtlglue_drvMutexUnlock();

	return ret;
}

int32 _rtl8651_domainBlocking_flushDomainBlockingEntry(void)
{
	while( _rtl8651_domainBlocking_info_inused )
	{
		if (_rtl8651_domainBlocking_delBlockInfo(_rtl8651_domainBlocking_info_inused, TRUE) != SUCCESS)
		{
			return FAILED;
		}
	}
	return SUCCESS;
}


void rtl8651_domainBlocking_dump(void)
{
	uint32 inusedCnt = 0, freeCnt = 0;
	_rtl8651_domainBlockingInfo_t *info;
	_rtl8651_domainBlockingBlockList_t *blkList;

	rtlglue_drvMutexLock();

	info = _rtl8651_domainBlocking_info_inused;
	while (info)
	{
		inusedCnt ++;
		info = info->next;
	}
	info = _rtl8651_domainBlocking_info_freeList;
	while (info)
	{
		freeCnt ++;
		info = info->next;
	}

	printk("\nDOMAIN BLOCK table:\n\n");
	printk("\tInused:\t\t%d\n", inusedCnt);
	printk("\tFree:\t\t%d\n", freeCnt);

	if (inusedCnt)
	{
		int32 idx;
		char *domain;

		printk("\n\t[NO] [SIP/MASK]\t\t\t\t\t[PTR]\n");
		printk("\t============================================================\n");
		info = _rtl8651_domainBlocking_info_inused;
		idx = 1;
		while (info)
		{
			domain = _rtl8651_dns_getProcessEntryDomainInfo(info->ctrlEntry);
			printk("\t%2d    %03u.%03u.%03u.%03u/%03u.%03u.%03u.%03u\t\t%p\n\t\t\t=> (%s)\n",
				idx,
				NIPQUAD(info->info.sip),
				NIPQUAD(info->info.sipMask),
				info,
				domain?domain:"(NULL)");
			idx ++;
			info = info->next;
		}
		printk("\t============================================================\n");
	}

	inusedCnt = 0;
	freeCnt = 0;

	blkList = _rtl8651_domainBlocking_blockList_h;
	while (blkList)
	{
		inusedCnt ++;
		blkList = blkList->next;
	}
	blkList = _rtl8651_domainBlocking_blockList_freeList;
	while (blkList)
	{
		freeCnt ++;
		blkList = blkList->next;
	}

	printk("\nBlocking list:\n\n");
	printk("\tInused:\t\t%d\n", inusedCnt);
	printk("\tFree:\t\t%d\n", freeCnt);

	if (inusedCnt)
	{
		char *domain;

		printk("\n\t[TOUT]\t[SIP/MASK]\t\t\t[DIP]\n");
		printk("\t============================================================\n");
		blkList = _rtl8651_domainBlocking_blockList_h;
		while (blkList)
		{
			info = blkList->info;
			domain = _rtl8651_dns_getProcessEntryDomainInfo(info->ctrlEntry);
			printk("\t%5d  %03u.%03u.%03u.%03u/%03u.%03u.%03u.%03u\t(%03u.%03u.%03u.%03u)\n\t\t\t=> (%s)\n",
				blkList->updateTime + _rtl8651_domainBlocking_blockListTimeout - _sysUpSeconds,
				NIPQUAD(info->info.sip),
				NIPQUAD(info->info.sipMask),
				NIPQUAD(blkList->dip),
				domain?domain:"(NULL)");
			blkList = blkList->next;
		}
		printk("\t============================================================\n");
	}

	rtlglue_drvMutexUnlock();
}


#define SET_DMBLKACL_MATCHLAYER(rule) \
	rule.pktOpApp = RTL8651_ACLTBL_L3_AND_L4;

uint32 _rtl8651_tryToAddAclRuleForDnsDomainBlocking(rtl8651_tblDrv_vlanTable_t *vlanPtr, uint32 aclStart, uint32 aclEnd, int8 *enoughFlag)
{
	_rtl8651_tblDrvAclRule_t rule;
	uint32 addCount = 0, aclIdx = aclStart;

	if ( vlanPtr->internal )	
	{
		_rtl8651_domainBlockingBlockList_t *blkList = _rtl8651_domainBlocking_blockList_h;

		while ( blkList )
		{
			if ( aclIdx > aclEnd )
			{
				*enoughFlag = 0;	
				goto out;
			}

			bzero(&rule, sizeof(_rtl8651_tblDrvAclRule_t));

			rule.ruleType_			= RTL8651_ACL_IP;
			rule.actionType_		= RTL8651_ACL_DROP;
			rule.srcIpAddr_		= blkList->info->info.sip;
			rule.srcIpAddrMask_	= blkList->info->info.sipMask;
			rule.dstIpAddr_		= blkList->dip;
			rule.dstIpAddrMask_	= 0xffffffff;

			SET_DMBLKACL_MATCHLAYER(rule);
			rtl8651_setAsicAclRule(aclIdx, &rule);

			addCount ++;
			aclIdx ++;
			blkList = blkList->next;
		}
	}

out:
	return addCount;
}



#define RTL8651_DOMAINBLOCK_INTERNAL


static _rtl8651_domainBlockingBlockList_t *_rtl8651_domainBlocking_findBlockList(_rtl8651_domainBlockingInfo_t *info, ipaddr_t mapped_ip)
{
	_rtl8651_domainBlockingBlockList_t *ptr;

	ptr = _rtl8651_domainBlocking_blockList_h;

	while (ptr)
	{
		if ((ptr->info == info) && (ptr->dip == mapped_ip))
		{
			return ptr;
		}
		ptr = ptr->next;
	}

	return NULL;
}


static int32 _rtl8651_domainBlocking_delBlockInfo(_rtl8651_domainBlockingInfo_t *info, uint8 fromDbModule)
{
	_rtl8651_domainBlockingBlockList_t * ptr, *prev;
	uint8 refreshAcl = FALSE;

	if (fromDbModule == TRUE)
	{
		

		
		if (_rtl8651_dns_delProcessEntry(	_rtl8651_domainBlocking_procSequence,
										TRUE,
										info->ctrlEntry,
										NULL,
										NULL) != SUCCESS)
		{
			return FAILED;
		}
	} else
	{
		

		
		HLIST_REMOVE(info, next, prev, _rtl8651_domainBlocking_info_inused);
		HLIST_ADD2HDR(info, next, prev, _rtl8651_domainBlocking_info_freeList);
	}

	
	prev = NULL;
	ptr = _rtl8651_domainBlocking_blockList_h;
	while (ptr)
	{
		if (ptr->info == info)
		{
			_rtl8651_domainBlocking_delBlockList(ptr, FALSE);
			refreshAcl = TRUE;
			ptr = prev;
		} else
		{
			prev = ptr;
			ptr = ptr->next;
		}
	}

	if (refreshAcl == TRUE)
	{
		_rtl8651_arrangeAllACL();
	}

	return SUCCESS;

}


static int32 _rtl8651_domainBlocking_delBlockList(_rtl8651_domainBlockingBlockList_t *ptr, uint8 refreshACL)
{
	if (ptr == NULL)
	{
		return FAILED;
	}
	
	HTLIST_REMOVE(ptr, next, prev, _rtl8651_domainBlocking_blockList_h, _rtl8651_domainBlocking_blockList_t);
	HLIST_ADD2HDR(ptr, next, prev, _rtl8651_domainBlocking_blockList_freeList);
	if (refreshACL == TRUE)
	{
		_rtl8651_arrangeAllACL();
	}
	return SUCCESS;
}



