

#include "rtl_types.h"
#include "rtl8651_multicast_local.h"
#include "mbuf.h"


#define MAX(a,b) ((a>b)?a:b)
#define MIN(a,b) ((a>b)?b:a)


static uint16					system_id							= 1;		
#ifdef RTL865X_TEST
#define						SYS_ID								system_id
#else
#define						SYS_ID								(system_id ++)
#endif 


static rtl8651_tblDrvMCastGlobal_t rtl8651_GlobalVariable;
static rtl8651_tblDrvMCast_t *	rtl8651_MCastTable					= NULL;									
static uint32					rtl8651_drvMaxMCastCnt				= RTL8651_DefaultDrvMaxMCastCnt;			
static uint32					rtl8651_drvMaxMCastMemberCnt		= RTL8651_DefaultDrvMaxMCastMemberCnt;	
static uint32					rtl8651_drvMaxTotalMemberCnt		= RTL8651_DefaultDrvMaxTotalMemberCnt;	
static uint32					rtl8651_drvMaxTotalSourceCnt		= RTL8651_DefaultDrvMaxTotalSourceCnt;		

static uint32					rtl8651_drvMCastMember_ExpireTime	= RTL8651_DefaultDrvMcastMemberTimeout;	
static uint32					rtl8651_drvMCastUpstream_Timeout	= RTL8651_DefaultDrvMCastUpstreamTimeout;	
static uint8					rtl8651_drvMCastVersion				= RTL8651_MCAST_IGMP_MAX_VERSION;

static uint32					rtl8651_igmp_qi						= RTL8651_IGMP_QI;
static uint32					rtl8651_igmp_qri					= RTL8651_IGMP_QRI;
static uint32					rtl8651_igmp_oqpi					= RTL8651_IGMP_OQPI;


static uint32					rtl8651_igmp_sqc					= RTL8651_IGMP_SQC;
static uint32					rtl8651_igmp_sqi					= RTL8651_IGMP_SQI;

static uint32					rtl8651_igmp_lmqc					= RTL8651_IGMP_LMQC;
static uint32					rtl8651_igmp_lmqi					= RTL8651_IGMP_LMQI;

static uint32					rtl8651_igmp_gmi					= RTL8651_IGMP_GMI;
static uint32					rtl8651_igmp_lmqt					= (RTL8651_IGMP_LMQC * RTL8651_IGMP_LMQI);

#define						rtl8651_drvMCastMember_DestroyTime	(((rtl8651_igmp_lmqc - 1) * rtl8651_igmp_lmqi) + rtl8651_igmp_qri + rtl8651_igmp_response_tolerance_delay)

static uint32					rtl8651_igmp_general_query			= RTL8651_IGMP_GENERAL_QUERY;
static uint32					rtl8651_igmp_group_query_interval	= RTL8651_IGMP_GROUP_QUERY_INTERVAL;
static uint32					rtl8651_igmp_response_tolerance_delay	= RTL8651_IGMP_RESPONSE_TOLERANCE_DELAY;

static uint32					rtl8651_MCastEntryCnt				= 0;				
static uint32					rtl8651_MemberCnt					= 0;				
static uint32					rtl8651_SourceListCnt				= 0;				
static uint8					system_enable						= TRUE;


#define						rtl8651_addedMCastSourceCnt		(4 * rtl8651_drvMaxTotalSourceCnt)
static uint32					rtl8651_SystemMCastMemberCnt		= MIN(RTL8651_DefaultDrvMaxTotalMemberCnt, (RTL8651_DefaultDrvMaxMCastCnt*RTL8651_DefaultDrvMaxMCastMemberCnt));
static uint32					rtl8651_SystemMCastSourceCnt		= (5 * RTL8651_DefaultDrvMaxTotalSourceCnt);



static uint32					rtl8651_multicastProcessOptions		= RTL8651_MCAST_PROCOPT_DEFAULT;


#define	MCast_DATA_TYPE_MCastEntry	0xff
#define	MCast_DATA_TYPE_MCastMember	0xfe
#define	MCast_DATA_TYPE_MCastSource	0xfd


static rtl8651_tblDrvMCast_t			*MCastEntryPool			= NULL;
static rtl8651_tblDrvMCastMember_t	*MCastMemberPool		= NULL;
static rtl8651_tblDrvMCastSource_t	*MCastSourcePool		= NULL;

static rtl8651_tblDrvMCast_t			*MCastEntryPoolPtr		= NULL;
static rtl8651_tblDrvMCastMember_t	*MCastMemberPoolPtr	= NULL;
static rtl8651_tblDrvMCastSource_t	*MCastSourcePoolPtr		= NULL;



static rtl8651_tblDrvMCastFilter_t	*MCastFilter[RTL8651_MCastFilterHashSize]	= {NULL};
static rtl8651_tblDrvMCastFilter_t *MCastFilter_FreeHdr						= NULL;



static __inline__ int32 _rtl8651_multicastSetInitPara(rtl8651_fwdEngineInitPara_t *);

static __inline__ int32 MCastPool_init(void);
static __inline__ void MCastPool_reinit(void);
static void *MCast_malloc(uint32);
static void MCast_free(void *);

static uint16 MCast_igmpChecksum(struct ip *);
static uint8 MCast_multicastQqic_Fix2Ft(	uint32,
											uint32,
											uint32,
											uint32);
static void MCast_BroadCastQuery(void);
static int32 MCast_setPkthdrProperty(struct rtl_pktHdr *, int32);


static rtl8651_tblDrvMCast_t *MCast_FindMCastEntry(ipaddr_t, uint8, uint8);
static rtl8651_tblDrvMCast_t *MCast_CreateMCastEntry(ipaddr_t);
static __inline__ uint8 MCast_MCastEntryMinVersion(rtl8651_tblDrvMCast_t *);
static __inline__ void MCast_SetMCastEntryUpStream(rtl8651_tblDrvMCast_t *,rtl8651_tblDrv_vlanTable_t *, uint16, uint8);
static void MCast_ClrMCastEntryUpStream(uint32);
static void MCast_RemoveMCastEntry(rtl8651_tblDrvMCast_t *);
static __inline__ void MCast_Entrylink(rtl8651_tblDrvMCast_t *);
static __inline__ void MCast_EntryUnlink(rtl8651_tblDrvMCast_t *);

static rtl8651_tblDrvMCastMember_t *MCast_FindMember(rtl8651_tblDrvMCast_t *, rtl8651_tblDrv_vlanTable_t *, uint16, uint8, uint8, uint8);
static rtl8651_tblDrvMCastMember_t *MCast_CreateMember(rtl8651_tblDrvMCast_t *, rtl8651_tblDrv_vlanTable_t *, uint16, uint8);
static void MCast_RemoveMember(rtl8651_tblDrvMCast_t *, rtl8651_tblDrvMCastMember_t *);
static void MCast_DestroyMember(rtl8651_tblDrvMCastMember_t *);
static __inline__ void MCast_MemberLink(rtl8651_tblDrvMCast_t *, rtl8651_tblDrvMCastMember_t *);
static __inline__ int32 MCast_MemberUnLink(rtl8651_tblDrvMCastMember_t *);
static void MCast_MemberExcludeExpire(uint32);
static void MCast_MemberExpire(uint32);
static void MCast_MemberSendQuery(rtl8651_tblDrvMCastMember_t *);

static __inline__ rtl8651_tblDrvMCastSource_t *MCast_SourceList_Create(void);
static __inline__ void MCast_SourceList_Link(rtl8651_tblDrvMCast_t *, rtl8651_tblDrvMCastMember_t *, rtl8651_tblDrvMCastSource_t *);
static __inline__ void MCast_SourceList_UnLink(rtl8651_tblDrvMCastSource_t *);
static rtl8651_tblDrvMCastSource_t* MCast_SourceList_inList(ipaddr_t , rtl8651_tblDrvMCastSource_t *);
static __inline__ uint32 MCast_SourceList_EntryCnt(rtl8651_tblDrvMCastSource_t *);
static rtl8651_tblDrvMCastSource_t *MCast_SourceList_freeCurrent(rtl8651_tblDrvMCastSource_t *);
static void MCast_SourceList_freeAll(rtl8651_tblDrvMCastSource_t *);
static uint32 MCast_SourceList_removal(rtl8651_tblDrvMCastSource_t *,uint8 ,uint8 ,rtl8651_tblDrvMCastSource_t *,uint8 ,uint8 ,rtl8651_tblDrvMCastSource_t **);
static void MCast_SourceList_addList(rtl8651_tblDrvMCastMember_t *, rtl8651_tblDrvMCastSource_t *, uint8, uint8);
static int32 MCast_SourceList_dup(rtl8651_tblDrvMCastMember_t *, rtl8651_tblDrvMCastSource_t *, rtl8651_tblDrvMCastSource_t **);
static void MCast_SourceList_Expire(rtl8651_tblDrvMCastSource_t *);
static uint32 MCast_SourceList_getCache(rtl8651_tblDrvMCast_t *, rtl8651_tblDrvMCastSource_t **);

static __inline__ ipaddr_t MCast_GetIpByVlan(rtl8651_tblDrv_vlanTable_t *, uint32);
static __inline__ int32 MCast_SetRouteInfoAndGetIntfIp(ipaddr_t *, rtl8651_tblDrvMCastRouteInfo_t *);
static int32 MCast_query(struct rtl_pktHdr *, struct ip *, struct igmp *);
static int32 MCast_report(struct rtl_pktHdr *, struct ip *, struct igmp *);
static int32 MCast_leave(struct rtl_pktHdr *, struct ip *, struct igmp *);

static int32 MCast_QueryPack(rtl8651_tblDrvMCastMember_t *,rtl8651_tblDrvMCastRouteInfo_t *, uint8, rtl8651_tblDrvMCastSource_t *, uint16, struct rtl_pktHdr *);
static int32 MCast_ReportPack(rtl8651_tblDrvMCast_t *, uint8, uint8, rtl8651_tblDrvMCastSource_t *, uint8, rtl8651_tblDrvMCastSource_t *, struct rtl_pktHdr *);

static int32 MCast_LeavePack(rtl8651_tblDrvMCast_t *, struct rtl_pktHdr *);
static int32 MCast_l3ActRouting(struct rtl_pktHdr *, rtl8651_tblDrvMCastRouteInfo_t *);

static __inline__ void MCast_Proc_ISIN(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);
static __inline__ void MCast_Proc_ISEX(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);
static __inline__ void MCast_Proc_TOIN(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);
static __inline__ void MCast_Proc_TOEX(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);
static __inline__ void MCast_Proc_ALLOW(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);
static __inline__ void MCast_Proc_BLOCK(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);


static __inline__ uint32 _rtl8651_multicast_Filter_Hash(ipaddr_t);
rtl8651_tblDrvMCastFilter_t *_rtl8651_multicast_FindFilter(ipaddr_t);


static __inline__ int32 MCastPool_init(void)
{
	int32 retval = SUCCESS;	

	rtlglue_drvMutexLock();	

	if (	(rtl8651_drvMaxMCastCnt == 0) ||
		(rtl8651_SystemMCastMemberCnt == 0) ||
		(rtl8651_SystemMCastSourceCnt == 0)	)
	{
		goto out;	
	}

	
	MCastEntryPool = (rtl8651_tblDrvMCast_t *)rtlglue_malloc(sizeof(rtl8651_tblDrvMCast_t) * rtl8651_drvMaxMCastCnt);
	MCastMemberPool =
		(rtl8651_tblDrvMCastMember_t *)rtlglue_malloc(sizeof(rtl8651_tblDrvMCastMember_t) * rtl8651_SystemMCastMemberCnt);
	MCastSourcePool = 
		(rtl8651_tblDrvMCastSource_t *)rtlglue_malloc(sizeof(rtl8651_tblDrvMCastSource_t) * rtl8651_SystemMCastSourceCnt);

	if (	(MCastEntryPool == NULL) ||
		(MCastMemberPool == NULL) ||
		(MCastSourcePool == NULL)	)
	{
		
		retval = FAILED;
		goto out;
	}

	MCastPool_reinit();

	

out:
	rtlglue_drvMutexUnlock();	

	return retval;
}
void MCastPool_reinit(void)
{
	uint32 idx;
	rtl8651_tblDrvMCast_t *mCastPtr;
	rtl8651_tblDrvMCastMember_t *mCastMemberPtr;
	rtl8651_tblDrvMCastSource_t *mCastSourcePtr;

	rtlglue_drvMutexLock();

	
	bzero(MCastEntryPool, (rtl8651_drvMaxMCastCnt * sizeof(rtl8651_tblDrvMCast_t)));
	MCastEntryPoolPtr = MCastEntryPool;
	mCastPtr = MCastEntryPool;
	for (idx = 0 ; idx < rtl8651_drvMaxMCastCnt ; idx ++, mCastPtr ++)
	{
		if (idx == (rtl8651_drvMaxMCastCnt - 1))
			mCastPtr->next = NULL;
		else
			mCastPtr->next = mCastPtr + 1;
	}
	
	bzero(MCastMemberPool, (rtl8651_SystemMCastMemberCnt * sizeof(rtl8651_tblDrvMCastMember_t)));
	MCastMemberPoolPtr = MCastMemberPool;
	mCastMemberPtr = MCastMemberPool;
	for (idx = 0 ; idx < rtl8651_SystemMCastMemberCnt ; idx ++, mCastMemberPtr ++)
	{
		if (idx == (rtl8651_SystemMCastMemberCnt - 1))
			mCastMemberPtr->next = NULL;
		else
			mCastMemberPtr->next = mCastMemberPtr + 1;
	}
	
	bzero(MCastSourcePool, (rtl8651_SystemMCastSourceCnt * sizeof(rtl8651_tblDrvMCastSource_t)));
	MCastSourcePoolPtr = MCastSourcePool;
	mCastSourcePtr = MCastSourcePool;
	for (idx = 0 ; idx < rtl8651_SystemMCastSourceCnt ; idx ++, mCastSourcePtr ++)
	{
		if (idx == (rtl8651_SystemMCastSourceCnt - 1))
			mCastSourcePtr->next = NULL;
		else
			mCastSourcePtr->next = mCastSourcePtr + 1;
	}

	rtlglue_drvMutexUnlock();
}

#define RTL8651_MULTICAST_EXTERNAL_APIS

int32 rtl8651_multicastSetPara(rtl8651_tblDrvMCastPara_t *para)
{
	if (!para)
		return FAILED;

	rtlglue_drvMutexLock();

	if (para->MCastMember_ExpireTime)
		rtl8651_drvMCastMember_ExpireTime = para->MCastMember_ExpireTime;
	if (para->MCastUpstream_Timeout)
		rtl8651_drvMCastUpstream_Timeout = para->MCastUpstream_Timeout;
	if (para->igmp_qi)
		rtl8651_igmp_qi = para->igmp_qi;
	if (para->igmp_qri)
		rtl8651_igmp_qri = para->igmp_qri;
	if (para->igmp_gmi)
		rtl8651_igmp_gmi = para->igmp_gmi;
	if (para->igmp_group_query_interval)
		rtl8651_igmp_group_query_interval = para->igmp_group_query_interval;

	rtl8651_igmp_oqpi = para->igmp_oqpi;
	rtl8651_igmp_sqc = para->igmp_sqc;
	rtl8651_igmp_sqi = para->igmp_sqi;
	rtl8651_igmp_lmqc = para->igmp_lmqc;
	rtl8651_igmp_lmqi = para->igmp_lmqi;
	rtl8651_igmp_lmqt = (rtl8651_igmp_lmqc * rtl8651_igmp_lmqi);

	rtl8651_igmp_response_tolerance_delay = para->igmp_response_tolerance_delay;

	if (rtl8651_GlobalVariable.remainingQuery)	
		rtl8651_GlobalVariable.remainingQuery = rtl8651_igmp_sqc;

	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

#ifdef RTL865X_TEST
void _rtl8651_setMcastProcessOptions(uint32 option)
{
	rtl8651_multicastProcessOptions = option;
}
#endif


int32 _rtl8651_multicastSetInitPara(rtl8651_fwdEngineInitPara_t *para)
{
	uint32 c1, c2;
	
	if (!para)
		return FAILED;

	rtl8651_multicastProcessOptions		= para->MCastProcessOption;
	rtl8651_drvMaxMCastCnt				= para->MaxMCastCnt; 
	rtl8651_drvMaxMCastMemberCnt		= para->MaxMCastMemberCnt;
	rtl8651_drvMaxTotalMemberCnt		= para->MaxMCastTotalMemberCnt;
	rtl8651_drvMaxTotalSourceCnt		= para->MaxMCastTotalSourceCnt;
	rtl8651_SystemMCastSourceCnt		= rtl8651_drvMaxTotalSourceCnt + rtl8651_addedMCastSourceCnt;

	if ((para->MCastDefaultVersion > RTL8651_MCAST_IGMP_MAX_VERSION) ||
			(para->MCastDefaultVersion == 0))
	{
		rtl8651_drvMCastVersion = RTL8651_MCAST_IGMP_MAX_VERSION;
	}else
	{
		rtl8651_drvMCastVersion = para->MCastDefaultVersion;
	}
		

	if (para->MCastMember_ExpireTime)
		rtl8651_drvMCastMember_ExpireTime	= para->MCastMember_ExpireTime;
	if (para->MCastUpstream_Timeout)
		rtl8651_drvMCastUpstream_Timeout	= para->MCastUpstream_Timeout;
	if (para->igmp_qi)
		rtl8651_igmp_qi						= para->igmp_qi;
	if (para->igmp_qri)
		rtl8651_igmp_qri					= para->igmp_qri;
	if (para->igmp_gmi)
		rtl8651_igmp_gmi					= para->igmp_gmi;
	if (para->igmp_group_query_interval)
		rtl8651_igmp_group_query_interval	= para->igmp_group_query_interval;

	rtl8651_igmp_oqpi						= para->igmp_oqpi;
	rtl8651_igmp_sqc						= para->igmp_sqc;
	rtl8651_igmp_sqi						= para->igmp_sqi;
	rtl8651_igmp_lmqc						= para->igmp_lmqc;
	rtl8651_igmp_lmqi						= para->igmp_lmqi;
	rtl8651_igmp_lmqt = (rtl8651_igmp_lmqc * rtl8651_igmp_lmqi);

	rtl8651_igmp_response_tolerance_delay	= para->igmp_response_tolerance_delay;

	{
		
		c1 = rtl8651_drvMaxTotalMemberCnt;
		c2 = (rtl8651_drvMaxMCastCnt * rtl8651_drvMaxMCastMemberCnt);

		if (c1 == 0)
			rtl8651_SystemMCastMemberCnt = c2;
		else if (c2 == 0)
			rtl8651_SystemMCastMemberCnt = c1;
		else
			rtl8651_SystemMCastMemberCnt = MIN(c1, c2);
	}

	return SUCCESS;
}


void rtl8651_multicastReinit(void)
{
	if (system_enable == FALSE)
		return;

	rtlglue_drvMutexLock();
	
	while (rtl8651_MCastTable)
		MCast_RemoveMCastEntry(rtl8651_MCastTable);

	
	_rtl8651_delMulticast(0, 0, 0);

	assert(rtl8651_MCastEntryCnt == 0);
	assert(rtl8651_MemberCnt == 0);

	
	system_id = 1;
	
	MCastPool_reinit();

	
	rtl8651_multicastProcessOptions = RTL8651_MCAST_PROCOPT_DEFAULT;

	
	rtl8651_GlobalVariable.remainingQuery = rtl8651_igmp_sqc;	
	
	if (rtl8651_igmp_sqc)
	{
		rtl8651_GlobalVariable.query_timer.data	= (uint32)0;
		rtl8651_GlobalVariable.query_timer.function	= (void*)MCast_BroadCastQuery;
		_rtl8651_mod_timer(&(rtl8651_GlobalVariable.query_timer), _sysUpSeconds + RTL8651_IGMP_STARTUP_QUERY_DELAY);
	}

	rtlglue_drvMutexUnlock();
}

int32 _rtl8651_multicastInit(rtl8651_fwdEngineInitPara_t * para)
{
	int32 retval = SUCCESS;

	system_enable = FALSE;

	
	rtl8651_MCastTable = NULL;
	bzero(&rtl8651_GlobalVariable, sizeof(rtl8651_tblDrvMCastGlobal_t));
	rtl8651_GlobalVariable.MCastUpStream.isSet	= FALSE;
	_rtl8651_init_timer(&(rtl8651_GlobalVariable.query_timer));

	
	if (para)
		retval = _rtl8651_multicastSetInitPara(para);

	
	if (
		(!rtl8651_drvMaxMCastCnt) ||
		(!rtl8651_SystemMCastMemberCnt) ||
		(!rtl8651_SystemMCastSourceCnt)
	)
		goto out;

	
	if ((retval = MCastPool_init()) != SUCCESS)
	{
		goto out;
	}

	
	system_enable = TRUE;

	
	rtl8651_multicastReinit();
out:
	return retval;
}

int32 rtl8651_multicastSetUpStream(int8* netifName, uint32 para)
{
	rtl8651_tblDrvMCastUpStreamInfo_t fake_UpStream;
	rtl8651_tblDrv_networkIntfTable_t *netif;

	rtlglue_printf("\nSet IGMP Default Upstream interface ");

	rtlglue_drvMutexLock();

	memset(&fake_UpStream, 0, sizeof(rtl8651_tblDrvMCastUpStreamInfo_t));
	if ((!netifName) ||
		((netif = GET_NETWORK_INTERFACE(netifName, NULL)) == NULL))
	{
		goto err;
	}
	rtlglue_printf("(%s) ... ", netifName);

	fake_UpStream.netif = netif;
	fake_UpStream.vlan = &DrvTbl.vlan[netif->vid & (RTL8651_VLAN_NUMBER - 1)];
	fake_UpStream.portMask = fake_UpStream.vlan->memberPortMask;
	
	switch (netif->linkLayerType)
	{
		case RTL8651_LL_VLAN:
			{
				rtl8651_tblDrv_ipIntfEntry_t * ipIntfPtr;
				rtl8651_tblDrv_ipIntfIpAddrEntry_t * ipAddrPtr;

				if (para)
				{	
					SLIST_FOREACH(ipIntfPtr, &(netif->IpHead), nextIp)
					{
						if (!ipIntfPtr)
						{
							goto err;
						}
						SLIST_FOREACH(ipAddrPtr, &ipIntfPtr->ipAddr, nextIpAddr)
						{
							if (!ipAddrPtr)
							{
								goto err;
							}
							if (ipAddrPtr->ipAddr == para)
							{
								fake_UpStream.M_intfIp= para;
								goto set_para;
							}
						}
					}
					goto err;
				}else
				{
#if 0
					ipIntfPtr = SLIST_FIRST(&netif->IpHead);
					if (!ipIntfPtr)
					{
						fake_UpStream.M_intfIp = 0;
						goto set_para;
					}
					ipAddrPtr = SLIST_FIRST(&ipIntfPtr->ipAddr);
					if (!ipAddrPtr)
					{
						fake_UpStream.M_intfIp = 0;
						goto set_para;
					}
					assert(ipAddrPtr->ipAddr);
					fake_UpStream.M_intfIp = ipAddrPtr->ipAddr;
#else
					fake_UpStream.M_intfIp = 0;	
#endif
				}
			}
			break;
		case RTL8651_LL_PPPOE:
			{
				rtl8651_tblDrvSession_t	*sessionPtr;
			
				if (para)
				{	
					CTAILQ_FOREACH(sessionPtr, &(netif->sessionHead), nextSession)
					{
						assert(sessionPtr->session_t);
						if (sessionPtr->session_t->sessionId == para)
						{
							fake_UpStream.M_pppoeSessionPtr = sessionPtr;
							goto set_para;
						}
					}
					goto err;
				}else
				{
					fake_UpStream.M_pppoeSessionPtr = CTAILQ_FIRST(&netif->sessionHead);
				}
			}
			break;
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		case RTL8651_LL_PPTP:
		case RTL8651_LL_L2TP:
			{
				rtl8651_tblDrvSession_t	*sessionPtr;
				if (para)
				{
					CTAILQ_FOREACH(sessionPtr, &(netif->sessionHead), nextSession)
					{
						assert(sessionPtr->session_t);
						if (sessionPtr->session_t->sessionId == para)
						{
							fake_UpStream.M_sessionPtr = sessionPtr;
							goto set_para;
						}
					}
					goto err;
				}else
				{
					fake_UpStream.M_sessionPtr = CTAILQ_FIRST(&netif->sessionHead);
				}
			}
			break;
		#endif 
		default:
			goto err;
	}
set_para:
	rtl8651_GlobalVariable.MCastUpStream		= fake_UpStream;
	rtl8651_GlobalVariable.MCastUpStream.isSet	= TRUE;

	rtlglue_drvMutexUnlock();

	rtlglue_printf("SUCCESS!!\n");
	return SUCCESS;
err:
	rtlglue_drvMutexUnlock();
	rtlglue_printf("FAILED!!\n");
	return FAILED;
}

int32 rtl8651_multicastPeriodicQuery(int8 enable)
{
	if (enable)
	{
		
		if (rtl8651_igmp_general_query == FALSE)
		{
			if (rtl8651_GlobalVariable.remainingQuery == 0)
			{
				_rtl8651_mod_timer(&(rtl8651_GlobalVariable.query_timer), _sysUpSeconds + rtl8651_igmp_qi);
			}
			rtl8651_igmp_general_query = TRUE;
			
			MCast_BroadCastQuery();
		}
	}else
	{
		rtl8651_igmp_general_query = FALSE;
		if (rtl8651_GlobalVariable.remainingQuery == 0)
			_rtl8651_del_timer(&(rtl8651_GlobalVariable.query_timer));
	}
	return SUCCESS;
}

void *MCast_malloc(uint32 size)
{
	void *ret = NULL;

	switch (size)
	{
		case (sizeof (rtl8651_tblDrvMCast_t)):
			if (MCastEntryPoolPtr)
			{
				ret = MCastEntryPoolPtr;
				MCastEntryPoolPtr = MCastEntryPoolPtr->next;
				bzero(ret, sizeof(rtl8651_tblDrvMCast_t));
				((rtl8651_tblDrvMCast_t *)ret)->_type = MCast_DATA_TYPE_MCastEntry;
			}
			break;
		case (sizeof (rtl8651_tblDrvMCastMember_t)):
			if (MCastMemberPoolPtr)
			{
				ret = MCastMemberPoolPtr;
				MCastMemberPoolPtr = MCastMemberPoolPtr->next;
				bzero(ret, sizeof(rtl8651_tblDrvMCastMember_t));
				((rtl8651_tblDrvMCastMember_t *)ret)->_type = MCast_DATA_TYPE_MCastMember;
			}
			break;
		case (sizeof (rtl8651_tblDrvMCastSource_t)):
			if (MCastSourcePoolPtr)
			{
				ret = MCastSourcePoolPtr;
				MCastSourcePoolPtr = MCastSourcePoolPtr->next;
				bzero(ret, sizeof(rtl8651_tblDrvMCastSource_t));
				((rtl8651_tblDrvMCastSource_t *)ret)->_type = MCast_DATA_TYPE_MCastSource;
			}
			break;
		default:
			assert(0);
	}

	return ret;
}
void MCast_free(void *ptr)
{
	if (!ptr)
		return;

	switch (((rtl8651_tblDrvMCast_t *)ptr)->_type)
	{
		case MCast_DATA_TYPE_MCastEntry:
			((rtl8651_tblDrvMCast_t *)ptr)->next = MCastEntryPoolPtr;
			MCastEntryPoolPtr = (rtl8651_tblDrvMCast_t *)ptr;
			break;
		case MCast_DATA_TYPE_MCastMember:
			((rtl8651_tblDrvMCastMember_t *)ptr)->next = MCastMemberPoolPtr;
			MCastMemberPoolPtr = (rtl8651_tblDrvMCastMember_t *)ptr;
			break;
		case MCast_DATA_TYPE_MCastSource:
			((rtl8651_tblDrvMCastSource_t *)ptr)->next = MCastSourcePoolPtr;
			MCastSourcePoolPtr = (rtl8651_tblDrvMCastSource_t *)ptr;
			break;
		default:
			assert(0);
	}
}


uint16 MCast_igmpChecksum(struct ip * pip)
{
	uint16   *ptr;
	int32  nhdr, nbytes;
	int32  sum, i;


	nhdr =  ((pip->ip_vhl &0xf) << 2);
	nbytes = ntohs(pip->ip_len) - nhdr;
	ptr = (uint16 *) ((char *) pip + nhdr);
	sum = 0;

	
	*(((uint8 *)pip) + nhdr + nbytes) = (uint8)0;
	for (i=0; i<nbytes ; i=i+2){
		sum +=  (unsigned long) *ptr++;
	}

	
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);

	
	return ((uint16) ~ sum);
}


static uint8 MCast_multicastQqic_Fix2Ft(	uint32 qqi,
											uint32 nbmant,
											uint32 nbexp,
											uint32 sign)
{
	uint8 fp_return = 0;	
	uint8 mant = 0;
	uint8 exp = 0;
	uint32 msb_position = 0, maxMsb_position = 0;

	
 
	
	if (sign)
	{
		fp_return |= 1 << (nbmant + nbexp);
	}

	
	maxMsb_position = sizeof(qqi) * 8 ;
	for (msb_position = (maxMsb_position - 1) ; msb_position >= 0 ; msb_position --)
	{
		if (qqi & (1 << msb_position))
		{
			break;
		}
	}
	
	if (msb_position < (nbmant + nbexp))
	{
		#ifdef RTL8651_MCAST_DEBUG
			rtlglue_printf(	"[MCAST] Error occurred when QQIC generation! (QQI: %d, nbmant %d, nbexp %d, sign %d)\n",
						qqi,
						nbmant,
						nbexp,
						sign);
		#endif
		return 0;
	}

	
	mant = (qqi >> (msb_position - nbmant)) & ((1 << nbmant) - 1);

	
	exp = msb_position - nbexp - nbmant;

	
	fp_return |= exp << nbmant;
	fp_return |= mant;

	return fp_return;
}


void MCast_BroadCastQuery(void)
{

	if ((system_enable == FALSE) || (rtl8651_drvMulticastProcessEnable == FALSE))
		return;
	
	if ((rtl8651_igmp_general_query == TRUE) ||
		(rtl8651_GlobalVariable.remainingQuery > 0))
	{
		
		rtl8651_tblDrvMCastRouteInfo_t	info;
		uint8						idx;
		rtl8651_tblDrv_vlanTable_t *		vlanp;

		#ifdef RTL8651_MCAST_DEBUG
		if (rtl8651_GlobalVariable.remainingQuery > 0)
			rtlglue_printf("[MCAST] Start-up Broadcast General Query: remaining[%d]\n", rtl8651_GlobalVariable.remainingQuery);
		else
			rtlglue_printf("[MCAST] Periodically broadcast General Query .....\n");
		#endif

		info.m_addr	= IGMP_ALL_HOSTS;

		vlanp = DrvTbl.vlan;
		for (idx = 0 ; idx < RTL8651_VLAN_NUMBER ; idx ++)
		{	
			if ((vlanp->valid) && (vlanp->internal))
			{
				info.netif		= NULL;
				info.ifInfo		= 0;
				info.vlan		= vlanp;
				info.port		= vlanp->memberPortMask;
				info.isPortMask= TRUE;
				if ((rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE) &&
					(rtl8651_GlobalVariable.MCastUpStream.vlan == vlanp))
				{
					info.port	&= ~(rtl8651_GlobalVariable.MCastUpStream.portMask);
					if (info.port == 0)
						goto next_vlan;
				}
				MCast_QueryPack(NULL, &info, rtl8651_drvMCastVersion, NULL, 0, NULL);
			}
		
		next_vlan:
			vlanp ++;
		}

		
		if (rtl8651_GlobalVariable.remainingQuery != 0)
		{	
			_rtl8651_mod_timer(&(rtl8651_GlobalVariable.query_timer), _sysUpSeconds + rtl8651_igmp_sqi);
			
			rtl8651_GlobalVariable.remainingQuery --;
		}
	}

	if (rtl8651_GlobalVariable.remainingQuery == 0)
	{
		if (rtl8651_igmp_general_query == TRUE)
			_rtl8651_mod_timer(&(rtl8651_GlobalVariable.query_timer), _sysUpSeconds + rtl8651_igmp_qi);
		else
			_rtl8651_del_timer(&(rtl8651_GlobalVariable.query_timer));
	}
	return;
}



void MCast_DumpMCast(void)
{
	rtl8651_tblDrvMCast_t *m_entry;
	rtl8651_tblDrvMCastMember_t *member;
	rtl8651_tblDrvMCastSource_t *source;

	rtlglue_printf("\nIGMP System report:\n");
	
	rtlglue_printf("System support latest IGMP version:\t\t[ IGMPv%d ]\n", RTL8651_MCAST_IGMP_MAX_VERSION);
	rtlglue_printf("System default IGMP version:\t\t\t[ IGMPv%d ]\n", rtl8651_drvMCastVersion);

	if (system_enable == FALSE)
	{
		rtlglue_printf("\nIP-multicast system Enable:\t\t[ NO ]\n\n");
		return;
	}

	
	rtlglue_printf("Timer:\n");
	rtlglue_printf("\tSystem specific:\n");
	if (rtl8651_drvMCastUpstream_Timeout == 0xffffffff)
		rtlglue_printf("\t\tUp-stream No-response timeout:\t\t[INFINITE]\n");
	else
		rtlglue_printf("\t\tUp-stream No-response timeout:\t\t%d sec.\n", rtl8651_drvMCastUpstream_Timeout);

	rtlglue_printf("\t\tDown-stream No-report timeout:\t\t%d sec.\n", rtl8651_drvMCastMember_ExpireTime);

	rtlglue_printf("\tIGMP msg related:\n");
	rtlglue_printf("\t\tQuery Interval:\t\t\t\t%d sec.\n", rtl8651_igmp_group_query_interval);
	rtlglue_printf("\t\tQuery Response Interval:\t\t%d sec.\n", rtl8651_igmp_qri);
	rtlglue_printf("\t\tOther Querier Present Interval:\t\t%d sec.\n", rtl8651_igmp_oqpi);
	rtlglue_printf("\t\tIGMP msg tolerance delay:\t\t%d sec\n", rtl8651_igmp_response_tolerance_delay);
	rtlglue_printf("\t\tSystem-up Query Count:\t\t\t%d\n", rtl8651_igmp_sqc);
	rtlglue_printf("\t\tSystem-up Query Interval:\t\t%d sec.\n", rtl8651_igmp_sqi);
	rtlglue_printf("\t\tLast Member Query Count:\t\t%d\n", rtl8651_igmp_lmqc);
	rtlglue_printf("\t\tLast Member Query Interval:\t\t%d sec.\n", rtl8651_igmp_lmqi);

	rtlglue_printf("\n\t\tPeriodic General Query:\t");
	if (rtl8651_igmp_general_query == TRUE)
		rtlglue_printf("[ on ]\n\t\t\tQuery Interval:\t\t\t%d sec.\n\n", rtl8651_igmp_qi);
	else
		rtlglue_printf("[ off ]\n\n");

	
	rtlglue_printf("System status:\n");
	rtlglue_printf("\t\tRemaining Start-up Query Count:\t\t%d\n", rtl8651_GlobalVariable.remainingQuery);
	if (rtl8651_igmp_general_query == TRUE)
		rtlglue_printf("\t\tNext general query remaining time:\t%d sec.\n", (rtl8651_GlobalVariable.query_timer.expires - _sysUpSeconds));
	rtlglue_printf("\n");

	
	rtlglue_printf("Resource Utilization:\n");
	rtlglue_printf("\t\t\t\tTotal\tUsage\tFree\n");
	rtlglue_printf("\tMulticast Entry\t\t%d\t%d\t%d\n", rtl8651_drvMaxMCastCnt, rtl8651_MCastEntryCnt, (rtl8651_drvMaxMCastCnt-rtl8651_MCastEntryCnt));
	rtlglue_printf("\tMulticast Member\t%d\t%d\t%d\n", rtl8651_SystemMCastMemberCnt, rtl8651_MemberCnt, (rtl8651_SystemMCastMemberCnt-rtl8651_MemberCnt));
	rtlglue_printf("\tSource List\t\t%d\t%d\t%d\n", rtl8651_SystemMCastSourceCnt, rtl8651_SourceListCnt, (rtl8651_SystemMCastSourceCnt-rtl8651_SourceListCnt));

	rtlglue_printf("Multicast Entries:\n");
	rtlglue_printf("===============================================================\n");

	rtlglue_drvMutexLock();

	m_entry = rtl8651_MCastTable;
	while (m_entry)
	{	
		rtlglue_printf("(%3u.%3u.%3u.%3u) [IGMPv%d] <Up ", NIPQUAD(m_entry->mcast_addr), m_entry->MulticastRouterVersion);
		if (m_entry->up_vlan)	
			rtlglue_printf("Vid[0x%02X] pMsk[0x%02X] ", m_entry->up_vlan->vid, m_entry->up_portMask);
		else
			rtlglue_printf("NOT Set ");

		if ((rtl8651_drvMCastUpstream_Timeout == 0xffffffff) || !(m_entry->up_vlan))
			rtlglue_printf("(NO Exp) >");
		else
			rtlglue_printf("(Exp: %d) >", (m_entry->timer_upstream.expires - _sysUpSeconds));

		if (rtl8651_drvMaxMCastMemberCnt)
			rtlglue_printf(" MEMBER[%d:%d]\n", m_entry->memberCnt, rtl8651_drvMaxMCastMemberCnt);
		else
			rtlglue_printf(" MEMBER[%d]\n", m_entry->memberCnt);

		
		member = m_entry->member;
		while (member)
		{	
			if (member->isForwarder == TRUE)
				rtlglue_printf("\t=> [O] ");
			else
				rtlglue_printf("\t=> [X] ");
			rtlglue_printf("VID[0x%02X] MBR[0x%02X] [v%d] %s(%d) ",
				member->vlan->vid, member->portMask, member->MulticastRouterVersion,
				((member->FilterMode == IGMP_FILTER_INCLUDE)?"Inc":"Exc"),
				((member->FilterMode == IGMP_FILTER_INCLUDE)?0:(member->timer_filterMode.expires - _sysUpSeconds)));

			if (member->state == IGMP_MEMBER_STATE_NORMAL)
			{
				rtlglue_printf("(NORMAL) NextQ(%d) ", (member->timer_query.expires - _sysUpSeconds));
				if (member->timer_query_expire.list.next)
					rtlglue_printf("Exp:(%d)\n", (member->timer_query_expire.expires - _sysUpSeconds));
				else
					rtlglue_printf("Exp:(%d)\n", (member->timer_state.expires - _sysUpSeconds));
			}else
				rtlglue_printf("(EXPIRE) Rem Q(%d) Exp(%d)\n", member->remainingQuery, (member->timer_state.expires - _sysUpSeconds));
			if (member->FilterMode == IGMP_FILTER_INCLUDE)
			{
				source = member->slist;
				while (source)
				{
					rtlglue_printf("\t\t+ %u.%u.%u.%u (Exp %d)\n", NIPQUAD(source->source), (source->timer.expires - _sysUpSeconds));
					source = source->next;
				}
			}else if (member->FilterMode == IGMP_FILTER_EXCLUDE)
			{
				source = member->slist;
				while (source)
				{
					if (source->timer.expires <= _sysUpSeconds)
						rtlglue_printf("\t\t- %u.%u.%u.%u\n", NIPQUAD(source->source));
					source = source->next;
				}
				source = member->slist;
				while (source)
				{
					if (source->timer.expires > _sysUpSeconds)
						rtlglue_printf("\t\t+ %u.%u.%u.%u (Exp %d)\n", NIPQUAD(source->source), (source->timer.expires - _sysUpSeconds));
					source = source->next;
				}
			}

			
			member = member->next;
		}

		
		m_entry = m_entry->next;
	}

	rtlglue_drvMutexUnlock();

	rtlglue_printf("===============================================================\n\n");
	return;
}

rtl8651_tblDrvMCast_t *MCast_FindMCastEntry(ipaddr_t mcast_addr, uint8 create, uint8 refresh)
{
	rtl8651_tblDrvMCast_t *m_entry = rtl8651_MCastTable;

	if (!IN_MULTICAST(mcast_addr))
		return NULL;

	while (m_entry)
	{
		if (m_entry->mcast_addr == mcast_addr)
		{	
			if (refresh == TRUE)
			{	
				if (rtl8651_drvMCastUpstream_Timeout == 0xffffffff)
					_rtl8651_del_timer(&m_entry->timer_upstream);
				else if (m_entry->up_vlan)
					_rtl8651_mod_timer(&m_entry->timer_upstream, _sysUpSeconds + rtl8651_drvMCastUpstream_Timeout);
			}
			return m_entry;
		}
		m_entry = m_entry->next;
	}

	if (create == TRUE)
		return MCast_CreateMCastEntry(mcast_addr);	
	else
		return NULL;
}

rtl8651_tblDrvMCast_t *MCast_CreateMCastEntry(ipaddr_t mcast_addr)
{
	rtl8651_tblDrvMCast_t *new_entry;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Create MCast Entry (%u.%u.%u.%u) System Entry Count: [%d]\n", NIPQUAD(mcast_addr), rtl8651_MCastEntryCnt);
	#endif

	rtlglue_drvMutexLock();

	
	if (rtl8651_MCastEntryCnt >= rtl8651_drvMaxMCastCnt)
		goto errout;

	
	new_entry = (rtl8651_tblDrvMCast_t *)MCast_malloc(sizeof(rtl8651_tblDrvMCast_t));
	if (!new_entry)
		goto errout;

	
	new_entry->mcast_addr				= mcast_addr;
	new_entry->MulticastRouterVersion		= rtl8651_drvMCastVersion;	
	new_entry->memberCnt				= 0;
	new_entry->member					= NULL;
	new_entry->up_vlan					= NULL;
	new_entry->up_portMask				= 0;

	
	new_entry->FilterMode					= IGMP_FILTER_NONE;
	new_entry->slist						= NULL;

	new_entry->next						= NULL;
	new_entry->prev						= NULL;

	
	_rtl8651_init_timer(&new_entry->timer_upstream);
	new_entry->timer_upstream.data = (uint32)new_entry;
	new_entry->timer_upstream.function =(void*)MCast_ClrMCastEntryUpStream;

	
	MCast_Entrylink(new_entry);

	
	rtl8651_MCastEntryCnt ++;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Create MCast Entry OK!! System Entry Count: [%d]\n", rtl8651_MCastEntryCnt);
	#endif

	rtlglue_drvMutexUnlock();
	return new_entry;
errout:
	rtlglue_drvMutexUnlock();
	return NULL;
}

uint8 MCast_MCastEntryMinVersion(rtl8651_tblDrvMCast_t *m_entry)
{
	rtl8651_tblDrvMCastMember_t *mptr;
	uint8 v = RTL8651_MCAST_IGMP_MAX_VERSION;

	if (!m_entry)
		goto out;
	mptr = m_entry->member;
	while (mptr)
	{
		if (mptr->MulticastRouterVersion < v)
			v = mptr->MulticastRouterVersion;
		mptr = mptr->next;
	}
out:
	return v;
}

void MCast_SetMCastEntryUpStream(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrv_vlanTable_t *vlan, uint16 up_port, uint8 RouterVersion)
{
	m_entry->up_vlan = vlan;
	m_entry->up_portMask = (1<<up_port);
	m_entry->MulticastRouterVersion = RouterVersion;
	if (rtl8651_drvMCastUpstream_Timeout != 0xffffffff)
		_rtl8651_mod_timer(&m_entry->timer_upstream, _sysUpSeconds + rtl8651_drvMCastUpstream_Timeout);
	m_entry->timer_upstream.data = (uint32)m_entry;
	m_entry->timer_upstream.function = (void*)MCast_ClrMCastEntryUpStream;
}

void MCast_ClrMCastEntryUpStream(uint32 expire)
{
	rtl8651_tblDrvMCast_t *m_entry = (rtl8651_tblDrvMCast_t *)expire;
	m_entry->up_vlan = NULL;
	m_entry->up_portMask = 0;
	m_entry->MulticastRouterVersion = rtl8651_drvMCastVersion;
}

void MCast_RemoveMCastEntry(rtl8651_tblDrvMCast_t *m_entry)
{
	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Remove MCast Entry (%u.%u.%u.%u) System Entry Count: [%d]\n", NIPQUAD(m_entry->mcast_addr), rtl8651_MCastEntryCnt);
	#endif

	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("[MCAST] *FLOW* Remove {%u.%u.%u.%u}\n",
		NIPQUAD(m_entry->mcast_addr));
	#endif
	rtlglue_drvMutexLock();

	_rtl8651_delMulticast(m_entry->mcast_addr, 0, 0);
	
	if (m_entry->MulticastRouterVersion == IGMPV2)
		MCast_LeavePack(m_entry, NULL);
	if (m_entry->MulticastRouterVersion == IGMPV3)
		MCast_ReportPack(m_entry, m_entry->MulticastRouterVersion, IGMP_FILTER_EXCLUDE, NULL, IGMP_FILTER_INCLUDE, NULL, NULL);
	
	while (m_entry->member)
		MCast_RemoveMember(m_entry, m_entry->member);
	
	MCast_SourceList_freeAll(m_entry->slist);
	
	_rtl8651_del_timer(&m_entry->timer_upstream);
	MCast_EntryUnlink(m_entry);
	MCast_free(m_entry);
	
	rtl8651_MCastEntryCnt --;
	rtlglue_drvMutexUnlock();

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Remove MCast Entry OK!! System Entry Count: [%d]\n", rtl8651_MCastEntryCnt);
	#endif
}


void MCast_Entrylink(rtl8651_tblDrvMCast_t *entry)
{
	rtlglue_drvMutexLock();
	entry->next = rtl8651_MCastTable;
	if (rtl8651_MCastTable)
		rtl8651_MCastTable->prev = entry;
	rtl8651_MCastTable = entry;
	rtlglue_drvMutexUnlock();
}

void MCast_EntryUnlink(rtl8651_tblDrvMCast_t *entry)
{
	rtlglue_drvMutexLock();
	if (rtl8651_MCastTable == entry)
		rtl8651_MCastTable = entry->next;
	if (entry->next)
		entry->next->prev = entry->prev;
	if (entry->prev)
		entry->prev->next = entry->next;
	entry->prev = NULL;
	entry->next = NULL;
	rtlglue_drvMutexUnlock();
}


rtl8651_tblDrvMCastMember_t *MCast_FindMember(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrv_vlanTable_t *vlan, uint16 port,  uint8 create, uint8 refresh, uint8 ModifyDataFlow)
{
	rtl8651_tblDrvMCastMember_t *member = m_entry->member;
	uint16 portMask = (1<<port);

	while (member)
	{
		if ((member->vlan == vlan) &&
			(member->portMask == portMask))
		{
			if (refresh == TRUE)
			{
				uint8 oldState = member->state;
				member->state = IGMP_MEMBER_STATE_NORMAL;

				
				_rtl8651_mod_timer(&(member->timer_state), _sysUpSeconds + rtl8651_drvMCastMember_ExpireTime);
				member->timer_state.data = (uint32)member;
				member->timer_state.function =(void*)MCast_MemberExpire;

				
				_rtl8651_mod_timer(&(member->timer_query), _sysUpSeconds + rtl8651_igmp_group_query_interval);
				member->timer_query.data = (uint32)member;
				member->timer_query.function = (void*)MCast_MemberSendQuery;

				
				_rtl8651_del_timer(&(member->timer_query_expire));

				
				if ((oldState != IGMP_MEMBER_STATE_NORMAL) && (rtl8651_multicastProcessOptions & MCAST_PROC_EXPIRE_STATE_DONT_FWD))
				{
					#ifdef RTL8651_MCAST_DATA_DEBUG
						rtlglue_printf("[MCAST] *FLOW* Add Member(0x%X)[0x%X] to {%u.%u.%u.%u}\n",
						member->vlan->vid, member->portMask, NIPQUAD(m_entry->mcast_addr));
					#endif
					_rtl8651_addMulticastMember(m_entry->mcast_addr, (1<<((member->vlan->vid)&(RTL8651_VLAN_NUMBER-1))), member->portMask, FALSE);
				}

			}
			return member;
		}
		member = member->next;
	}
	if (create == TRUE)
		
		return MCast_CreateMember(m_entry, vlan, portMask, ModifyDataFlow);
	else
		return NULL;
}

rtl8651_tblDrvMCastMember_t *MCast_CreateMember(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrv_vlanTable_t *vlan, uint16 portMask, uint8 ModifyDataFlow)
{
	rtl8651_tblDrvMCastMember_t *member;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Create Member (%u.%u.%u.%u) Member Count: [%d]\n", NIPQUAD(m_entry->mcast_addr), m_entry->memberCnt);
	#endif

	rtlglue_drvMutexLock();

	if ((!m_entry) ||
		((rtl8651_drvMaxMCastMemberCnt)&&(m_entry->memberCnt >= rtl8651_drvMaxMCastMemberCnt)) ||
		((rtl8651_drvMaxTotalMemberCnt)&&(rtl8651_MemberCnt >= rtl8651_drvMaxTotalMemberCnt)))
		goto errout;

	if ((member = (rtl8651_tblDrvMCastMember_t *)MCast_malloc(sizeof(rtl8651_tblDrvMCastMember_t))) == NULL)
		goto errout;

	
	member->MulticastRouterVersion	= rtl8651_drvMCastVersion;		
	member->vlan				= vlan;							
	member->state				= IGMP_MEMBER_STATE_NORMAL;	
	member->isForwarder			= TRUE;							

	member->next				= NULL;
	member->prev				= NULL;
	
	member->portMask			= portMask;						

	member->FilterMode			= IGMP_FILTER_NONE;
	member->slist				= NULL;

	
	MCast_MemberLink(m_entry, member);

	
	_rtl8651_init_timer(&member->timer_state);
	member->timer_state.data = (uint32) member;						
	member->timer_state.function = (void*)MCast_MemberExpire;		
	_rtl8651_mod_timer(&member->timer_state, _sysUpSeconds + rtl8651_drvMCastMember_ExpireTime);

	_rtl8651_init_timer(&member->timer_query);
	member->timer_query.data = (uint32) member;
	member->timer_query.function = (void*)MCast_MemberSendQuery;	
	_rtl8651_mod_timer(&member->timer_query, _sysUpSeconds + rtl8651_igmp_group_query_interval);

	_rtl8651_init_timer(&member->timer_filterMode);
	member->timer_filterMode.data = (uint32)member;
	member->timer_filterMode.function = (void*)MCast_MemberExcludeExpire;	
	_rtl8651_mod_timer(&member->timer_filterMode, _sysUpSeconds + rtl8651_igmp_gmi);

	
	_rtl8651_init_timer(&(member->timer_query_expire));
	member->timer_query_expire.data = (uint32) member;
	member->timer_query_expire.function = (void*)MCast_MemberExpire;		


	
	m_entry->memberCnt ++;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Create Member OK !! Member Count: [%d]\n", m_entry->memberCnt);
	#endif

	
	rtl8651_MemberCnt ++;

	if (ModifyDataFlow == TRUE)
	{
	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("[MCAST] *FLOW* Add Member(0x%X)[0x%X] to {%u.%u.%u.%u}\n",
		member->vlan->vid, member->portMask, NIPQUAD(m_entry->mcast_addr));
	#endif
	_rtl8651_addMulticastMember(m_entry->mcast_addr, (1<<((member->vlan->vid)&(RTL8651_VLAN_NUMBER-1))), member->portMask, FALSE);
	}

	rtlglue_drvMutexUnlock();
	return member;
errout:
	rtlglue_drvMutexUnlock();
	return NULL;
}

void MCast_RemoveMember(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrvMCastMember_t *member)
{
	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Remove Member (%u.%u.%u.%u) Member Count: [%d]\n", NIPQUAD(m_entry->mcast_addr), m_entry->memberCnt);
	#endif

	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("[MCAST] *FLOW* Remove Member(0x%X)[0x%X] from {%u.%u.%u.%u}\n",
		member->vlan->vid, member->portMask, NIPQUAD(m_entry->mcast_addr));
	#endif

	rtlglue_drvMutexLock();

	_rtl8651_delMulticast(m_entry->mcast_addr, (1<<((member->vlan->vid)&(RTL8651_VLAN_NUMBER-1))), member->portMask);

	
	_rtl8651_del_timer(&member->timer_filterMode);
	_rtl8651_del_timer(&member->timer_state);
	_rtl8651_del_timer(&member->timer_query);
	_rtl8651_del_timer(&member->timer_query_expire);

	MCast_MemberUnLink(member);

	
	MCast_SourceList_freeAll(member->slist);

	MCast_free(member);

	
	m_entry->memberCnt --;
	rtl8651_MemberCnt --;

	rtlglue_drvMutexUnlock();

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Remove Member OK !! Member Count: [%d]\n", m_entry->memberCnt);
	#endif
}

void MCast_DestroyMember(rtl8651_tblDrvMCastMember_t *member)
{
	rtl8651_tblDrvMCast_t *m_entry;

	rtlglue_drvMutexLock();

	m_entry = member->m_entry;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Destroy Member! (%u.%u.%u.%u) Member Count: [%d]\n", NIPQUAD(m_entry->mcast_addr), m_entry->memberCnt);
	#endif

	MCast_RemoveMember(m_entry, member);
	if (m_entry->memberCnt == 0)	
		MCast_RemoveMCastEntry(m_entry);

	rtlglue_drvMutexUnlock();
}

void MCast_MemberLink(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrvMCastMember_t *member)
{
	rtlglue_drvMutexLock();
	member->m_entry = m_entry;
	if (m_entry->member)
	{
		member->next = m_entry->member;
		m_entry->member->prev = member;
	}
	m_entry->member = member;
	rtlglue_drvMutexUnlock();
}

int32 MCast_MemberUnLink(rtl8651_tblDrvMCastMember_t *member)
{
	rtl8651_tblDrvMCast_t *m_entry;

	rtlglue_drvMutexLock();
	m_entry = member->m_entry;
	if (!m_entry)
		goto errout;

	if (m_entry->member == member)
		m_entry->member = member->next;
	if (member->next)
		member->next->prev = member->prev;
	if (member->prev)
		member->prev->next = member->next;

	member->m_entry = NULL;
	member->prev = NULL;
	member->next = NULL;

	rtlglue_drvMutexUnlock();
	return SUCCESS;
errout:
	rtlglue_drvMutexUnlock();
	return FAILED;
}


void MCast_MemberExcludeExpire(uint32 expire)
{
	rtl8651_tblDrvMCastMember_t *member = (rtl8651_tblDrvMCastMember_t *)expire;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Member FilterMode Change! (%u.%u.%u.%u) time [%d] \n", NIPQUAD(member->m_entry->mcast_addr), _sysUpSeconds);
	#endif

	if (member->FilterMode == IGMP_FILTER_EXCLUDE)
	{
		rtl8651_tblDrvMCastSource_t *ptr;
		member->FilterMode = IGMP_FILTER_INCLUDE;
		ptr = member->slist;
		while (ptr)
		{
			if (ptr->timer.expires <= _sysUpSeconds)
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
				ptr = ptr->next;
		}
		if (member->MulticastRouterVersion == IGMPV3)
		{
			_rtl8651_delMulticast(member->m_entry->mcast_addr, 0, 0);
			if (!member->slist)
				MCast_MemberExpire((uint32)member);	
		}
	}

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Member FilterMode Change OK! time [%d] \n", _sysUpSeconds);
	#endif

}


void MCast_MemberExpire(uint32 expire)
{
	rtl8651_tblDrvMCastMember_t *member = (rtl8651_tblDrvMCastMember_t *)expire;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Member EXPIRE! (%u.%u.%u.%u) time [%d] --------> migth be Destroyed at [%d]\n", NIPQUAD(member->m_entry->mcast_addr), _sysUpSeconds, _sysUpSeconds + rtl8651_drvMCastMember_DestroyTime);
	#endif

	member->state			= IGMP_MEMBER_STATE_WILLTIMEOUT;
	member->remainingQuery	= rtl8651_igmp_lmqc;

	
	_rtl8651_del_timer(&(member->timer_query_expire));

	if (!member->remainingQuery)
	{
		MCast_DestroyMember(member);
		return;
	}

	
	if (rtl8651_multicastProcessOptions & MCAST_PROC_EXPIRE_STATE_DONT_FWD)
	{
		#ifdef RTL8651_MCAST_DATA_DEBUG
		rtlglue_printf("[MCAST] *FLOW* Remove Member(0x%X)[0x%X] from {%u.%u.%u.%u}\n",
		member->vlan->vid, member->portMask, NIPQUAD(member->m_entry->mcast_addr));
		#endif
		_rtl8651_delMulticast(member->m_entry->mcast_addr, (1<<((member->vlan->vid)&(RTL8651_VLAN_NUMBER-1))), member->portMask);
	}



	member->timer_state.data = (uint32) member;						
	member->timer_state.function = (void*)MCast_DestroyMember;		
	_rtl8651_mod_timer(&member->timer_state, _sysUpSeconds + rtl8651_drvMCastMember_DestroyTime);

	if (member->isForwarder == TRUE)
		MCast_MemberSendQuery(member);
}

void MCast_MemberSendQuery(rtl8651_tblDrvMCastMember_t *member)
{
	uint32 expire_time;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Send Query to (%u.%u.%u.%u) time [%d]\n", NIPQUAD(member->m_entry->mcast_addr), _sysUpSeconds);
	#endif

	
	member->isForwarder = TRUE;

	if (member->state == IGMP_MEMBER_STATE_NORMAL)
	{	
		expire_time = rtl8651_igmp_group_query_interval;
		
		_rtl8651_mod_timer(&(member->timer_query_expire), _sysUpSeconds + rtl8651_igmp_qri + rtl8651_igmp_response_tolerance_delay);
		member->timer_query_expire.data = (uint32)member;
		member->timer_query_expire.function = (void*)MCast_MemberExpire;		
	}else
	{	
		if (!member->remainingQuery)
			return;
		expire_time = rtl8651_igmp_lmqi;
		member->remainingQuery --;
	}

	
	member->timer_query.data = (uint32)member;
	member->timer_query.function = (void*)MCast_MemberSendQuery;
	_rtl8651_mod_timer(&member->timer_query, _sysUpSeconds + expire_time);

	
	if ((_sysUpSeconds + rtl8651_igmp_qri + rtl8651_igmp_response_tolerance_delay) > (member->timer_state.expires))
		_rtl8651_mod_timer(&member->timer_state, _sysUpSeconds + rtl8651_igmp_qri + rtl8651_igmp_response_tolerance_delay);

	
	if (member->MulticastRouterVersion == IGMPV1)
	{	
		rtl8651_tblDrvMCastRouteInfo_t info;

		info.netif		= NULL;
		info.ifInfo		= 0;
		info.m_addr	= IGMP_ALL_HOSTS;
		info.netif		= NULL;
		info.ifInfo		= 0;
		info.vlan		= member->vlan;
		info.port		= member->portMask;
		info.isPortMask= TRUE;

		MCast_QueryPack(NULL, &info, IGMPV1, NULL, 0, NULL);
	}else
		
		MCast_QueryPack(member, NULL, member->MulticastRouterVersion, NULL, 0, NULL);
}



rtl8651_tblDrvMCastSource_t *MCast_SourceList_Create()
{
	rtl8651_tblDrvMCastSource_t *ret = NULL;
	if ((ret = MCast_malloc(sizeof(rtl8651_tblDrvMCastSource_t))) == NULL)
		return NULL;
	_rtl8651_init_timer(&ret->timer);
	ret->timer.data = (uint32)ret;
	ret->timer.function = (void*)MCast_SourceList_Expire;
	rtl8651_SourceListCnt ++;

	return ret;
}

void MCast_SourceList_Link(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrvMCastMember_t *member, rtl8651_tblDrvMCastSource_t *src)
{
	rtl8651_tblDrvMCastSource_t **listpp;
	if ((!src) || (m_entry && member) || (!m_entry && !member))
		return;

	if (m_entry)
		listpp = &m_entry->slist;
	else
	{
		listpp = &member->slist;
		src->mbr = member;
	}

	src->prev = NULL;
	src->next = (*listpp);
	if (*listpp)
		(*listpp)->prev = src;
	*listpp = src;
}


void MCast_SourceList_UnLink(rtl8651_tblDrvMCastSource_t *src)
{
	if (src->next)
		src->next->prev = src->prev;
	if (src->prev)
		src->prev->next = src->next;
	if (src->mbr && (src->mbr->slist == src))
		src->mbr->slist = src->next;

	src->prev = NULL;
	src->next = NULL;
}

rtl8651_tblDrvMCastSource_t* MCast_SourceList_inList(ipaddr_t addr, rtl8651_tblDrvMCastSource_t *slist)
{
	rtl8651_tblDrvMCastSource_t *ptr = slist;

	if (!ptr)
		return ptr;

	while (ptr)
	{
		if (ptr->source == addr)
			return ptr;
		ptr = ptr->next;
	}

	return NULL;
}


uint32 MCast_SourceList_EntryCnt(rtl8651_tblDrvMCastSource_t *slist)
{
	rtl8651_tblDrvMCastSource_t *ptr = slist;
	uint32 cnt = 0;

	while (ptr)
	{
		cnt ++;
		ptr = ptr->next;
	}

	return cnt;
}



rtl8651_tblDrvMCastSource_t *MCast_SourceList_freeCurrent(rtl8651_tblDrvMCastSource_t *s_entry)
{
	rtl8651_tblDrvMCastSource_t *ret;
	if (!s_entry)
		return NULL;

	
	ret = s_entry->next;

	MCast_SourceList_UnLink(s_entry);
	_rtl8651_del_timer(&s_entry->timer);
	MCast_free(s_entry);
	rtl8651_SourceListCnt --;

	return ret;
}


void MCast_SourceList_freeAll(rtl8651_tblDrvMCastSource_t *slist)
{
	rtl8651_tblDrvMCastSource_t *ptr = slist;
	if (!ptr)
		return;

	while(ptr->prev)		
		ptr = ptr->prev;

	if (ptr->mbr)
	{
		assert(ptr->mbr->slist == ptr);
		ptr->mbr->slist = NULL;
	}

	while (ptr)
		ptr = MCast_SourceList_freeCurrent(ptr);

	return;
}


uint32 MCast_SourceList_removal(
	rtl8651_tblDrvMCastSource_t *s1,
	uint8 s1constraint,
	uint8 s1action,
	rtl8651_tblDrvMCastSource_t *s2,
	uint8 s2constraint,
	uint8 newSrcAction,
	rtl8651_tblDrvMCastSource_t **removalResult)
{
	rtl8651_tblDrvMCastSource_t *tmp;
	rtl8651_tblDrvMCastSource_t *ptr = s1;
	rtl8651_tblDrvMCastSource_t *ret = NULL;
	rtl8651_tblDrvMCastSource_t* newp = NULL;
	uint32 count = 0;

	if (!ptr)
		goto out;

	while(ptr)
	{
		if (
			((s1constraint == IGMP_SRCLIST_CONSTRAINT_NONEXPIRE) && (ptr->timer.expires <= _sysUpSeconds)) ||
			((s1constraint == IGMP_SRCLIST_CONSTRAINT_EXPIRE) && (ptr->timer.expires > _sysUpSeconds))
		)
			goto next_s1;
		if ((tmp = MCast_SourceList_inList(ptr->source, s2)) == NULL)	
		{
match:
			if ((newp = MCast_SourceList_Create()) == NULL)
				goto out;	

			newp->next = ret;
			if (ret)
				ret->prev = newp;
			ret = newp;

			newp->source = ptr->source;
			newp->mbr = NULL;
			_rtl8651_init_timer(&newp->timer);

			
			if (s1action == IGMP_SRCLIST_ACTION_SETLMQT)
				_rtl8651_mod_timer(&ptr->timer, _sysUpSeconds + rtl8651_igmp_lmqt);

			if (newSrcAction== IGMP_SRCLIST_ACTION_SETLMQT)
				_rtl8651_mod_timer(&newp->timer, _sysUpSeconds + rtl8651_igmp_lmqt);

			count ++;
		}else
		{
			if (
				((s2constraint == IGMP_SRCLIST_CONSTRAINT_NONEXPIRE) && (tmp->timer.expires <= _sysUpSeconds)) ||
				((s2constraint == IGMP_SRCLIST_CONSTRAINT_EXPIRE) && (tmp->timer.expires > _sysUpSeconds))
				)
				goto match;
		}
next_s1:
		ptr = ptr->next;
	}
out:
	*removalResult = ret;
	return count;
}




void MCast_SourceList_addList(rtl8651_tblDrvMCastMember_t *member, rtl8651_tblDrvMCastSource_t *add, uint8 refreshAdded, uint8 refreshExist)
{
	rtl8651_tblDrvMCastSource_t *tmp, *ptr = add;

	while(ptr)
	{
		if ((tmp = MCast_SourceList_inList(ptr->source, member->slist)) != NULL)
		{	
			ptr = MCast_SourceList_freeCurrent(ptr);
			if (refreshExist == TRUE)
				_rtl8651_mod_timer(&tmp->timer, (_sysUpSeconds + rtl8651_igmp_gmi));
		}else
		{	
			
			tmp = ptr;
			ptr = ptr->next;
			MCast_SourceList_UnLink(tmp);
			MCast_SourceList_Link(NULL, member, tmp);
			if (refreshAdded == TRUE)
				_rtl8651_mod_timer(&tmp->timer, (_sysUpSeconds + rtl8651_igmp_gmi));
		}
	}
}




int32 MCast_SourceList_dup(rtl8651_tblDrvMCastMember_t *member, rtl8651_tblDrvMCastSource_t *org, rtl8651_tblDrvMCastSource_t **dup)
{
	rtl8651_tblDrvMCastSource_t *ptr, *newp;

	if ((!org) || (!dup && !member) || (dup && member))
		return FAILED;

	if (dup)
		*dup = NULL;

	ptr = org;

	while (ptr)
	{
		newp = MCast_SourceList_Create();
		if (!newp)
			return FAILED;
		newp->source = ptr->source;
		if (member)
			MCast_SourceList_addList(member, newp, TRUE, TRUE);
		else
		{
			newp->next = *dup;
			if (*dup)
				(*dup)->prev = newp;
			*dup = newp;
		}
		ptr = ptr->next;
	}
	return SUCCESS;
}


void MCast_SourceList_Expire(rtl8651_tblDrvMCastSource_t *s_entry)
{
	rtl8651_tblDrvMCastMember_t *mbr;
	if ((!s_entry) || (!s_entry->mbr))
		return;

	mbr = s_entry->mbr;

	if (mbr->FilterMode == IGMP_FILTER_EXCLUDE)
	{
		if (mbr->MulticastRouterVersion == IGMPV3)
			_rtl8651_delMulticast(mbr->m_entry->mcast_addr, 0, 0);
	}else if (mbr->FilterMode == IGMP_FILTER_INCLUDE)
	{
		
		MCast_SourceList_freeCurrent(s_entry);
		if (mbr->MulticastRouterVersion == IGMPV3)
		{
			
			_rtl8651_delMulticast(mbr->m_entry->mcast_addr, 0, 0);

			if (!mbr->slist)
				MCast_MemberExpire((uint32)mbr);
		}
	}else
	{
		assert(0);
	}
}


uint32 MCast_SourceList_getCache(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrvMCastSource_t **newCache)
{
	rtl8651_tblDrvMCastMember_t *mptr, *firstmptr, *firstExcludeMember;
	uint8 mFilterMode = IGMP_FILTER_NONE;
	uint8 allMemberNone;
	rtl8651_tblDrvMCastSource_t  *ret = NULL, *tmp, *tmp_2, *newp;

	if (!m_entry)
		goto list_add_done;

	assert(m_entry->memberCnt > 0);
	firstmptr = m_entry->member;
	
	firstExcludeMember = NULL;				
	allMemberNone = TRUE;	
	mptr = firstmptr;
	while (mptr)
	{	

		
		if ((mptr->FilterMode == IGMP_FILTER_INCLUDE) && (mptr->slist == NULL))
			mptr->FilterMode = IGMP_FILTER_EXCLUDE;

		if (mptr->FilterMode == IGMP_FILTER_EXCLUDE)
		{
			mFilterMode = IGMP_FILTER_EXCLUDE;
			if (!firstExcludeMember)
				firstExcludeMember = mptr;
		}
		if (mptr->FilterMode != IGMP_FILTER_NONE)
			allMemberNone = FALSE;
		mptr = mptr->next;
	}
	if (allMemberNone == TRUE)
	{
		mFilterMode = IGMP_FILTER_NONE;
		ret = NULL;
		goto list_add_done; 
	}
	
	if (mFilterMode == IGMP_FILTER_NONE)	
		mFilterMode = IGMP_FILTER_INCLUDE;

	
	ret = NULL;	
	if (mFilterMode == IGMP_FILTER_INCLUDE)
	{	
		
		
		mptr = firstmptr;
		while (mptr)
		{
			assert(mptr->MulticastRouterVersion >= IGMPV3);
			tmp = mptr->slist;
			while (tmp)
			{
				if (!MCast_SourceList_inList(tmp->source, ret))
				{
					if ((newp = MCast_SourceList_Create()) == NULL)
						goto list_add_done;
					
					newp->source = tmp->source;
					newp->next = ret;
					if (ret)
						ret->prev = newp;
					ret = newp;
				}
				tmp = tmp->next;
			}
			mptr = mptr->next;
		}
	}else
	{	
		
		
		
		
		
		
		

		
		assert(firstExcludeMember);
		tmp = firstExcludeMember->slist;
		while (tmp)
		{
			if ((newp = MCast_SourceList_Create()) == NULL)
				goto list_init_done;

			newp->source = tmp->source;
			newp->next = ret;
			if (ret)
				ret->prev = newp;
			ret = newp;

			tmp = tmp->next;			
		}
	list_init_done:
		
		mptr = m_entry->member;
		while (mptr)
		{
			if (mptr->MulticastRouterVersion < IGMPV3)	
				goto ex_next_member;
			if (!ret)
				goto list_add_done;
			if (mptr != firstExcludeMember)
			{
				if (mptr->FilterMode == IGMP_FILTER_INCLUDE)
				{
					tmp = mptr->slist;
					while (tmp)
					{
						if ((tmp_2 = MCast_SourceList_inList(tmp->source, ret)) != NULL)
						{
							if (tmp_2 == ret)
								ret = MCast_SourceList_freeCurrent(tmp_2);	
							else
								MCast_SourceList_freeCurrent(tmp_2);		
						}
						if (!ret)
							goto list_add_done;

						tmp = tmp->next;
					}
				}else
				{	
					tmp = ret;
					while (tmp)
					{
						if (!MCast_SourceList_inList(tmp->source, mptr->slist))
						{	
							if (tmp == ret)
							{
								ret = MCast_SourceList_freeCurrent(tmp);	
								tmp = ret;
							} else
								tmp = MCast_SourceList_freeCurrent(tmp);	
						} else
							tmp = tmp->next;
						if (!ret)
							goto list_add_done;
					}
				}
			}
	ex_next_member:
			mptr = mptr->next;
		}
	}
list_add_done:
	if (newCache)
		*newCache = ret;
	return mFilterMode;
}



ipaddr_t MCast_GetIpByVlan(rtl8651_tblDrv_vlanTable_t *vlanp, uint32 vid)
{
	rtl8651_tblDrv_networkIntfTable_t	*netif;
	rtl8651_tblDrv_ipIntfEntry_t			*ipIf;
	rtl8651_tblDrv_ipIntfIpAddrEntry_t	*ipEntry;

	if (!vlanp && !vid)
	{
		return 0;
	}

	if (vid)
		netif = _rtl8651_getNetworkIntfEntryByVid(vid);
	else
		netif = _rtl8651_getNetworkIntfEntryByVid(vlanp->vid);

	if (netif)
	{
		SLIST_FOREACH(ipIf, &netif->IpHead, nextIp)
		{
			SLIST_FOREACH(ipEntry, &ipIf->ipAddr, nextIpAddr)
			{
				if (ipEntry->ipAddr)
					return (ipEntry->ipAddr);
			}
		}
	}
	return 0;
}

void inline MCast_asicSetActualVlanPort(uint16 *pPortMask, uint32 *pVlanMask, rtl8651_tblDrv_networkIntfTable_t *netif, rtl8651_tblDrv_vlanTable_t* vlan, uint16 orgPortMask, uint32 intfInfo)
{
	assert(pPortMask && pVlanMask);
	assert(netif && vlan);

	switch (netif->linkLayerType) {
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		case RTL8651_LL_PPTP:
		case RTL8651_LL_L2TP:
		{
			rtl8651_tblDrv_pppTable_t *ptb;
			assert(intfInfo);
			
			ptb = ((rtl8651_tblDrvSession_t *)intfInfo)->session_t;

			*pVlanMask = (ptb->initiated)?(1<<(tunnel.loopBackVid & (RTL8651_VLAN_NUMBER - 1))):0;
			*pPortMask = (ptb->initiated)?(1<<CONFIG_RTL865XB_PPTPL2TP_PORT):0;
		}
			break;
#endif 
		default:
			*pPortMask = orgPortMask;
			*pVlanMask = 1<<(vlan->vid & (RTL8651_VLAN_NUMBER - 1));
	}
}


int32 MCast_SetRouteInfoAndGetIntfIp(ipaddr_t *gip, rtl8651_tblDrvMCastRouteInfo_t *info)
{	
	ipaddr_t ip;

	assert(info->vlan);	

	if ((!(info->netif)) && 
		((info->netif = _rtl8651_getNetworkIntfEntryByVid(info->vlan->vid)) == NULL))
	{
		return FAILED;
	}
	switch (info->netif->linkLayerType)
	{
		case RTL8651_LL_VLAN:
				if (info->ifInfo)
				{	
					ip = info->ifInfo;
				}else
				{
					if (SLIST_FIRST(&(info->netif->IpHead)) == NULL)
						return FAILED;
					if (SLIST_FIRST(&(SLIST_FIRST(&(info->netif->IpHead)))->ipAddr) == NULL)
						return FAILED;
					if ((SLIST_FIRST(&(SLIST_FIRST(&(info->netif->IpHead)))->ipAddr))->ipAddr == 0)
						return FAILED;
					ip = info->ifInfo = SLIST_FIRST(&(SLIST_FIRST(&(info->netif->IpHead))->ipAddr))->ipAddr;
				}
			break;
		case RTL8651_LL_PPPOE:
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		case RTL8651_LL_PPTP:
		case RTL8651_LL_L2TP:
#endif 
				if (!info->ifInfo)
				{
					if ((info->ifInfo = (uint32)(CTAILQ_FIRST(&(info->netif->sessionHead)))) == 0)
					{
						return FAILED;
					}
				}
				if ((!((rtl8651_tblDrvSession_t*)(info->ifInfo))->ip_t) ||
					(!((rtl8651_tblDrvSession_t*)(info->ifInfo))->ip_t->ipAddr))
				{
					return FAILED;
				}
				ip = ((rtl8651_tblDrvSession_t*)(info->ifInfo))->ip_t->ipAddr;
			break;
		default:
			return FAILED;
	}

	if (gip)
		*gip = ip;
	return SUCCESS;
}


void MCast_Proc_ISIN(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	
	
	MCast_SourceList_addList(member, slist, TRUE, TRUE);
}

static inline void MCast_Proc_ISEX(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	rtl8651_tblDrvMCastSource_t *ptr, *tmp;
	
	
	if (member->FilterMode == IGMP_FILTER_INCLUDE)
	{
		
		ptr = member->slist;
		while(ptr)
		{
			if (!MCast_SourceList_inList(ptr->source, slist))
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
				ptr = ptr->next;
		}
		
		ptr = slist;
		while(ptr)
		{
			if (!MCast_SourceList_inList(ptr->source, member->slist))
			{	
				tmp = ptr->next;
				MCast_SourceList_UnLink(ptr);
				MCast_SourceList_Link(NULL, member, ptr);
				_rtl8651_del_timer(&ptr->timer);	
				ptr->timer.expires	= _sysUpSeconds;
				ptr = tmp;
			}else
			{	
				ptr = MCast_SourceList_freeCurrent(ptr);
			}
		}
	}else		
	{
		
		ptr = member->slist;
		while(ptr)
		{
			if (!MCast_SourceList_inList(ptr->source, slist))
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
				ptr = ptr->next;
		}
		
		MCast_SourceList_addList(member, slist, TRUE, FALSE);
	}
	member->FilterMode = IGMP_FILTER_EXCLUDE;										
	_rtl8651_mod_timer(&member->timer_filterMode, (_sysUpSeconds + rtl8651_igmp_gmi));	
}

static inline void MCast_Proc_TOIN(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	rtl8651_tblDrvMCastSource_t *sendQsrc = NULL;
	uint32 sendQsrcCnt = 0;

	
	
	
	if (member->FilterMode == IGMP_FILTER_INCLUDE)	
		sendQsrcCnt = MCast_SourceList_removal(member->slist, 0, IGMP_SRCLIST_ACTION_SETLMQT, slist, 0, 0, &sendQsrc);
	else	
		sendQsrcCnt = MCast_SourceList_removal(member->slist, IGMP_SRCLIST_CONSTRAINT_NONEXPIRE, IGMP_SRCLIST_ACTION_SETLMQT, slist, 0, 0, &sendQsrc);
	
	MCast_SourceList_addList(member, slist, TRUE, TRUE);
	if (sendQ == TRUE)
	{
		if (member->FilterMode == IGMP_FILTER_INCLUDE)
		{
			if (sendQsrcCnt)	
				MCast_QueryPack(member, NULL, member->MulticastRouterVersion, sendQsrc, sendQsrcCnt, NULL);
		}else
		{	
			if (sendQsrcCnt > 0)	
				MCast_QueryPack(member, NULL, member->MulticastRouterVersion, sendQsrc, sendQsrcCnt, NULL);

			MCast_MemberExpire((uint32)member);
		}
	}
	
	MCast_SourceList_freeAll(sendQsrc);
}

static inline void MCast_Proc_TOEX(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	rtl8651_tblDrvMCastSource_t *ptr;
	rtl8651_tblDrvMCastSource_t *sendQsrc = NULL;
	uint32 sendQsrcCnt = 0;

	
	
	if (member->FilterMode == IGMP_FILTER_INCLUDE)
	{
		
		ptr = member->slist;
		while(ptr)
		{
			if (!MCast_SourceList_inList(ptr->source, slist))
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
				ptr = ptr->next;
		}

		sendQsrcCnt = 0;
		ptr = member->slist;
		while (ptr)
		{	
			_rtl8651_mod_timer(&ptr->timer, (_sysUpSeconds + rtl8651_igmp_lmqt));
			sendQsrcCnt ++;
			ptr = ptr->next;
		}
		if ((sendQ == TRUE) && (sendQsrcCnt > 0))	
			MCast_QueryPack(member, NULL, member->MulticastRouterVersion, member->slist, sendQsrcCnt, NULL);
		
		MCast_SourceList_addList(member, slist, FALSE, FALSE);
		
		member->FilterMode = IGMP_FILTER_EXCLUDE;
		_rtl8651_mod_timer(&member->timer_filterMode, (_sysUpSeconds + rtl8651_igmp_gmi));
	}else
	{	
		
		ptr = member->slist;
		while(ptr)
		{
			if (!MCast_SourceList_inList(ptr->source, slist))
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
				ptr = ptr->next;
		}
		
		
		sendQsrcCnt = MCast_SourceList_removal(slist, IGMP_SRCLIST_CONSTRAINT_NONE, IGMP_SRCLIST_ACTION_NONE, member->slist, IGMP_SRCLIST_CONSTRAINT_EXPIRE, IGMP_SRCLIST_ACTION_NONE, &sendQsrc);
		if ((sendQ == TRUE) && (sendQsrcCnt > 0))
		{
			MCast_QueryPack(member, NULL, member->MulticastRouterVersion, sendQsrc, sendQsrcCnt, NULL);
			MCast_SourceList_freeAll(sendQsrc);
		}
		
		ptr = member->slist;
		while (ptr)
		{
			if (ptr->timer.expires <= _sysUpSeconds)
				goto next_ptr;
			if (MCast_SourceList_inList(ptr->source, slist))
				_rtl8651_mod_timer(&ptr->timer, (_sysUpSeconds * rtl8651_igmp_lmqt));
			next_ptr:
				ptr = ptr->next;
		}
		
		ptr = slist;
		while(ptr)
		{
			if (MCast_SourceList_inList(ptr->source, member->slist))
			{	
				ptr = MCast_SourceList_freeCurrent(ptr);
			}else
			{	
				rtl8651_tblDrvMCastSource_t *tmp;
				tmp = ptr->next;
				MCast_SourceList_UnLink(ptr);
				MCast_SourceList_Link(NULL, member, ptr);
				if (member->timer_filterMode.expires > _sysUpSeconds)
					_rtl8651_mod_timer(&ptr->timer, (member->timer_filterMode.expires));
				else
					_rtl8651_mod_timer(&ptr->timer, (_sysUpSeconds + 1));	
				ptr = tmp;
			}
		}
		
		_rtl8651_mod_timer(&member->timer_filterMode, (_sysUpSeconds + rtl8651_igmp_gmi));
	}
}

static inline void MCast_Proc_ALLOW(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	
	
	MCast_SourceList_addList(member, slist, TRUE, TRUE);
}

static inline void MCast_Proc_BLOCK(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	rtl8651_tblDrvMCastSource_t *slistP = slist;
	rtl8651_tblDrvMCastSource_t *ptr, *tmp;
	rtl8651_tblDrvMCastSource_t *sendQsrc = NULL;
	uint32 sendQsrcCnt = 0;
	
	
	
	if (member->FilterMode == IGMP_FILTER_INCLUDE)
	{
		ptr = member->slist;
		sendQsrcCnt = 0;
		sendQsrc = NULL;
		while (ptr)
		{
			if ((tmp = MCast_SourceList_inList(ptr->source, slistP)) != NULL)
			{	
				if (slistP == tmp)
					slistP = tmp->next;
				MCast_SourceList_UnLink(tmp);
				tmp->next = sendQsrc;
				if (sendQsrc)
					sendQsrc->prev = tmp;
				sendQsrc = tmp;
				sendQsrcCnt ++;

				_rtl8651_mod_timer(&ptr->timer, (_sysUpSeconds + rtl8651_igmp_lmqt));
			}
			ptr = ptr->next;
		}
		if ((sendQ == TRUE) && (sendQsrcCnt > 0))		
			MCast_QueryPack(member, NULL, member->MulticastRouterVersion, sendQsrc, sendQsrcCnt, NULL);
		
		MCast_SourceList_freeAll(slistP);
		if (sendQsrcCnt > 0)
			MCast_SourceList_freeAll(sendQsrc);
	}else
	{	

		
		sendQsrcCnt = MCast_SourceList_removal(slist, IGMP_SRCLIST_CONSTRAINT_NONE, IGMP_SRCLIST_ACTION_NONE, member->slist, IGMP_SRCLIST_CONSTRAINT_EXPIRE, IGMP_SRCLIST_ACTION_NONE, &sendQsrc);
		if ((sendQ == TRUE) && (sendQsrcCnt>0))	
			MCast_QueryPack(member, NULL, member->MulticastRouterVersion, sendQsrc, sendQsrcCnt, NULL);

		if (sendQsrcCnt > 0)
		MCast_SourceList_freeAll(sendQsrc);

		ptr = slist;
		while (ptr)
		{	
			if (MCast_SourceList_inList(ptr->source, member->slist))
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
			{	
				tmp = ptr->next;
				MCast_SourceList_UnLink(ptr);
				MCast_SourceList_Link(NULL, member, ptr);
				if (member->timer_filterMode.expires > _sysUpSeconds)
					_rtl8651_mod_timer(&ptr->timer, (member->timer_filterMode.expires));
				else
					_rtl8651_mod_timer(&ptr->timer, (_sysUpSeconds + 1));	
				ptr = tmp;
			}
		}
	}
}


uint32 MCast_ProcessGroupRecord(
	rtl8651_tblDrvMCast_t *m_entry,
	rtl8651_tblDrvMCastMember_t *member,
	uint8 RecordType,
	rtl8651_tblDrvMCastSource_t *slist,
	uint16 srcCnt,
	uint8 sendQ,
	rtl8651_tblDrvMCastSource_t **newCache)
{
	uint8 mFilterMode = IGMP_FILTER_EXCLUDE;

	if (newCache)
		*newCache = NULL;

	if ((!m_entry) || (!member))
		goto out;

	if (member->FilterMode == IGMP_FILTER_NONE)
	{
		MCast_SourceList_freeAll(member->slist);
		MCast_SourceList_dup(member, slist, NULL);
		switch (RecordType)
		{
			case IGMP_MODE_ISIN:
			case IGMP_MODE_TOIN:
			case IGMP_MODE_ALLOW:
				member->FilterMode = IGMP_FILTER_INCLUDE;
				break;
			case IGMP_MODE_ISEX:
			case IGMP_MODE_TOEX:
			case IGMP_MODE_BLOCK:
				member->FilterMode = IGMP_FILTER_EXCLUDE;
				break;
		}
	}
	
	switch (RecordType)
	{
		case IGMP_MODE_ISIN:
			MCast_Proc_ISIN(member, slist, srcCnt, sendQ);
			break;
		case IGMP_MODE_ISEX:
			MCast_Proc_ISEX(member, slist, srcCnt, sendQ);
			break;
		case IGMP_MODE_TOIN:
			MCast_Proc_TOIN(member, slist, srcCnt, sendQ);
			break;
		case IGMP_MODE_TOEX:
			MCast_Proc_TOEX(member, slist, srcCnt, sendQ);
			break;
		case IGMP_MODE_ALLOW:
			MCast_Proc_ALLOW(member, slist, srcCnt, sendQ);
			break;
		case IGMP_MODE_BLOCK:
			MCast_Proc_BLOCK(member, slist, srcCnt, sendQ);
			break;
		default:
			goto out;
	}

	
	mFilterMode = MCast_SourceList_getCache(m_entry, newCache);
out:
	return mFilterMode;
}


int32 MCast_query(struct rtl_pktHdr *pkthdr, struct ip *ip, struct igmp *ig)
{
	uint16						igmp_length = 0;
	uint8						version = 0;
	ipaddr_t						dip = 0;
	ipaddr_t						mcast_addr = 0;
	rtl8651_tblDrvMCast_t *			m_entry;
	rtl8651_tblDrvMCastMember_t *	member;
	rtl8651_tblDrv_vlanTable_t *		vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	uint8						keep_pkt = FALSE;	

	

	igmp_length = ntohs(ip->ip_len) - ((ip->ip_vhl&0xf)<<2);

	if (igmp_length == 8)
	{
		if (ig->igmp_code)	
			version = IGMPV2;
		else
			version = IGMPV1;
	}else if (igmp_length >= 12)
		version = IGMPV3;
	else
		goto errout;	

	
	mcast_addr = PKTGET_UINT32_UNALIGNED(&ig->igmp_group);

	
	{
		rtl8651_tblDrvMCastFilter_t *filter;
		filter = _rtl8651_multicast_FindFilter(mcast_addr);
		if (filter)
		{
			switch (filter->action)
			{
				case (MCAST_FILTER_ACT_MIRROR|MCAST_FILTER_ACT_NOPROC):
					
					goto errout;
					break;
				case (MCAST_FILTER_ACT_MIRROR):
					
					keep_pkt = TRUE;
					break;
				case (MCAST_FILTER_ACT_NOPROC):
					
					goto freeout;
					break;
			}
		}
	}

	if (!IN_MULTICAST(mcast_addr))
	{
		#ifdef RTL8651_MCAST_DEBUG
		rtlglue_printf("##  [MCAST] Receive [General Query] <%d>\n", _sysUpSeconds);
		#endif
		dip = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);
		if (dip == IGMP_ALL_HOSTS)
		{	
			#ifdef RTL8651_MCAST_DEBUG
			rtlglue_printf("##  [MCAST] Receive [General Query] <%d>\n", _sysUpSeconds);
			#endif
			goto GeneralQuery;
		}else
		{	
			goto errout;
		}
	}

	if (IN_MULTICAST_RESV(mcast_addr))
		goto errout;		



	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("##  [MCAST] Receive [Group-Specific Query] for (%u.%u.%u.%u) <%d>\n", NIPQUAD(mcast_addr), _sysUpSeconds);
	#endif

	if ((m_entry = MCast_FindMCastEntry(mcast_addr, FALSE, TRUE)) == NULL)	
		goto freeout;

	if ((m_entry->up_vlan) &&
		((m_entry->up_vlan != vlanp) || (m_entry->up_portMask != (1<<pkthdr->ph_portlist))))
	{	
		member = MCast_FindMember(m_entry, vlanp, pkthdr->ph_portlist, FALSE, FALSE, FALSE);
		if (member)
		{	
			ipaddr_t sip;

			sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);

			if (sip && (MCast_GetIpByVlan(member->vlan, 0) > sip) && rtl8651_igmp_oqpi)
			{
				member->isForwarder = FALSE;
				
				member->timer_query.data = (uint32)member;
				member->timer_query.function = (void*)MCast_MemberSendQuery;
				_rtl8651_mod_timer(&member->timer_query, _sysUpSeconds + rtl8651_igmp_oqpi);
			}
			
			member->MulticastRouterVersion = version;
		}
		goto freeout;
	}else
	{	
		uint8 ver;
	
		member = MCast_FindMember(m_entry, vlanp, pkthdr->ph_portlist, FALSE, FALSE, FALSE);
		if (member)
		{	
			MCast_RemoveMember(m_entry, member);
			if (m_entry->memberCnt == 0)	
			{	
				MCast_RemoveMCastEntry(m_entry);
				goto freeout;
			}
		}
		
		MCast_SetMCastEntryUpStream(m_entry, vlanp, pkthdr->ph_portlist, version);
		if (version >= IGMPV3)
		{	
			MCast_SourceList_freeAll(m_entry->slist);
			m_entry->FilterMode = MCast_SourceList_getCache(m_entry, &(m_entry->slist));
		}
		
		ver = MCast_MCastEntryMinVersion(m_entry);
		if (ver < IGMPV3)
			MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, IGMP_FILTER_EXCLUDE, NULL, ((keep_pkt == TRUE)?NULL:pkthdr));
		else
			MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, m_entry->FilterMode, m_entry->slist, ((keep_pkt == TRUE)?NULL:pkthdr));
	}
	goto out;


GeneralQuery:
	{
		struct rtl_pktHdr *pkt = pkthdr;
		uint8 proc_entry = TRUE;
		uint16 portlist = pkthdr->ph_portlist;
		uint8 memberVersion = rtl8651_drvMCastVersion;

		
		if ((rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE) &&
			((rtl8651_GlobalVariable.MCastUpStream.vlan != vlanp) ||
			(!(rtl8651_GlobalVariable.MCastUpStream.portMask & (1<<pkt->ph_portlist)))))
			goto freeout;

		m_entry = rtl8651_MCastTable;
		while (m_entry)
		{	
			
			
			
			proc_entry = TRUE;
			if ((!m_entry->up_vlan) ||
				((m_entry->up_vlan) && (m_entry->up_vlan == vlanp) && (m_entry->up_portMask == (1<<portlist))))
			{	
				if (!m_entry->up_vlan)
				{	
					member = MCast_FindMember(m_entry, vlanp, portlist, FALSE, FALSE, FALSE);
					if (member)
					{	
						MCast_RemoveMember(m_entry, member);
						if (m_entry->memberCnt == 0)	
						{	
							rtl8651_tblDrvMCast_t *next = m_entry->next;
							MCast_RemoveMCastEntry(m_entry);
							m_entry = next;
							proc_entry = FALSE;
						}
					}
				}
				if (proc_entry == TRUE)
				{
				#ifdef RTL8651_MCAST_DEBUG
				rtlglue_printf("##  [MCAST] Process Multicast entry (%u.%u.%u.%u) for [General Query] <%d>\n", NIPQUAD(m_entry->mcast_addr), _sysUpSeconds);
				#endif
					
					MCast_SetMCastEntryUpStream(m_entry, vlanp, portlist, version);
					
					memberVersion = MCast_MCastEntryMinVersion(m_entry);
					if ((memberVersion >= IGMPV3) && (version >= IGMPV3))
					{	
						MCast_SourceList_freeAll(m_entry->slist);
						m_entry->FilterMode = MCast_SourceList_getCache(m_entry, &(m_entry->slist));
					}
					
					if (pkt && (keep_pkt == FALSE))
					{
						if (memberVersion < IGMPV3)
							MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, IGMP_FILTER_EXCLUDE, NULL, pkt);
						else
							MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, m_entry->FilterMode, m_entry->slist, pkt);
						pkt = NULL;
					}else
					{
						if (memberVersion < IGMPV3)
							MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, IGMP_FILTER_EXCLUDE, NULL, NULL);
						else
							MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, m_entry->FilterMode, m_entry->slist, NULL);
					}
				}			
			}
			if (proc_entry == TRUE)
			{
				
				if ((rtl8651_drvMCastUpstream_Timeout != 0xffffffff) && (m_entry->up_vlan))
					_rtl8651_mod_timer(&m_entry->timer_upstream, _sysUpSeconds + rtl8651_drvMCastUpstream_Timeout);
				
				m_entry = m_entry->next;
			}
		}
		
		if (pkt && (keep_pkt == FALSE))
			mBuf_freeMbufChain(pkt->ph_mbuf);
	}	
out:
	if (keep_pkt == TRUE)
		return FWDENG_IGMPPKT;
	return SUCCESS;
freeout:
	if (keep_pkt == TRUE)
		return FWDENG_IGMPPKT;

	mBuf_freeMbufChain(pkthdr->ph_mbuf);
	return SUCCESS;
errout:


	return FWDENG_IGMPPKT;
}



int32 MCast_report(struct rtl_pktHdr *pkthdr, struct ip *ip, struct igmp *ig)
{
	ipaddr_t						mcast_addr;
	uint8						version;
	rtl8651_tblDrvMCast_t *			m_entry;
	rtl8651_tblDrvMCastMember_t *	member;
	rtl8651_tblDrv_vlanTable_t *		vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	uint16						portlist = pkthdr->ph_portlist;
	uint8						keep_pkt = FALSE;
	
	uint32						l4Len = 0;
	uint32						l4ClaimLen = ntohs(ip->ip_len) - ((*(uint8*)ip&0xf) << 2); 
	struct igmp_gr *				gr = NULL;
	int32						LoopCnt = 0;
	uint8						mirrorThis = FALSE;
	
	rtl8651_tblDrvMCastSource_t *	slist;
	uint16						srcCnt;
	uint8						RecordType;
	uint8						newFilterMode = IGMP_FILTER_EXCLUDE;
	rtl8651_tblDrvMCastSource_t *	newSlist = NULL;
	

	
	if (!vlanp->internal)	
		goto freeout;

	switch (ig->igmp_type)
	{
		case IGMP_V1_REPORT:
			version = IGMPV1;
			break;
		case IGMP_V2_REPORT:
			version = IGMPV2;
			break;
		case IGMP_V3_REPORT:
			version = IGMPV3;
			break;
		default:
			
			goto errout;
	}

	
	if (version >= IGMPV3)
	{	
		LoopCnt	= ntohs(ig->igmp_rnofg);
		gr		= &(ig->igmp_grlist);
	}else
		LoopCnt	= 1;	

	l4Len	= 0;
	keep_pkt	= FALSE;

	while (LoopCnt > 0)
	{
		if (version < IGMPV3)
			mcast_addr = PKTGET_UINT32_UNALIGNED(&ig->igmp_group);		
		else
			mcast_addr = PKTGET_UINT32_UNALIGNED(&gr->igmp_gr_group);	

		
		mirrorThis = FALSE;
		{
			rtl8651_tblDrvMCastFilter_t *filter;
			filter = _rtl8651_multicast_FindFilter(mcast_addr);
			if (filter)
			{
				switch (filter->action)
				{
					case (MCAST_FILTER_ACT_MIRROR|MCAST_FILTER_ACT_NOPROC):
						
						keep_pkt = TRUE;
						mirrorThis = TRUE;
						goto next_loop;
						break;
					case (MCAST_FILTER_ACT_MIRROR):
						
						keep_pkt = TRUE;
						mirrorThis = TRUE;
						break;
					case (MCAST_FILTER_ACT_NOPROC):
						
						goto next_loop;
						break;
				}
			}
		}

		
		if ((!IN_MULTICAST(mcast_addr)) || IN_MULTICAST_RESV(mcast_addr))
		{	
			keep_pkt = TRUE;
			goto next_loop;
		}
		
		if ((m_entry = MCast_FindMCastEntry(mcast_addr, TRUE, FALSE)) == NULL)
			goto next_loop;
		
		if ((m_entry->up_vlan == vlanp) &&
			(m_entry->up_portMask == (1<<portlist)))
			goto next_loop;
		
		if ((member = MCast_FindMember(m_entry, vlanp, portlist, TRUE, TRUE, TRUE)) == NULL)
			goto next_loop;
		
		if (mirrorThis == TRUE)
			_rtl8651_setMGroupAttribute(mcast_addr, TRUE);

		
		member->MulticastRouterVersion = version;

		
		if (version < IGMPV3)
		{	
			slist			= NULL;
			srcCnt		= 0;
			RecordType	= IGMP_MODE_ISEX;
		}else
		{	
			rtl8651_tblDrvMCastSource_t *s;
			ipaddr_t*	addrPtr = &(gr->src_list);
			uint16 i = 0;

			RecordType	= gr->igmp_gr_rt;
			srcCnt		= ntohs(gr->igmp_gr_nofs);
			slist			= NULL;
			
			for (i = 0 ; i < srcCnt ; i ++)
			{
				
				if ((s = MCast_SourceList_Create()) == NULL)
					break;
				
				s->source	= PKTGET_UINT32_UNALIGNED(addrPtr);
				s->next		= slist;
				if (slist)
					slist->prev = s;
				slist			= s;
				
				addrPtr ++;
			}
		}
		
		MCast_SourceList_freeAll(m_entry->slist);
		m_entry->FilterMode = MCast_SourceList_getCache(m_entry, &m_entry->slist);
		
		newFilterMode = MCast_ProcessGroupRecord(m_entry, member, RecordType, slist, srcCnt, ((version == IGMPV3)?TRUE:FALSE), &newSlist);

		
		if ((!m_entry->up_vlan) &&										
			(rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE) &&			
			((rtl8651_GlobalVariable.MCastUpStream.vlan != vlanp) ||((rtl8651_GlobalVariable.MCastUpStream.portMask& ~(1<<portlist)) != 0)))	
			
		{
			
			uint8 ver = MCast_MCastEntryMinVersion(m_entry);

			if (ver < IGMPV3)
				MCast_ReportPack(m_entry, m_entry->MulticastRouterVersion, m_entry->FilterMode, m_entry->slist, IGMP_FILTER_EXCLUDE, NULL, NULL);
			else
				MCast_ReportPack(m_entry, m_entry->MulticastRouterVersion, m_entry->FilterMode, m_entry->slist, newFilterMode, newSlist, NULL);
		}

		
		MCast_SourceList_freeAll(m_entry->slist);
		m_entry->slist = newSlist;
		m_entry->FilterMode = newFilterMode;

		
		
		if (version >= IGMPV3)
			_rtl8651_delMulticast(mcast_addr, 0, 0);

next_loop:
		
		if (version < IGMPV3)
		{	
			l4Len += 8;	
		}else
		{	
			l4Len += (8 + (uint32)(gr->igmp_gr_auxlen) + (ntohs(gr->igmp_gr_nofs) * sizeof(ipaddr_t)));	
			
			if (LoopCnt > 1)	
				gr = (struct igmp_gr*)(8  + (uint32)(gr->igmp_gr_auxlen) + (uint32)(ntohs(gr->igmp_gr_nofs) * sizeof(ipaddr_t)) + (uint32)gr);
		}
		
		if (l4Len > l4ClaimLen)
		{
			assert(0);
			keep_pkt = FALSE;	
			goto freeout;
		}
		
		LoopCnt --;
	}


freeout:
	if (keep_pkt == TRUE)
		return FWDENG_IGMPPKT;

	mBuf_freeMbufChain(pkthdr->ph_mbuf);
	return SUCCESS;
errout:


	return FWDENG_IGMPPKT;
}


int32 MCast_leave(struct rtl_pktHdr *pkthdr, struct ip *ip, struct igmp *ig)
{
	rtl8651_tblDrvMCast_t *			m_entry;
	rtl8651_tblDrvMCastMember_t *	member;
	rtl8651_tblDrv_vlanTable_t *		vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	uint16						portlist = pkthdr->ph_portlist;
	ipaddr_t mcast_addr;
	uint8 keep_pkt = FALSE;

	
	if (!vlanp->internal)
		goto freeout;

	mcast_addr = PKTGET_UINT32_UNALIGNED(&(ig->igmp_group));

	
	{
		rtl8651_tblDrvMCastFilter_t *filter;
		filter = _rtl8651_multicast_FindFilter(mcast_addr);
		if (filter)
		{
			switch (filter->action)
			{
				case (MCAST_FILTER_ACT_MIRROR|MCAST_FILTER_ACT_NOPROC):
					
					goto errout;
					break;
				case (MCAST_FILTER_ACT_MIRROR):
					
					keep_pkt = TRUE;
					break;
				case (MCAST_FILTER_ACT_NOPROC):
					
					goto freeout;
					break;
			}
		}
	}

	if ((!IN_MULTICAST(mcast_addr)) || IN_MULTICAST_RESV(mcast_addr))	
		goto errout;

	if ((m_entry = MCast_FindMCastEntry(mcast_addr, TRUE, FALSE)) == NULL)	
		goto freeout;
	else if ((m_entry->up_vlan == vlanp) &&
			(m_entry->up_portMask== (1<<portlist)))		
		goto freeout;

	if ((member = MCast_FindMember(m_entry, vlanp, portlist, TRUE, FALSE, (rtl8651_multicastProcessOptions & MCAST_PROC_EXPIRE_STATE_DONT_FWD)?FALSE:TRUE)) == NULL)
		
		goto freeout;
	
	if (keep_pkt == TRUE)
		_rtl8651_setMGroupAttribute(mcast_addr, TRUE);

	member->MulticastRouterVersion = IGMPV2;

	
	{
		uint8 newFilterMode;
		rtl8651_tblDrvMCastSource_t *newSlist;
		newFilterMode = MCast_ProcessGroupRecord(m_entry, member, IGMP_MODE_TOIN, NULL, 0, FALSE, &newSlist);
	}

	
	MCast_MemberExpire((uint32) member);
#if 0
	
	{
		ipaddr_t sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);

		#ifdef RTL8651_MCAST_DATA_DEBUG
		rtlglue_printf("[MCAST] *FLOW* Try to remove Source(%u.%u.%u.%u)(0x%X)[0x%X] from {%u.%u.%u.%u}\n",
			NIPQUAD(sip), (1<<((vlanp->vid)&(RTL8651_VLAN_NUMBER-1))), (1<<portlist), NIPQUAD(mcast_addr));
		#endif
		_rtl8651_delMulticastUpStream(mcast_addr, sip, (1<<((vlanp->vid)&(RTL8651_VLAN_NUMBER-1))), (1<<portlist));
	}
#endif
	
freeout:
	if (keep_pkt == TRUE)
		return FWDENG_IGMPPKT;

	mBuf_freeMbufChain(pkthdr->ph_mbuf);
	return SUCCESS;
errout:


	return FWDENG_IGMPPKT;
}



static inline struct rtl_pktHdr *MCast_procPkt(struct rtl_pktHdr *pkt)
{
	if (pkt)
	{
		
		mBuf_adjTail(pkt->ph_mbuf, mBuf_getPktlen(pkt->ph_mbuf));
		return pkt;
	}else
	{
		struct rtl_mBuf *mbuf = NULL;
		if((mbuf = mBuf_get(MBUF_DONTWAIT, MBUFTYPE_DATA, 1)) == NULL)
			return NULL;
		if((mBuf_getPkthdr(mbuf, MBUF_DONTWAIT) == (struct rtl_mBuf *) NULL)||
		(mBuf_reserve(mbuf,rtl8651_fwdEnginePara.mbufHeadroom)))
		{
			mBuf_freeMbufChain(mbuf);
			return NULL;
		}
		return mbuf->m_pkthdr;
	}
}



int32 MCast_QueryPack(
	rtl8651_tblDrvMCastMember_t *member,
	rtl8651_tblDrvMCastRouteInfo_t *info,
	uint8 version,
	rtl8651_tblDrvMCastSource_t *sourceList,
	uint16 NumOfSrc,
	struct rtl_pktHdr *pkthdr)
{
	rtl8651_tblDrvMCastRouteInfo_t McastRouteInfo, *route_info = &McastRouteInfo;
	struct rtl_pktHdr	*pkt = pkthdr;
	struct rtl_mBuf	*mbuf;
	ipaddr_t			sip;
	struct ip			*pip;
	struct igmp		*ig;
	uint8			*trailer;
	uint8			isGeneral = FALSE;
	uint32			pktLen = 0;
	rtl8651_tblDrvMCastSource_t *sptr = sourceList;
	uint32			remNumOfSrc = NumOfSrc;

	
	if (
		(!member && !info) ||
		((version == IGMPV3) && sourceList && !member) ||
		((version == IGMPV3) && NumOfSrc && !member))
		goto errout;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Generate IGMPv%d [Query] <%d>\n", version, _sysUpSeconds);
	#endif

	
	if (member)
	{	
		isGeneral = FALSE;	

		route_info->netif		= NULL;
		route_info->ifInfo		= 0;
		route_info->m_addr	= member->m_entry->mcast_addr;
		route_info->vlan		= member->vlan;
		route_info->port		= member->portMask;
		route_info->isPortMask	= TRUE;
	}else
	{	
		isGeneral = TRUE;

		*route_info = *info;
		route_info->m_addr = IGMP_ALL_HOSTS;
	}
	
	if (MCast_SetRouteInfoAndGetIntfIp(&sip, route_info) != SUCCESS)
		goto errout;

sendNext:

	
	pkt = MCast_procPkt(pkt);
	if (pkt)
		mbuf = pkt->ph_mbuf;
	else
		goto errout;

	
	pktLen = 0;

	trailer = (uint8 *)(mbuf->m_data + 46);

	
	
	
	*(uint16 *)(mbuf->m_data + 12) = htons(0x0800);
	pktLen += 14;

	
	pip = (struct ip *)(mbuf->m_data + pktLen);
	if (version == IGMPV1)
		pip->ip_vhl	= 0x45;
	else
		pip->ip_vhl	= 0x46;
	pip->ip_tos	= 0;
	pip->ip_id	= htons(SYS_ID);
	pip->ip_off	= htons(0);
	pip->ip_ttl	= 1;
	pip->ip_p	= IPPROTO_IGMP;
	pip->ip_sum	= 0;

	PKTSET_UINT32_UNALIGNED(sip, &pip->ip_src);
	if (isGeneral == TRUE)
		PKTSET_UINT32_UNALIGNED(IGMP_ALL_HOSTS, &pip->ip_dst);
	else
		PKTSET_UINT32_UNALIGNED(route_info->m_addr, &pip->ip_dst);

	if (version == IGMPV1)
		pktLen += 20;		
	else
	{	
		*(uint16*)((uint8*)pip+20) = htons(0x9404);
		*(uint16*)((uint8*)pip+22) = htons(0x0000);
		pktLen += 24;		
	}

	
	ig = (struct igmp *)(mbuf->m_data + pktLen);
	ig->igmp_type = IGMP_QUERY;
	ig->igmp_cksum = 0;
	if (isGeneral)
		PKTSET_UINT32_UNALIGNED(0, &ig->igmp_group);
	else
		PKTSET_UINT32_UNALIGNED(route_info->m_addr, &ig->igmp_group);

	switch (version)
	{
		case IGMPV1:
			ig->igmp_code = 0;
			pktLen += 8;
			break;
		case IGMPV2:
			{
				if (rtl8651_igmp_qri)
					ig->igmp_code = ((rtl8651_igmp_qri * 10) & 0xff);		
				else
					ig->igmp_code = 10;	
				pktLen += 8;
			}
			break;
		case IGMPV3:
			{
				uint32 NumOfSrcThisPkt;
				ipaddr_t *pkt_srcPtr;
				
				if (rtl8651_igmp_qri)
					ig->igmp_code = ((rtl8651_igmp_qri * 10) & 0xff);		
				else
					ig->igmp_code = 10;	

				ig->igmp_qrsq	= 0x01;									
				if (member)
					ig->igmp_qqic	= (uint8)(IGMPV3_QQIC(rtl8651_igmp_group_query_interval));	
				else
					ig->igmp_qqic	= (uint8)(IGMPV3_QQIC(rtl8651_igmp_qi));	

				
				if (route_info->netif->mtu < (pktLen + 12 + (remNumOfSrc * sizeof(ipaddr_t)) - 14))
				{
					NumOfSrcThisPkt =	(( pktLen >= 2 )?
										((route_info->netif->mtu - (pktLen + 12 - 14)) / sizeof(ipaddr_t)) :
										(route_info->netif->mtu / sizeof(ipaddr_t)));
					remNumOfSrc -= NumOfSrcThisPkt;
				}else
				{
					NumOfSrcThisPkt = remNumOfSrc;
					remNumOfSrc = 0;
				}

				ig->igmp_qnofs	= htons(NumOfSrcThisPkt); 				
				pktLen += (12 + (NumOfSrcThisPkt * sizeof(ipaddr_t)));		

				
				pkt_srcPtr = (ipaddr_t *)(&(ig->igmp_qsrclist));
				while (NumOfSrcThisPkt)
				{
					assert(sptr);
					if (!sptr)
						goto errout;

					PKTSET_UINT32_UNALIGNED(sptr->source, pkt_srcPtr);
					sptr = sptr->next;
					pkt_srcPtr ++;
					NumOfSrcThisPkt --;
				}
			}
			break;
		default:
			goto errout;
	}

	
	pip->ip_len = htons((pktLen - 14));	

	
	if (pktLen < 60)
	{
		trailer = (uint8*)(mbuf->m_data + pktLen);
		memset(trailer, 0, (60 - pktLen));
		pktLen = 60;
	}

	
	ig->igmp_cksum	= MCast_igmpChecksum(pip);
	pip->ip_sum		= _rtl8651_ipChecksum(pip);
	pkt->ph_flags		&= ~(CSUM_IP|CSUM_L4);

	
	pkt->ph_len = pktLen;
	pkt->ph_mbuf->m_len = pktLen;
	pkt->ph_portlist = 0;
	pkt->ph_pppeTagged = 0;
	pkt->ph_iphdrOffset = 14;
	pkt->ph_proto = PKTHDR_IGMP;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("\t=> Send IGMPv%d [Query] to VLAN[0x%X] Mask[0x%X] <%d>\n", version, route_info->vlan->vid, route_info->port, _sysUpSeconds);
	#endif

	if (MCast_l3ActRouting(pkt, route_info) != SUCCESS)
		goto errout;

	if ((version == IGMPV3) && (remNumOfSrc > 0))
	{	
		pkt = NULL;	
		goto sendNext;
	}
	return SUCCESS;
errout:
	if (pkt)
		mBuf_freeMbufChain(pkt->ph_mbuf);
	return FAILED;
}



int32 MCast_ReportPack(
	rtl8651_tblDrvMCast_t *m_entry,
	uint8 version,
	uint8 oldFilterMode,
	rtl8651_tblDrvMCastSource_t *oldSourceList,
	uint8 newFilterMode,
	rtl8651_tblDrvMCastSource_t *newSourceList,
	struct rtl_pktHdr *pkthdr)
{
	rtl8651_tblDrvMCastRouteInfo_t McastRouteInfo, *route_info = &McastRouteInfo;
	struct rtl_pktHdr	*pkt = pkthdr;
	struct rtl_mBuf	*mbuf;
	ipaddr_t			sip = 0;
	struct ip			*pip;
	struct igmp		*ig;
	uint8			*trailer;
	uint32			pktLen;

	rtl8651_tblDrvMCastSource_t*	tempList[2] = {NULL, NULL};

	rtl8651_tblDrvMCastSource_t*	SourceList[3] = {NULL, NULL, NULL};
	uint8						RecordType[3] = {IGMP_MODE_NONE, IGMP_MODE_NONE, IGMP_MODE_NONE};
	uint32						SourceListCnt[3] = {0, 0, 0};

	rtl8651_tblDrvMCastSource_t	**CurrentSourceListp = NULL;
	uint32						*CurrentSourceListCntp = NULL;
	uint8						*CurrentRecordTypep = NULL;

	if 	((!m_entry) ||
		(
			(version == IGMPV3) &&
			(newFilterMode != IGMP_FILTER_EXCLUDE) &&
			(newFilterMode != IGMP_FILTER_INCLUDE))
		)
		goto release_out;

	
	if (version == IGMPV3)
	{
		if (oldFilterMode == IGMP_FILTER_NONE)
		{	
			RecordType[0]		= ((newFilterMode == IGMP_FILTER_EXCLUDE)?IGMP_MODE_ISEX:IGMP_MODE_ISIN);
			SourceList[0]			= newSourceList;
			SourceListCnt[0]		= MCast_SourceList_EntryCnt(newSourceList);
			RecordType[1]		= IGMP_MODE_NONE;
			SourceList[1]			= NULL;
			SourceListCnt[1]		= 0;
		}else if (oldFilterMode != newFilterMode)
		{	
			RecordType[0]		= ((newFilterMode == IGMP_FILTER_EXCLUDE)?IGMP_MODE_TOEX:IGMP_MODE_TOIN);
			SourceList[0]			= newSourceList;
			SourceListCnt[0]		= MCast_SourceList_EntryCnt(newSourceList);
			RecordType[1]		= IGMP_MODE_NONE;
			SourceList[1]			= NULL;
			SourceListCnt[1]		= 0;
		}else
		{	
			RecordType[0]	= IGMP_MODE_ALLOW;
			RecordType[1]	= IGMP_MODE_BLOCK;
			if (oldFilterMode == IGMP_FILTER_INCLUDE)
			{
				SourceListCnt[0] = MCast_SourceList_removal(newSourceList, 0, 0, oldSourceList, 0, 0, &tempList[0]);
				SourceListCnt[1] = MCast_SourceList_removal(oldSourceList, 0, 0, newSourceList, 0, 0, &tempList[1]);
			}else
			{
				SourceListCnt[0] = MCast_SourceList_removal(oldSourceList, 0, 0, newSourceList, 0, 0, &tempList[0]);
				SourceListCnt[1] = MCast_SourceList_removal(newSourceList, 0, 0, oldSourceList, 0, 0, &tempList[1]);
			}
			SourceList[0] = tempList[0];
			SourceList[1] = tempList[1];
		}

		
		CurrentSourceListp		= &SourceList[0];
		CurrentSourceListCntp	= &SourceListCnt[0];
		CurrentRecordTypep	= &RecordType[0];
	}

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Generate IGMPv%d [Report] <%d>\n", version, _sysUpSeconds);
	#endif

	if (version < IGMPV3)
		route_info->m_addr	= m_entry->mcast_addr;
	else
		route_info->m_addr	= IGMP_V3_REPORT_ADDR;

	
	if (m_entry->up_vlan)
	{
		route_info->netif		= NULL;
		route_info->ifInfo		= 0;
		route_info->vlan		= m_entry->up_vlan;
		route_info->port		= m_entry->up_portMask;
		route_info->isPortMask	= TRUE;
	}else if (rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE)	
	{
		rtl8651_tblDrvMCastMember_t *member;

		route_info->netif	= rtl8651_GlobalVariable.MCastUpStream.netif;
		route_info->vlan	= rtl8651_GlobalVariable.MCastUpStream.vlan;
		route_info->port	= (rtl8651_GlobalVariable.MCastUpStream.portMask & phyPortLinkStatus);
		if (!route_info->port)
			goto release_out;
		{
			uint16 tmp = ~(route_info->port);
			route_info->port = route_info->port & (tmp + 1);	
		}
		route_info->ifInfo	= (uint32)(rtl8651_GlobalVariable.MCastUpStream.M_intfIp);	

		
		member = m_entry->member;
		while (member)
		{
			if (member->vlan == route_info->vlan)
			{
				route_info->port &= ~(member->portMask);
				if (!(route_info->port))
					goto release_out;
			}
			member = member->next;
		}
		route_info->isPortMask	= TRUE;
	}else
		goto release_out;

	if (MCast_SetRouteInfoAndGetIntfIp(&sip, route_info) != SUCCESS)
		goto errout;

sendNext:
	
	pkt = MCast_procPkt(pkt);
	if (pkt)
		mbuf = pkt->ph_mbuf;
	else
		goto errout;

	
	pktLen = 0;
	
	
	
	*(uint16 *)(mbuf->m_data + 12) = htons(0x0800);
	pktLen += 14;
	pip = (struct ip *)(mbuf->m_data + pktLen);

	
	if (version == IGMPV1)
		pip->ip_vhl	= 0x45;
	else
		pip->ip_vhl	= 0x46;
	pip->ip_tos	= 0;
	pip->ip_id	= htons(SYS_ID);
	pip->ip_off	= htons(0);
	pip->ip_ttl	= 1;
	pip->ip_p	= IPPROTO_IGMP;
	pip->ip_sum	= 0;

	PKTSET_UINT32_UNALIGNED(sip, &pip->ip_src);
	PKTSET_UINT32_UNALIGNED(route_info->m_addr, &pip->ip_dst);

	if (version == IGMPV1)
	{
		pktLen += 20;		
	}else
	{
		
		*(uint16*)((uint8*)pip+20) = htons(0x9404);
		*(uint16*)((uint8*)pip+22) = htons(0x0000);
		pktLen += 24;		
	}

	ig = (struct igmp *)(mbuf->m_data + pktLen);
	
	ig->igmp_cksum = 0;
	ig->igmp_code = 0;

	switch (version)
	{
		case IGMPV1:
			PKTSET_UINT32_UNALIGNED(route_info->m_addr, &ig->igmp_group);
			ig->igmp_type = IGMP_V1_REPORT;
			pktLen += 8;
			break;
		case IGMPV2:
			PKTSET_UINT32_UNALIGNED(route_info->m_addr, &ig->igmp_group);
			ig->igmp_type = IGMP_V2_REPORT;
			pktLen += 8;
			break;
		case IGMPV3:
			{
				struct igmp_gr *	currGr, *gr;
				ipaddr_t *		addrPtr;
				uint32			mtu;
				uint32			nofg;
				uint32			nofs;
				uint32			genNextGroup;	
				
				ig->igmp_type	= IGMP_V3_REPORT;
				ig->igmp_rresv	= 0;
				pktLen			+= 8;		

				mtu			= route_info->netif->mtu;
				currGr		= &(ig->igmp_grlist);			
				nofg			= 0;
				genNextGroup	= TRUE;						

				while ((*CurrentRecordTypep != IGMP_MODE_NONE) && (genNextGroup == TRUE))
				{
					if (((*CurrentRecordTypep == IGMP_MODE_ALLOW) || (*CurrentRecordTypep == IGMP_MODE_BLOCK)) &&
						(*CurrentSourceListCntp == 0))
					{	
						
						CurrentSourceListp ++;
						CurrentSourceListCntp ++;
						CurrentRecordTypep ++;
						continue;
					}
						
					gr		= currGr;
					addrPtr	= &(gr->src_list);
					nofs		= 0;
					nofg		++;
					
					gr->igmp_gr_rt = *CurrentRecordTypep;
					
					gr->igmp_gr_auxlen = 0;
					PKTSET_UINT32_UNALIGNED(m_entry->mcast_addr, &(gr->igmp_gr_group));
					pktLen += 8;	
					while(*CurrentSourceListCntp > 0)
					{
						assert(*CurrentSourceListp);
						if (((pktLen + sizeof(ipaddr_t)-14) > mtu) || (nofs == 0xff))
						{	
							genNextGroup		= FALSE;		
							goto SlistAdd_End;
						}
						PKTSET_UINT32_UNALIGNED((*CurrentSourceListp)->source, addrPtr);	
						addrPtr ++;
						nofs ++;
						(*CurrentSourceListCntp) --;
						(*CurrentSourceListp) = (*CurrentSourceListp)->next;
					}
					
					CurrentSourceListp ++;
					CurrentSourceListCntp ++;
					CurrentRecordTypep ++;
					currGr = (struct igmp_gr *)addrPtr;	
	SlistAdd_End:
					gr->igmp_gr_nofs = htons(nofs);
					pktLen += (nofs * sizeof(ipaddr_t));
				}
				ig->igmp_rnofg = htons(nofg);	
				if (nofg == 0)
					goto release_out;
			}
			break;
		default:
			assert(0);
			goto errout;
	}

	pip->ip_len	= htons(pktLen - 14);	

	
	if (pktLen < 60)
	{
		trailer = (uint8*)(mbuf->m_data + pktLen);
		memset(trailer, 0, (60 - pktLen));
		pktLen = 60;
	}

	
	ig->igmp_cksum	= MCast_igmpChecksum(pip);
	pip->ip_sum		= _rtl8651_ipChecksum(pip);
	pkt->ph_flags		&= ~(CSUM_IP|CSUM_L4);

	
	pkt->ph_len = pktLen;
	pkt->ph_mbuf->m_len = pktLen;
	pkt->ph_portlist = 0;
	pkt->ph_pppeTagged = 0;
	pkt->ph_iphdrOffset = 14;
	pkt->ph_proto = PKTHDR_IGMP;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("\t=> Send IGMPv%d [Report] to VLAN[0x%X] Mask[0x%X] <%d>\n", version, route_info->vlan->vid, route_info->port, _sysUpSeconds);
	#endif

	if (MCast_l3ActRouting(pkt, route_info) != SUCCESS)
		goto errout;

	if (
		(version == IGMPV3) &&
		(*CurrentSourceListCntp > 0) &&
		(*CurrentRecordTypep != IGMP_MODE_ISEX) &&	
		(*CurrentRecordTypep != IGMP_MODE_TOEX))		
	{	
		pkt = NULL;	
		goto sendNext;
	}
	
	goto out;

release_out:
	if (pkt)
		mBuf_freeMbufChain(pkt->ph_mbuf);
out:
	if (tempList[0])
			MCast_SourceList_freeAll(tempList[0]);
	if (tempList[1])
			MCast_SourceList_freeAll(tempList[1]);
	return SUCCESS;
errout:
	if (pkt)
		mBuf_freeMbufChain(pkt->ph_mbuf);
	if (tempList[0])
			MCast_SourceList_freeAll(tempList[0]);
	if (tempList[1])
			MCast_SourceList_freeAll(tempList[1]);
	return FAILED;
}


int32 MCast_LeavePack(rtl8651_tblDrvMCast_t *m_entry, struct rtl_pktHdr *pkthdr)
{
	struct rtl_pktHdr	*pkt = pkthdr;
	struct rtl_mBuf	*mbuf;
	ipaddr_t			sip;
	struct ip			*pip;
	struct igmp		*ig;
	uint8			*trailer;
	rtl8651_tblDrvMCastRouteInfo_t McastRouteInfo, *route_info = &McastRouteInfo;

	if (!m_entry)
		goto errout;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Generate IGMP [Leave] <%d>\n", _sysUpSeconds);
	#endif

	
	route_info->m_addr	= m_entry->mcast_addr;
	if (m_entry->up_vlan)
	{
		route_info->netif		= NULL;
		route_info->ifInfo		= 0;
		route_info->vlan		= m_entry->up_vlan;
		route_info->port		= m_entry->up_portMask;
		route_info->isPortMask = TRUE;
	}else
	{	
		if (rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE)
		{
			rtl8651_tblDrvMCastMember_t *member;

			route_info->netif		= rtl8651_GlobalVariable.MCastUpStream.netif;
			route_info->vlan		= rtl8651_GlobalVariable.MCastUpStream.vlan;
			route_info->port		= rtl8651_GlobalVariable.MCastUpStream.portMask;
			if (!route_info->port)
				goto release_out;
			{
				uint16 tmp = ~(route_info->port);
				route_info->port = route_info->port & (tmp + 1);	
			}
			route_info->ifInfo	= (uint32)(rtl8651_GlobalVariable.MCastUpStream.M_intfIp);	

			
			member = m_entry->member;
			while (member)
			{
				if (member->vlan == route_info->vlan)
				{
					route_info->port = route_info->port & ~(member->portMask);
					if (!(route_info->port))
					{
						goto release_out;
					}
				}
				member = member->next;
			}
			route_info->isPortMask	= TRUE;
		}else
		{
			goto errout;
		}
	}

	if (MCast_SetRouteInfoAndGetIntfIp(&sip, route_info) != SUCCESS)
	{
		goto errout;
	}

	
	pkt = MCast_procPkt(pkthdr);
	if (pkt)
		mbuf = pkt->ph_mbuf;
	else
		goto errout;

	
	pkt->ph_len = 60;
	mbuf->m_len = 60;
	pip = (struct ip *)(mbuf->m_data + 14);
	ig = (struct igmp *)(mbuf->m_data + 38);
	trailer = (uint8 *)(mbuf->m_data + 46);

	
	
	
	*(uint16 *)(mbuf->m_data + 12) = htons(0x0800);

	
	pip->ip_vhl	= 0x46;
	pip->ip_tos	= 0;
	pip->ip_len	= htons((24+8));	
	pip->ip_id	= htons(SYS_ID);
	pip->ip_off	= htons(0);
	pip->ip_ttl	= 1;
	pip->ip_p	= IPPROTO_IGMP;
	pip->ip_sum	= 0;

	PKTSET_UINT32_UNALIGNED(sip, &pip->ip_src);
	PKTSET_UINT32_UNALIGNED(IGMP_ALL_ROUTER, &pip->ip_dst);
	
	*(uint16*)((uint8*)pip+20) = htons(0x9404);
	*(uint16*)((uint8*)pip+22) = htons(0x0000);

	
	ig->igmp_type = IGMP_V2_LEAVE;
	ig->igmp_code = 0;
	ig->igmp_cksum = 0;
	PKTSET_UINT32_UNALIGNED(route_info->m_addr, &ig->igmp_group);


	
	memset(trailer, 0, 14);

	
	ig->igmp_cksum	= MCast_igmpChecksum(pip);
	pip->ip_sum		= _rtl8651_ipChecksum(pip);
	pkt->ph_flags		&= ~(CSUM_IP|CSUM_L4);

	
	pkt->ph_portlist = 0;
	pkt->ph_pppeTagged = 0;
	pkt->ph_iphdrOffset = 14;
	pkt->ph_proto = PKTHDR_IGMP;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("\t=> Send IGMPv%d [Leave] to VLAN[0x%X] Mask[0x%X] <%d>\n", IGMPV2, route_info->vlan->vid, route_info->port, _sysUpSeconds);
	#endif

	if (MCast_l3ActRouting(pkt, route_info) != SUCCESS)
	{
		goto errout;
	}

	return SUCCESS;
release_out:
	if (pkt)
		mBuf_freeMbufChain(pkt->ph_mbuf);
	return SUCCESS;
errout:
	if (pkt)
		mBuf_freeMbufChain(pkt->ph_mbuf);
	return FAILED;
}

int32 MCast_l3ActRouting(struct rtl_pktHdr* pkthdr, rtl8651_tblDrvMCastRouteInfo_t *info)
{
	int8 *ptr;
	rtl8651_tblDrv_networkIntfTable_t *netif;
	uint16 srcPortList;

	if ((!(info->vlan)) ||
		(!(info->port)) ||
		(!(info->m_addr)))
	{
		return FAILED;
	}

	
	if (info->netif)
	{
		netif = info->netif;
	}else if ((netif = _rtl8651_getNetworkIntfEntryByVid(info->vlan->vid)) == NULL)
	{
		return FAILED;
	}

	if(pkthdr->ph_pppeTagged)
	{
		uint16 etherType = htons(0x0800);
		
		mBuf_adjHead(pkthdr->ph_mbuf, 8);
		
		ptr = (int8*)(pkthdr->ph_mbuf->m_data);
		
		rtl8651_memcpy(ptr+12, &etherType, 2);
		
		pkthdr->ph_pppeTagged = 0;
		pkthdr->ph_iphdrOffset = 14;
	}
	assert(pkthdr->ph_iphdrOffset == 14);

	
	ptr = (int8*)(pkthdr->ph_mbuf->m_data);

	switch (netif->linkLayerType)
	{
		case RTL8651_LL_VLAN:
			
			{
				ether_addr_t	ether;

				ether.octet[0] = 0x01;
				ether.octet[1] = 0x00;
				ether.octet[2] = 0x5e;
				ether.octet[3] = (uint8)((info->m_addr >> 16) & 0xff);
				ether.octet[4] = (uint8)((info->m_addr >> 8) & 0xff);
				ether.octet[5] = (uint8)(info->m_addr & 0xff);
				ether.octet[3] &= 0x7f;
				rtl8651_memcpy(ptr, &ether, 6);
			}

			break;
		case RTL8651_LL_PPPOE:
			
			{
				rtl8651_tblDrv_pppTable_t *poe_tb;

				if (info->ifInfo)
				{
					poe_tb = ((rtl8651_tblDrvSession_t*)(info->ifInfo))->session_t;
				}else
				{
					poe_tb = CTAILQ_FIRST(&netif->sessionHead)->session_t;
				}

				
				pkthdr->ph_pppoeIdx = PPPOE_TABLE_INDEX(poe_tb); 
				pkthdr->ph_dsid=poe_tb->dsid;
				
				pkthdr->ph_flags |= (PKTHDR_PPPOE_AUTOADD);
				
				rtl8651_memcpy(ptr, &(poe_tb->macInfo->macAddr), 6);
			}
			break;
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		case RTL8651_LL_PPTP:
		case RTL8651_LL_L2TP:
			{
				rtl8651_tblDrv_pppTable_t *ptb;
				if (info->ifInfo)
				{
					ptb = ((rtl8651_tblDrvSession_t *)(info->ifInfo))->session_t;
				}else
				{
					ptb = ((rtl8651_tblDrvSession_t *)CTAILQ_FIRST(&(netif->sessionHead)))->session_t;
				}
				
				rtl8651_memcpy(ptr, &(ptb->macInfo->macAddr), 6);

				
				info->vlan = &DrvTbl.vlan[netif->vid & (RTL8651_VLAN_NUMBER - 1)];
				info->port = info->vlan->memberPortMask;
				info->isPortMask = TRUE;
			}
			break;
		#endif 
		default:
			return FAILED;
	}
	
	if (info->vlan->macAddrNumber == 0)
	{
		return FAILED;
	}
	rtl8651_memcpy(ptr+6, &(info->vlan->macAddr), 6);

	
	srcPortList = pkthdr->ph_portlist;
	if (info->isPortMask == TRUE)
	{
		pkthdr->ph_portlist = (info->port);
	}else
	{
		pkthdr->ph_portlist = (1<<(info->port));
	}
	if (pkthdr->ph_portlist == 0)
	{
		return FAILED;
	}
	pkthdr->ph_flags &= ~(CSUM_IP|CSUM_L4);
	MCast_setPkthdrProperty(pkthdr, PKTHDR_ASLOOPBACK);	

	{
		uint32 mtu, tol_len;
		mtu = netif->mtu;
		tol_len = mBuf_getPktlen(pkthdr->ph_mbuf);	
		if (tol_len > (mtu+14))
		{
			int32 retval;
			int8 old_IcmpRoutingMsgEnable;

			
			old_IcmpRoutingMsgEnable = rtl8651_drvIcmpRoutingMsgEnable;
			
			rtl8651_drvIcmpRoutingMsgEnable = FALSE;
			
			retval = _rtl8651_mtuFragment(FWDENG_L34PROC|FWDENG_FROMDRV, pkthdr, mtu, (info->vlan->vid), FALSE, srcPortList);
			
			rtl8651_drvIcmpRoutingMsgEnable = old_IcmpRoutingMsgEnable;
			return retval;
		}
	}
	return _rtl8651_fwdEngineSend(FWDENG_L34PROC|FWDENG_FROMDRV, (void*)pkthdr, (info->vlan->vid),-1);
}



int32 MCast_setPkthdrProperty(struct rtl_pktHdr *ph, int32 property)
{
	uint32 swNicRxIndex = 0;

	if (!ph)
		return FAILED;

#ifdef CONFIG_RTL865XC
	rtlglue_getRingSize(&swNicRxIndex, NULL,0); 
#else
	rtlglue_getRingSize(&swNicRxIndex, NULL);
#endif

	if (!swNicRxIndex)
		return FAILED;

	switch (property)
	{
		case PKTHDR_ASLOOPBACK:	
			if ((ph->ph_rxdesc>=0)&&(ph->ph_rxdesc<swNicRxIndex))
				return FAILED;
			ph->ph_rxdesc = swNicRxIndex;
			break;
		default:
			return FAILED;
	}
	return SUCCESS;
}


int32 _rtl8651_igmp_process(struct rtl_pktHdr *pkthdr, struct ip *ip, struct igmp *ig)
{
	if ((system_enable == FALSE) || (rtl8651_drvMulticastProcessEnable == FALSE))
	{
		return FWDENG_IGMPPKT;
	}

	switch (ig->igmp_type)
	{
		case IGMP_V1_REPORT:
		case IGMP_V2_REPORT:
		case IGMP_V3_REPORT:
			#ifdef RTL8651_MCAST_DEBUG
			rtlglue_printf("##  [MCAST] Receive IGMP [Report] Port [0x%02X] <%d>\n", (1<<pkthdr->ph_portlist), _sysUpSeconds);
			#endif
			return MCast_report(pkthdr, ip, ig);
		case IGMP_V2_LEAVE:
			#ifdef RTL8651_MCAST_DEBUG
			rtlglue_printf("##  [MCAST] Receive IGMP [Leave] Port [0x%02X] <%d>\n", (1<<pkthdr->ph_portlist), _sysUpSeconds);
			#endif
			return MCast_leave(pkthdr, ip, ig);
		case IGMP_QUERY:
			#ifdef RTL8651_MCAST_DEBUG
			rtlglue_printf("##  [MCAST] Receive IGMP [Query] Port [0x%02X] <%d>\n", (1<<pkthdr->ph_portlist), _sysUpSeconds);
			#endif
			return MCast_query(pkthdr, ip, ig);
	}
	return FWDENG_IGMPPKT;
}

#define RTL8651_MULTICAST_FWD_FUNCTIONS
int32 _rtl8651_multicast_process(struct rtl_pktHdr *pkthdr, struct ip *ip)
{
	ipaddr_t						sip_org, sip;
	uint16						portSrc = pkthdr->ph_portlist;
	uint16						portMask = (1<<pkthdr->ph_portlist);
	uint16						MemberPortMask;
	ipaddr_t						mcast_addr;
	ipaddr_t						extIp;
	rtl8651_tblDrvMCast_t *			m_entry;
	rtl8651_tblDrvMCastMember_t *	member;
	rtl8651_tblDrv_vlanTable_t *		vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	rtl8651_tblDrvMCastRouteInfo_t	info;
	rtl8651_tblDrv_vlanTable_t *		FwdVlan[RTL8651_VLAN_NUMBER];		
	uint16						FwdPortMask[RTL8651_VLAN_NUMBER];	
	uint32						CacheVlanMask = 0;	
	uint32						CachePortMask = 0;	
	uint8						idx;
	uint8						fwdPktCnt;	
#ifndef RTL8651_MCAST_ALWAYS2UPSTREAM
	uint8						totalUpCnt = 0;
#endif 
	rtl8651_tblDrvMCastUpStreamInfo_t upstreamInfo, *up = &upstreamInfo;
	struct rtl_pktHdr				*fwdHdr = pkthdr;
	struct rtl_mBuf				*fwdMbuf = pkthdr->ph_mbuf;
	struct ip						*fwdIp = ip;
	uint8						l4proto;
	
	uint8						keep_pkt = FALSE;
	
	uint8						ignoreCacheUpStream;
	uint16						upPortMask = 0;

	if ((system_enable == FALSE) || (rtl8651_drvMulticastProcessEnable == FALSE))
		return FWDENG_IP_MULTICAST;

	
	CacheVlanMask= 0;
	CachePortMask = 0;
	extIp = 0;
	ignoreCacheUpStream = FALSE;
	
	up->isSet = FALSE;
	
	l4proto = ip->ip_p;

	
	switch(l4proto)
	{
		case IPPROTO_TCP:
			rtl8651_fwdEngineCounter.tcpReceive ++;
			break;
		case IPPROTO_UDP:
			rtl8651_fwdEngineCounter.udpReceive ++;
			break;
	}

	mcast_addr = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);
	sip_org = PKTGET_UINT32_UNALIGNED(&ip->ip_src);
	assert(IN_MULTICAST(mcast_addr));

	
	{
		rtl8651_tblDrvMCastFilter_t *filter;
		filter = _rtl8651_multicast_FindFilter(mcast_addr);
		if (filter)
		{
			switch (filter->action)
			{
				case (MCAST_FILTER_ACT_MIRROR|MCAST_FILTER_ACT_NOPROC):
					
					return FWDENG_IP_MULTICAST;
				case (MCAST_FILTER_ACT_MIRROR):
					
					{
						struct rtl_mBuf *nbuf = NULL;
						nbuf = mBuf_dupPacket(pkthdr->ph_mbuf, MBUF_DONTWAIT);
						if (nbuf)
						{
							fwdMbuf = nbuf;
							fwdHdr= nbuf->m_pkthdr;
							fwdIp = (struct ip *)((uint32)nbuf->m_data + ((uint32)ip - (uint32)(pkthdr->ph_mbuf->m_data)));
							keep_pkt = TRUE;
						}
					}
					break;
				case (MCAST_FILTER_ACT_NOPROC):
					
					goto release_out;
			}
		}
	}

	
	if (IN_MULTICAST_RESV(mcast_addr))
		goto release_out;

	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("[MCAST] *FLOW* Slow path: Flow{%u.%u.%u.%u} from (%u.%u.%u.%u)vlan[0x%X]/port[%X]\n",
		NIPQUAD(mcast_addr), NIPQUAD(sip_org), vlanp->vid, fwdHdr->ph_portlist);
	#endif

	
	if ((m_entry = MCast_FindMCastEntry(mcast_addr, FALSE, TRUE)) == NULL)
	{	
		if ((rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE) && ((rtl8651_GlobalVariable.MCastUpStream.vlan != vlanp) || !(rtl8651_GlobalVariable.MCastUpStream.portMask & portMask)))
		{
				up = &(rtl8651_GlobalVariable.MCastUpStream);
		}
	}else
	{	

		
		
		if ((m_entry->up_vlan) && ((m_entry->up_vlan != vlanp) || (m_entry->up_portMask != portMask)))
		{	
			up->isSet = TRUE;
			up->vlan = m_entry->up_vlan;
			up->portMask = m_entry->up_portMask;
			up->netif = _rtl8651_getNetworkIntfEntryByVid(m_entry->up_vlan->vid);
			up->M_intfIp = 0;
		}else if ((rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE) && ((rtl8651_GlobalVariable.MCastUpStream.vlan != vlanp) || !(rtl8651_GlobalVariable.MCastUpStream.portMask & portMask)))
		{
			up = &(rtl8651_GlobalVariable.MCastUpStream);
		}

		
		
	member = m_entry->member;
	memset(FwdPortMask, 0, (sizeof(uint16)*RTL8651_VLAN_NUMBER));
	memset(FwdVlan, 0, (sizeof(uint32)*RTL8651_VLAN_NUMBER));
	fwdPktCnt = 0;
	while (member)
	{
			
		if (member->isForwarder == FALSE)
			goto next_member;

			
			if ((rtl8651_multicastProcessOptions & MCAST_PROC_EXPIRE_STATE_DONT_FWD) &&
				(member->state != IGMP_MEMBER_STATE_NORMAL))
				goto next_member;

			
		if (member->MulticastRouterVersion == IGMPV3)
		{
			rtl8651_tblDrvMCastSource_t *tmp;
			if ((member->FilterMode == IGMP_FILTER_INCLUDE) && (!MCast_SourceList_inList(sip_org, member->slist)))
				goto next_member;
			if (member->FilterMode == IGMP_FILTER_EXCLUDE)
			{
				tmp = MCast_SourceList_inList(sip_org, member->slist);
				if (tmp && (tmp->timer.expires <= _sysUpSeconds))
					goto next_member;
			}
		}

			
		MemberPortMask = member->portMask;
		if (member->vlan == vlanp)
			MemberPortMask &= ~(portMask);
		if (MemberPortMask == 0)
			goto next_member;

			
		if (FwdVlan[((member->vlan->vid) & (RTL8651_VLAN_NUMBER - 1))] == NULL)
		{
				fwdPktCnt ++; 
			FwdVlan[((member->vlan->vid) & (RTL8651_VLAN_NUMBER - 1))] = member->vlan;
		}
			FwdPortMask[((member->vlan->vid) & (RTL8651_VLAN_NUMBER - 1))] |= MemberPortMask;

			#ifdef RTL8651_MCAST_DATA_DEBUG
			rtlglue_printf("[MCAST] *FLOW* Slow path: Flow{%u.%u.%u.%u} can to: Vlan[%d]portMask{0x%x}\n",
				NIPQUAD(mcast_addr), member->vlan->vid, MemberPortMask);
			#endif
		
	next_member:
		member = member->next;
	}

#ifndef RTL8651_MCAST_ALWAYS2UPSTREAM
		totalUpCnt = fwdPktCnt;
#endif 

		
	if (fwdPktCnt == 0)
		goto fwd_to_upstream;

		
	idx = 0;
	info.m_addr		= mcast_addr;
	info.isPortMask	= TRUE;
	info.netif			= NULL;
	info.ifInfo			= 0;

		
	fwdHdr->ph_flags &= ~(CSUM_IP|CSUM_L4);

		
	while (fwdPktCnt > 0)
	{
			uint8 needDup = FALSE;
			uint8 needDecrTTL = FALSE;
			uint8 needIgnoreThisFwd = FALSE;
		
		while (FwdVlan[idx] == NULL)
		{
			idx ++;
			assert((idx < RTL8651_VLAN_NUMBER));
		}

			
			CacheVlanMask |= (1<<idx);
			CachePortMask|= FwdPortMask[idx];

			
		info.vlan	= FwdVlan[idx];
		info.port	= FwdPortMask[idx];
			assert(info.vlan);
			assert(info.port);	


			
			if ((fwdPktCnt > 1)
#ifdef RTL8651_MCAST_ALWAYS2UPSTREAM
				|| (up->isSet == TRUE)
#endif 
			)
		{
				needDup = TRUE;
		}


			#if 0	
		if ((!info.vlan->internal) && (vlanp->internal))	
		{
				ipaddr_t sip;
				if ((sip = MCast_GetIpByVlan(info.vlan, 0)) == 0)
				{	
					goto fail_loop;
				}
				
				PKTSET_UINT32_UNALIGNED(sip, &fwdIp->ip_src);
		}
		#endif

			
			if (FwdVlan[idx] != vlanp)
			{	
				if (fwdIp->ip_ttl > 1)
				{
					needDecrTTL = TRUE;
				}else
				{
					#ifdef RTL8651_MCAST_DATA_DEBUG
					rtlglue_printf("[MCAST] \t\tTTL(%d) too low ==> Ignore it!!!\n", fwdIp->ip_ttl);
					#endif
					needIgnoreThisFwd = TRUE;
				}
			}

			#ifdef RTL8651_MCAST_DATA_DEBUG
			rtlglue_printf("[MCAST] *FLOW* Send {%u.%u.%u.%u} to: Vlan[%d]/portMask{0x%x} -- ttl[%d]\n",
				NIPQUAD(mcast_addr), info.vlan->vid, info.port, fwdIp->ip_ttl);
			#endif

			if (needIgnoreThisFwd == TRUE)
			{	
				if (needDup == FALSE)
				{
					assert(fwdPktCnt == 1);
					mBuf_freeMbufChain(fwdMbuf);
				}
			} else
			{	
				if (needDup == TRUE)
		{
			struct rtl_mBuf *newMbuf;

					if ((newMbuf = mBuf_dupPacket(fwdMbuf, MBUF_DONTWAIT)) != NULL)
					{
						
						if (needDecrTTL == TRUE)
						{
							struct ip *newIp = (struct ip*)(((uint32)fwdIp - (uint32)(fwdMbuf->m_data)) + (uint32)(newMbuf->m_data)); 

							_RTL8651_TTL_DEC_1(newIp->ip_ttl, newIp->ip_sum);

							#ifdef RTL8651_MCAST_DATA_DEBUG
							rtlglue_printf("[MCAST] \t\t==> decrease TTL to %d ...\n", newIp->ip_ttl);
							#endif
						}
						#ifdef RTL8651_MCAST_DATA_DEBUG
						else
							rtlglue_printf("[MCAST] \t\t==> We Keep TTL ...\n");
						#endif
										
			if (MCast_l3ActRouting(newMbuf->m_pkthdr, &info) != SUCCESS)
			{
				mBuf_freeMbufChain(newMbuf);
							#ifdef RTL8651_MCAST_DATA_DEBUG
							rtlglue_printf("[MCAST] \t\t==> FAILED!!!\n");
							#endif
						}
						#ifdef RTL8651_MCAST_DATA_DEBUG
						else
							rtlglue_printf("[MCAST] \t\t==> SUCCESS\n");
						#endif
			}
		}else
		{
					
					if (needDecrTTL == TRUE)
					{
						_RTL8651_TTL_DEC_1(fwdIp->ip_ttl, fwdIp->ip_sum);
						#ifdef RTL8651_MCAST_DATA_DEBUG
						rtlglue_printf("[MCAST] \t\t==> decrease TTL to %d ...\n", fwdIp->ip_ttl);
						#endif
					}
					#ifdef RTL8651_MCAST_DATA_DEBUG
					else
						rtlglue_printf("[MCAST] \t\t==> We Keep TTL ...\n");
					#endif
					

					if (MCast_l3ActRouting(fwdHdr, &info) != SUCCESS)
					{	
						mBuf_freeMbufChain(fwdMbuf);
						#ifdef RTL8651_MCAST_DATA_DEBUG
						rtlglue_printf("[MCAST] \t\t==> FAILED!!!\n");
						#endif
					}
					#ifdef RTL8651_MCAST_DATA_DEBUG
					else
						rtlglue_printf("[MCAST] \t\t==> SUCCESS\n");
					#endif
				}
			}

			fwdPktCnt --;
			idx ++;
		}
	}

	
fwd_to_upstream:
#ifndef RTL8651_MCAST_ALWAYS2UPSTREAM
	if (totalUpCnt == 0)	
	{
#endif	
		if (up->isSet == TRUE)
		{
			#ifdef RTL8651_MCAST_DATA_DEBUG
			rtlglue_printf("[MCAST] *FLOW* Slow path: Flow{%u.%u.%u.%u} from{%u.%u.%u.%u},vlan[%d]/port[%d] to upstream...\n",
				NIPQUAD(mcast_addr), NIPQUAD(sip_org), vlanp->vid, fwdHdr->ph_portlist);
			rtlglue_printf("\t\t==> Upstream :\n\t\t\tVid [%d]\n\t\t\tPort [0x%x]\n", up->vlan->vid, up->portMask);
			#endif

			
			info.port		= ((up->portMask & ~(portMask)) & phyPortLinkStatus);
			if (info.port == 0)	
			{
				#ifdef RTL8651_MCAST_DATA_DEBUG
				rtlglue_printf("[MCAST] *FLOW* Source port[0x%x] is same as destination port[0x%x], ignore.\n", portMask, up->portMask);
				#endif
				ignoreCacheUpStream = TRUE;
				goto upstream_fwd_fail;
			} else
			{
				uint16 tmp = ~(info.port);
				info.port = info.port & (tmp + 1);	
				upPortMask = info.port;
			}
			info.m_addr		= mcast_addr;
			info.vlan			= up->vlan;
			info.netif			= up->netif;
			info.ifInfo 		= (uint32)(up->M_intfIp);		
			info.isPortMask	= TRUE;

			
			if (MCast_SetRouteInfoAndGetIntfIp(&sip, &info) != SUCCESS)
			{
				ignoreCacheUpStream = TRUE;
				goto upstream_fwd_fail;
			}

			if ((!(portMask&rtl8651_externalMulticastPortMask)) &&
					((info.port)&rtl8651_externalMulticastPortMask))
			{	

				
				extIp = sip;

				
				PKTSET_UINT32_UNALIGNED(sip, &fwdIp->ip_src);
				
				_RTL8651_L3_ADJUST_CHECKSUM(sip, sip_org, fwdIp->ip_sum);
				if (!(ntohs(fwdIp->ip_off) & IP_OFFMASK))
				{
					uint8 *l4hdr = (uint8 *) ((int8 *) fwdIp + ((*(uint8*)fwdIp&0xf) << 2));
					uint16 *cksum = NULL;

					switch(fwdHdr->ph_proto)
					{
						case PKTHDR_TCP:
	 						cksum = &(((struct tcphdr*)l4hdr)->th_sum);
							break;
						case PKTHDR_UDP:
							cksum = &(((struct udphdr*)l4hdr)->uh_sum);
							break;
						default:
							goto skip_l4_checksum;
					}
					
					_RTL8651_L3_ADJUST_CHECKSUM(sip, sip_org, *cksum);
				skip_l4_checksum:
					; 
				}
			}

			
			if (up->vlan != vlanp)
			{	
				if (fwdIp->ip_ttl > 1)
				{
					_RTL8651_TTL_DEC_1(fwdIp->ip_ttl, fwdIp->ip_sum);
					#ifdef RTL8651_MCAST_DATA_DEBUG
					rtlglue_printf("[MCAST] \t\t==> decrease TTL to %d ...\n", fwdIp->ip_ttl);
					#endif
				}else
				{	
					#ifdef RTL8651_MCAST_DATA_DEBUG
					rtlglue_printf("[MCAST] \t\tTTL(%d) too low ==> Ignore it!!!\n", fwdIp->ip_ttl);
					#endif
					goto upstream_fwd_fail;
				}
			}

			if (MCast_l3ActRouting(fwdHdr, &info) != SUCCESS)
			{
	upstream_fwd_fail:
				mBuf_freeMbufChain(fwdMbuf);
			}
		}else
		{
			goto release_out;
		}
	}
	

	

#ifdef RTL8651_MCAST_ALWAYS2UPSTREAM
	if (((up->isSet == TRUE) && (ignoreCacheUpStream == FALSE)) || (CacheVlanMask && CachePortMask))
	{
		uint16 recUpPortMask;
		uint32 recUpVlanMask;
		if ((up->isSet == TRUE) && (ignoreCacheUpStream == FALSE))
		{	
			MCast_asicSetActualVlanPort(&recUpPortMask, &recUpVlanMask, up->netif, up->vlan, upPortMask, (uint32)(up->M_intfIp));

			if ((recUpPortMask = (recUpPortMask & ~(portMask))) != 0)
			{	
				CacheVlanMask |= recUpVlanMask;
				CachePortMask |= recUpPortMask;
			}				
		}
		}
#else	
	
	if ((totalUpCnt == 0) && (up->isSet == TRUE) && (ignoreCacheUpStream == FALSE))
	{
		uint16 recUpPortMask;
		uint32 recUpVlanMask;

		assert(!CacheVlanMask && !CachePortMask);

		MCast_asicSetActualVlanPort(&recUpPortMask, &recUpVlanMask, up->netif, up->vlan, upPortMask, (uint32)(up->M_intfIp));

		if ((recUpPortMask = (recUpPortMask & ~(portMask))) != 0)
		{
			CacheVlanMask |= recUpVlanMask;
			CachePortMask |= recUpPortMask;
		}
	}
#endif	
	if (CacheVlanMask && CachePortMask)
	{	
		#ifdef RTL8651_MCAST_DATA_DEBUG
		{
			int32 retval;

			rtlglue_printf("[MCAST] *FLOW* Add Flow{%u.%u.%u.%u} (%u.%u.%u.%u)(0x%X)[0x%X]=>(0x%X)[0x%X]\n",
				NIPQUAD(mcast_addr), NIPQUAD(sip_org), vlanp->vid, portSrc, CacheVlanMask, CachePortMask);
		#ifdef RTL8651_MCAST_ALWAYS2UPSTREAM
				retval = _rtl8651_addMulticast(mcast_addr, sip_org, vlanp->vid, portSrc, CacheVlanMask, CachePortMask, extIp, ((keep_pkt == TRUE)?TRUE:FALSE), 0);
		#else
				retval = _rtl8651_addMulticast(mcast_addr, sip_org, vlanp->vid, portSrc, CacheVlanMask, CachePortMask, extIp, ((keep_pkt == TRUE)?TRUE:FALSE), (totalUpCnt == 0)?RTL865X_MULTICAST_UPLOADONLY:0);
		#endif 
			if (retval == SUCCESS)
				rtlglue_printf("\t=>SUCCESS!\n");
			else
				rtlglue_printf("\t=>FAILED!\n");
		}
	#else
		#ifdef RTL8651_MCAST_ALWAYS2UPSTREAM
		_rtl8651_addMulticast(mcast_addr, sip_org, vlanp->vid, portSrc, CacheVlanMask, CachePortMask, extIp, ((keep_pkt == TRUE)?TRUE:FALSE), 0);
		#else
		_rtl8651_addMulticast(mcast_addr, sip_org, vlanp->vid, portSrc, CacheVlanMask, CachePortMask, extIp, ((keep_pkt == TRUE)?TRUE:FALSE),  (totalUpCnt == 0)?RTL865X_MULTICAST_UPLOADONLY:0);
		#endif 
	#endif
	}
ret:
	
	{
		switch(l4proto)
		{
			case IPPROTO_TCP:
				rtl8651_fwdEngineCounter.tcpSend ++;
				break;
			case IPPROTO_UDP:
				rtl8651_fwdEngineCounter.udpSend ++;
				break;
		}
	}
	if (keep_pkt == TRUE)
		return FWDENG_IP_MULTICAST;
	else
		return SUCCESS;

release_out:
	mBuf_freeMbufChain(fwdMbuf);
	goto ret;
}


int32 _rtl8651_multicast_Fastfwd(struct rtl_pktHdr *pkthdr, struct ip *ip, rtl8651_tblDrv_vlanTable_t *vlanp)
{
	rtl8651_tblDrvMCastRouteInfo_t info;
	rtl8651_tblDrv_mCast_t *m_data;
	ipaddr_t mcast_addr;
	ipaddr_t sip;
	uint32 vlanMask;
	uint16 idx;
	uint8 l4proto;
	uint8 keep_pkt = FALSE;

	if ((system_enable == FALSE) || (rtl8651_drvMulticastProcessEnable == FALSE))
		return FAILED;

	if (IN_MULTICAST_RESV(pkthdr->ph_dip))	
		return FAILED;

	mcast_addr = pkthdr->ph_dip;
	sip = pkthdr->ph_sip;
	l4proto = ip->ip_p;

	
	{
		rtl8651_tblDrvMCastFilter_t *filter;
		filter = _rtl8651_multicast_FindFilter(mcast_addr);
		if (filter)
		{
			switch (filter->action)
			{
				case (MCAST_FILTER_ACT_MIRROR|MCAST_FILTER_ACT_NOPROC):
					
					return FWDENG_IP_MULTICAST;
				case (MCAST_FILTER_ACT_MIRROR):
					
					keep_pkt = TRUE;
					break;
				case (MCAST_FILTER_ACT_NOPROC):
					
					goto errout;
			}
		}
	}
	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("[MCAST] *FLOW* Find Flow{%u.%u.%u.%u} from (%u.%u.%u.%u)(0x%X)[0x%X]\n",
		NIPQUAD(mcast_addr), NIPQUAD(sip), vlanp->vid, pkthdr->ph_portlist);
	#endif

	m_data = (rtl8651_tblDrv_mCast_t *)_rtl8651_findMCast(mcast_addr, sip, vlanp->vid, pkthdr->ph_portlist);

	if(m_data == NULL )
	{
		#ifdef RTL8651_MCAST_DATA_DEBUG
		rtlglue_printf("\t=>NOT Found......... To Slow Path\n");
		#endif
		return FAILED;
	}
	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("\t=>Entry FOUND: (0x%X) [0x%X]\n", m_data->vlanMask, m_data->mbr);
	#endif

	
	if (m_data->flag & RTL865X_MULTICAST_EXTIP_SET)
	{
		#ifdef RTL8651_MCAST_DATA_DEBUG
		rtlglue_printf("\t=>Skip due to source IP modification...... To Slow Path\n");
		#endif
		return FAILED;
	}
	
	
	if (((1<<(pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1))) & m_data->vlanMask)
		&& (m_data->vlanMask & ~(1<<(pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)))))
	{
		#ifdef RTL8651_MCAST_DATA_DEBUG
		rtlglue_printf("\t=>Skip due to complex TTL handling...... To Slow Path\n");
		#endif
		return FAILED;
	}else
	{	
		if (m_data->vlanMask & ~(1<<(pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1))))
		{	
			if (ip->ip_ttl > 1)
			{
				_RTL8651_TTL_DEC_1(ip->ip_ttl, ip->ip_sum);					
			}else
			{	
				goto freeout;
			}
		}
	}
	
	
	info.m_addr		= mcast_addr;
	info.port			= m_data->mbr;
	info.isPortMask	= TRUE;
	info.netif			= NULL;
	info.ifInfo			= 0;

	vlanMask = m_data->vlanMask;
	idx = 0;

	if ((!vlanMask) || (!(m_data->mbr)))
		return FAILED;

	
	pkthdr->ph_flags &= ~(CSUM_IP|CSUM_L4);

	while (vlanMask)
	{
		if (vlanMask & 1)	
		{
			info.vlan = &DrvTbl.vlan[idx];

			
			if ((keep_pkt == FALSE) && (vlanMask == 1))
			{	
				if (MCast_l3ActRouting(pkthdr, &info) != SUCCESS)
					goto errout;
			}else
			{	
				struct rtl_mBuf *newMbuf;
				newMbuf = mBuf_dupPacket(pkthdr->ph_mbuf, MBUF_DONTWAIT);

				if (MCast_l3ActRouting(newMbuf->m_pkthdr, &info) != SUCCESS)
					mBuf_freeMbufChain(newMbuf);
					
			}
		}
		
		vlanMask = vlanMask >> 1;
		idx ++;
	}

	
	{
		switch(l4proto)
		{
			case IPPROTO_TCP:
				rtl8651_fwdEngineCounter.tcpSend ++;
				break;
			case IPPROTO_UDP:
				rtl8651_fwdEngineCounter.udpSend ++;
				break;
		}
	}

	if (keep_pkt == TRUE)
		return FWDENG_IP_MULTICAST;
	else
		return SUCCESS;
errout:

freeout:
	mBuf_freeMbufChain(pkthdr->ph_mbuf);
	return SUCCESS;
}

#define RTL8651_MULTICAST_FILTER_FUNCTIONS


int32 _rtl8651_multicast_FilterInit(rtl8651_fwdEngineInitPara_t *para)
{
	uint32 idx;
	rtl8651_tblDrvMCastFilter_t *ptr;

	rtlglue_drvMutexLock();

	
	for (idx = 0 ; idx < RTL8651_MCastFilterHashSize ; idx ++)
		MCastFilter[idx] = NULL;

	
	MCastFilter_FreeHdr = NULL;

	if (!para->MCastFilterEntryCnt)
		goto out;

	
	MCastFilter_FreeHdr = (rtl8651_tblDrvMCastFilter_t *)rtlglue_malloc(para->MCastFilterEntryCnt * sizeof(rtl8651_tblDrvMCastFilter_t));

	if (!MCastFilter_FreeHdr)
		goto out;

	
	bzero(MCastFilter_FreeHdr, (para->MCastFilterEntryCnt * sizeof(rtl8651_tblDrvMCastFilter_t)));
	
	ptr = MCastFilter_FreeHdr;
	for (idx = 0 ; idx < para->MCastFilterEntryCnt ; idx ++, ptr++)
	{
		if (idx < (para->MCastFilterEntryCnt - 1))
			ptr->next = ptr + 1;
		else
			ptr->next = NULL;
	}

out:
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}
void rtl8651_multicast_FilterReinit(void)
{
	uint32 idx;
	
	for (idx = 0 ; idx < RTL8651_MCastFilterHashSize ; idx ++)
	{
		while (MCastFilter[idx])	
			rtl8651_multicast_RemoveFilter(MCastFilter[idx]->gip, MCAST_FILTER_ACT_ALL);
		assert(MCastFilter[idx] == NULL);
	}
}

uint32 _rtl8651_multicast_Filter_Hash(ipaddr_t gip)
{
	return (gip % RTL8651_MCastFilterHashSize);
}

rtl8651_tblDrvMCastFilter_t *_rtl8651_multicast_FindFilter(ipaddr_t gip)
{
	rtl8651_tblDrvMCastFilter_t *ptr;

	ptr = MCastFilter[_rtl8651_multicast_Filter_Hash(gip)];

	while(ptr && (ptr->gip != gip))
		ptr = ptr->next;

	return ptr;
}

int32 rtl8651_multicast_AddFilter(ipaddr_t gip, uint32 action)
{
	rtl8651_tblDrvMCastFilter_t *ptr;
	uint32 idx;
	uint32 orgAction = 0;	

	
	MCASTFILTER_UNLOCK_ACTIONCHECK(action, (MCAST_FILTER_ACT_NOPROC|MCAST_FILTER_ACT_L2FWD), FAILED);

	rtlglue_drvMutexLock();

	ptr = _rtl8651_multicast_FindFilter(gip);
	if (ptr)
	{	
		orgAction = ptr->action;
		ptr->action |= action;
	}else
	{	
		if (!action)	
			goto out;
		if (!MCastFilter_FreeHdr)	
			goto errout;

		
		ptr = MCastFilter_FreeHdr; 
		MCastFilter_FreeHdr = MCastFilter_FreeHdr->next;

		
		ptr->gip		= gip;
		ptr->action	= action;

		
		idx = _rtl8651_multicast_Filter_Hash(gip);
		ptr->next = MCastFilter[idx];
		MCastFilter[idx] = ptr;
	}

	
	_rtl8651_setMGroupAttribute(gip, ((ptr->action & MCAST_FILTER_ACT_MIRROR)?TRUE:FALSE));
	if (ptr->action & (MCAST_FILTER_ACT_NOPROC|MCAST_FILTER_ACT_L2FWD))
	{
		rtl8651_tblDrvMCast_t *m_entry = MCast_FindMCastEntry(gip, FALSE, FALSE);
		if (m_entry)
		{
			MCast_RemoveMCastEntry(m_entry);
			_rtl8651_delMulticast(gip, 0, 0);
		}
	}

out:
	rtlglue_drvMutexUnlock();
	return SUCCESS;
errout:
	rtlglue_drvMutexUnlock();
	return FAILED;
}
int32 rtl8651_multicast_RemoveFilter(ipaddr_t gip, uint32 action)
{
	rtl8651_tblDrvMCastFilter_t *ptr, *prev;
	uint32	idx;
	uint8	TurnMirrorOff = FALSE;

	rtlglue_drvMutexLock();

	
	idx = _rtl8651_multicast_Filter_Hash(gip);
	prev = NULL;
	ptr = MCastFilter[idx];
	while (ptr && (ptr->gip != gip))
	{
		ptr	= ptr->next;
		prev	= ptr;
	}
	

	
	if (ptr)
	{
		uint32 orgAction = ptr->action;
	
		ptr->action &= ~action;

		if ((orgAction & MCAST_FILTER_ACT_MIRROR) && (!(ptr->action & MCAST_FILTER_ACT_MIRROR)))
			TurnMirrorOff = TRUE;
		
		if (ptr->action)
			goto out;

		
		
		if (prev)
			prev->next = ptr->next;

		if (MCastFilter[idx] == ptr)
			MCastFilter[idx] = ptr->next;

		
		ptr->next = MCastFilter_FreeHdr;
		MCastFilter_FreeHdr = ptr;
	}

out:

	
	if (TurnMirrorOff == TRUE)
		_rtl8651_setMGroupAttribute(gip, FALSE);

	rtlglue_drvMutexUnlock();
	return SUCCESS;
}


