//==========================================================================
//
//      hal_misc.c
//
//      HAL miscellaneous functions
//
//==========================================================================

// Purpose:      HAL miscellaneous functions
// Description:  This file contains miscellaneous functions provided by the
//               HAL.
//
//========================================================================*/

/*
#include <include/pkgconf/hal.h>

#include <include/infra/rt_type.h>         // Base types
#include <include/infra/rt_trac.h>         // tracing macros
#include <include/infra/rt_ass.h>          // assertion macros

#define RTARC_HAL_COMMON_EXPORT_CPU_MACROS
#include <include/hal/hal_arch.h>           // architectural definitions

#ifdef	RTPKG_DEVS_REALTEK_8601
#include <include\pkgconf\devs_realtek_8601.h>
#endif

#include <include/hal/hal_intr.h>           // Interrupt handling

#include <include/hal/hal_cache.h>          // Cache handling

#include <include/hal/mips-regs.h>          // FPU cause register definitions
*/

#include "include/pkgconf/hal.h"

#include "include/infra/rt_type.h"         // Base types
#include "include/infra/rt_trac.h"         // tracing macros
#include "include/infra/rt_ass.h"          // assertion macros

#define RTARC_HAL_COMMON_EXPORT_CPU_MACROS
#include "include/hal/hal_arch.h"           // architectural definitions

#ifdef	RTPKG_DEVS_REALTEK_8601
#include "include\pkgconf\devs_realtek_8601.h"
#endif

#include "include/hal/hal_intr.h"           // Interrupt handling

#include "include/hal/hal_cache.h"          // Cache handling

#include "include/hal/mips-regs.h"          // FPU cause register definitions

/*------------------------------------------------------------------------*/
/* If required, define a variable to store the clock period.              */

#ifdef RTHWR_HAL_CLOCK_PERIOD_DEFINED

RT_WORD32 rt_hal_clock_period;

#endif

#if 0

externC unsigned int _byte_swap(unsigned int uValue)
{
	return (uValue >> 24) | ((uValue >> 8) & 0xff00) |
		((uValue << 8) & 0xff0000) | (uValue << 24);
}
#endif
/*------------------------------------------------------------------------*/
/* First level C exception handler.                                       */

externC void __handle_exception (void);

externC HAL_SavedRegisters *_hal_registers;

externC void* volatile __mem_fault_handler;

externC rt_uint8 rt_hal_mips_process_fpe( HAL_SavedRegisters *regs );

externC rt_uint32 rt_hal_exception_handler(HAL_SavedRegisters *regs)
{
    int vec = regs->vector>>2;

	vec = vec;	// avoid compiler complianing unused variable

#if defined(RTHWR_HAL_MIPS_FPU)
    // Special handling of FPU exceptions
    if (RTNUM_HAL_VECTOR_FPE == vec) {

#if defined(RTSEM_HAL_MIPS_EMULATE_UNIMPLEMENTED_FPU_OPS)
        // We may be required to emulate certain unimplemented Floating Point
        // operations

        // rt_hal_mips_process_fpe() returns non-zero if it could handle
        // the exception successfully. If so, we just return

        if ( rt_hal_mips_process_fpe(regs) )
            return 0;
#endif

        // Find the cause of the FPU exception, clear the flag and use
        // the decoded vector number.
        {
            rt_uint32 cause = regs->fcr31;

            if (cause & FCR31_CAUSE_I) {
                vec = RTNUM_HAL_EXCEPTION_FPU_INEXACT;
                cause &= ~(FCR31_FLAGS_I | FCR31_CAUSE_I);
            } else if (cause & FCR31_CAUSE_U) {
                vec = RTNUM_HAL_EXCEPTION_FPU_UNDERFLOW;
                cause &= ~(FCR31_FLAGS_U | FCR31_CAUSE_U);
            } else if (cause & FCR31_CAUSE_O) {
                vec = RTNUM_HAL_EXCEPTION_FPU_OVERFLOW;
                cause &= ~(FCR31_FLAGS_O | FCR31_CAUSE_O);
            } else if (cause & FCR31_CAUSE_Z) {
                vec = RTNUM_HAL_EXCEPTION_FPU_DIV_BY_ZERO;
                cause &= ~(FCR31_FLAGS_Z | FCR31_CAUSE_Z);
            } else if (cause & FCR31_CAUSE_V) {
                vec = RTNUM_HAL_EXCEPTION_FPU_INVALID;
                cause &= ~(FCR31_FLAGS_V | FCR31_CAUSE_V);
            }
            regs->fcr31 = cause;

#if 1
            // Update the FPU status register with the new
            // setting. This is a workaround for the signal2 test
            // which does a longjump in the exception handler and thus
            // never gets around to executing the register restore
            // code.
            asm ("ctc1	%0,$31" : : "r" (cause));
#endif
        }
    }
#endif // RTHWR_HAL_MIPS_FPU


#if defined(RTDBG_HAL_DEBUG_GDB_INCLUDE_STUBS)

    // If we caught an exception inside the stubs, see if we were expecting it
    // and if so jump to the saved address

#ifndef RTDBG_HAL_STUB_WITH_VIRTUAL_VECTOR

    if (__mem_fault_handler) {
        regs->pc = (RT_ADDRWORD)__mem_fault_handler;
        return 0; // Caught an exception inside stubs
    }

    // Set the pointer to the registers of the current exception
    // context. At entry the GDB stub will expand the
    // HAL_SavedRegisters structure into a (bigger) register array.
    _hal_registers = regs;
    __handle_exception();

#endif

#elif defined(RTFUN_HAL_COMMON_KERNEL_SUPPORT) && defined(RTPKG_HAL_EXCEPTIONS)

    // We should decode the vector and pass a more appropriate
    // value as the second argument. For now we simply pass a
    // pointer to the saved registers. We should also divert
    // breakpoint and other debug vectors into the debug stubs.

//    rt_hal_deliver_exception( vec, (RT_ADDRWORD)regs );

#else

    RT_FAIL("Exception!!!");

#endif
    return 0;
}


#if 0

/*------------------------------------------------------------------------*/
/* default ISR                                                            */

#ifndef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT
externC rt_uint32 hal_default_isr(RT_ADDRWORD vector, RT_ADDRWORD data)
{
#if defined(RTDBG_HAL_MIPS_DEBUG_GDB_CTRLC_SUPPORT) &&      \
    defined(RTHWR_HAL_GDB_PORT_VECTOR) &&              \
    defined(HAL_CTRLC_ISR)

#ifndef RTIMP_HAL_COMMON_INTERRUPTS_CHAIN
    if( vector == RTHWR_HAL_GDB_PORT_VECTOR )
#endif
    {
        rt_uint32 result = HAL_CTRLC_ISR( vector, data );
        if( result != 0 ) return result;
    }

#if defined(RTSEM_HAL_USE_ROM_MONITOR_RtMon)
#if defined(HAL_DIAG_IRQ_CHECK)
    {
        rt_uint32 ret;
        /* let ROM monitor handle unexpected interrupts */
        HAL_DIAG_IRQ_CHECK(vector, ret);
        if (ret<=0)
            return ret;
    }
#endif // def HAL_DIAG_IRQ_CHECK
#endif // def RTSEM_HAL_USE_ROM_MONITOR_RtMon
#endif

    RT_TRACE1(true, "Interrupt: %d", vector);
    RT_FAIL("Spurious Interrupt!!!");
    return 0;
}

#else // RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT

externC rt_uint32 hal_arch_default_isr(RT_ADDRWORD vector, RT_ADDRWORD data)
{
#if defined(RTDBG_HAL_MIPS_DEBUG_GDB_CTRLC_SUPPORT) &&      \
    defined(RTHWR_HAL_GDB_PORT_VECTOR) &&              \
    defined(HAL_CTRLC_ISR)

#if defined(RTSEM_HAL_USE_ROM_MONITOR_RtMon)
#if defined(HAL_DIAG_IRQ_CHECK)
    {
        rt_uint32 ret;
        /* let ROM monitor handle unexpected interrupts */
        HAL_DIAG_IRQ_CHECK(vector, ret);
        if (ret<=0)
            return ret;
    }
#endif // def HAL_DIAG_IRQ_CHECK
#endif // def RTSEM_HAL_USE_ROM_MONITOR_RtMon
#endif

#ifdef RTPKG_REDBOOT
    hal_ctrlc_isr( vector, data );
#endif

    return 0;
}

#endif // RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT

/*------------------------------------------------------------------------*/
/* data copy and bss zero functions                                       */

typedef void (RT_SYM_ADDRESS)(void);

// All these must use this type of address to stop them being given relocations
// relative to $gp (i.e. assuming they would be in .sdata)
extern RT_SYM_ADDRESS __ghsbegin_data;
extern RT_SYM_ADDRESS __ghsend_sdata;
extern RT_SYM_ADDRESS __ghsbegin_romdata;

#ifdef RT_HAL_STARTUP_ROM
void hal_copy_data(void)
{
    char *p = (char *)&__ghsbegin_data;
    char *q = (char *)&__ghsbegin_romdata;

    while( p != (char *)&__ghsend_sdata )
        *p++ = *q++;
}
#endif


/*------------------------------------------------------------------------*/

#ifdef RTSEM_HAL_STOP_CONSTRUCTORS_ON_FLAG
rt_bool rt_hal_stop_constructors;
#endif


/*------------------------------------------------------------------------*/
/* Determine the index of the ls bit of the supplied mask.                */

rt_uint32 hal_lsbit_index(rt_uint32 mask)
{
    rt_uint32 n = mask;

    static const signed char tab[64] =
    { -1, 0, 1, 12, 2, 6, 0, 13, 3, 0, 7, 0, 0, 0, 0, 14, 10,
      4, 0, 0, 8, 0, 0, 25, 0, 0, 0, 0, 0, 21, 27 , 15, 31, 11,
      5, 0, 0, 0, 0, 0, 9, 0, 0, 24, 0, 0 , 20, 26, 30, 0, 0, 0,
      0, 23, 0, 19, 29, 0, 22, 18, 28, 17, 16, 0
    };

    n &= ~(n-1UL);
    n = (n<<16)-n;
    n = (n<<6)+n;
    n = (n<<4)+n;

    return tab[n>>26];
}

/*------------------------------------------------------------------------*/
/* Determine the index of the ms bit of the supplied mask.                */

rt_uint32 hal_msbit_index(rt_uint32 mask)
{
    rt_uint32 x = mask;
    rt_uint32 w;

    /* Phase 1: make word with all ones from that one to the right */
    x |= x >> 16;
    x |= x >> 8;
    x |= x >> 4;
    x |= x >> 2;
    x |= x >> 1;

    /* Phase 2: calculate number of "1" bits in the word        */
    w = (x & 0x55555555) + ((x >> 1) & 0x55555555);
    w = (w & 0x33333333) + ((w >> 2) & 0x33333333);
    w = w + (w >> 4);
    w = (w & 0x000F000F) + ((w >> 8) & 0x000F000F);
    return (rt_uint32)((w + (w >> 16)) & 0xFF);

}

/*------------------------------------------------------------------------*/
/* Delay for some number of useconds.                                     */
void
hal_delay_us(int us)
{
	rt_uint32 val1, val2;
	int diff;
	long ticks;

    // Scale the desired number of microseconds to be a number of
    // incrementer ticks
    
    // note that the following multiplication may cause the (intermedia) result
    // out of 32-bit range. 
    //
    // ticks = us * (RTNUM_HAL_RTC_PERIOD * RTNUM_HAL_RTC_DENOMINATOR) / 1000000;
    ticks = us * (RTNUM_HAL_RTC_PERIOD / 10000);

	HAL_CLOCK_READ(&val1);
	while (ticks > 0)
	{
		HAL_CLOCK_READ(&val2);
		diff = val2 - val1;
		if (diff < 0)
			diff += RTNUM_HAL_RTC_PERIOD;
		ticks -= diff;
		val1 = val2;
	}
}

/*------------------------------------------------------------------------*/
/* Idle thread action                                                     */

#include <include/infra/diag.h>

void hal_idle_thread_action( rt_uint32 count )
{
#if defined (RTPAG_DEVS_SLEEP_8601)
	asm volatile (         
	        ".word   0x42000038;"           //enable sleep mode
	        "nop;nop;nop;"                  //add by chteng
            );        
#endif
}

/*------------------------------------------------------------------------*/
/* End of hal_misc.c                                                      */

#endif
