//========================================================================
//
//      dbg-thread-syscall.h
//
//      Supports thread-aware debugging
//
//========================================================================

//========================================================================
//
// Purpose:       
// Description:   Supports thread-aware debugging
// Usage:         This header is not to be included by user code.
//                This file should be included only by
//                thread-syscall-relay.c and dbg-thread-demux.c
//
//========================================================================


enum dbg_syscall_ids 
  {
    dbg_null_func ,
    dbg_capabilities_func,
    dbg_currthread_func,
    dbg_threadlist_func,
    dbg_threadinfo_func,
    dbg_getthreadreg_func,
    dbg_setthreadreg_func,
    dbg_scheduler_func,
  } ;


union dbg_thread_syscall_parms
{
  struct
  {
    struct dbg_capabilities * abilities ;
  } cap_parms ;

  struct
  {
    threadref * ref ;
  } currthread_parms ;
  
  struct
  {
    int startflag ;
    threadref * lastid ;
    threadref * nextthreadid ;
  } threadlist_parms ;

  struct
  {
    threadref * ref ;
    struct rtmon_thread_debug_info * info ;
  } info_parms ;
  
  struct
  {
    threadref * thread ;
    int regcount ;
    void * registers ;
  } reg_parms ;
  struct
  {
    threadref * thread ;        /* 64-bit thread identifier */
    int lock;                   /* 0 == unlock, 1 == lock */
    int mode;                   /* 0 == short (step), 1 == continue */
  } scheduler_parms ;
} ;


typedef int (*dbg_syscall_func) (enum dbg_syscall_ids id,
                                 union dbg_thread_syscall_parms  * p ) ;
