#ifndef RTONCE_HAL_HAL_ARCH_H
#define RTONCE_HAL_HAL_ARCH_H

//==========================================================================
//
//      hal_arch.h
//
//      Architecture specific abstractions
//
//==========================================================================


#ifndef __ASSEMBLER__
/*
#include <include/pkgconf/hal.h>
*/
#include "../pkgconf/hal.h"

#ifndef	ENTER_ASSEMBLY_GREENHILLS
/*
#include <include/infra/rt_type.h>
#include <include/hal/var_arch.h>
*/
#include "../infra/rt_type.h"
#include "var_arch.h"
#endif


//--------------------------------------------------------------------------
// Processor saved states:
// The layout of this structure is also defined in "arch.inc", for assembly
// code. Do not change this without changing that (or vice versa).
// Notes: This structure is carefully laid out. It is a multiple of 8
// bytes and the pc and badvr fields are positioned to ensure that
// they are on 8 byte boundaries. 

#ifdef RTHWR_HAL_MIPS_64BIT
# define RT_HAL_MIPS_REG RT_WORD64
# define RT_HAL_MIPS_REG_SIZE 8
#else
# define RT_HAL_MIPS_REG RT_WORD32
# define RT_HAL_MIPS_REG_SIZE 4
#endif

#if defined(RTHWR_HAL_MIPS_FPU)
# if defined(RTHWR_HAL_MIPS_FPU_64BIT)
#  define RT_HAL_FPU_REG RT_WORD64
# elif defined(RTHWR_HAL_MIPS_FPU_32BIT)
#  define RT_HAL_FPU_REG RT_WORD32
# else
# error MIPS FPU register size not defined
# endif
#endif

typedef struct 
{
    // These are common to all saved states
    RT_HAL_MIPS_REG    d[32];          /* Data regs                    */
    RT_HAL_MIPS_REG    hi;             /* hi word of mpy/div reg       */
    RT_HAL_MIPS_REG    lo;             /* lo word of mpy/div reg       */
#ifdef RTHWR_HAL_MIPS_FPU
    RT_HAL_FPU_REG     f[32];          /* FPU registers                */
    RT_WORD32          fcr31;          /* FPU control/status register  */
    RT_WORD32          fppad;          /* Dummy location to make this  */
                                        /* structure a multiple of 8    */
                                        /* bytes long.                  */
#endif
    
    // These are only saved for exceptions and interrupts
    RT_WORD32          vector;         /* Vector number                */
    RT_WORD32          sr;             /* Status Reg                   */
    RT_HAL_MIPS_REG    pc;             /* Program Counter              */
    RT_WORD32          cache;          /* Cache control register       */


    // These are only saved for exceptions, and are not restored
    // when continued.
    RT_WORD32          cause;          /* Exception cause register     */
    RT_HAL_MIPS_REG    badvr;          /* Bad virtual address reg      */
    RT_WORD32          prid;           /* Processor Version            */
    RT_WORD32          config;         /* Config register              */
#if   LX5280_PATCH                     /* Guevara add this to support LX5280  at 20031029  */
   RT_WORD32          estatus;
   RT_WORD32          ecause;
   RT_WORD32          intvec;
   RT_WORD32          cbs0;
   RT_WORD32          cbs1;
   RT_WORD32          cbs2;
   RT_WORD32          cbe0;
   RT_WORD32          cbe1;
   RT_WORD32          cbe2;
   RT_WORD32          lps0;
   RT_WORD32          lpe0;
   RT_WORD32          lpc0;
   RT_WORD32          mmd;
   RT_WORD32          m0ll;
   RT_WORD32          m0lh;
   RT_WORD32          m0hl;
   RT_WORD32          m0hh;
   RT_WORD32          m1ll;
   RT_WORD32          m1lh;
   RT_WORD32          m1hl;
   RT_WORD32          m1hh;
   RT_WORD32          m2ll;
   RT_WORD32          m2lh;
   RT_WORD32          m2hl;
   RT_WORD32          m2hh;
   RT_WORD32          m3ll;
   RT_WORD32          m3lh;
   RT_WORD32          m3hl;
   RT_WORD32          m3hh;


#endif
} HAL_SavedRegisters;

//--------------------------------------------------------------------------
// Exception handling function.
// This function is defined by the kernel according to this prototype. It is
// invoked from the HAL to deal with any CPU exceptions that the HAL does
// not want to deal with itself. It usually invokes the kernel's exception
// delivery mechanism.

externC void rt_hal_deliver_exception( RT_WORD code, RT_ADDRWORD data );

//--------------------------------------------------------------------------
// Bit manipulation macros

externC rt_uint32 hal_lsbit_index(rt_uint32 mask);
externC rt_uint32 hal_msbit_index(rt_uint32 mask);

#define HAL_LSBIT_INDEX(index, mask) index = hal_lsbit_index(mask);

#define HAL_MSBIT_INDEX(index, mask) index = hal_msbit_index(mask);

//--------------------------------------------------------------------------
// Context Initialization


// Optional FPU context initialization
#ifdef RTHWR_HAL_MIPS_FPU
#define HAL_THREAD_INIT_FPU_CONTEXT( _regs_, _id_ )                             \
{                                                                               \
    for( _i_ = 0; _i_ < 32; _i_++ ) (_regs_)->f[_i_] = (_id_)|0xFF00|_i_;       \
    (_regs_)->fcr31 = 0x01000000;                                               \
}
#else
#define HAL_THREAD_INIT_FPU_CONTEXT( _regs_, _id_ )
#endif


// Initialize the context of a thread.
// Arguments:
// _sparg_ name of variable containing current sp, will be written with new sp
// _thread_ thread object address, passed as argument to entry point
// _entry_ entry point address.
// _id_ bit pattern used in initializing registers, for debugging.
#define HAL_THREAD_INIT_CONTEXT( _sparg_, _thread_, _entry_, _id_ )                     \
{                                                                                       \
    register RT_WORD _sp_ = ((RT_WORD)_sparg_)-56;                                    \
    register HAL_SavedRegisters *_regs_;                                                \
    int _i_;                                                                            \
    _sp_ = _sp_ & 0xFFFFFFF0;                                                           \
    _regs_ = (HAL_SavedRegisters *)(((_sp_) - sizeof(HAL_SavedRegisters))&0xFFFFFFF0);  \
    for( _i_ = 0; _i_ < 32; _i_++ ) (_regs_)->d[_i_] = (_id_)|_i_;                      \
    HAL_THREAD_INIT_FPU_CONTEXT( _regs_, _id_ );                                        \
    (_regs_)->d[29] = (RT_HAL_MIPS_REG)(_sp_);       /* SP = top of stack      */      \
    (_regs_)->d[04] = (RT_HAL_MIPS_REG)(_thread_);   /* R4 = arg1 = thread ptr */      \
    (_regs_)->lo = 0;                                 /* LO = 0                 */      \
    (_regs_)->hi = 0;                                 /* HI = 0                 */      \
    (_regs_)->d[30] = (RT_HAL_MIPS_REG)(_sp_);       /* FP = top of stack      */      \
    (_regs_)->d[31] = (RT_HAL_MIPS_REG)(_entry_);    /* RA(d[31]) = entry point*/      \
    (_regs_)->pc = (RT_WORD)(_entry_);               /* PC = entry point       */      \
    (_regs_)->sr  = 0x00000001;                       /* SR = ls 3 bits only    */      \
    _sparg_ = (RT_ADDRESS)_regs_;                                                      \
}

//--------------------------------------------------------------------------
// Context switch macros.
// The arguments are pointers to locations where the stack pointer
// of the current thread is to be stored, and from where the sp of the
// next thread is to be fetched.

externC void hal_thread_switch_context( RT_ADDRESS to, RT_ADDRESS from );
externC void hal_thread_load_context( RT_ADDRESS to )
    __attribute__ ((noreturn));

#define HAL_THREAD_SWITCH_CONTEXT(_fspptr_,_tspptr_)                    \
        hal_thread_switch_context( (RT_ADDRESS)_tspptr_,               \
                                   (RT_ADDRESS)_fspptr_);

#define HAL_THREAD_LOAD_CONTEXT(_tspptr_)                               \
        hal_thread_load_context( (RT_ADDRESS)_tspptr_ );

//--------------------------------------------------------------------------
// Execution reorder barrier.
// When optimizing the compiler can reorder code. In multithreaded systems
// where the order of actions is vital, this can sometimes cause problems.
// This macro may be inserted into places where reordering should not happen.
// The "memory" keyword is potentially unnecessary, but it is harmless to
// keep it.
#if (CROSS_COMPILER == COMPILER_GNU_GCC)
#define HAL_REORDER_BARRIER() asm volatile ( "" : : : "memory" )
#endif

#if (CROSS_COMPILER == COMPILER_GREEN_HILLS)
/*
 *	Green Hills compiler can't recognize all the inline assembly syntax of GNU.
 *	In fact, it also defines its own inline assembly macro facility.
 */
asm void HAL_REORDER_BARRIER(void)
{
	.set	noreorder
}
#endif

//--------------------------------------------------------------------------
// Breakpoint support
// HAL_BREAKPOINT() is a code sequence that will cause a breakpoint to
// happen if executed.
// HAL_BREAKINST is the value of the breakpoint instruction and
// HAL_BREAKINST_SIZE is its size in bytes.
// HAL_BREAKINST_TYPE is the type.

#define HAL_BREAKPOINT(_label_)                 \
asm volatile (" .globl  " #_label_ ";"          \
              #_label_":"                       \
              " break   5"                      \
    );

#define HAL_BREAKINST           0x0005000d		// That is, "break 5" in MIPS-I assembly.

#define HAL_BREAKINST_SIZE      4

#define HAL_BREAKINST_TYPE      rt_uint32

//--------------------------------------------------------------------------
// Thread register state manipulation for GDB support.

// Default to a 32 bit register size for GDB register dumps.
#ifndef RT_HAL_GDB_REG
#define RT_HAL_GDB_REG RT_WORD32
#endif

// Translate a stack pointer as saved by the thread context macros above into
// a pointer to a HAL_SavedRegisters structure.
#define HAL_THREAD_GET_SAVED_REGISTERS( _sp_, _regs_ )          \
        (_regs_) = (HAL_SavedRegisters *)(_sp_)

// If the CPU has an FPU, we also need to move the FPU registers.
#ifdef RTHWR_HAL_MIPS_FPU
#define HAL_GET_GDB_FPU_REGISTERS( _regval_ , _regs_ )  \
RT_MACRO_START                                         \
    int _i_;                                            \
    for( _i_ = 0; _i_ < 32; _i_++ )                     \
        _regval_[38+_i_] = (_regs_)->f[_i_];            \
    _regval_[70] = (_regs_)->fcr31;                     \
RT_MACRO_END
#define HAL_SET_GDB_FPU_REGISTERS( _regs_ , _regval_ )  \
RT_MACRO_START                                         \
    int _i_;                                            \
    for( _i_ = 0; _i_ < 32; _i_++ )                     \
        (_regs_)->f[_i_] = _regval_[38+_i_];            \
    (_regs_)->fcr31 = _regval_[70];                     \
RT_MACRO_END
#else
#define HAL_GET_GDB_FPU_REGISTERS( _regval_ , _regs_ )
#define HAL_SET_GDB_FPU_REGISTERS( _regs_ , _regval_ )
#endif

#if LX5280_PATCH       //add by shao
// Copy a set of registers from a HAL_SavedRegisters structure into a
// GDB ordered array.    
#define HAL_GET_GDB_REGISTERS( _aregval_ , _regs_ )             \
{                                                               \
    RT_HAL_GDB_REG *_regval_ = (RT_HAL_GDB_REG *)(_aregval_); \
    int _i_;                                                    \
                                                                \
    for( _i_ = 0; _i_ < 32; _i_++ )                             \
        _regval_[_i_] = (_regs_)->d[_i_];                       \
                                                                \
    HAL_GET_GDB_FPU_REGISTERS( _regval_, _regs_ );              \
                                                                \
    _regval_[32] = (_regs_)->sr;                                \
    _regval_[33] = (_regs_)->lo;                                \
    _regval_[34] = (_regs_)->hi;                                \
    _regval_[35] = (_regs_)->badvr;                             \
    _regval_[36] = (_regs_)->cause;                             \
    _regval_[37] = (_regs_)->pc;                                \
    _regval_[73] = (_regs_)->estatus;                           \
    _regval_[74] = (_regs_)->ecause;                           \
    _regval_[75] = (_regs_)->intvec;                           \
    _regval_[76] = (_regs_)->cbs0;                           \
    _regval_[77] = (_regs_)->cbs1;                           \
    _regval_[78] = (_regs_)->cbs2;                           \
    _regval_[79] = (_regs_)->cbe0;                           \
    _regval_[80] = (_regs_)->cbe1;                           \
    _regval_[81] = (_regs_)->cbe2;                           \
    _regval_[82] = (_regs_)->lps0;                           \
    _regval_[83] = (_regs_)->lpe0;                           \
    _regval_[84] = (_regs_)->lpc0;                           \
    _regval_[85] = (_regs_)->mmd;                           \
	_regval_[86] = (_regs_)->m0ll;                           \
    _regval_[87] = (_regs_)->m0lh;                           \
    _regval_[88] = (_regs_)->m0hl;                           \
    _regval_[89] = (_regs_)->m0hh;                           \
	_regval_[90] = (_regs_)->m1ll;                           \
    _regval_[91] = (_regs_)->m1lh;                           \
    _regval_[92] = (_regs_)->m1hl;                           \
    _regval_[93] = (_regs_)->m1hh;                           \
	_regval_[94] = (_regs_)->m2ll;                           \
    _regval_[95] = (_regs_)->m2lh;                           \
    _regval_[96] = (_regs_)->m2hl;                           \
    _regval_[97] = (_regs_)->m2hh;                           \
	_regval_[98] = (_regs_)->m3ll;                           \
    _regval_[99] = (_regs_)->m3lh;                           \
    _regval_[100] = (_regs_)->m3hl;                           \
    _regval_[101] = (_regs_)->m3hh;                           \
}

// Copy a GDB ordered array into a HAL_SavedRegisters structure.
#define HAL_SET_GDB_REGISTERS( _regs_ , _aregval_ )             \
{                                                               \
    RT_HAL_GDB_REG *_regval_ = (RT_HAL_GDB_REG *)(_aregval_);	\
    int _i_;                                                    \
                                                                \
    for( _i_ = 0; _i_ < 32; _i_++ )                             \
        (_regs_)->d[_i_] = _regval_[_i_];                       \
                                                                \
    HAL_SET_GDB_FPU_REGISTERS( _regs_, _regval_ );              \
                                                                \
    (_regs_)->sr = _regval_[32];                                \
    (_regs_)->lo = _regval_[33];                                \
    (_regs_)->hi = _regval_[34];                                \
    (_regs_)->badvr = _regval_[35];                             \
    (_regs_)->cause = _regval_[36];                             \
    (_regs_)->pc = _regval_[37];                                \
    (_regs_)->estatus=_regval_[73];                           \
    (_regs_)->ecause = _regval_[74];                           \
    (_regs_)->intvec = _regval_[75] ;                           \
    (_regs_)->cbs0 = _regval_[76] ;                           \
    (_regs_)->cbs1 =_regval_[77] ;                           \
    (_regs_)->cbs2 = _regval_[78];                           \
    (_regs_)->cbe0 =_regval_[79];                           \
    (_regs_)->cbe1 =_regval_[80];                           \
    (_regs_)->cbe2 =_regval_[81];                           \
    (_regs_)->lps0 =_regval_[82];                           \
    (_regs_)->lpe0 =_regval_[83];                           \
    (_regs_)->lpc0 =_regval_[84];                           \
    (_regs_)->mmd = _regval_[85] ;                           \
	(_regs_)->m0ll=_regval_[86];                           \
    (_regs_)->m0lh=_regval_[87];                           \
    (_regs_)->m0hl=_regval_[88] ;                           \
    (_regs_)->m0hh=_regval_[89] ;                           \
	(_regs_)->m1ll=_regval_[90] ;                           \
    (_regs_)->m1lh=_regval_[91];                           \
    (_regs_)->m1hl=_regval_[92];                           \
    (_regs_)->m1hh=_regval_[93] ;                           \
	(_regs_)->m2ll=_regval_[94] ;                           \
    (_regs_)->m2lh=_regval_[95] ;                           \
    (_regs_)->m2hl=_regval_[96] ;                           \
    (_regs_)->m2hh=_regval_[97] ;                           \
	(_regs_)->m3ll=_regval_[98] ;                           \
    (_regs_)->m3lh=_regval_[99] ;                           \
    (_regs_)->m3hl=_regval_[100]  ;                           \
    (_regs_)->m3hh=_regval_[101]  ;                           \
}

#else

// Copy a set of registers from a HAL_SavedRegisters structure into a
// GDB ordered array.    
#define HAL_GET_GDB_REGISTERS( _aregval_ , _regs_ )             \
{                                                               \
    RT_HAL_GDB_REG *_regval_ = (RT_HAL_GDB_REG *)(_aregval_); \
    int _i_;                                                    \
                                                                \
    for( _i_ = 0; _i_ < 32; _i_++ )                             \
        _regval_[_i_] = (_regs_)->d[_i_];                       \
                                                                \
    HAL_GET_GDB_FPU_REGISTERS( _regval_, _regs_ );              \
                                                                \
    _regval_[32] = (_regs_)->sr;                                \
    _regval_[33] = (_regs_)->lo;                                \
    _regval_[34] = (_regs_)->hi;                                \
    _regval_[35] = (_regs_)->badvr;                             \
    _regval_[36] = (_regs_)->cause;                             \
    _regval_[37] = (_regs_)->pc;                                \
 
}

// Copy a GDB ordered array into a HAL_SavedRegisters structure.
#define HAL_SET_GDB_REGISTERS( _regs_ , _aregval_ )             \
{                                                               \
    RT_HAL_GDB_REG *_regval_ = (RT_HAL_GDB_REG *)(_aregval_);	\
    int _i_;                                                    \
                                                                \
    for( _i_ = 0; _i_ < 32; _i_++ )                             \
        (_regs_)->d[_i_] = _regval_[_i_];                       \
                                                                \
    HAL_SET_GDB_FPU_REGISTERS( _regs_, _regval_ );              \
                                                                \
    (_regs_)->sr = _regval_[32];                                \
    (_regs_)->lo = _regval_[33];                                \
    (_regs_)->hi = _regval_[34];                                \
    (_regs_)->badvr = _regval_[35];                             \
    (_regs_)->cause = _regval_[36];                             \
    (_regs_)->pc = _regval_[37];                                \
}
#endif
//--------------------------------------------------------------------------
// HAL setjmp
// Note: These definitions are repeated in hal_arch.h. If changes are
// required remember to update both sets.

#define RTARC_JMP_BUF_SP        0
#define RTARC_JMP_BUF_R16       1
#define RTARC_JMP_BUF_R17       2
#define RTARC_JMP_BUF_R18       3
#define RTARC_JMP_BUF_R19       4
#define RTARC_JMP_BUF_R20       5
#define RTARC_JMP_BUF_R21       6
#define RTARC_JMP_BUF_R22       7
#define RTARC_JMP_BUF_R23       8
#define RTARC_JMP_BUF_R28       9
#define RTARC_JMP_BUF_R30      10
#define RTARC_JMP_BUF_R31      11

#define RTARC_JMP_BUF_SIZE     12

typedef rt_uint32 hal_jmp_buf[RTARC_JMP_BUF_SIZE];

externC int hal_setjmp(hal_jmp_buf env);
externC void hal_longjmp(hal_jmp_buf env, int val);

//-------------------------------------------------------------------------
// Idle thread code.
// This macro is called in the idle thread loop, and gives the HAL the
// chance to insert code. Typical idle thread behaviour might be to halt the
// processor.

externC void hal_idle_thread_action(rt_uint32 loop_count);

#define HAL_IDLE_THREAD_ACTION(_count_) hal_idle_thread_action(_count_)

//--------------------------------------------------------------------------
// Minimal and sensible stack sizes: the intention is that applications
// will use these to provide a stack size in the first instance prior to
// proper analysis.  Idle thread stack should be this big.

//    THESE ARE NOT INTENDED TO BE MICROMETRICALLY ACCURATE FIGURES.
//           THEY ARE HOWEVER ENOUGH TO START PROGRAMMING.
// YOU MUST MAKE YOUR STACKS LARGER IF YOU HAVE LARGE "AUTO" VARIABLES!

// This is not a config option because it should not be adjusted except
// under "enough rope" sort of disclaimers.

// Typical case stack frame size: return link + 4 pushed registers + some locals.
#define RTNUM_HAL_STACK_FRAME_SIZE (48)

// Stack needed for a context switch:
#if defined(RTHWR_HAL_MIPS_FPU)
# if defined(RTHWR_HAL_MIPS_FPU_64BIT)
#define RTNUM_HAL_STACK_CONTEXT_SIZE (((32+12)*RT_HAL_MIPS_REG_SIZE)+(32*8))
# elif defined(RTHWR_HAL_MIPS_FPU_32BIT)
#define RTNUM_HAL_STACK_CONTEXT_SIZE (((32+12)*RT_HAL_MIPS_REG_SIZE)+(32*4))
# else
# error MIPS FPU register size not defined
# endif
#else
#define RTNUM_HAL_STACK_CONTEXT_SIZE ((32+10)*RT_HAL_MIPS_REG_SIZE)
#endif



// Interrupt + call to ISR, interrupt_end() and the DSR
#define RTNUM_HAL_STACK_INTERRUPT_SIZE (4+2*RTNUM_HAL_STACK_CONTEXT_SIZE) 

#ifdef RTIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK

// An interrupt stack which is large enough for all possible interrupt
// conditions (and only used for that purpose) exists.  "User" stacks
// can be much smaller

#define RTNUM_HAL_STACK_SIZE_MINIMUM (RTNUM_HAL_STACK_CONTEXT_SIZE+      \
                                       RTNUM_HAL_STACK_INTERRUPT_SIZE*2+  \
                                       RTNUM_HAL_STACK_FRAME_SIZE*8)
#define RTNUM_HAL_STACK_SIZE_TYPICAL (RTNUM_HAL_STACK_SIZE_MINIMUM+1024)

#else // RTIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK 

// No separate interrupt stack exists.  Make sure all threads contain
// a stack sufficiently large.

#define RTNUM_HAL_STACK_SIZE_MINIMUM (4096)
#define RTNUM_HAL_STACK_SIZE_TYPICAL (4096)

#endif

#endif /* __ASSEMBLER__ */

// Convenience macros for accessing memory cached or uncached
#define RTARC_KSEG_MASK                            (0xE0000000)
#define RTARC_KSEG_CACHED							(0x80000000)
#define RTARC_KSEG_UNCACHED                        (0xA0000000)
#define RTARC_KSEG_CACHED_BASE                     (0x80000000)
#define RTARC_KSEG_UNCACHED_BASE                   (0xA0000000)

#define ROS_KSEG_CACHED_BASE						(0x80000000)
#define ROS_KSEG_UNCACHED_BASE		                (0xA0000000)

#define RTARC_CACHED_ADDRESS(x)                    (((x) & ~RTARC_KSEG_MASK) | RTARC_KSEG_CACHED)
#define RTARC_UNCACHED_ADDRESS(x)                  (((x) & ~RTARC_KSEG_MASK) | RTARC_KSEG_UNCACHED)
#define RTARC_PHYSICAL_ADDRESS(x)                  ((x) & ~RTARC_KSEG_MASK)
#ifdef __ASSEMBLER__
#define RTARC_ADDRESS_REG_CACHED(reg)              \
        and     reg, reg, ~RTARC_KSEG_MASK;        \
        or      reg, reg, RTARC_KSEG_CACHED
#define RTARC_ADDRESS_REG_UNCACHED(reg)            \
        and     reg, reg, ~RTARC_KSEG_MASK;        \
        or      reg, reg, RTARC_KSEG_UNCACHED
#endif /* __ASSEMBLER__ */



#if (CROSS_COMPILER == COMPILER_GREEN_HILLS)

asm void RTARC_HAL_SAVE_GP(__gp_save)
{
%reg __gp_save
; # __gp_save is declared as a "register" variable
	move	__gp_save,$28
	.extern _gp
	la		$gp,_gp
%error

}

asm void RTARC_HAL_RESTORE_GP(__gp_save)
{
%reg __gp_save
; # __gp_save is declared as a "register" variable
	move	$gp,__gp_save
%error
}
#endif


#if (CROSS_COMPILER == COMPILER_GNU_GCC)

//--------------------------------------------------------------------------
// Macros for switching context between two ROS instances (jump from
// code in ROM to code in RAM or vice versa).
#define RTARC_HAL_SAVE_GP(__gp_save)		   \
    RT_MACRO_START                             \
    asm volatile ( "move   %0,$28;"            \
                   ".extern _gp;"              \
                   "la     $gp,_gp;"           \
                    : "=r"(__gp_save))

#define RTARC_HAL_RESTORE_GP(__gp_save)                                 \
    asm volatile ( "move   $gp,%0;" :: "r"(__gp_save) );        \
    RT_MACRO_END

#endif // (CROSS_COMPILER == COMPILER_GNU_GCC)

//--------------------------------------------------------------------------
// Macro for finding return address. 
#define RTARC_HAL_GET_RETURN_ADDRESS(_x_) \
  asm volatile ( "move %0,$31;" : "=r" (_x_) )

#define RTARC_HAL_GET_RETURN_ADDRESS_BACKUP()

//--------------------------------------------------------------------------
#endif // RTONCE_HAL_HAL_ARCH_H
// End of hal_arch.h
