#ifndef RTONCE_INFRA_RT_TRAC_H
#define RTONCE_INFRA_RT_TRAC_H

//==========================================================================
//
//      rt_trac.h
//
//      Macros and prototypes for the tracing system
//
//==========================================================================
//####COPYRIGHTBEGIN####
//                                                                          
// -------------------------------------------                              
// The contents of this file are subject to the Red Hat eCos Public License 
// Version 1.1 (the "License"); you may not use this file except in         
// compliance with the License.  You may obtain a copy of the License at    
// http://www.redhat.com/                                                   
//                                                                          
// Software distributed under the License is distributed on an "AS IS"      
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the 
// License for the specific language governing rights and limitations under 
// the License.                                                             
//                                                                          
// The Original Code is eCos - Embedded Configurable Operating System,      
// released September 30, 1998.                                             
//                                                                          
// The Initial Developer of the Original Code is Red Hat.                   
// Portions created by Red Hat are                                          
// Copyright (C) 1998, 1999, 2000 Red Hat, Inc.                             
// All Rights Reserved.                                                     
// -------------------------------------------                              
//                                                                          
//####COPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):   nickg from an original by hmt
// Contributors:        nickg
// Date:        1998-04-23
// Purpose:     Use traces to log procedure entry, and "print" stuff
// Description: Runtime logging messages that compile to nothing in
//              release versions of the code, to allow
//              as-you-go tracing of alternate builds.
// Usage:       #include <rt/infra/rt_trac.h>
//              ...
//              RT_TRACE2( PIPE_TRACE, "pipe %x, data %d", ppipe, oword );
//
//      which can result, for example, in a message of the form:
//      "TRACE: pipemgr.cxx:1340, write_pipe(): pipe 0x8004c, data 17"
//
//####DESCRIPTIONEND####
//

/****************************************************************************

Explicit tracing
================

RT_TRACE0( bool, msg );
RT_TRACE1( bool, msg, arg1 );
RT_TRACE2( bool, msg, arg1, arg2 );
....
RT_TRACE8( bool, msg, .... [with 8 args] );

In general, the bool controls whether or not the tracing occurs for a
particular invocation of the macro.  The msg is a printf-style string,
though exactly which formats are supported depends on the underlying
implementation.  Typically, at least %d, %x, %08x, %c and %s will be
supported.  Of course a most compact implementation might print

  TRACE:z1dbuff.c[92]get_nextdata(): data pointer %x offset %d: 42BD8 1C

or some such, leaving you to work it out for yourself.

It is expected that the boolean would rarely actually be a complex
expression; it is more likely that it would either be "1", tracing being
controlled for the whole compilation unit or subsystem by means of the
RTDBG_USE_TRACING symbol, or a local symbol for tracing over the whole
file, defined to 0 or to 1.  For runtime control of tracing in a debugging
session, it is typical to use symbols defined to expressions such as:

    static int xxx_trace = 0;
    #define TL1 (0 < xxx_trace)
    #define TL2 (1 < xxx_trace)

so you set xxx_trace to 1 to enable those messages conditioned by TL1
(trace level 1) and so on.

    RT_TRACE1( TL1, "Major argument is %d", zz );
    RT_TRACE4( TL2, "...minor details %d %d %d %d", m1, m2, m3 ,m4 );

To assist with the case where the same symbol or expression is used
throughout a compilation unit, the programmer can define the symbol
RT_TRACE_USER_BOOL as they choose and then use convenience macros with the
suffix 'B' in the obvious manner:

    #define RT_TRACE_USER_BOOL (xxx_trace > 0)
    RT_TRACE2B( "Counters are %d, %d", countlo, counthi );

For the case where you just want to print a load of numbers in hex, or
decimal, convenience suffices X, D and Y are provided.  X uses %08x, D %d
and Y an unadorned %x for each argument.

    RT_TRACE3D( TL2, m1, m2, d );
   
If you want to do something similar but with a little more comment, the
names (strictly spellings) of the variables you are printing can be used by
appending a V to the X, D or Y.

    RT_TRACE3DV( TL2, m1, m2, d );

might output:

  TRACE:z1dbuff.c[92]get_nextdata(): m1=23 m2=-4 d=55

These conveniences can be combined, and they apply equally to tracing with
up to 8 variables; the B for Bool goes last:

     RT_TRACE4DVB( i, i*i, i*i*i, i*i*i*i );

might output:

  TRACE:table.c[12]main(): i=3 i*i=9 i*i*i=27 i*i*i*i=81


Function Tracing
================

There are also facities for easily reporting function entry and exit,
printing the function arguments, and detecting returns without logging (or
without a value!).

The basic facility is

        RT_REPORT_FUNCTION();

In C, place this between the local variable declarations and the first
statement or errors will ensue.  C++ is more flexible; place the macro as
the first line of all functions you wish to trace.  The following
variations are also provided:

  RT_REPORT_FUNCTYPE( exitmsg )  provide a printf string for the type
                                  of the returned value
  RT_REPORT_FUNCNAME( name )     supply a function name explicitly, for
                                  if __FUNCTION__ is not supported
  RT_REPORT_FUNCNAMETYPE( name, exitmsg ) both of the above extensions

These are unconditional; the assumption is that if function reporting is
used at all it will be used for all functions within a compilation unit.
However, it is useful to be able to control function reporting at finer
grain without editing the source files concerned, at compile time or at
runtime.  To support this, conditioned versions (with suffix 'C') are
provided for the above four macros, which only procduce trace output if the
macro RT_REPORT_USER_BOOL evaluates true.

  RT_REPORT_FUNCTIONC()
  RT_REPORT_FUNCNAMEC( name )
  RT_REPORT_FUNCTYPEC( exitmsg )
  RT_REPORT_FUNCNAMETYPEC( name, exitmsg )

You can define RT_REPORT_USER_BOOL to anything you like before invoking
these macros; using a simple -DRT_REPORT_USER_BOOL=0 or ...=1 on the
compiler command line would do the trick, but there is more flexibility to
be gained by something like:

  #define RT_REPORT_USER_BOOL (reporting_bool_FOO)
  #ifdef TRACE_FOO
  int reporting_bool_FOO = 1;
  #else
  int reporting_bool_FOO = 0;
  #endif

where FOO relates to the module name.  Thus an external symbol sets the
default, but it can be overridden in a debugging session by setting the
variable reporting_bool_FOO.

Note that the condition applied to the initial RT_REPORT_FUNC...() macro
controls all function-related reporting (not tracing) from that function;
the underlying mechanisms still operate even if no output is created.  Thus
no conditioned variants of RT_REPORT_FUNCARG[s] nor of RT_REPORT_RETURN
are needed.

Examples:
    int myfunction()
    {
        RT_REPORT_FUNCTYPE( "recode is %d" );

A function return is traced using

    RT_REPORT_RETURN()         a void return
    RT_REPORT_RETVAL( value )  returning a value

With the RT_REPORT_FUNCTYPE example, the latter might produce a message
like:

  TRACE:myprog.c[40]fact(): enter
  TRACE:myprog.c[53]fact(): retcode is 24

It is also useful to trace the values of the arguments to a function:
        RT_REPORT_FUNCARGVOID          confirms that the function is void
        RT_REPORT_FUNCARG1( format, arg )              printf-style
                to
        RT_REPORT_FUNCARG8( format, arg1...arg8 )      printf-style

The RT_REPORT_FUNCARG[1-8] macros are also offered with the convenience
extensions: D, X, or Y, and V like the explicit tracing macros.  For
example:

    int fact( int number )
    {
        RT_REPORT_FUNCTYPE( "recode is %d" );
        RT_REPORT_FUNCARG1DV( number );
        int result = number;
        while ( --number > 1 )  result *= number
        RT_REPORT_RETVAL( result );
        return result;
    }

might produce:

  TRACE:myprog.c[40]fact(): enter
  TRACE:myprog.c[40]fact(): number=4
  TRACE:myprog.c[53]fact(): retcode is 24

If no exit message is provided, a default of %08x is used.


General Configury
=================

If RTDBG_INFRA_DEBUG_FUNCTION_PSEUDOMACRO is *not* defined, it is assumed
that __PRETTY_FUNCTION__ or equivalents do not exist, so no function name
tracing is possible; only file and line number.

If RTDBG_INFRA_DEBUG_TRACE_MESSAGE is *not* defined, the message and
arguments to all tracing macros are not used; only "execution was here"
type information, by file, function and line number, is available.  This
can greatly reduce the size of an image with tracing disabled, which may be
crucial in debugging on actual shipped hardware with limited memory.

If configured for buffered tracing then RT_TRACE_PRINT() can be used to
output the contents of the trace buffer on demand.

RT_TRACE_DUMP() outputs a form of "core dump" containing info on the
scheduler and threads at the time. This information will be invalid if
the kernel is not running.

C/C++: in C++ the function reporting is implemented using a class object
with a destructor; this allows reporting of a return which has not been
explicitly reported, and detection of accidental multiple return reports.
This helps you write the function reporting correctly.  In C it is not
possible to be so sophisticated, so the implementation is not so helpful in
detecting errors in the use of the tracing system.

Note that for all of the above variations, the internal API to the
functions which are called in consequence of tracing remains the same, so
these variations can be mixed in the same executable, by configuring the
tracing macros differently in different compilation units or subsystems.


Summary
=======

Explicit tracing
----------------

RT_TRACE0( bool, msg )                         if bool, print msg
RT_TRACE1( bool, msg, arg )                    if bool, printf-style
        to
RT_TRACE8( bool, msg, arg1...arg8 )            if bool, printf-style

RT_TRACE0B( msg, args... ) to RT_TRACE8B()    use RT_TRACE_USER_BOOL

RT_TRACE1X( bool, args... ) to RT_TRACE8X()   print args using %08x
RT_TRACE1Y( bool, args... ) to RT_TRACE8Y()   print args using %x
RT_TRACE1D( bool, args... ) to RT_TRACE8D()   print args using %d

RT_TRACE1XV( bool, args... ) to RT_TRACE8XV() print args using "arg=%08x"
RT_TRACE1YV( bool, args... ) to RT_TRACE8YV() print args using "arg=%x"
RT_TRACE1DV( bool, args... ) to RT_TRACE8DV() print args using "arg=%d"

RT_TRACE1XB( args... ) to RT_TRACE8XB()       print using %08x, no bool
RT_TRACE1YB( args... ) to RT_TRACE8YB()       print using %x, no bool
RT_TRACE1DB( args... ) to RT_TRACE8DB()       print using %d, no bool

RT_TRACE1XVB( args... ) to RT_TRACE8XVB()     use "arg=%08x", no bool
RT_TRACE1YVB( args... ) to RT_TRACE8YVB()     use "arg=%x", no bool
RT_TRACE1DVB( args... ) to RT_TRACE8DVB()     use "arg=%d", no bool

Function tracing
----------------

RT_REPORT_FUNCTION()                           default function entry
RT_REPORT_FUNCNAME( name )                     name the function
RT_REPORT_FUNCTYPE( exitmsg )                  printf for retval
RT_REPORT_FUNCNAMETYPE( name, exitmsg )        both

RT_REPORT_FUNCTIONC()                          as above, but conditional
RT_REPORT_FUNCNAMEC( name )                    on RT_REPORT_USER_BOOL
RT_REPORT_FUNCTYPEC( exitmsg )                 however it is defined
RT_REPORT_FUNCNAMETYPEC( name, exitmsg )       ...

RT_REPORT_RETURN()                             void function exit
RT_REPORT_RETVAL( value )                      returning value

RT_REPORT_FUNCARGVOID()                        void function entry
RT_REPORT_FUNCARG1( format, arg )              printf-style
        to
RT_REPORT_FUNCARG8( format, arg1...arg8 )      printf-style

RT_REPORT_FUNCARG1X( arg )
        to
RT_REPORT_FUNCARG8X( arg1...arg8 )             use %08x
RT_REPORT_FUNCARG1Y...                         use %x
RT_REPORT_FUNCARG1D...                         use %d

RT_REPORT_FUNCARG1XV...                        use "arg=%08x"
RT_REPORT_FUNCARG1YV...                        use "arg=%x"
RT_REPORT_FUNCARG1DV...                        use "arg=%d"

Other
-----

RT_TRACE_DUMP()                                dumps kernel state
RT_TRACE_PRINT()                               prints buffered tracing


---------------------------------------------------------------------------

Internal Documentation
======================

The required functions which are used by the tracing macros are

    externC void
    rt_tracenomsg( const char *psz_func, const char *psz_file, 
                    rt_uint32 linenum );
    
    externC void
    rt_tracemsg( rt_uint32 what, const char *psz_func, const char *psz_file,
                  rt_uint32 linenum, char *psz_msg );
    
    externC void
    rt_tracemsg2( rt_uint32 what, 
                   const char *psz_func, const char *psz_file,
                   rt_uint32 linenum, char *psz_msg,
                   RT_ADDRWORD arg0,  RT_ADDRWORD arg1 );
    // extended in the obvious way for 4,6,8 arguments

These functions should expect psz_func and psz_file to possibly be NULL in
case those facilities are not available in the compilation environment, and
do something safe in such cases.  A NULL message should really be dealt
with safely also, just logging "execution here" info like rt_tracenomsg().

Discussion of possible underlying implementations
-------------------------------------------------

It is intended that the functions that get called can simply print the info
they are given in as fancy a format as you like, or they could do the
printf-type formatting and log the resulting text in a buffer.  They get
told the type of event (function-entry, function-arguments, function-exit
or plain tracing info) and so can perform fancy indenting, for example, to
make call stack inspection more obvious to humans.  It is also intended
that a more compact logging arrangement be possible, for example one which
records, in 32-bit words (RT_ADDRWORDs), the addresses of the file,
function and msg strings, the line number and the arguments.  This has the
implication that the msg string should not be constructed dynamically but
be static ie. a plain quoted C string.  The number of arguments also must
be recorded, and if it is chosen to save string arguments in the buffer
rather than just their addresses (which could be invalid by the time the
logged information is processed) some flagging of which arguments are
strings must be provided.  The system could also be extended to deal with
floats of whichever size fir in a RT_ADDRWORD; these would probably
require special treatment also.  With these considerations in mind, the
maximum number of parameters in a single trace message has been set to 8,
so that a byte bitset could be used to indicate which arguments are
strings, another for those which are floats, and the count of arguments
also fits in a byte as number or a bitset.


****************************************************************************/

/*
#include <include/pkgconf/infra.h>
#include <include/infra/rt_ass.h>
*/

#include "../pkgconf/infra.h"
#include "rt_ass.h"

// -------------------------------------------------------------------------
// RTDBG_INFRA_DEBUG_FUNCTION_PSEUDOMACRO is dealt with in rt_ass.h.
// -------------------------------------------------------------------------

#ifdef RTDBG_USE_TRACING

// -------------------------------------------------------------------------
// We define macros and appropriate prototypes for the trace/fail
// system.  These are:
//      RT_TRACE0..8     - trace if boolean
//      RT_TRACEPROC     - default no comment proc entry
//      RT_TRACEPROCOUT  - default no comment proc exit
//      RT_TRACE_DUMP    - outputs a form of "core dump", including the state
//                          of the kernel scheduler, threads, etc.
//      RT_TRACE_PRINT   - Forces manual output of any trace info that has
//                          been buffered up.

// these are executed to deal with tracing - breakpoint?

externC void
rt_tracenomsg( const char *psz_func, const char *psz_file, rt_uint32 linenum );

externC void
rt_trace_dump(void);

#define RT_TRACE_DUMP() rt_trace_dump()

#ifdef RTDBG_INFRA_DEBUG_TRACE_ASSERT_BUFFER

externC void
rt_trace_print(void);

#define RT_TRACE_PRINT() rt_trace_print()

#else
#define RT_TRACE_PRINT() RT_EMPTY_STATEMENT
#endif

// provide every other one of these as a space/caller bloat compromise.

# ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE

enum rt_trace_what{
    rt_trace_trace = 0,
    rt_trace_enter,
    rt_trace_args,
    rt_trace_return,
//    rt_trace_,
//    rt_trace_,
};

externC void
rt_tracemsg( rt_uint32 what, 
              const char *psz_func, const char *psz_file, rt_uint32 linenum, 
              char *psz_msg );

externC void
rt_tracemsg2( rt_uint32 what, 
               const char *psz_func, const char *psz_file, rt_uint32 linenum, 
               char *psz_msg,
               RT_ADDRWORD arg0,  RT_ADDRWORD arg1 );
externC void
rt_tracemsg4( rt_uint32 what, 
               const char *psz_func, const char *psz_file, rt_uint32 linenum, 
               char *psz_msg,
               RT_ADDRWORD arg0,  RT_ADDRWORD arg1,
               RT_ADDRWORD arg2,  RT_ADDRWORD arg3 );
externC void
rt_tracemsg6( rt_uint32 what, 
               const char *psz_func, const char *psz_file, rt_uint32 linenum, 
               char *psz_msg,
               RT_ADDRWORD arg0,  RT_ADDRWORD arg1,
               RT_ADDRWORD arg2,  RT_ADDRWORD arg3,
               RT_ADDRWORD arg4,  RT_ADDRWORD arg5 );
externC void
rt_tracemsg8( rt_uint32 what, 
               const char *psz_func, const char *psz_file, rt_uint32 linenum, 
               char *psz_msg,
               RT_ADDRWORD arg0,  RT_ADDRWORD arg1,
               RT_ADDRWORD arg2,  RT_ADDRWORD arg3,
               RT_ADDRWORD arg4,  RT_ADDRWORD arg5,
               RT_ADDRWORD arg6,  RT_ADDRWORD arg7 );

#endif // RTDBG_INFRA_DEBUG_TRACE_MESSAGE

// -------------------------------------------------------------------------

# ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE

#  define RT_TRACE_docall0( _msg_ )                                    \
    rt_tracemsg( rt_trace_trace,                                      \
                  __PRETTY_FUNCTION__, __FILE__, __LINE__, _msg_ );

#  define RT_TRACE_docall2( _msg_, _arg0_, _arg1_ )                    \
    rt_tracemsg2( rt_trace_trace,                                     \
                   __PRETTY_FUNCTION__, __FILE__, __LINE__, _msg_,      \
                 (RT_ADDRWORD)_arg0_, (RT_ADDRWORD)_arg1_ );

#  define RT_TRACE_docall4( _msg_, _arg0_, _arg1_ , _arg2_, _arg3_  )  \
    rt_tracemsg4( rt_trace_trace,                                     \
                   __PRETTY_FUNCTION__, __FILE__, __LINE__, _msg_,      \
                 (RT_ADDRWORD)_arg0_, (RT_ADDRWORD)_arg1_,            \
                 (RT_ADDRWORD)_arg2_, (RT_ADDRWORD)_arg3_ );

#  define RT_TRACE_docall6( _msg_, _arg0_, _arg1_ , _arg2_, _arg3_,    \
                                    _arg4_, _arg5_                   )  \
    rt_tracemsg6( rt_trace_trace,                                     \
                   __PRETTY_FUNCTION__, __FILE__, __LINE__, _msg_,      \
                 (RT_ADDRWORD)_arg0_, (RT_ADDRWORD)_arg1_,            \
                 (RT_ADDRWORD)_arg2_, (RT_ADDRWORD)_arg3_,            \
                 (RT_ADDRWORD)_arg4_, (RT_ADDRWORD)_arg5_ );

#  define RT_TRACE_docall8( _msg_, _arg0_, _arg1_ , _arg2_, _arg3_,    \
                                    _arg4_,  _arg5_, _arg6_, _arg7_ )   \
    rt_tracemsg8( rt_trace_trace,                                     \
                   __PRETTY_FUNCTION__, __FILE__, __LINE__, _msg_,      \
                 (RT_ADDRWORD)_arg0_, (RT_ADDRWORD)_arg1_,            \
                 (RT_ADDRWORD)_arg2_, (RT_ADDRWORD)_arg3_,            \
                 (RT_ADDRWORD)_arg4_, (RT_ADDRWORD)_arg5_,            \
                 (RT_ADDRWORD)_arg6_, (RT_ADDRWORD)_arg7_ );

# else // do not RTDBG_INFRA_DEBUG_TRACE_MESSAGE

#  define RT_TRACE_docall0( _msg_ )                                    \
    rt_tracenomsg( __PRETTY_FUNCTION__, __FILE__, __LINE__ );

#  define RT_TRACE_docall2( _msg_, _arg0_, _arg1_ )                    \
    rt_tracenomsg( __PRETTY_FUNCTION__, __FILE__, __LINE__ );

#  define RT_TRACE_docall4( _msg_, _arg0_, _arg1_ , _arg2_, _arg3_  )  \
    rt_tracenomsg( __PRETTY_FUNCTION__, __FILE__, __LINE__ );

#  define RT_TRACE_docall6( _msg_, _arg0_, _arg1_ , _arg2_, _arg3_,    \
                                    _arg4_, _arg5_                   )  \
    rt_tracenomsg( __PRETTY_FUNCTION__, __FILE__, __LINE__ );

#  define RT_TRACE_docall8( _msg_, _arg0_, _arg1_, _arg2_, _arg3_,     \
                                    _arg4_, _arg5_, _arg6_, _arg7_   )  \
    rt_tracenomsg( __PRETTY_FUNCTION__, __FILE__, __LINE__ );

#endif

// -------------------------------------------------------------------------
// Conditioned trace; if the condition is false, fail.

#define RT_TRACE0( _bool_, _msg_ )                             \
    RT_MACRO_START                                             \
    if ( ( _bool_ ) )                                           \
        RT_TRACE_docall0( _msg_ );                             \
    RT_MACRO_END

#define RT_TRACE1( _bool_, _msg_, a )                          \
    RT_MACRO_START                                             \
    if ( ( _bool_ ) )                                           \
        RT_TRACE_docall2( _msg_, a, 0 );                       \
    RT_MACRO_END
 
#define RT_TRACE2( _bool_, _msg_, a, b )                       \
    RT_MACRO_START                                             \
    if ( ( _bool_ ) )                                           \
        RT_TRACE_docall2( _msg_, a, b );                       \
    RT_MACRO_END

#define RT_TRACE3( _bool_, _msg_, a, b, c )                    \
    RT_MACRO_START                                             \
    if ( ( _bool_ ) )                                           \
        RT_TRACE_docall4( _msg_, a, b, c, 0 );                 \
    RT_MACRO_END
 
#define RT_TRACE4( _bool_, _msg_, a, b, c, d )                 \
    RT_MACRO_START                                             \
    if ( ( _bool_ ) )                                           \
        RT_TRACE_docall4( _msg_, a, b, c, d );                 \
    RT_MACRO_END

#define RT_TRACE5( _bool_, _msg_, a, b, c, d, e )              \
    RT_MACRO_START                                             \
    if ( ( _bool_ ) )                                           \
        RT_TRACE_docall6( _msg_, a, b, c, d, e, 0 );           \
    RT_MACRO_END
 
#define RT_TRACE6( _bool_, _msg_, a, b, c, d, e, f )           \
    RT_MACRO_START                                             \
    if ( ( _bool_ ) )                                           \
        RT_TRACE_docall6( _msg_, a, b, c, d, e, f );           \
    RT_MACRO_END

#define RT_TRACE7( _bool_, _msg_, a, b, c, d, e, f, g )        \
    RT_MACRO_START                                             \
    if ( ( _bool_ ) )                                           \
        RT_TRACE_docall8( _msg_, a, b, c, d, e, f, g, 0 );     \
    RT_MACRO_END
 
#define RT_TRACE8( _bool_, _msg_, a, b, c, d, e, f, g, h )     \
    RT_MACRO_START                                             \
    if ( ( _bool_ ) )                                           \
        RT_TRACE_docall8( _msg_, a, b, c, d, e, f, g, h );     \
    RT_MACRO_END

// -------------------------------------------------------------------------
// Report function entry and exit.
// In C++ the macro RT_REPORT_FUNCTION should appear as the first line of
// any function. It will generate a message whenever the function is entered
// and when it is exited.
// In C the macro should appear as the first statement after any local variable
// definitions. No exit message will be generated unless RT_REPORT_RETURN is
// placed just before each return.
// Where a piece of code is to be compiled with both C and C++, the above
// rules for C should be followed.

#ifdef RTDBG_INFRA_DEBUG_FUNCTION_REPORTS
        
#ifdef __cplusplus

class Rt_TraceFunction_Report_
{
public:
    int   cond;
    const char *func;
    const char *file;
    rt_uint32 lnum;
#ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE
    char *exitmsg;
    RT_ADDRWORD exitvalue;
    enum { UNSET = 0, SET, VOID } exitset;
#endif

    Rt_TraceFunction_Report_(
        int condition, const char *psz_func, const char *psz_file,
        rt_uint32 linenum)
    {
        cond = condition;
        func = psz_func;
        file = psz_file;
        lnum = linenum;
        
#ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE
        exitmsg = NULL;
        exitset  = UNSET;
        if ( cond )
            rt_tracemsg( rt_trace_enter, func, file, lnum, "enter");
#else
        if ( cond )
            rt_tracenomsg( func, file, lnum );
#endif
    };

    Rt_TraceFunction_Report_(
        int condition, const char *psz_func, const char *psz_file, 
        rt_uint32 linenum, char *psz_exitmsg )
    {
        cond = condition;
        func = psz_func;
        file = psz_file;
        lnum = linenum;
#ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE
        exitmsg = psz_exitmsg;
        exitset  = UNSET;
        if ( cond )
            rt_tracemsg( rt_trace_enter, func, file, lnum, "enter");
#else
        RT_UNUSED_PARAM( char *, psz_exitmsg );
        if ( cond )
            rt_tracenomsg( func, file, lnum );
#endif
    };

    inline void set_exitvoid( rt_uint32 linenum )
    {
        lnum = linenum;
#ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE
        RT_ASSERT( NULL == exitmsg, "exitvoid used in typed function" );
        RT_ASSERT( UNSET == exitset, "exitvoid used when arg already set" );
        exitset = VOID;
#endif
    }

    inline void set_exitvalue( rt_uint32 linenum, RT_ADDRWORD retcode )
    {
        lnum = linenum;
#ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE
        RT_ASSERT( UNSET == exitset, "exitvalue used when arg already set" );
        exitvalue = retcode;
        exitset = SET;
#else
        RT_UNUSED_PARAM( RT_ADDRWORD, retcode );
#endif
    }

    ~Rt_TraceFunction_Report_()
    {
        if ( cond ) {
#ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE
            if ( VOID == exitset )
                rt_tracemsg( rt_trace_return, func, file, lnum,
                              "return void");
            else if ( UNSET == exitset )
                rt_tracemsg( rt_trace_return, func, file, lnum,
                              "RETURNING UNSET!");
            else if ( NULL == exitmsg )
                rt_tracemsg2( rt_trace_return, func, file, lnum,
                               "return %08x", exitvalue, 0 );
            else
                rt_tracemsg2( rt_trace_return, func, file, lnum,
                               exitmsg, exitvalue, 0 );
#else
            rt_tracenomsg( func, file, lnum );
#endif
        }
    }
};

// These have no RT_MACRO_START,END around because it is required
// that the scope of the object be the whole function body.  Get it?

// These are the unconditional versions:
#define RT_REPORT_FUNCTION()                           \
  Rt_TraceFunction_Report_ rt_tracefunction_report_(  \
        1, __PRETTY_FUNCTION__,                         \
        __FILE__, __LINE__ )

#define RT_REPORT_FUNCTYPE( _exitmsg_ )                \
  Rt_TraceFunction_Report_ rt_tracefunction_report_(  \
        1, __PRETTY_FUNCTION__,                         \
        __FILE__, __LINE__, _exitmsg_ )

#define RT_REPORT_FUNCNAME( _name_ )                   \
  Rt_TraceFunction_Report_ rt_tracefunction_report_(  \
        1, _name_,                                      \
        __FILE__, __LINE__ )

#define RT_REPORT_FUNCNAMETYPE( _name_, _exitmsg_  )   \
  Rt_TraceFunction_Report_ rt_tracefunction_report_(  \
        1, _name_,                                      \
        __FILE__, __LINE__, _exitmsg_ )

// These are conditioned on macro RT_REPORT_USER_BOOL
// (which you better have defined)
#define RT_REPORT_FUNCTIONC()                          \
  Rt_TraceFunction_Report_ rt_tracefunction_report_(  \
        RT_REPORT_USER_BOOL, __PRETTY_FUNCTION__,      \
        __FILE__, __LINE__ )

#define RT_REPORT_FUNCTYPEC( _exitmsg_ )               \
  Rt_TraceFunction_Report_ rt_tracefunction_report_(  \
        RT_REPORT_USER_BOOL, __PRETTY_FUNCTION__,      \
        __FILE__, __LINE__, _exitmsg_ )

#define RT_REPORT_FUNCNAMEC( _name_ )                  \
  Rt_TraceFunction_Report_ rt_tracefunction_report_(  \
        RT_REPORT_USER_BOOL, _name_,                   \
        __FILE__, __LINE__ )

#define RT_REPORT_FUNCNAMETYPEC( _name_, _exitmsg_  )  \
  Rt_TraceFunction_Report_ rt_tracefunction_report_(  \
        RT_REPORT_USER_BOOL, _name_,                   \
        __FILE__, __LINE__, _exitmsg_ )


#define RT_REPORT_RETURN() RT_MACRO_START             \
    rt_tracefunction_report_.set_exitvoid( __LINE__ ); \
RT_MACRO_END

#define RT_REPORT_RETVAL( _value_) RT_MACRO_START     \
    rt_tracefunction_report_.set_exitvalue(            \
        __LINE__, (RT_ADDRWORD)(_value_) );            \
RT_MACRO_END


#else   // not __cplusplus


struct Rt_TraceFunction_Report_
{
    int   cond;
    char *func;
    char *file; /* not strictly needed in plain 'C' */
    rt_uint32 lnum; /* nor this */
#ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE
    char *exitmsg;
    RT_ADDRWORD exitvalue;
    int exitset;
#endif

};

#ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE

#define RT_REPORT_FUNCTION_ENTER_INTERNAL() RT_MACRO_START            \
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracemsg( rt_trace_enter,                                      \
                  rt_tracefunction_report_.func,                       \
                  rt_tracefunction_report_.file,                       \
                  rt_tracefunction_report_.lnum,                       \
                  "enter" );                                            \
RT_MACRO_END

#define RT_REPORT_FUNCTION_CONSTRUCT( _c_, _fn_,_fl_,_l_,_xm_,_xv_,_xs_ ) \
        { _c_, _fn_, _fl_, _l_, _xm_, _xv_, _xs_ }

#else // do not RTDBG_INFRA_DEBUG_TRACE_MESSAGE

#define RT_REPORT_FUNCTION_ENTER_INTERNAL() RT_MACRO_START            \
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracenomsg( rt_tracefunction_report_.func,                     \
                    rt_tracefunction_report_.file,                     \
                    rt_tracefunction_report_.lnum );                   \
RT_MACRO_END

#define RT_REPORT_FUNCTION_CONSTRUCT( _c_, _fn_,_fl_,_l_,_xm_,_xv_,_xs_ ) \
        { _c_, _fn_, _fl_, _l_ }

#endif // not RTDBG_INFRA_DEBUG_TRACE_MESSAGE

// These have no RT_MACRO_START,END around because it is required
// that the scope of the object be the whole function body.  Get it?

// These are the unconditional versions:
#define RT_REPORT_FUNCTION()                                           \
    struct Rt_TraceFunction_Report_ rt_tracefunction_report_ =        \
    RT_REPORT_FUNCTION_CONSTRUCT(                                      \
        1, __PRETTY_FUNCTION__, __FILE__, __LINE__, NULL, 0, 0 );       \
    RT_REPORT_FUNCTION_ENTER_INTERNAL()

#define RT_REPORT_FUNCTYPE( _exitmsg_ )                                \
    struct Rt_TraceFunction_Report_ rt_tracefunction_report_ =        \
    RT_REPORT_FUNCTION_CONSTRUCT(                                      \
        1, __PRETTY_FUNCTION__, __FILE__, __LINE__, _exitmsg_, 0, 0 );  \
    RT_REPORT_FUNCTION_ENTER_INTERNAL()

#define RT_REPORT_FUNCNAME( _name_ )                                   \
    struct Rt_TraceFunction_Report_ rt_tracefunction_report_ =        \
    RT_REPORT_FUNCTION_CONSTRUCT(                                      \
        1, _name_, __FILE__, __LINE__, NULL, 0, 0 );                    \
    RT_REPORT_FUNCTION_ENTER_INTERNAL()

#define RT_REPORT_FUNCNAMETYPE( _name_, _exitmsg_  )                   \
    struct Rt_TraceFunction_Report_ rt_tracefunction_report_ =        \
    RT_REPORT_FUNCTION_CONSTRUCT(                                      \
        1, _name_, __FILE__, __LINE__, _exitmsg_, 0, 0 );               \
    RT_REPORT_FUNCTION_ENTER_INTERNAL()

// These are conditioned on macro RT_REPORT_USER_BOOL
// (which you better have defined)
#define RT_REPORT_FUNCTIONC()                                          \
    struct Rt_TraceFunction_Report_ rt_tracefunction_report_ =        \
    RT_REPORT_FUNCTION_CONSTRUCT(                                      \
        RT_REPORT_USER_BOOL,                                           \
        __PRETTY_FUNCTION__, __FILE__, __LINE__, NULL, 0, 0 );          \
    RT_REPORT_FUNCTION_ENTER_INTERNAL()

#define RT_REPORT_FUNCTYPEC( _exitmsg_ )                               \
    struct Rt_TraceFunction_Report_ rt_tracefunction_report_ =        \
    RT_REPORT_FUNCTION_CONSTRUCT(                                      \
        RT_REPORT_USER_BOOL,                                           \
        __PRETTY_FUNCTION__, __FILE__, __LINE__, _exitmsg_, 0, 0 );     \
    RT_REPORT_FUNCTION_ENTER_INTERNAL()

#define RT_REPORT_FUNCNAMEC( _name_ )                                  \
    struct Rt_TraceFunction_Report_ rt_tracefunction_report_ =        \
    RT_REPORT_FUNCTION_CONSTRUCT(                                      \
        RT_REPORT_USER_BOOL,                                           \
        _name_, __FILE__, __LINE__, NULL, 0, 0 );                       \
    RT_REPORT_FUNCTION_ENTER_INTERNAL()

#define RT_REPORT_FUNCNAMETYPEC( _name_, _exitmsg_  )                  \
    struct Rt_TraceFunction_Report_ rt_tracefunction_report_ =        \
    RT_REPORT_FUNCTION_CONSTRUCT(                                      \
        RT_REPORT_USER_BOOL,                                           \
        _name_, __FILE__, __LINE__, _exitmsg_, 0, 0 );                  \
    RT_REPORT_FUNCTION_ENTER_INTERNAL()

#ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE

#define RT_REPORT_RETURN() RT_MACRO_START                             \
    RT_ASSERT( NULL == rt_tracefunction_report_.exitmsg,              \
                "exitvoid used in typed function" );                    \
    RT_ASSERT( 0 == rt_tracefunction_report_.exitset,                 \
                "exitvoid used when arg already set" );                 \
    rt_tracefunction_report_.lnum = __LINE__;                          \
    rt_tracefunction_report_.exitset = 2;                              \
    if ( rt_tracefunction_report_.cond )                               \
      rt_tracemsg( rt_trace_return,                                   \
                    rt_tracefunction_report_.func,                     \
                    rt_tracefunction_report_.file,                     \
                    rt_tracefunction_report_.lnum,                     \
                    "return void" );                                    \
RT_MACRO_END

#define RT_REPORT_RETVAL( _value_ ) RT_MACRO_START                    \
    RT_ASSERT( 0 == rt_tracefunction_report_.exitset,                 \
                "exitvalue used when arg already set" );                \
    rt_tracefunction_report_.lnum = __LINE__;                          \
    rt_tracefunction_report_.exitvalue = (RT_ADDRWORD)(_value_);      \
    rt_tracefunction_report_.exitset = 1;                              \
    if ( rt_tracefunction_report_.cond )                               \
      rt_tracemsg2( rt_trace_return,                                  \
                     rt_tracefunction_report_.func,                    \
                     rt_tracefunction_report_.file,                    \
                     rt_tracefunction_report_.lnum,                    \
                     rt_tracefunction_report_.exitmsg ?                \
                        rt_tracefunction_report_.exitmsg :             \
                        "return %08x",                                  \
                     rt_tracefunction_report_.exitvalue, 0 );          \
RT_MACRO_END

#else // do not RTDBG_INFRA_DEBUG_TRACE_MESSAGE

#define RT_REPORT_RETURN() RT_MACRO_START                             \
    rt_tracefunction_report_.lnum = __LINE__;                          \
    if ( rt_tracefunction_report_.cond )                               \
      rt_tracenomsg( rt_tracefunction_report_.func,                   \
                      rt_tracefunction_report_.file,                   \
                      rt_tracefunction_report_.lnum );                 \
RT_MACRO_END

#define RT_REPORT_RETVAL( _value_ ) RT_MACRO_START                    \
    RT_REPORT_RETURN();                                                \
RT_MACRO_END

#endif // not RTDBG_INFRA_DEBUG_TRACE_MESSAGE

#endif // not __cplusplus

#ifdef RTDBG_INFRA_DEBUG_TRACE_MESSAGE

#define RT_REPORT_FUNCARGVOID() RT_MACRO_START                        \
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracemsg(  rt_trace_args,                                      \
                   rt_tracefunction_report_.func,                      \
                   rt_tracefunction_report_.file,                      \
                   rt_tracefunction_report_.lnum,                      \
                   "(void)"                                             \
                   );                                                   \
RT_MACRO_END

#define RT_REPORT_FUNCARG1( _format_, a ) RT_MACRO_START              \
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracemsg2( rt_trace_args,                                      \
                   rt_tracefunction_report_.func,                      \
                   rt_tracefunction_report_.file,                      \
                   rt_tracefunction_report_.lnum,                      \
                   (_format_),                                          \
                   (RT_ADDRWORD)a      , 0                             \
                   );                                                   \
RT_MACRO_END
    
#define RT_REPORT_FUNCARG2( _format_, a,b ) RT_MACRO_START            \
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracemsg2( rt_trace_args,                                      \
                   rt_tracefunction_report_.func,                      \
                   rt_tracefunction_report_.file,                      \
                   rt_tracefunction_report_.lnum,                      \
                   (_format_),                                          \
                   (RT_ADDRWORD)a, (RT_ADDRWORD)b                     \
                   );                                                   \
RT_MACRO_END

#define RT_REPORT_FUNCARG3( _format_, a,b,c ) RT_MACRO_START          \
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracemsg4( rt_trace_args,                                      \
                   rt_tracefunction_report_.func,                      \
                   rt_tracefunction_report_.file,                      \
                   rt_tracefunction_report_.lnum,                      \
                   (_format_),                                          \
                   (RT_ADDRWORD)a, (RT_ADDRWORD)b,                    \
                   (RT_ADDRWORD)c      , 0                             \
                   );                                                   \
RT_MACRO_END

#define RT_REPORT_FUNCARG4( _format_, a,b,c,d ) RT_MACRO_START        \
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracemsg4( rt_trace_args,                                      \
                   rt_tracefunction_report_.func,                      \
                   rt_tracefunction_report_.file,                      \
                   rt_tracefunction_report_.lnum,                      \
                   (_format_),                                          \
                   (RT_ADDRWORD)a, (RT_ADDRWORD)b,                    \
                   (RT_ADDRWORD)c, (RT_ADDRWORD)d                     \
                   );                                                   \
RT_MACRO_END

#define RT_REPORT_FUNCARG5( _format_, a,b,c,d,e ) RT_MACRO_START      \
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracemsg6( rt_trace_args,                                      \
                   rt_tracefunction_report_.func,                      \
                   rt_tracefunction_report_.file,                      \
                   rt_tracefunction_report_.lnum,                      \
                   (_format_),                                          \
                   (RT_ADDRWORD)a, (RT_ADDRWORD)b,                    \
                   (RT_ADDRWORD)c, (RT_ADDRWORD)d,                    \
                   (RT_ADDRWORD)e      , 0                             \
                   );                                                   \
RT_MACRO_END

#define RT_REPORT_FUNCARG6( _format_, a,b,c,d,e,f ) RT_MACRO_START    \
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracemsg6( rt_trace_args,                                      \
                   rt_tracefunction_report_.func,                      \
                   rt_tracefunction_report_.file,                      \
                   rt_tracefunction_report_.lnum,                      \
                   (_format_),                                          \
                   (RT_ADDRWORD)a, (RT_ADDRWORD)b,                    \
                   (RT_ADDRWORD)c, (RT_ADDRWORD)d,                    \
                   (RT_ADDRWORD)e, (RT_ADDRWORD)f                     \
                   );                                                   \
RT_MACRO_END

#define RT_REPORT_FUNCARG7( _format_, a,b,c,d,e,f,g ) RT_MACRO_START  \
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracemsg8( rt_trace_args,                                      \
                   rt_tracefunction_report_.func,                      \
                   rt_tracefunction_report_.file,                      \
                   rt_tracefunction_report_.lnum,                      \
                   (_format_),                                          \
                   (RT_ADDRWORD)a, (RT_ADDRWORD)b,                    \
                   (RT_ADDRWORD)c, (RT_ADDRWORD)d,                    \
                   (RT_ADDRWORD)e, (RT_ADDRWORD)f,                    \
                   (RT_ADDRWORD)g      , 0                             \
                   );                                                   \
RT_MACRO_END

#define RT_REPORT_FUNCARG8( _format_, a,b,c,d,e,f,g,h ) RT_MACRO_START\
  if ( rt_tracefunction_report_.cond )                                 \
    rt_tracemsg8( rt_trace_args,                                      \
                   rt_tracefunction_report_.func,                      \
                   rt_tracefunction_report_.file,                      \
                   rt_tracefunction_report_.lnum,                      \
                   (_format_),                                          \
                   (RT_ADDRWORD)a, (RT_ADDRWORD)b,                    \
                   (RT_ADDRWORD)c, (RT_ADDRWORD)d,                    \
                   (RT_ADDRWORD)e, (RT_ADDRWORD)f,                    \
                   (RT_ADDRWORD)g, (RT_ADDRWORD)h                     \
                   );                                                   \
RT_MACRO_END


#else // do not RTDBG_INFRA_DEBUG_TRACE_MESSAGE

#define RT_REPORT_FUNCARGVOID() RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG1( _format_, a ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG2( _format_, a,b ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG3( _format_, a,b,c ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG4( _format_, a,b,c,d ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG5( _format_, a,b,c,d,e ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG6( _format_, a,b,c,d,e,f ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG7( _format_, a,b,c,d,e,f,g ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG8( _format_, a,b,c,d,e,f,g,h ) RT_EMPTY_STATEMENT

#endif  // not RTDBG_INFRA_DEBUG_TRACE_MESSAGE

#else   // no RTDBG_INFRA_DEBUG_FUNCTION_REPORTS

#define RT_REPORT_FUNCTION()                           RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCTYPE( _exitmsg_ )                RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCNAME( _name_ )                   RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCNAMETYPE( _name_, _exitmsg_  )   RT_EMPTY_STATEMENT

#define RT_REPORT_FUNCTIONC()                          RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCTYPEC( _exitmsg_ )               RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCNAMEC( _name_ )                  RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCNAMETYPEC( _name_, _exitmsg_  )  RT_EMPTY_STATEMENT

#define RT_REPORT_FUNCARGVOID() RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG1( _format_, a ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG2( _format_, a,b ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG3( _format_, a,b,c ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG4( _format_, a,b,c,d ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG5( _format_, a,b,c,d,e ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG6( _format_, a,b,c,d,e,f ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG7( _format_, a,b,c,d,e,f,g ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG8( _format_, a,b,c,d,e,f,g,h ) RT_EMPTY_STATEMENT

#define RT_REPORT_RETURN()                             RT_EMPTY_STATEMENT
#define RT_REPORT_RETVAL( _value_ )                    RT_EMPTY_STATEMENT
    
#endif  // RTDBG_INFRA_DEBUG_FUNCTION_REPORTS
    
#else   // ! RTDBG_USE_TRACING

// -------------------------------------------------------------------------
// No traces: we define empty statements for trace macros.

#define RT_TRACE0( _bool_, _msg_  ) RT_EMPTY_STATEMENT
#define RT_TRACE1( _bool_, _msg_, a ) RT_EMPTY_STATEMENT
#define RT_TRACE2( _bool_, _msg_, a,b ) RT_EMPTY_STATEMENT
#define RT_TRACE3( _bool_, _msg_, a,b,c ) RT_EMPTY_STATEMENT
#define RT_TRACE4( _bool_, _msg_, a,b,c,d ) RT_EMPTY_STATEMENT
#define RT_TRACE5( _bool_, _msg_, a,b,c,d,e ) RT_EMPTY_STATEMENT
#define RT_TRACE6( _bool_, _msg_, a,b,c,d,e,f ) RT_EMPTY_STATEMENT
#define RT_TRACE7( _bool_, _msg_, a,b,c,d,e,f,g ) RT_EMPTY_STATEMENT
#define RT_TRACE8( _bool_, _msg_, a,b,c,d,e,f,g,h ) RT_EMPTY_STATEMENT

#define RT_REPORT_FUNCTION()                           RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCTYPE( _exitmsg_ )                RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCNAME( _name_ )                   RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCNAMETYPE( _name_, _exitmsg_  )   RT_EMPTY_STATEMENT

#define RT_REPORT_FUNCTIONC()                          RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCTYPEC( _exitmsg_ )               RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCNAMEC( _name_ )                  RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCNAMETYPEC( _name_, _exitmsg_  )  RT_EMPTY_STATEMENT

#define RT_REPORT_FUNCARGVOID() RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG1( _format_, a ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG2( _format_, a,b ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG3( _format_, a,b,c ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG4( _format_, a,b,c,d ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG5( _format_, a,b,c,d,e ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG6( _format_, a,b,c,d,e,f ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG7( _format_, a,b,c,d,e,f,g ) RT_EMPTY_STATEMENT
#define RT_REPORT_FUNCARG8( _format_, a,b,c,d,e,f,g,h ) RT_EMPTY_STATEMENT

#define RT_REPORT_RETURN()                             RT_EMPTY_STATEMENT
#define RT_REPORT_RETVAL( _value_ )                    RT_EMPTY_STATEMENT

#define RT_TRACE_PRINT() RT_EMPTY_STATEMENT
#define RT_TRACE_DUMP()  RT_EMPTY_STATEMENT

#endif // ! RTDBG_USE_TRACING

// -------------------------------------------------------------------------
//
// RT_TRACEn{[XDY]{V}}{B}
//
// Convenience macros: these fall into a few dimensions, with suffix letters:
// First option:
//     X: user need not supply a format string, %08x is used
//     D: ditto but signed decimal, %d
//     Y: ditto but just plain %x
// Second option, only meaningful with one of XDY:
//     V: "<var> = %..." is used, by stringifying the argument
// Third option:
//     B: user need not supply a bool; the symbol RT_TRACE_USER_BOOL is
//        used (which we do not define, user must do this)

#define RT_TRACE0B( _msg_  ) \
        RT_TRACE0( RT_TRACE_USER_BOOL, _msg_  ) 
#define RT_TRACE1B( _msg_, a ) \
        RT_TRACE1( RT_TRACE_USER_BOOL, _msg_, a ) 
#define RT_TRACE2B( _msg_, a,b ) \
        RT_TRACE2( RT_TRACE_USER_BOOL, _msg_, a,b ) 
#define RT_TRACE3B( _msg_, a,b,c ) \
        RT_TRACE3( RT_TRACE_USER_BOOL, _msg_, a,b,c ) 
#define RT_TRACE4B( _msg_, a,b,c,d ) \
        RT_TRACE4( RT_TRACE_USER_BOOL, _msg_, a,b,c,d ) 
#define RT_TRACE5B( _msg_, a,b,c,d,e ) \
        RT_TRACE5( RT_TRACE_USER_BOOL, _msg_, a,b,c,d,e ) 
#define RT_TRACE6B( _msg_, a,b,c,d,e,f ) \
        RT_TRACE6( RT_TRACE_USER_BOOL, _msg_, a,b,c,d,e,f ) 
#define RT_TRACE7B( _msg_, a,b,c,d,e,f,g ) \
        RT_TRACE7( RT_TRACE_USER_BOOL, _msg_, a,b,c,d,e,f,g ) 
#define RT_TRACE8B( _msg_, a,b,c,d,e,f,g,h ) \
        RT_TRACE8( RT_TRACE_USER_BOOL, _msg_, a,b,c,d,e,f,g,h ) 

// long hex versions

#define RT_TRACE1X( _bool_, a ) \
        RT_TRACE1( _bool_, "%08x", a ) 
#define RT_TRACE2X( _bool_, a,b ) \
        RT_TRACE2( _bool_, "%08x %08x", a,b ) 
#define RT_TRACE3X( _bool_, a,b,c ) \
        RT_TRACE3( _bool_, "%08x %08x %08x", a,b,c ) 
#define RT_TRACE4X( _bool_, a,b,c,d ) \
        RT_TRACE4( _bool_, "%08x %08x %08x %08x", a,b,c,d ) 
#define RT_TRACE5X( _bool_, a,b,c,d,e ) \
        RT_TRACE5( _bool_, "%08x %08x %08x %08x %08x", a,b,c,d,e ) 
#define RT_TRACE6X( _bool_, a,b,c,d,e,f ) \
        RT_TRACE6( _bool_, "%08x %08x %08x %08x %08x %08x", \
                    a,b,c,d,e,f ) 
#define RT_TRACE7X( _bool_, a,b,c,d,e,f,g ) \
        RT_TRACE7( _bool_, "%08x %08x %08x %08x %08x %08x %08x", \
                    a,b,c,d,e,f,g ) 
#define RT_TRACE8X( _bool_, a,b,c,d,e,f,g,h ) \
        RT_TRACE8( _bool_, "%08x %08x %08x %08x %08x %08x %08x %08x", \
                    a,b,c,d,e,f,g,h )

#define RT_TRACE1XV( _bool_, a ) \
        RT_TRACE1( _bool_, # a "=%08x ", a ) 
#define RT_TRACE2XV( _bool_, a,b ) \
        RT_TRACE2( _bool_, \
                    # a "=%08x " # b "=%08x " , a,b )
#define RT_TRACE3XV( _bool_, a,b,c ) \
        RT_TRACE3( _bool_, \
                    # a "=%08x " # b "=%08x " # c "=%08x " , a,b,c )
#define RT_TRACE4XV( _bool_, a,b,c,d ) \
        RT_TRACE4( _bool_, \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    , a,b,c,d )
#define RT_TRACE5XV( _bool_, a,b,c,d,e ) \
        RT_TRACE5( _bool_, \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " \
                    , a,b,c,d,e )
#define RT_TRACE6XV( _bool_, a,b,c,d,e,f ) \
        RT_TRACE6( _bool_, \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " # f "=%08x " \
                    , a,b,c,d,e,f ) 
#define RT_TRACE7XV( _bool_, a,b,c,d,e,f,g ) \
        RT_TRACE7( _bool_, \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " # f "=%08x " # g "=%08x " \
                    , a,b,c,d,e,f,g ) 
#define RT_TRACE8XV( _bool_, a,b,c,d,e,f,g,h ) \
        RT_TRACE8( _bool_, \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " # f "=%08x " # g "=%08x " # h "=%08x " \
                    , a,b,c,d,e,f,g,h )

#define RT_TRACE1XB( a ) \
        RT_TRACE1( RT_TRACE_USER_BOOL, "%08x", a ) 
#define RT_TRACE2XB( a,b ) \
        RT_TRACE2( RT_TRACE_USER_BOOL, "%08x %08x", a,b ) 
#define RT_TRACE3XB( a,b,c ) \
        RT_TRACE3( RT_TRACE_USER_BOOL, "%08x %08x %08x", a,b,c ) 
#define RT_TRACE4XB( a,b,c,d ) \
        RT_TRACE4( RT_TRACE_USER_BOOL, "%08x %08x %08x %08x", a,b,c,d ) 
#define RT_TRACE5XB( a,b,c,d,e ) \
        RT_TRACE5( RT_TRACE_USER_BOOL, "%08x %08x %08x %08x %08x", a,b,c,d,e ) 
#define RT_TRACE6XB( a,b,c,d,e,f ) \
        RT_TRACE6( RT_TRACE_USER_BOOL, "%08x %08x %08x %08x %08x %08x", \
                    a,b,c,d,e,f ) 
#define RT_TRACE7XB( a,b,c,d,e,f,g ) \
        RT_TRACE7( RT_TRACE_USER_BOOL, "%08x %08x %08x %08x %08x %08x %08x", \
                    a,b,c,d,e,f,g ) 
#define RT_TRACE8XB( a,b,c,d,e,f,g,h ) \
        RT_TRACE8( RT_TRACE_USER_BOOL, "%08x %08x %08x %08x %08x %08x %08x %08x", \
                    a,b,c,d,e,f,g,h )

#define RT_TRACE1XVB( a ) \
        RT_TRACE1( RT_TRACE_USER_BOOL, # a "=%08x ", a ) 
#define RT_TRACE2XVB( a,b ) \
        RT_TRACE2( RT_TRACE_USER_BOOL, \
                    # a "=%08x " # b "=%08x " , a,b )
#define RT_TRACE3XVB( a,b,c ) \
        RT_TRACE3( RT_TRACE_USER_BOOL, \
                    # a "=%08x " # b "=%08x " # c "=%08x " , a,b,c )
#define RT_TRACE4XVB( a,b,c,d ) \
        RT_TRACE4( RT_TRACE_USER_BOOL, \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    , a,b,c,d )
#define RT_TRACE5XVB( a,b,c,d,e ) \
        RT_TRACE5( RT_TRACE_USER_BOOL, \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " \
                    , a,b,c,d,e )
#define RT_TRACE6XVB( a,b,c,d,e,f ) \
        RT_TRACE6( RT_TRACE_USER_BOOL, \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " # f "=%08x " \
                    , a,b,c,d,e,f ) 
#define RT_TRACE7XVB( a,b,c,d,e,f,g ) \
        RT_TRACE7( RT_TRACE_USER_BOOL, \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " # f "=%08x " # g "=%08x " \
                    , a,b,c,d,e,f,g ) 
#define RT_TRACE8XVB( a,b,c,d,e,f,g,h ) \
        RT_TRACE8( RT_TRACE_USER_BOOL, \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " # f "=%08x " # g "=%08x " # h "=%08x " \
                    , a,b,c,d,e,f,g,h )

// decimal versions

#define RT_TRACE1D( _bool_, a ) \
        RT_TRACE1( _bool_, "%d", a ) 
#define RT_TRACE2D( _bool_, a,b ) \
        RT_TRACE2( _bool_, "%d %d", a,b ) 
#define RT_TRACE3D( _bool_, a,b,c ) \
        RT_TRACE3( _bool_, "%d %d %d", a,b,c ) 
#define RT_TRACE4D( _bool_, a,b,c,d ) \
        RT_TRACE4( _bool_, "%d %d %d %d", a,b,c,d ) 
#define RT_TRACE5D( _bool_, a,b,c,d,e ) \
        RT_TRACE5( _bool_, "%d %d %d %d %d", a,b,c,d,e ) 
#define RT_TRACE6D( _bool_, a,b,c,d,e,f ) \
        RT_TRACE6( _bool_, "%d %d %d %d %d %d", \
                    a,b,c,d,e,f ) 
#define RT_TRACE7D( _bool_, a,b,c,d,e,f,g ) \
        RT_TRACE7( _bool_, "%d %d %d %d %d %d %d", \
                    a,b,c,d,e,f,g ) 
#define RT_TRACE8D( _bool_, a,b,c,d,e,f,g,h ) \
        RT_TRACE8( _bool_, "%d %d %d %d %d %d %d %d", \
                    a,b,c,d,e,f,g,h )

#define RT_TRACE1DV( _bool_, a ) \
        RT_TRACE1( _bool_, # a "=%d ", a ) 
#define RT_TRACE2DV( _bool_, a,b ) \
        RT_TRACE2( _bool_, \
                    # a "=%d " # b "=%d " , a,b )
#define RT_TRACE3DV( _bool_, a,b,c ) \
        RT_TRACE3( _bool_, \
                    # a "=%d " # b "=%d " # c "=%d " , a,b,c )
#define RT_TRACE4DV( _bool_, a,b,c,d ) \
        RT_TRACE4( _bool_, \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    , a,b,c,d )
#define RT_TRACE5DV( _bool_, a,b,c,d,e ) \
        RT_TRACE5( _bool_, \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " \
                    , a,b,c,d,e )
#define RT_TRACE6DV( _bool_, a,b,c,d,e,f ) \
        RT_TRACE6( _bool_, \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " # f "=%d " \
                    , a,b,c,d,e,f ) 
#define RT_TRACE7DV( _bool_, a,b,c,d,e,f,g ) \
        RT_TRACE7( _bool_, \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " # f "=%d " # g "=%d " \
                    , a,b,c,d,e,f,g ) 
#define RT_TRACE8DV( _bool_, a,b,c,d,e,f,g,h ) \
        RT_TRACE8( _bool_, \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " # f "=%d " # g "=%d " # h "=%d " \
                    , a,b,c,d,e,f,g,h )

#define RT_TRACE1DB( a ) \
        RT_TRACE1( RT_TRACE_USER_BOOL, "%d", a ) 
#define RT_TRACE2DB( a,b ) \
        RT_TRACE2( RT_TRACE_USER_BOOL, "%d %d", a,b ) 
#define RT_TRACE3DB( a,b,c ) \
        RT_TRACE3( RT_TRACE_USER_BOOL, "%d %d %d", a,b,c ) 
#define RT_TRACE4DB( a,b,c,d ) \
        RT_TRACE4( RT_TRACE_USER_BOOL, "%d %d %d %d", a,b,c,d ) 
#define RT_TRACE5DB( a,b,c,d,e ) \
        RT_TRACE5( RT_TRACE_USER_BOOL, "%d %d %d %d %d", a,b,c,d,e ) 
#define RT_TRACE6DB( a,b,c,d,e,f ) \
        RT_TRACE6( RT_TRACE_USER_BOOL, "%d %d %d %d %d %d", \
                    a,b,c,d,e,f ) 
#define RT_TRACE7DB( a,b,c,d,e,f,g ) \
        RT_TRACE7( RT_TRACE_USER_BOOL, "%d %d %d %d %d %d %d", \
                    a,b,c,d,e,f,g ) 
#define RT_TRACE8DB( a,b,c,d,e,f,g,h ) \
        RT_TRACE8( RT_TRACE_USER_BOOL, "%d %d %d %d %d %d %d %d", \
                    a,b,c,d,e,f,g,h )

#define RT_TRACE1DVB( a ) \
        RT_TRACE1( RT_TRACE_USER_BOOL, # a "=%d ", a ) 
#define RT_TRACE2DVB( a,b ) \
        RT_TRACE2( RT_TRACE_USER_BOOL, \
                    # a "=%d " # b "=%d " , a,b )
#define RT_TRACE3DVB( a,b,c ) \
        RT_TRACE3( RT_TRACE_USER_BOOL, \
                    # a "=%d " # b "=%d " # c "=%d " , a,b,c )
#define RT_TRACE4DVB( a,b,c,d ) \
        RT_TRACE4( RT_TRACE_USER_BOOL, \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    , a,b,c,d )
#define RT_TRACE5DVB( a,b,c,d,e ) \
        RT_TRACE5( RT_TRACE_USER_BOOL, \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " \
                    , a,b,c,d,e )
#define RT_TRACE6DVB( a,b,c,d,e,f ) \
        RT_TRACE6( RT_TRACE_USER_BOOL, \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " # f "=%d " \
                    , a,b,c,d,e,f ) 
#define RT_TRACE7DVB( a,b,c,d,e,f,g ) \
        RT_TRACE7( RT_TRACE_USER_BOOL, \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " # f "=%d " # g "=%d " \
                    , a,b,c,d,e,f,g ) 
#define RT_TRACE8DVB( a,b,c,d,e,f,g,h ) \
        RT_TRACE8( RT_TRACE_USER_BOOL, \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " # f "=%d " # g "=%d " # h "=%d " \
                    , a,b,c,d,e,f,g,h )

// short hex versions

#define RT_TRACE1Y( _bool_, a ) \
        RT_TRACE1( _bool_, "%x", a ) 
#define RT_TRACE2Y( _bool_, a,b ) \
        RT_TRACE2( _bool_, "%x %x", a,b ) 
#define RT_TRACE3Y( _bool_, a,b,c ) \
        RT_TRACE3( _bool_, "%x %x %x", a,b,c ) 
#define RT_TRACE4Y( _bool_, a,b,c,d ) \
        RT_TRACE4( _bool_, "%x %x %x %x", a,b,c,d ) 
#define RT_TRACE5Y( _bool_, a,b,c,d,e ) \
        RT_TRACE5( _bool_, "%x %x %x %x %x", a,b,c,d,e ) 
#define RT_TRACE6Y( _bool_, a,b,c,d,e,f ) \
        RT_TRACE6( _bool_, "%x %x %x %x %x %x", \
                    a,b,c,d,e,f ) 
#define RT_TRACE7Y( _bool_, a,b,c,d,e,f,g ) \
        RT_TRACE7( _bool_, "%x %x %x %x %x %x %x", \
                    a,b,c,d,e,f,g ) 
#define RT_TRACE8Y( _bool_, a,b,c,d,e,f,g,h ) \
        RT_TRACE8( _bool_, "%x %x %x %x %x %x %x %x", \
                    a,b,c,d,e,f,g,h )

#define RT_TRACE1YV( _bool_, a ) \
        RT_TRACE1( _bool_, # a "=%x ", a ) 
#define RT_TRACE2YV( _bool_, a,b ) \
        RT_TRACE2( _bool_, \
                    # a "=%x " # b "=%x " , a,b )
#define RT_TRACE3YV( _bool_, a,b,c ) \
        RT_TRACE3( _bool_, \
                    # a "=%x " # b "=%x " # c "=%x " , a,b,c )
#define RT_TRACE4YV( _bool_, a,b,c,d ) \
        RT_TRACE4( _bool_, \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    , a,b,c,d )
#define RT_TRACE5YV( _bool_, a,b,c,d,e ) \
        RT_TRACE5( _bool_, \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " \
                    , a,b,c,d,e )
#define RT_TRACE6YV( _bool_, a,b,c,d,e,f ) \
        RT_TRACE6( _bool_, \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " # f "=%x " \
                    , a,b,c,d,e,f ) 
#define RT_TRACE7YV( _bool_, a,b,c,d,e,f,g ) \
        RT_TRACE7( _bool_, \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " # f "=%x " # g "=%x " \
                    , a,b,c,d,e,f,g ) 
#define RT_TRACE8YV( _bool_, a,b,c,d,e,f,g,h ) \
        RT_TRACE8( _bool_, \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " # f "=%x " # g "=%x " # h "=%x " \
                    , a,b,c,d,e,f,g,h )

#define RT_TRACE1YB( a ) \
        RT_TRACE1( RT_TRACE_USER_BOOL, "%x", a ) 
#define RT_TRACE2YB( a,b ) \
        RT_TRACE2( RT_TRACE_USER_BOOL, "%x %x", a,b ) 
#define RT_TRACE3YB( a,b,c ) \
        RT_TRACE3( RT_TRACE_USER_BOOL, "%x %x %x", a,b,c ) 
#define RT_TRACE4YB( a,b,c,d ) \
        RT_TRACE4( RT_TRACE_USER_BOOL, "%x %x %x %x", a,b,c,d ) 
#define RT_TRACE5YB( a,b,c,d,e ) \
        RT_TRACE5( RT_TRACE_USER_BOOL, "%x %x %x %x %x", a,b,c,d,e ) 
#define RT_TRACE6YB( a,b,c,d,e,f ) \
        RT_TRACE6( RT_TRACE_USER_BOOL, "%x %x %x %x %x %x", \
                    a,b,c,d,e,f ) 
#define RT_TRACE7YB( a,b,c,d,e,f,g ) \
        RT_TRACE7( RT_TRACE_USER_BOOL, "%x %x %x %x %x %x %x", \
                    a,b,c,d,e,f,g ) 
#define RT_TRACE8YB( a,b,c,d,e,f,g,h ) \
        RT_TRACE8( RT_TRACE_USER_BOOL, "%x %x %x %x %x %x %x %x", \
                    a,b,c,d,e,f,g,h )

#define RT_TRACE1YVB( a ) \
        RT_TRACE1( RT_TRACE_USER_BOOL, # a "=%x ", a ) 
#define RT_TRACE2YVB( a,b ) \
        RT_TRACE2( RT_TRACE_USER_BOOL, \
                    # a "=%x " # b "=%x " , a,b )
#define RT_TRACE3YVB( a,b,c ) \
        RT_TRACE3( RT_TRACE_USER_BOOL, \
                    # a "=%x " # b "=%x " # c "=%x " , a,b,c )
#define RT_TRACE4YVB( a,b,c,d ) \
        RT_TRACE4( RT_TRACE_USER_BOOL, \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    , a,b,c,d )
#define RT_TRACE5YVB( a,b,c,d,e ) \
        RT_TRACE5( RT_TRACE_USER_BOOL, \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " \
                    , a,b,c,d,e )
#define RT_TRACE6YVB( a,b,c,d,e,f ) \
        RT_TRACE6( RT_TRACE_USER_BOOL, \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " # f "=%x " \
                    , a,b,c,d,e,f ) 
#define RT_TRACE7YVB( a,b,c,d,e,f,g ) \
        RT_TRACE7( RT_TRACE_USER_BOOL, \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " # f "=%x " # g "=%x " \
                    , a,b,c,d,e,f,g ) 
#define RT_TRACE8YVB( a,b,c,d,e,f,g,h ) \
        RT_TRACE8( RT_TRACE_USER_BOOL, \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " # f "=%x " # g "=%x " # h "=%x " \
                    , a,b,c,d,e,f,g,h )

// -------------------------------------------------------------------------
//
// RT_REPORT_FUNCARGn{[XDY]{V}}
//
// Convenience macros two: these fall into a few dimensions, with suffix letters:
// First option:
//     X: user need not supply a format string, %08x is used
//     D: ditto but signed decimal, %d
//     Y: ditto but just plain %x
// Second option, only meaningful with one of XDY:
//     V: "<var> = %..." is used, by stringifying the argument

// long hex versions

#define RT_REPORT_FUNCARG1X( a ) \
        RT_REPORT_FUNCARG1( "%08x", a ) 
#define RT_REPORT_FUNCARG2X( a,b ) \
        RT_REPORT_FUNCARG2( "%08x %08x", a,b ) 
#define RT_REPORT_FUNCARG3X( a,b,c ) \
        RT_REPORT_FUNCARG3( "%08x %08x %08x", a,b,c ) 
#define RT_REPORT_FUNCARG4X( a,b,c,d ) \
        RT_REPORT_FUNCARG4( "%08x %08x %08x %08x", a,b,c,d ) 
#define RT_REPORT_FUNCARG5X( a,b,c,d,e ) \
        RT_REPORT_FUNCARG5( "%08x %08x %08x %08x %08x", a,b,c,d,e ) 
#define RT_REPORT_FUNCARG6X( a,b,c,d,e,f ) \
        RT_REPORT_FUNCARG6( "%08x %08x %08x %08x %08x %08x", \
                    a,b,c,d,e,f ) 
#define RT_REPORT_FUNCARG7X( a,b,c,d,e,f,g ) \
        RT_REPORT_FUNCARG7( "%08x %08x %08x %08x %08x %08x %08x", \
                    a,b,c,d,e,f,g ) 
#define RT_REPORT_FUNCARG8X( a,b,c,d,e,f,g,h ) \
        RT_REPORT_FUNCARG8( "%08x %08x %08x %08x %08x %08x %08x %08x", \
                    a,b,c,d,e,f,g,h )

#define RT_REPORT_FUNCARG1XV( a ) \
        RT_REPORT_FUNCARG1( # a "=%08x ", a ) 
#define RT_REPORT_FUNCARG2XV( a,b ) \
        RT_REPORT_FUNCARG2( \
                    # a "=%08x " # b "=%08x " , a,b )
#define RT_REPORT_FUNCARG3XV( a,b,c ) \
        RT_REPORT_FUNCARG3( \
                    # a "=%08x " # b "=%08x " # c "=%08x " , a,b,c )
#define RT_REPORT_FUNCARG4XV( a,b,c,d ) \
        RT_REPORT_FUNCARG4( \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    , a,b,c,d )
#define RT_REPORT_FUNCARG5XV( a,b,c,d,e ) \
        RT_REPORT_FUNCARG5( \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " \
                    , a,b,c,d,e )
#define RT_REPORT_FUNCARG6XV( a,b,c,d,e,f ) \
        RT_REPORT_FUNCARG6( \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " # f "=%08x " \
                    , a,b,c,d,e,f ) 
#define RT_REPORT_FUNCARG7XV( a,b,c,d,e,f,g ) \
        RT_REPORT_FUNCARG7( \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " # f "=%08x " # g "=%08x " \
                    , a,b,c,d,e,f,g ) 
#define RT_REPORT_FUNCARG8XV( a,b,c,d,e,f,g,h ) \
        RT_REPORT_FUNCARG8( \
                    # a "=%08x " # b "=%08x " # c "=%08x " # d "=%08x " \
                    # e "=%08x " # f "=%08x " # g "=%08x " # h "=%08x " \
                    , a,b,c,d,e,f,g,h )

// decimal versions


#define RT_REPORT_FUNCARG1D( a ) \
        RT_REPORT_FUNCARG1( "%d", a ) 
#define RT_REPORT_FUNCARG2D( a,b ) \
        RT_REPORT_FUNCARG2( "%d %d", a,b ) 
#define RT_REPORT_FUNCARG3D( a,b,c ) \
        RT_REPORT_FUNCARG3( "%d %d %d", a,b,c ) 
#define RT_REPORT_FUNCARG4D( a,b,c,d ) \
        RT_REPORT_FUNCARG4( "%d %d %d %d", a,b,c,d ) 
#define RT_REPORT_FUNCARG5D( a,b,c,d,e ) \
        RT_REPORT_FUNCARG5( "%d %d %d %d %d", a,b,c,d,e ) 
#define RT_REPORT_FUNCARG6D( a,b,c,d,e,f ) \
        RT_REPORT_FUNCARG6( "%d %d %d %d %d %d", \
                    a,b,c,d,e,f ) 
#define RT_REPORT_FUNCARG7D( a,b,c,d,e,f,g ) \
        RT_REPORT_FUNCARG7( "%d %d %d %d %d %d %d", \
                    a,b,c,d,e,f,g ) 
#define RT_REPORT_FUNCARG8D( a,b,c,d,e,f,g,h ) \
        RT_REPORT_FUNCARG8( "%d %d %d %d %d %d %d %d", \
                    a,b,c,d,e,f,g,h )

#define RT_REPORT_FUNCARG1DV( a ) \
        RT_REPORT_FUNCARG1( # a "=%d ", a ) 
#define RT_REPORT_FUNCARG2DV( a,b ) \
        RT_REPORT_FUNCARG2( \
                    # a "=%d " # b "=%d " , a,b )
#define RT_REPORT_FUNCARG3DV( a,b,c ) \
        RT_REPORT_FUNCARG3( \
                    # a "=%d " # b "=%d " # c "=%d " , a,b,c )
#define RT_REPORT_FUNCARG4DV( a,b,c,d ) \
        RT_REPORT_FUNCARG4( \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    , a,b,c,d )
#define RT_REPORT_FUNCARG5DV( a,b,c,d,e ) \
        RT_REPORT_FUNCARG5( \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " \
                    , a,b,c,d,e )
#define RT_REPORT_FUNCARG6DV( a,b,c,d,e,f ) \
        RT_REPORT_FUNCARG6( \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " # f "=%d " \
                    , a,b,c,d,e,f ) 
#define RT_REPORT_FUNCARG7DV( a,b,c,d,e,f,g ) \
        RT_REPORT_FUNCARG7( \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " # f "=%d " # g "=%d " \
                    , a,b,c,d,e,f,g ) 
#define RT_REPORT_FUNCARG8DV( a,b,c,d,e,f,g,h ) \
        RT_REPORT_FUNCARG8( \
                    # a "=%d " # b "=%d " # c "=%d " # d "=%d " \
                    # e "=%d " # f "=%d " # g "=%d " # h "=%d " \
                    , a,b,c,d,e,f,g,h )

// short hex versions

#define RT_REPORT_FUNCARG1Y( a ) \
        RT_REPORT_FUNCARG1( "%x", a ) 
#define RT_REPORT_FUNCARG2Y( a,b ) \
        RT_REPORT_FUNCARG2( "%x %x", a,b ) 
#define RT_REPORT_FUNCARG3Y( a,b,c ) \
        RT_REPORT_FUNCARG3( "%x %x %x", a,b,c ) 
#define RT_REPORT_FUNCARG4Y( a,b,c,d ) \
        RT_REPORT_FUNCARG4( "%x %x %x %x", a,b,c,d ) 
#define RT_REPORT_FUNCARG5Y( a,b,c,d,e ) \
        RT_REPORT_FUNCARG5( "%x %x %x %x %x", a,b,c,d,e ) 
#define RT_REPORT_FUNCARG6Y( a,b,c,d,e,f ) \
        RT_REPORT_FUNCARG6( "%x %x %x %x %x %x", \
                    a,b,c,d,e,f ) 
#define RT_REPORT_FUNCARG7Y( a,b,c,d,e,f,g ) \
        RT_REPORT_FUNCARG7( "%x %x %x %x %x %x %x", \
                    a,b,c,d,e,f,g ) 
#define RT_REPORT_FUNCARG8Y( a,b,c,d,e,f,g,h ) \
        RT_REPORT_FUNCARG8( "%x %x %x %x %x %x %x %x", \
                    a,b,c,d,e,f,g,h )

#define RT_REPORT_FUNCARG1YV( a ) \
        RT_REPORT_FUNCARG1( # a "=%x ", a ) 
#define RT_REPORT_FUNCARG2YV( a,b ) \
        RT_REPORT_FUNCARG2( \
                    # a "=%x " # b "=%x " , a,b )
#define RT_REPORT_FUNCARG3YV( a,b,c ) \
        RT_REPORT_FUNCARG3( \
                    # a "=%x " # b "=%x " # c "=%x " , a,b,c )
#define RT_REPORT_FUNCARG4YV( a,b,c,d ) \
        RT_REPORT_FUNCARG4( \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    , a,b,c,d )
#define RT_REPORT_FUNCARG5YV( a,b,c,d,e ) \
        RT_REPORT_FUNCARG5( \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " \
                    , a,b,c,d,e )
#define RT_REPORT_FUNCARG6YV( a,b,c,d,e,f ) \
        RT_REPORT_FUNCARG6( \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " # f "=%x " \
                    , a,b,c,d,e,f ) 
#define RT_REPORT_FUNCARG7YV( a,b,c,d,e,f,g ) \
        RT_REPORT_FUNCARG7( \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " # f "=%x " # g "=%x " \
                    , a,b,c,d,e,f,g ) 
#define RT_REPORT_FUNCARG8YV( a,b,c,d,e,f,g,h ) \
        RT_REPORT_FUNCARG8( \
                    # a "=%x " # b "=%x " # c "=%x " # d "=%x " \
                    # e "=%x " # f "=%x " # g "=%x " # h "=%x " \
                    , a,b,c,d,e,f,g,h )


#endif // RTONCE_INFRA_RT_TRAC_H multiple inclusion protection
// EOF rt_trac.h
