#ifndef RTONCE_KERNEL_KTYPES_H
#define RTONCE_KERNEL_KTYPES_H

//==========================================================================
//
//      ktypes.h
//
//      Standard types used in the kernel and its interfaces
//
//==========================================================================
//
// Purpose:     Define kernel specific types
// Description: Kernel specific types
// Usage:       #include <rt/kernel/ktypes.h>
//
//==========================================================================

// -------------------------------------------------------------------------
// Check that a configuration file is present.
#define	RTONCE_PKGCONF_KERNEL_H

#ifndef RTONCE_PKGCONF_KERNEL_H
#error "No Configuration file included"
#endif

// -------------------------------------------------------------------------

/*
#include <include\infra\rt_type.h>
#include <stddef.h>              // for size_t
*/

#include "../infra/rt_type.h"
#include <stddef.h>              // for size_t

// -------------------------------------------------------------------------
// Integer types:

typedef rt_int32       rt_code;       // General return/error/status code

typedef rt_count32     rt_priority;   // priority value

typedef rt_uint32      rt_vector;     // vector number/descriptor

typedef rt_uint64      rt_tick_count; // clock tick count value


// -------------------------------------------------------------------------
// Class and structure conversion macros.
// RT_CLASSFROMFIELD translates a pointer to a field of a struct or
// class into a pointer to the class.
// RT_OFFSETOFBASE yields the offset of a base class of a derived
// class.
// RT_CLASSFROMBASE translates a pointer to a base class into a pointer
// to a selected derived class. The base class object _must_ be part of
// the specified derived class. This is essentially a poor mans version
// of the RTTI dynamic_cast operator.
// Caveat: These macros do not work for virtual base classes.

#define RT_CLASSFROMFIELD(_type_,_member_,_ptr_)\
    ((_type_ *)((char *)(_ptr_)-((char *)&(((_type_ *)0)->_member_))))

#ifdef __cplusplus

#define RT_OFFSETOFBASE(_type_,_base_)\
    ((char *)((_base_ *)((_type_ *)4)) - (char *)4)

# define RT_CLASSFROMBASE(_class_,_base_,_ptr_)\
    ((_class_ *)((char *)(_ptr_) - RT_OFFSETOFBASE(_class_,_base_)))

#endif

// -------------------------------------------------------------------------
#endif // RTONCE_KERNEL_KTYPES_H
// EOF ktypes.h

