/*
* ----------------------------------------------------------------
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* $Header: /local/cvsroot/uClinux-dist/user/boa/src/rtl865x/flashmtd.c,v 1.6 2006/01/03 07:03:24 alva_zhang Exp $
*
* Abstract: Flash driver (based on MTD) source code .
*
* $Author: alva_zhang $
*
* $Log: flashmtd.c,v $
* Revision 1.6  2006/01/03 07:03:24  alva_zhang
* +: add a important flag clearbits in struct mtd_dev to indicate whether to do block-erased before flash writing
* *: modify function flashdrv_write so that it can do flash write with block-erased and bits-clear without block-erased
* +: add functin flashdrv_clearbits which can do flash bits-clear by calling function flashdrv_write
*
* Revision 1.5  2005/12/23 03:25:51  alva_zhang
* *: change flash info structure & initial process
* *: refine and optimize flash_write function
*
* Revision 1.4  2005/12/05 09:27:51  alva_zhang
* *: during boa initializing,it used flash detail info ,but mtd_dev is do not initialized;so we have to specify a static value for those
*
* Revision 1.3  2005/12/02 01:16:03  alva_zhang
* *: refine the functions refer to flash base information
*
* Revision 1.2  2005/11/30 07:59:38  alva_zhang
* *: refine the defination of those functions evolved from flashdrv.c
*
* Revision 1.1  2005/11/24 10:05:48  alva_zhang
* +: evolved from flashdrv.c,and do work based on MTD;this is just an alpha version
*
*
*
*
*---------------------------------------------------------------
*/


#include <linux/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <linux/fcntl.h>
#include <linux/ioctl.h>

#include <linux/mtd/mtd.h>

#include "flashdrv.h"

#define SEEK_SET 0

#if 0 //list here for reference
typedef struct mtd_flashdev
{
  unsigned char  flash_init;
  unsigned char  partcnt;
  unsigned short erasize;
  unsigned short buswidth;
  unsigned short clearbits;
  
  unsigned int  chipbase;
  unsigned long chipsize; 
  
  unsigned int bdinfobase;
  unsigned int bdinfosize;
  
  unsigned int ccfgbase;
  unsigned int ccfgsize;
  
  unsigned int kernbase;
  unsigned int kernsize;
  
  unsigned int rootfsbase;
  unsigned int rootfssize;
  
  // struct partinfo p[6];

  unsigned int p0base;
  unsigned int p0size;
  
  unsigned int p1base;
  unsigned int p1size;

  unsigned int p2base;
  unsigned int p2size;

  unsigned int p3base;
  unsigned int p3size;

  unsigned int p4base;
  unsigned int p4size;

  unsigned int p5base;
  unsigned int p5size;

}MTD_FlashDev;
#endif

// global variable mtd_dev;
MTD_FlashDev mtd_dev={ 0, 0, 0, 0x02, 0, /*if clearbits is set 0,flash will be block-erased before it is written  */
                       0, 0x0400000,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
/*
 *  bdinfo instance in SDRAM.
 */
bdinfo_t    bdinfo;

void flashdrv_init(void)
{  
  FILE* fp;
  
  if(mtd_dev.flash_init==0)
    {
      mtd_dev.flash_init=1;
      
      //printf("\nFunction flashdrv_init runing.......... \n\n");
      
      mtd_dev.bdinfobase = FLASH_MAP_BOARD_INFO_ADDR;
      mtd_dev.bdinfosize = FLASH_MAP_BOARD_INFO_SIZE;
      
      mtd_dev.ccfgbase = FLASH_MAP_CCFG_IMAGE_ADDR;
      mtd_dev.ccfgsize = FLASH_MAP_CCFG_IMAGE_SIZE;
      
      mtd_dev.kernbase = FLASH_MAP_RUN_IMAGE_ADDR;
      mtd_dev.kernsize = FLASH_MAP_RUN_IMAGE_SIZE;
      
      mtd_dev.rootfsbase = FLASH_MAP_RUN_IMAGE_ADDR+CONFIG_FLASH_SIZE_OF_KERNEL;     
      
      if((fp = open( "/dev/mtd0", O_RDWR))!=-1)
        {
          mtd_dev.partcnt = mtd_dev.partcnt + 1;                      
          mtd_dev.p0base = FLASH_BASE + CONFIG_RTL865X_CUSTOM_LOADER_SEG1_ADDRESS;
          mtd_dev.p0size = CONFIG_RTL865X_CUSTOM_LOADER_SEG1_SIZE;                
          mtd_dev.chipbase = FLASH_BASE;
        } 
      close(fp);
      
      if((fp = open( "/dev/mtd1", O_RDWR))!=-1) 
        {
          mtd_dev.partcnt = mtd_dev.partcnt+1;
          mtd_dev.p1base = mtd_dev.bdinfobase;
          mtd_dev.p1size = mtd_dev.bdinfosize+mtd_dev.ccfgsize;
        }
      close(fp);
     
      if((fp = open( "/dev/mtd2", O_RDWR)) !=-1) 
        {
          mtd_dev.partcnt = mtd_dev.partcnt+1;          
          mtd_dev.p2base =FLASH_BASE + CONFIG_RTL865X_CUSTOM_LOADER_SEG2_ADDRESS;
          mtd_dev.p2size = CONFIG_RTL865X_CUSTOM_LOADER_SEG2_SIZE;      
        }
      close(fp);

      if((fp = open( "/dev/mtd3", O_RDWR)) !=-1) 
        {
          mtd_dev.partcnt = mtd_dev.partcnt+1;                  
          mtd_dev.p3base = mtd_dev.kernbase;
          mtd_dev.p3size = mtd_dev.kernsize;
        }
      close(fp);
                  
      if((fp = open( "/dev/mtd4", O_RDWR)) !=-1) 
        {
          mtd_dev.partcnt = mtd_dev.partcnt+1;
          mtd_dev.p4base = mtd_dev.rootfsbase;
          mtd_dev.rootfssize = FLASH_MAP_RUN_IMAGE_SIZE -CONFIG_FLASH_SIZE_OF_KERNEL;
          mtd_dev.p4size = mtd_dev.rootfssize;
        }
       close(fp);
              
       if((fp = open( "/dev/mtd5", O_RDWR)) !=-1) 
         {
           mtd_dev.partcnt = mtd_dev.partcnt+1;              
           mtd_dev.p5base = FLASH_BASE + CONFIG_RTL865X_CUSTOM_LOADER_SEG3_ADDRESS;
           mtd_dev.p5size = CONFIG_RTL865X_CUSTOM_LOADER_SEG3_SIZE;      
         }
       close(fp);
#if 0       
       printf("\n\n");
       printf("flashdrv_init: Partition 5-0 Base = 0x%08x  0x%08x  0x%08x  0x%08x  0x%08x  0x%08x \n\n",mtd_dev.p5base,mtd_dev.p4base,mtd_dev.p3base,mtd_dev.p2base,mtd_dev.p1base,mtd_dev.p0base);
#endif
       
#ifdef _SUPPORT_LARGE_FLASH_
       /* read bdinfo before any read to flash (to recover flash JUMP code) */
       flashdrv_read( &bdinfo, (void*)(FLASH_MAP_BOARD_INFO_ADDR), sizeof(bdinfo_t) );
       /*
         rtlglue_printf( "\n*** read bdinfo ***\n0x%08x\n0x%08x\n\n", bdinfo.BackupInst[0], bdinfo.BackupInst[1] );
       */
#endif
    } //end of  if(mtd_dev.flash_init==0)
  else
    {
      mtd_dev.flash_init++;
    } 
}

#ifdef _SUPPORT_LARGE_FLASH_
/*
 *  flashdrv_isOver5xBBootInstructions(), flashdrv_recoverBootInstructions(),
 *  flashdrv_isInjected() and flashdrv_backupAndModifyBootCode() are used for 865xB's bug.
 *
 *  Before read/write flash memory, program should call flashdrv_isOver5xBBootInstructions() to ensure that
 *    the [startOffset,endOffset) range is not overlapped with the boot J instruction.
 *    The returned pInstructionOffset is used to pass to flashdrv_backupAndModifyBootCode().
 */
int32 flashdrv_isOver5xBBootInstructions( uint32 startOffset, uint32 endOffset, uint32 *pInstructionOffset )
{
  *pInstructionOffset = 0;
  
  /* Only 865XB's revB needs to check. */
  if ( !IS_865XB() )
    {
      /* Not 865xB, it is ok. */
      return FALSE;
    }
  else if ( ( REG32( CRMR_ADDR ) & 0x000f0000 ) != 0x00010000 )
    {
      /* 865XB, but not revB, it is ok. */
      return FALSE;
    }
  
  if ( mtd_dev.chipsize == 0x0800000 ) /* 8MB flash */
    {
      *pInstructionOffset = 0x0400000;
      if ( startOffset < 0x0400008 && endOffset > 0x0400000 )
        {
          return TRUE;
        }
    }
  else if ( mtd_dev.chipsize == 0x1000000 ) /* 16MB flash */
    {
      *pInstructionOffset = 0x0C00000;
      if ( startOffset < 0x0C00008 &&  endOffset > 0x0C00000 )
        {
          return TRUE;
        }
    }
  
  return FALSE;
}


/*
 *  flashdrv_recoverBootInstructions() is used to recover code that overwriten by JUMP instructions.
 *  Typically, this function is called by flashdrv_read().
 *
 *  Warning: User must read bdinfo structure from flash before calling this function.
 */
int32 flashdrv_recoverBootInstructions( uint32* pStartAddrToRecover )
{
  pStartAddrToRecover[0] = bdinfo.BackupInst[0];
  pStartAddrToRecover[1] = bdinfo.BackupInst[1];
#ifdef _DEBUG_
  rtlglue_printf("\n*** Recover Instruction Mechanism Fired ***\nbdinfo.BackupInst[0]=0x%08x\nbdinfo.BackupInst[1]=0x%08x\n\n",
                 bdinfo.BackupInst[0] ,bdinfo.BackupInst[1] );
#endif//_DEBUG_
  return SUCCESS;
}
#endif //_SUPPORT_LARGE_FLASH_

uint32 flashdrv_read(void *dstAddr_P, void *srcAddr_P, uint32 size)
{
#if 0
  printf("\n\n");
  printf("flashdrv_read  dst=0x%x,  src=0x%x,  size=0x%x\n\n",dstAddr_P,srcAddr_P,size);
#endif

#ifdef _SUPPORT_LARGE_FLASH_
  uint32 startOffset; /* start offset to read */
  uint32 endOffset;   /* end offset to read */
  uint32 InstructionOffset; /* offset of boot instruction */
#endif/*_SUPPORT_LARGE_FLASH_*/
  
  ASSERT_CSP( srcAddr_P );
  ASSERT_CSP( dstAddr_P );
  
  memcpy(dstAddr_P, srcAddr_P, size);
  
#ifdef _SUPPORT_LARGE_FLASH_
  startOffset = ((uint32)srcAddr_P) - FLASH_BASE;
  endOffset = startOffset + size;
  if ( TRUE == flashdrv_isOver5xBBootInstructions( startOffset, endOffset, &InstructionOffset ) )
    {
      uint32* pStartAddrToRecover;
      
      pStartAddrToRecover = (uint32*)( ((uint32)dstAddr_P+(InstructionOffset-startOffset)) );
      flashdrv_recoverBootInstructions( pStartAddrToRecover );
    }
#endif/*_SUPPORT_LARGE_FLASH_*/
  
  return 0;
}

#if 0 //table list here for reference
static struct mtd_partition rtl865x_parts[] = {
  { name: "boot1", offset: CONFIG_RTL865X_CUSTOM_LOADER_SEG1_ADDRESS, size: CONFIG_RTL865X_CUSTOM_LOADER_SEG1_SIZE, mask_flags: MTD_WRITEABLE },
  { name: "bdinfo", offset: CONFIG_RTL865X_CUSTOM_LOADER_SEG1_ADDRESS+CONFIG_RTL865X_CUSTOM_LOADER_SEG1_SIZE, size: CONFIG_RTL865X_CUSTOM_LOADER_SEG2_ADDRESS - CONFIG_RTL865X_CUSTOM_LOADER_SEG1_ADDRESS - CONFIG_RTL865X_CUSTOM_LOADER_SEG1_SIZE, mask_flags: MTD_WRITEABLE },
  { name: "boot2", offset: CONFIG_RTL865X_CUSTOM_LOADER_SEG2_ADDRESS, size: CONFIG_RTL865X_CUSTOM_LOADER_SEG2_SIZE, mask_flags: MTD_WRITEABLE },
  { name: "kernel", offset: CONFIG_RTL865X_CUSTOM_RUNTIME_ADDRESS, size: CONFIG_FLASH_SIZE_OF_KERNEL, mask_flags: MTD_WRITEABLE },
  { name: "rootfs", offset: CONFIG_RTL865X_CUSTOM_RUNTIME_ADDRESS+CONFIG_FLASH_SIZE_OF_KERNEL, size: CONFIG_RTL865X_CUSTOM_LOADER_SEG3_ADDRESS-CONFIG_RTL865X_CUSTOM_RUNTIME_ADDRESS-CONFIG_FLASH_SIZE_OF_KERNEL, mask_flags: MTD_WRITEABLE },
  { name: "boot3", offset: CONFIG_RTL865X_CUSTOM_LOADER_SEG3_ADDRESS, size: CONFIG_RTL865X_CUSTOM_LOADER_SEG3_SIZE, mask_flags: MTD_WRITEABLE },
};
#endif

/* Here we only consider write  partition 5,4,3,because we do NOT write bootloader  in uclinux environment */
uint32 flashdrv_write(void *dstAddr_P, void *srcAddr_P, uint32 size)
{
  unsigned int poffset;
  FILE*  fp;
  int cnt=0;
#if 0  
  printf("\n\n");
  printf("flashdrv_write  dst=0x%x,  src=0x%x,  size=0x%x\n\n",dstAddr_P,srcAddr_P,size);  
  printf("Partition Base = 0x%08x  0x%08x  0x%08x  0x%08x  0x%08x  0x%08x \n\n",mtd_dev.p5base,mtd_dev.p4base,mtd_dev.p3base,mtd_dev.p2base,mtd_dev.p1base,mtd_dev.p0base);
#endif   
  
  //Notice: in all mtd char devices, the odd-number devices can be read only,
  //thus,device mtd0/1 map to partition 0 and device mtd2/3  map to partition 1,so on.....
  
  // Here partition1 bdinfo&ccfg is mostly common writen !
  if(((unsigned int)dstAddr_P>=mtd_dev.p1base)&&(((unsigned int)dstAddr_P+size)<=(mtd_dev.p1base+mtd_dev.p1size)))
    {           
      poffset=(unsigned int)dstAddr_P -mtd_dev.p1base;     
      if((fp = open( "/dev/mtd2", O_RDWR))!=-1)
        { // printf("Flash_Write  fp open successed,fp=0x%08x \n\n",fp);
          lseek(fp, poffset, SEEK_SET);
          if(mtd_dev.clearbits==FALSE)
            {
              cnt=write(fp, srcAddr_P, size);
            }
          else
            { //  printf("------FLASH_WRITE: ioctl operate \n\n");
              unsigned int eraseinfo[3];
              eraseinfo[0]=srcAddr_P;
              eraseinfo[1]=size;
              eraseinfo[2]=poffset;
              cnt=ioctl(fp,MTDCLERABIT,eraseinfo);
              //printf("------FLASH_WRITE: cnt=0x%08x \n\n",cnt);
            }
          // printf("Flashdrv_Write:  dstAddr_P=0x%08x, return value =%d\n\n",dstAddr_P,cnt);
        }     
      close(fp);
    }
  
  // partition 3 -4 kernel&rootfs may be writen cross boundary,
  // so here we just check whether dst locates in partitions 3 kernerl space
  else if(((unsigned int)dstAddr_P>=mtd_dev.p3base)&&((unsigned int)dstAddr_P<=(mtd_dev.p3base+mtd_dev.p3size)))
    {
      poffset=(unsigned int)dstAddr_P -mtd_dev.p3base;
      //here  check whether partition 3-4 can hold all content need be writen
      if(size<(mtd_dev.p3size-poffset+mtd_dev.p4size))
        {
          if((fp = open( "/dev/mtd6", O_RDWR))!=-1)  
            {
              lseek(fp, poffset, SEEK_SET);
              if(mtd_dev.clearbits==FALSE)
                {
                  cnt=write(fp, srcAddr_P, size);  
                }
              else
                {  
                  unsigned int eraseinfo[3];
                  eraseinfo[0]=srcAddr_P;
                  eraseinfo[1]=size;
                  eraseinfo[2]=&poffset;
                  cnt=ioctl(fp,MTDCLERABIT,eraseinfo);
                }
            }
          close(fp);
          //here check whether partition 3 can hold all content need be writen
          if(size>mtd_dev.p3size-poffset)
            {
              if((fp = open( "/dev/mtd8", O_RDWR))!=-1)
                {
                  lseek(fp, 0, SEEK_SET);
                  if(mtd_dev.clearbits==FALSE)
                    {
                      cnt=cnt+write(fp, srcAddr_P, (size-mtd_dev.p3size+poffset));
                    }
                  else
                    { 
                      unsigned int eraseinfo[3];
                      eraseinfo[0]=srcAddr_P+mtd_dev.p3size-poffset;
                      eraseinfo[1]=size-mtd_dev.p3size+poffset;
                      poffset=0; 
                      eraseinfo[2]=&poffset;
                      cnt=cnt+ioctl(fp,MTDCLERABIT,eraseinfo);
                    }
                }
              close(fp); 
            }
          else
            {
              printf("both partition 3-4 can NOT hold all content need be writen, check it pls!\n\n");
              cnt=-1;
            }
        }
    }
  // partition 4rootfs can be writen within boundary
  else if(((unsigned int)dstAddr_P>=mtd_dev.p4base)&&(((unsigned int)dstAddr_P+size)<=(mtd_dev.p4base+mtd_dev.p4size)))
    {
      poffset=(unsigned int)dstAddr_P -mtd_dev.p4base;
      if((fp = open( "/dev/mtd8", O_RDWR))!=-1)
        {
          lseek(fp, poffset, SEEK_SET);
          
          if(mtd_dev.clearbits==FALSE)
            {
              cnt=write(fp, srcAddr_P, size); 
            }
          else
            {  
              unsigned int eraseinfo[3];
              eraseinfo[0]=srcAddr_P;
              eraseinfo[1]=size;
              eraseinfo[2]=&poffset;
              cnt=ioctl(fp,MTDCLERABIT,eraseinfo);
            }           
        }
      close(fp);
    }
    
  // here we do consider that loader can be writen into partition 0-2-4 boot1-3 and not beyond partition 4 boot3 
  else if(((unsigned int)dstAddr_P>=mtd_dev.p0base)&&((unsigned int)dstAddr_P<=(mtd_dev.p0base+mtd_dev.p0size)))
    {
      poffset=(unsigned int)dstAddr_P -mtd_dev.p0base;
      //here  check whether partition 0-2-4 can hold loader need be writen
      if(size<=( mtd_dev.p0size-poffset + mtd_dev.p2size +mtd_dev.p4size))
        {
          if((fp = open( "/dev/mtd0", O_RDWR))!=-1)
            {
              lseek(fp, poffset, SEEK_SET);
              if(mtd_dev.clearbits==FALSE)
                {
                  cnt=write(fp, srcAddr_P, size);
                }
              else
                {  
                  unsigned int eraseinfo[3];
                  eraseinfo[0]=srcAddr_P;
                  eraseinfo[1]=size;
                  eraseinfo[2]=&poffset;
                  cnt=ioctl(fp,MTDCLERABIT,eraseinfo);
                }           
            }
          close(fp);
          if(size>mtd_dev.p0size-poffset)
            { 
              if((fp = open( "/dev/mtd4", O_RDWR))!=-1)
                {
                  lseek(fp, 0, SEEK_SET);
                  
                  if(mtd_dev.clearbits==FALSE)
                    {
                      cnt=cnt+write(fp, (srcAddr_P+mtd_dev.p0size-poffset), (size-mtd_dev.p0size+poffset));
                    }
                  else
                    { 
                      unsigned int eraseinfo[3];
                      eraseinfo[0]=srcAddr_P+mtd_dev.p0size-poffset;
                      eraseinfo[1]=size-mtd_dev.p0size+poffset;
                      poffset=0; 
                      eraseinfo[2]=&poffset;
                      cnt=cnt+ioctl(fp,MTDCLERABIT,eraseinfo);
                    }
                }
              close(fp);
              if(size>(mtd_dev.p0size-((unsigned int)dstAddr_P -mtd_dev.p0base)+mtd_dev.p2size))
                { 
                  if((fp = open( "/dev/mtd8", O_RDWR))!=-1)
                    {      
                      lseek(fp, 0, SEEK_SET);
                      if(mtd_dev.clearbits==FALSE)
                        {
                          cnt=cnt+write(fp, (srcAddr_P+mtd_dev.p0size-((unsigned int)dstAddr_P -mtd_dev.p0base)+mtd_dev.p2size), (size-mtd_dev.p0size+((unsigned int)dstAddr_P -mtd_dev.p0base) -mtd_dev.p2size));  
                        }
                      else
                        { 
                          unsigned int eraseinfo[3];
                          eraseinfo[0]=srcAddr_P+mtd_dev.p0size-((unsigned int)dstAddr_P -mtd_dev.p0base)+mtd_dev.p2size;
                          eraseinfo[1]=size-mtd_dev.p0size+((unsigned int)dstAddr_P -mtd_dev.p0base) -mtd_dev.p2size;
                          poffset=0; 
                          eraseinfo[2]=&poffset;
                          cnt=cnt+ioctl(fp,MTDCLERABIT,eraseinfo);
                        }                      
                    }
                  close(fp);
                }
            }
        } //end of if(size<=( mtd_dev.p0size-poffset + mtd_dev.p2size +mtd_dev.p4size))
      else
        { 
          printf("all partition 0-2-4 can NOT hold loader need be writen, check it pls!\n\n");
          cnt=-1;
        }
    } //end of else if(((unsigned int)dstAddr_P>=mtd_dev.p0base)&&((unsigned int)dstAddr_P<=(mtd_dev.p0base+mtd_dev.p0size)))
  else { printf("Error,dest location or size is not correct, please check it!\n"); cnt=-1;}
  
  if(cnt==size)return SUCCESS;
  else return FAILED;
  
}

//Write check is done by flashdrv_write, so wo do NOT need  function flashdrv_isLoaderOverwritten ;
int8 flashdrv_isLoaderOverwritten( uint8* dst, uint32 len )
{
  return FALSE;
}

/*
 *  flashdrv_backupAndModifyBootCode() is used to inject JUMP instructions.
 *  Before injection, it will backup original instructions in BDINFO (by calling flashdrv_updateImg()).
 *  pStartAddrToBackup is pointed to the address used to inject JUMP instructions.
 *  Typically, this function is called by flashdrv_updateImg().
 *
 *  Warning: User must read bdinfo structure from flash before calling this function.
 */
int32 flashdrv_backupAndModifyBootInstructions( uint32* pStartAddrToBackup )
{
  // will overwrite the 4/8MB'th byte, backup them.  
  bdinfo.BackupInst[0] =  pStartAddrToBackup[0];
  bdinfo.BackupInst[1] =  pStartAddrToBackup[1];
  if( flashdrv_updateImg((void *)&bdinfo, (void*)mtd_dev.bdinfobase, sizeof(bdinfo_t) ) )
    {
      rtlglue_printf("\nupdate bdinfo FAIL when backuping instructions!\n\n");
      return 1;
    }    
  
  /* modify the source code */
  pStartAddrToBackup[0] = 0x0B800080; /* j  0xbe000200 */
  pStartAddrToBackup[1] = 0x00000000; /* nop           */
  
#ifdef _DEBUG_
  rtlglue_printf("\n*** Backup Instruction Mechanism Fired ***\nbdinfo.BackupInst[0]=0x%08x\nbdinfo.BackupInst[1]=0x%08x\n\n",
                 bdinfo.BackupInst[0] ,bdinfo.BackupInst[1] );
#endif//_DEBUG_
  return SUCCESS;
}


/*
 *  flashdrv_isInjected() is used to check if JUMP instructions are injected.
 *  pStartAddrToCheck is pointed to the address to be check.
 *  Typically, this function is called by fileSys_updateBootImage().
 */
int32 flashdrv_isInjected( uint32* pStartAddrToCheck )
{
  if ( pStartAddrToCheck[0] == 0x0B800080 && /* j  0xbe000200 */
       pStartAddrToCheck[1] == 0x00000000 )  /* nop           */
    {
      return TRUE;
    }
  else
    {
      return FALSE;
    }
}

uint32 flashdrv_clearbits(void *dstAddr_P, void *srcAddr_P, uint32 size)
{
  int ret=0;    
  mtd_dev.clearbits=TRUE;
  ret=flashdrv_write(dstAddr_P, srcAddr_P,  size);
  mtd_dev.clearbits=FALSE;
  return ret;
}

unsigned int flashdrv_updateImg(void * srcAddr_P, void * dstAddr_P, uint32 size)
{
  // printf("flashdrv_updateImg: dst=0x%08x,  src=0x%08x \n\n",dstAddr_P,srcAddr_P);
  unsigned int retval=-1;
  unsigned char* pNew=NULL;
  unsigned char* pOld =NULL;
  int flag,i=0;
  pNew=malloc(size);
  pOld =malloc(size);

#if 0
  for(i=0;i<size/256+1;i++)
      {
      flashdrv_read(pOld,dstAddr_P,size);
      memcpy(pNew,srcAddr_P,size);
      flag=memcmp(pOld,pNew,size);
      
      printf(":::::::::::::::::::::  Before buffer counter i=%d  flag=%d \n",i,flag);            
    }
#endif
 
#ifdef _SUPPORT_LARGE_FLASH_
  if ( mtd_dev.chipsize > 0x400000 )
    {
      /* If the first chip is larger than 4MB (not including 4MB),
       * 865xB has bug to boot system.
       * We must backup instructions at 4MB or 8MB position into bdinfo, and recover them in flashdrv_read().
       */
      uint32 startOffset; /* start offset to write */
      uint32 endOffset;   /* end offset to write */
      uint32 InstructionOffset; /* offset of boot instruction */
      
      startOffset = ((uint32)dstAddr_P) - mtd_dev.chipbase;
      endOffset = startOffset + size;
      
      /* 865xB has bug when booting from a single 8MB or 16MB flash. */
      if ( TRUE == flashdrv_isOver5xBBootInstructions( startOffset, endOffset, &InstructionOffset ) )
        {
          uint32 *pStartAddrToBackup;
          void *pAddrToInject;
          
          pStartAddrToBackup = (uint32*)( ((uint32)srcAddr_P+(InstructionOffset-startOffset)) );
          rtlglue_printf("before backup ...\n%p=0x%08x\n%p=0x%08x\n",&pStartAddrToBackup[0],pStartAddrToBackup[0],&pStartAddrToBackup[1],pStartAddrToBackup[1]);
          flashdrv_backupAndModifyBootInstructions( pStartAddrToBackup );
          
          /* Inject right now! To reduce risk .... */
          pAddrToInject = (void*)( (uint32)((uint32)dstAddr_P+(InstructionOffset-startOffset)) );
          rtlglue_printf("flashdrv_write(dst=%p,src=%p,size=0x%08x)\n",pAddrToInject, pStartAddrToBackup, sizeof(bdinfo.BackupInst));
          
          // printf("dst=0x%08x,  src=0x%08x \n\n",pAddrToInject,srcAddr_P);
          flashdrv_write(pAddrToInject, pStartAddrToBackup, sizeof(bdinfo.BackupInst) );
        }
    }
#endif/*_SUPPORT_LARGE_FLASH_*/
  
  retval = flashdrv_write(dstAddr_P, srcAddr_P, size);
#if 0
  for(i=0;i<size/256+1;i++)
      {
      flashdrv_read(pOld,dstAddr_P+256*i,256);
      memcpy(pNew,srcAddr_P+256*i,256);
      flag=strncmp(pOld,pNew,256);
      printf(":::::::::::::::::::::  After buffer counter i=%d  flag=%d \n",i,flag);        
    }
#endif

  return retval;
}

unsigned int flashdrv_filewrite(FILE *fp,int size,void *dstP)
{
  int i;
  int nWritten; /* completed bytes */
  volatile unsigned char data[4096];
  volatile uint16 *dstAddr;
  
  nWritten = 0;
  dstAddr = dstP;
  //printf("flashdrv_filewrite: dst=0x%08x,  size=0x%08x \n\n",dstP,size);
  while ( nWritten < size )
    {
      int nRead;
      
      /* fill buffer with file */
      for( i = 0, nRead = 0; i < sizeof(data) && ( i < (size-nWritten) ); i++, nRead++ )
        {
          data[i] = fgetc(fp);
        }
      
      if ( 0 != flashdrv_updateImg( (void*)data, (void*)dstAddr, nRead ) )
        return 1;
      
      dstAddr += nRead >> 1;
      nWritten += nRead;
    }
  return 0;
}

#if 0
uint32 flashdrv_getBlockBase( uint32 ChipSeq )
{
  return mtd_dev.chipbase; 
}

/* 
 * Offset is related to flashbase, such as: 0x00004567.
 * blockOffset is returned with related to flashbase, for example: 0x00004000.
 *
 */

int32 flashdrv_getBlockSizeAndBlockOffset( uint32 Offset, uint32 ChipSeq, uint32 *blockNum, uint32* blockOffset, uint32 *blockSize )
{
  uint32 BlockSeq;
  
  assert( flash_chip_info[ChipSeq].BlockOffset != NULL );
  
  for( BlockSeq = 0; BlockSeq < flash_chip_info[ChipSeq].NumOfBlock; BlockSeq++ )
    {
      //rtlglue_printf("[%d][%d]: 0x%08x\n", ChipSeq, BlockSeq, blockbase + (uint32)flash_chip_info[ChipSeq].BlockOffset[ BlockSeq + 1 ] );
      if ( (uint32)flash_chip_info[ChipSeq].BlockOffset[ BlockSeq + 1 ] > Offset )
        {
          goto found;
        }
    }
  // Not found 
  return 1;
  
 found:
  if ( blockNum ) *blockNum = BlockSeq;
  if ( blockOffset ) *blockOffset = flash_chip_info[ChipSeq].BlockOffset[BlockSeq];
  if ( blockSize ) *blockSize = flash_chip_info[ChipSeq].BlockOffset[BlockSeq+1] - flash_chip_info[ChipSeq].BlockOffset[BlockSeq];
  return 0;
}
#endif


/*
 *  The following functions are used to return the addresses of BDINFO, CCFG, RUNTIME IMAGE.
 *
 *    These addresses are decided by the flash type and size.
 *  By calling these functions, we no more need compile flags.
 *
 */

uint32 flashdrv_getBoardInfoAddr()
{
#ifdef CONFIG_MTD_RTL865X
  return  FLASH_BASE+CONFIG_RTL865X_CUSTOM_BDINFO_ADDRESS; 
#else
  return mtd_dev.bdinfobase;
#endif
}

uint32 flashdrv_getCcfgImageAddr()
{
#ifdef CONFIG_MTD_RTL865X
  return  FLASH_BASE+CONFIG_RTL865X_CUSTOM_CCFG_ADDRESS; 
#else
  return mtd_dev.ccfgbase;
#endif
}

uint32 flashdrv_getRunImageAddr()
{
#ifdef CONFIG_MTD_RTL865X
  return  FLASH_BASE+CONFIG_RTL865X_CUSTOM_RUNTIME_ADDRESS; 
#else
  return  mtd_dev.kernbase;
#endif
  
}

uint32 flashdrv_getFlashBase()
{
#ifdef CONFIG_MTD_RTL865X
  return  FLASH_BASE ;
#else
  return mtd_dev.chipbase;
#endif
}


